/**
 * Copyright (C) 1997-2000 Dallas Semiconductor Corporation.
 * All rights Reserved. Printed in U.S.A.
 * This software is protected by copyright laws of
 * the United States and of foreign countries.
 * This material may also be protected by patent laws of the United States
 * and of foreign countries.
 * This software is furnished under a license agreement and/or a
 * nondisclosure agreement and may only be used or copied in accordance
 * with the terms of those agreements.
 * The mere transfer of this software does not imply any licenses
 * of trade secrets, proprietary technology, copyrights, patents,
 * trademarks, maskwork rights, or any other form of intellectual
 * property whatsoever. Dallas Semiconductor retains all ownership rights.
 */

public class IniFileException extends Exception
{
    /**
     * The version specified was not the same as the version in the file.
     */
    public static final int VERSION_ERROR       = 0;
    /**
     * Unable to open this file
     */
    public static final int FILEOPEN_ERROR      = 1;
    /**
     * Cannot read and/or write to this file
     */
    public static final int READWRITE_ERROR     = 2;
    /**
     * Error encountered while parsing file
     */
    public static final int PARSE_ERROR         = 3;

    /**
     * @serial the error number
     */
    private int errorNo = 0;
    /**
     * @serial the line number where a parse error occurred
     */
    private int parseLineNo = -1;
    /**
     * @serial the version string obtained from the file
     */
    private String version = "";

    /**
     * Class Constructor to use for <code>VERION_ERROR</code>s.
     * If this constructor is used with an <code>errorId</code> other than
     * <code>VERSION_ERROR</code>, the <code>version</code> parameter is ignored
     * @param description A description of the error that occurred.
     * @param errorId <code>VERSION_ERROR</code>.
     * @param version The version denoted in the file
     */
    public IniFileException(String description, int errorId, String version)
    {
        super(description);

        errorNo = errorId;

        // version is "" if not a VERSION_ERROR
        this.version = (errorId == VERSION_ERROR) ? version : "" ;
    }

    /**
     * Class Constructor to use for <code>PARSE_ERROR</code>s.
     * If this constructor is used with an <code>errorId</code> other than
     * <code>PARSE_ERROR</code>, the <code>lineNo</code> parameter is ignored
     * @param description A description of the error that occurred.
     * @param errorId <code>PARSE_ERROR<code>.
     * @param lineNo The line where a parse error occurred
     */
    public IniFileException(String description, int errorId, int lineNo)
    {
        super(description);

        errorNo = errorId;

        // lineNo is -1 if not a PARSE_ERROR
        parseLineNo = (errorId == PARSE_ERROR) ? lineNo : -1 ;
    }

    /**
     * Class Constructor with a descripiton and error ID.
     * @param description A description of the error that occurred.
     * @param errorId The predefined error number used to identify this type of error.
     */
    public IniFileException(String description, int errorId)
    {
        super(description);

        errorNo = errorId;
    }

    /**
     * Gets the error number for the type of error that occurred.
     * @return the error number constant for this class
     */
    public int getErrorId()
    {
        return errorNo;
    }

    /**
     * Gets the line where the parse error occurred in the file.
     * <p>NOTE: value is -1 if a parse error did not occur
     * @return the line number where the parse error occurred
     */
     public int getParseLineNumber()
     {
        return parseLineNo;
     }

    /**
     * Gets the version string (top line) in the file.
     * <p>NOTE: value is "" if there was no VERSION_ERROR
     * @return the line number where the parse error occurred
     */
     public String getFileVersion()
     {
        return version;
     }

}