/**
 * Copyright (C) 1997-2000 Dallas Semiconductor Corporation.
 * All rights Reserved. Printed in U.S.A.
 * This software is protected by copyright laws of
 * the United States and of foreign countries.
 * This material may also be protected by patent laws of the United States
 * and of foreign countries.
 * This software is furnished under a license agreement and/or a
 * nondisclosure agreement and may only be used or copied in accordance
 * with the terms of those agreements.
 * The mere transfer of this software does not imply any licenses
 * of trade secrets, proprietary technology, copyrights, patents,
 * trademarks, maskwork rights, or any other form of intellectual
 * property whatsoever. Dallas Semiconductor retains all ownership rights.
 */

import java.util.Vector;
import java.util.Enumeration;

/**
 * There are no duplicate keys.
 * Duplicate value names are treated individually.
 * All keys and values are case sensitive.
 * @author   MGP
 * @version  ???
 */


public class IniGroup extends Object
{
    private String group = null;
    private Vector keys = null;
    private Vector values = null;

    /**
     * Class Constructor.
     * @param group The name used to identify this group
     */
    public IniGroup( String group )
    {
        this.group = group;
        keys = new Vector(4);
        values = new Vector(4);
    }

    /**
     * Gets the name used to identify this group.
     * @return group name
     */
    public String getName()
    {
        return group;
    }

    /**
     * Adds a key-value pair to this group.
     * <p>If the key already exists, the value of the pair is replaced with the new value.
     * @param key the key name to be added to this group
     * @param value the value to be assigned to the key
     */
    public void addPair( String key, String value )
    {
        int index=keys.indexOf(key);

        if ( index == -1 )
        {   // this key not in vector
            // .add()
            keys.addElement( key );
            values.addElement( value );
        }
        else
        {   // this key in vector
            setValue( key, value );
        }
    }

    /**
     * Gets an ordered array of all keys contained in this group.
     * <p><code>null</code> is returned if there are no keys.
     * @return an array of keys
     */
    public String[] getKeys()
    {
        String[] arr = new String[keys.size()];
        int i=0;

        for (Enumeration e = keys.elements() ; e.hasMoreElements() ;)
        {
            arr[i] = (String)e.nextElement();
            i++;
        }

        return arr;
    }

    /**
     * Gets an ordered array of all values contained in this group.
     * <p><code>null</code> is returned if there are no values.
     * @return an array of values
     */
    public String[] getValues()
    {
        String[] arr = new String[values.size()];
        int i=0;

        for (Enumeration e = values.elements() ; e.hasMoreElements() ;)
        {
            arr[i] = (String)e.nextElement();
            i++;
        }

        return arr;
    }

    /**
     * Sets the value for the specified key.
     * <p>The value of the key is set only if the key already exists.
     * @param key the key used for identification of the key/value pair
     * @param value the value for the specified key
     */
    public void setValue(String key, String value)
    {
        int index=keys.indexOf(key);

        if ( index != -1 )
            // .set()
            values.setElementAt( value, index );
    }

    /**
     * Gets the value of for the specified key.
     * <p><code>null</code> is returned if the key was not found in the group.
     * @return value of the specified key
     */
    public String getValue(String key)
    {
        int index=keys.indexOf(key);

        if ( index != -1 )
            // .get()
            return (String) values.elementAt(index);
        else
            return null;
    }

    /**
     * Remove the key/value pair specified by the key.
     * <p>The pair is only removed if the key is found.
     * @param key the key used for identification of the key/value pair
     */
    public void removePair(String key)
    {
        int index=keys.indexOf(key);

        if ( index != -1 )
        {   // key found
            // .remove()
            keys.removeElementAt(index);
            values.removeElementAt(index);
        }
    }

    /**
     * Creates a string that can be used to write the group to a file.
     * <p>The string is formatted with an "=" between each key/value pair
     * followed by a newline after each pair.
     * @return a string useful for initialization file creation
     */
    public String toString()
    {
        String str = new String("");
        String[] kStrs = getKeys();
        String[] vStrs = getValues();

        for (int i=0; i<kStrs.length; i++)
            str += kStrs[i]
                + "="
                + vStrs[i]
                + System.getProperty("line.separator");

        return str;
    }

    /**
     * <code>IniGroup<code>s are equal when the group names are lexicographically equal.
     * @returns group equivalence
     */
    public boolean equals(Object o)
    {
        return ( this.group.equals( ((IniGroup)o).group ) );
    }

}
