
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



/**
 * Class allowing access to DS80C390 CAN Bus Controllers.
 * Each controller can access 15 message centers numbered 1 - 15.
 * Each message center can be programmed to either transmit or receive, except message center 15.
 * Message center 15 can only receive (not transmit) and can also buffer two back-to-back
 * frames.<br>
 * With an 18.432MHz system clock, CAN bit rates of 20, 50, and 125Kbps can be achieved.<br>
 * <pre>
 *         Prescaler  tqu      tseg1   tseg2
 * 20Kbps  71         3.852us  5       7
 * 50Kbps  41         2.224us  3       5
 * 125Kbps 7          3.798us  13      7
 * </pre>
 *
 * @see CanFrame
 */
public class CanBus
{

   /**
    * Passed to constructor to create object to access CAN Bus 0
    */
   public static final byte CANBUS0 = 0;

   /**
    * Passed to constructor to create object to access CAN Bus 1
    */
   public static final byte CANBUS1 = 1;

   private byte             portNum;
   private boolean          isOpen        = false;
   private CanFrame         transmitFrame = new CanFrame();

   // CAN Controller specific commands
   private static final int CMD_Open                  = 0;
   private static final int CMD_Close                 = 1;
   private static final int CMD_EnableController      = 2;
   private static final int CMD_DisableController     = 3;
   private static final int CMD_ResetController       = 4;
   private static final int CMD_SetBaudRatePrescaler  = 5;
   private static final int CMD_SetSJW                = 6;
   private static final int CMD_SetSMP                = 7;
   private static final int CMD_SetTSEG1              = 8;
   private static final int CMD_SetTSEG2              = 9;
   private static final int CMD_SetMediaIDMask        = 10;
   private static final int CMD_SetMediaIDArbitration = 11;
   private static final int CMD_Set11BitGlobalIDMask  = 12;
   private static final int CMD_Set29BitGlobalIDMask  = 13;
   private static final int CMD_Set29BitMC15IDMask    = 14;
   private static final int CMD_GetTXErrorCount       = 15;
   private static final int CMD_GetRXErrorCount       = 16;
   private static final int CMD_SetSiestaMode         = 17;
   private static final int CMD_EnableTransmit        = 18;
   private static final int CMD_DisableTransmit       = 19;

   // Message Center specific commands
   private static final int CMD_EnableMessageCenter     = 20;
   private static final int CMD_DisableMessageCenter    = 21;
   private static final int CMD_Set11BitMCArbitrationID = 22;
   private static final int CMD_Set29BitMCArbitrationID = 23;
   private static final int CMD_SetMCWriteOverEnable    = 24;
   private static final int CMD_SetMCRXTXMode           = 25;
   private static final int CMD_SetMCIDMaskEnable       = 26;
   private static final int CMD_SetMCMediaIDMaskEnable  = 27;
   private static final int CMD_SetAutoAnswerRFR        = 28;

   // Send/Receive commands
   private static final int CMD_Receive               = 29;
   private static final int CMD_ReceivePoll           = 30;
   private static final int CMD_SendFrame             = 31;
   private static final int CMD_SetReceiveQueueLimit  = 32;
   private static final int CMD_FramesAvailable       = 33;
   private static final int CMD_SetTransmitQueueLimit = 34;

   // Error return codes
   private static final int ERROR_BUSOFF            = -1;
   private static final int ERROR_TIMEOUT           = -2;
   private static final int ERROR_OPEN_ALREADYOPEN  = -3;
   private static final int ERROR_CLOSE_NOTOWNER    = -4;
   private static final int ERROR_PORT_ENABLED      = -5;
   private static final int ERROR_ALLOCATION_ERROR  = -6;
   private static final int ERROR_BIT_STUFF         = -7;
   private static final int ERROR_FORMAT            = -8;
   private static final int ERROR_TRANSMIT_NO_ACK   = -9;
   private static final int ERROR_BIT_ONE           = -10;
   private static final int ERROR_BIT_ZERO          = -11;
   private static final int ERROR_CRC               = -12;
   private static final int ERROR_COUNT_EXCEEDED    = -13;
   private static final int ERROR_PORT_DISABLED     = -14;

   /**
    * This is the native method that does it all.
    */
   private static native int canNative (int controller, int command,
                                        int messageCenter, byte[] params,
                                        CanFrame frame);

   protected CanBus ()
   {
   }

   /**
    * Creates a new CanBus instance on the specified port.
    * This opens and resets the specified CAN Controller and readies
    * it for operation.  The controller will be in a disabled state.
    * Calling <code>enableController</code> will place the controller on the bus.
    * All message centers start up disabled.
    *
    * @param portnum either <code>CANBUS0</code> or <code>CANBUS1</code>
    * @throws CanBusException if port is already open, or the controller will not reset
    */
   public CanBus (byte portnum)
      throws CanBusException
   {
      portNum = portnum;

      open();
      resetController();
   }

   /**
    * Opens this CAN Bus Controller.
    * Assigns ownership of the controller to the calling process.
    * The constructor calls this automatically when a new CanBus object is created.
    *
    * @throws CanBusException if controller is already open
    */
   public void open ()
      throws CanBusException
   {
      if (!isOpen)
      {
         if (canNative(portNum, CMD_Open, 0, null, null) < 0)
            throw new CanBusException("Cannot open CAN Controller",CanBusException.OPEN_ALREADYOPEN);

         isOpen = true;
      }
   }

   /**
    * Closes this CAN Bus Controller.
    * Reliquishes ownership of the controller and
    * releases all resources for this CAN controller.
    *
    * @throws CanBusException if controller is not open
    */
   public void close ()
      throws CanBusException
   {
      if (isOpen)
      {
         if (canNative(portNum, CMD_Close, 0, null, null) < 0)
            throw new CanBusException("Cannot close CAN Controller",CanBusException.CLOSE_NOTOWNER);

         isOpen = false;
      }
   }

   /**
    * Sets the baud rate prescaler from the CPU crystal.
    * This method sets the basic time quantum (tqu) necessary for
    * CAN communication.
    * The divisor divides straight off the external crystal on the processor.<br>
    * For instance, at 18.432MHz, a divisor of 7 will give you a
    * tqu of 379.774ns.<br>
    * tqu = 7 * 1/18.432MHz<br>
    *
    * @param prescaler 1 to 256
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if prescaler is out of range
    */
   public void setBaudRatePrescaler (int prescaler)
      throws CanBusException
   {
      checkOpen();
      if ((prescaler < 0) || (prescaler > 256))
         throw new IllegalArgumentException("Argument out of range.");

      byte[] paramArray = new byte [4];

      paramArray [0] = ( byte ) prescaler;

      canNative(portNum, CMD_SetBaudRatePrescaler, 0, paramArray, null);
   }

   /**
    * Sets the SJW (Synchronization Jump Width) necessary for adjusting TSEG1
    * and TSEG2 to compensate for sync problems during CAN communication. <br>
    * TSEG1 may be lengthened during resync.<br>
    * TSEG2 may be shortened during resync.
    *
    * @param jumpWidth 1 to 4 (1tqu to 4tqu)
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if SJW is out of range
    */
   public void setSynchronizationJumpWidth (int jumpWidth)
      throws CanBusException
   {
      checkOpen();
      if ((jumpWidth < 1) || (jumpWidth > 4))
         throw new IllegalArgumentException("Argument out of range.");

      // SJW in MOVX
      byte[] paramArray = new byte [4];

      paramArray [0] = ( byte ) jumpWidth;

      canNative(portNum, CMD_SetSJW, 0, paramArray, null);
   }

   /**
    * Sets SMP (Sample Rate) which is whether to use one or three samples per bit time
    * during CAN communication.
    *
    * @param sampleRate 1 or 3
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if SMP is out of range
    */
   public void setSampleRate (int sampleRate)
      throws CanBusException
   {
      checkOpen();
      if ((sampleRate != 1) && (sampleRate != 3))
         throw new IllegalArgumentException("Argument out of range.");

      // SMP = 1, three samples per bit, two tqu will be added to TSEG1
      // SMP = 0, one sample per bit
      byte[] paramArray = new byte [4];

      paramArray [0] = ( byte ) sampleRate;

      canNative(portNum, CMD_SetSMP, 0, paramArray, null);
   }

   /**
    * Sets TSEG1 (Timing Segment 1 = PROP_SEG + PHASE_SEG1) to a specified number
    * of time quanta.  This is the timing segment before the bit sample.
    *
    * @param tseg1 1 to 16 (1tqu to 16tqu)
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if TSEG1 is out of range
    */
   public void setTSEG1 (int tseg1)
      throws CanBusException
   {
      checkOpen();
      if ((tseg1 < 1) || (tseg1 > 16))
         throw new IllegalArgumentException("Argument out of range.");

      // PROP_SEG + PHASE_SEG1
      // Valid values 1tqu to 16tqu
      byte[] paramArray = new byte [4];

      paramArray [0] = ( byte ) tseg1;

      canNative(portNum, CMD_SetTSEG1, 0, paramArray, null);
   }

   /**
    * Sets TSEG2 (Timing Segment 2 = PHASE_SEG2) to a specified number of time quanta.
    * This is the timing segment after the bit sample.
    *
    * @param tseg2 1 to 8 (1tqu to 8tqu)
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if TSEG2 is out of range
    */
   public void setTSEG2 (int tseg2)
      throws CanBusException
   {
      checkOpen();
      if ((tseg2 < 1) || (tseg2 > 8))
         throw new IllegalArgumentException("Argument out of range.");

      // PHASE_SEG2
      // Valid values 1tqu to 8tqu
      byte[] paramArray = new byte [4];

      paramArray [0] = ( byte ) tseg2;

      canNative(portNum, CMD_SetTSEG2, 0, paramArray, null);
   }

   /**
    * Resets this CAN controller.
    * No more transmissions, or receptions will be
    * processed.  The CAN controller is disconnected from the
    * bus.  All message centers reset to power on defaults.<br>
    * <b>ALL SETTINGS WILL BE LOST!</b>
    *
    * @throws CanBusException if controller is not open
    */
   public void resetController ()
      throws CanBusException
   {
      checkOpen();

      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_ResetController, 0, paramArray, null);
   }

   /**
    * Shuts down this CAN controller.
    * No more transmissions, or receptions will be
    * processed.  The CAN controller is disconnected from the
    * bus.  User is free to change timing and other communication
    * critical parameters.
    * Message centers remain intact.
    *
    * @throws CanBusException if controller is not open
    */
   public void disableController ()
      throws CanBusException
   {
      checkOpen();

      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_DisableController, 0, paramArray, null);
   }

   /**
    * Starts up this CAN controller, and connects to the bus.
    * All critical timing parameters must already be set.
    * Message centers remain intact.
    *
    * @throws CanBusException if controller is not open
    */
   public void enableController ()
      throws CanBusException
   {
      checkOpen();

      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_EnableTransmit, 0, paramArray, null);
      canNative(portNum, CMD_EnableController, 0, paramArray, null);
   }

   /**
    * Starts up this CAN controller, but doesn't connect
    * CAN transmit to the bus.  Become a quiet listener
    * on the bus.
    * All critical timing parameters must already be set.
    * Message centers remain intact.
    *
    * @throws CanBusException if controller is not open
    */
   public void enableControllerPassive ()
      throws CanBusException
   {
      checkOpen();
      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_DisableTransmit, 0, paramArray, null);
      canNative(portNum, CMD_EnableController, 0, paramArray, null);
   }

   /**
    * Starts up this message center.
    * All mode changes must be complete.
    * Message center becomes operational and able to send
    * or receive messages depending on mode.
    *
    * @param messageCenter message center to enable
    * @throws CanBusException if controller is not open
    */
   public void enableMessageCenter (int messageCenter)
      throws CanBusException
   {
      checkOpen();

      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_EnableMessageCenter, messageCenter, paramArray, null);
   }

   /**
    * Shuts down this message center for changes.
    * No more transmissions, or receptions will be
    * processed at this message center.
    *
    * @param messageCenter message center to enable
    * @throws CanBusException if controller is not open
    */
   public void disableMessageCenter (int messageCenter)
      throws CanBusException
   {
      checkOpen();

      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_DisableMessageCenter, messageCenter, paramArray, null);
   }

   /**
    * Sets receive queue frame limit.
    * If the limit is set to zero, reads will block until a frame is
    * received, or bus error.  Maximum limit is 255 frames.
    *
    * @param numframes maximum number of frames to store in queue
    * @throws CanBusException if controller is not open, or if queue could not be
    * allocated
    * @throws IllegalArgumentException if number of frames is out of range
    */
   public void setReceiveQueueLimit (int numframes)
      throws CanBusException
   {
      checkOpen();
      if ((numframes < 256) && (numframes >= 0))
      {
         byte[] paramArray = new byte [4];

         paramArray [0] = ( byte ) numframes;

         if (canNative(portNum, CMD_SetReceiveQueueLimit, 0, paramArray, null) < 0)
            throw new CanBusException("Could not allocate queue",CanBusException.ALLOCATION_ERROR);
      }
      else
         throw new IllegalArgumentException("Argument out of range.");
   }

   /**
    * Gets number of frames in the receive queue.
    * If the receive queue limit is set to zero, the return value
    * will be zero.
    *
    * @return number of frames pending in receive queue
    * @throws CanBusException if controller is not open
    */
   public int receiveFramesAvailable ()
      throws CanBusException
   {
      checkOpen();
      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_FramesAvailable, 0, paramArray, null);

      return paramArray [0] & 0xFF;
   }

   /**
    * Receives a CAN data frame.
    * This method blocks until a frame is received.
    *
    * @param frame a frame in which to place the incoming data
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void receive (CanFrame frame)
      throws CanBusException
   {
      if (isOpen)
      {
          int    retVal  = canNative(portNum, CMD_Receive, 0, null, frame);

          if (retVal < 0)
          {
             if (retVal == ERROR_BUSOFF)
                throw new CanBusException("Receive Failed: BusOff",CanBusException.PORT_DISABLED);
             if (retVal == ERROR_PORT_DISABLED)
                throw new CanBusException("Receive Failed: Controller Disabled",CanBusException.PORT_DISABLED);
             else
                throw new CanBusException("failed in receive " + retVal,CanBusException.NONE);
          }
      }
      else
        throw new CanBusException("Controller not open",CanBusException.PORT_NOTOPENED);
   }

   /*
   Send/Receive Message Format
   +----------------+----------------------+-------------+--------+--------+------------+
   | Message Center | Remote Frame Request | Extended ID | ID (4) | Length | Data (1-8) |
   +----------------+----------------------+-------------+--------+--------+------------+
   */

   /**
    * Receives a CAN data frame.
    * This method returns immediately and does not block.
    *
    * @param frame frame in which to place the incoming data
    *
    * @return <code>true</code> if frame has been received
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public boolean receivePoll (CanFrame frame)
      throws CanBusException
   {
      if (isOpen)
      {
          int    retVal  = canNative(portNum, CMD_ReceivePoll, 0, null, frame);

          if (retVal > 0)
          {
             return true;
          }
          else if (retVal < 0)
          {
             if (retVal == ERROR_BUSOFF)
                throw new CanBusException("ReceivePoll Failed: BusOff",CanBusException.PORT_DISABLED);
             if (retVal == ERROR_PORT_DISABLED)
                throw new CanBusException("Receive Failed: Controller Disabled",CanBusException.PORT_DISABLED);
             else
                throw new CanBusException("failed in receive " + retVal,CanBusException.NONE);
          }
          else
             return false;
      }
      else
        throw new CanBusException("Controller not open",CanBusException.PORT_NOTOPENED);
   }

   /* Parse the array and fill the frame */

   /*
   Frame format

   message center | remote frame request | extended id | id (4) | length | data (8)
   */

   /**
    * Sets transmit queue frame limit.
    * If the limit is set to zero, writes will block until frame is
    * transmitted, or bus error.  Maximum limit is 255 frames.
    *
    * @param numframes  maximum number of frames to store in queue
    * @throws CanBusException if controller is not open
    * @throws IllegalArgumentException if number of frames is out of range
    */
   public void setTransmitQueueLimit (int numframes)
      throws CanBusException
   {
      checkOpen();
      if ((numframes < 256) && (numframes >= 0))
      {
         byte[] paramArray = new byte [4];

         paramArray [0] = ( byte ) numframes;

         if (canNative(portNum, CMD_SetTransmitQueueLimit, 0, paramArray, null) < 0)
            throw new CanBusException("Could not allocate queue",CanBusException.ALLOCATION_ERROR);
      }
      else
         throw new IllegalArgumentException("Argument out of range.");
   }

   /**
    * Sends a CAN frame from message center requested in CanFrame object.
    * If transmit queue size is one, this method will block until the
    * message has been successfully transmitted, or an error has occurred.
    * If transmit queue size is greater than one, messages will be queued
    * for transmit and method will return immediately.
    *
    * @param frame frame to send
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void sendFrame (CanFrame frame)
      throws CanBusException
   {
      if (isOpen)
      {
          int retVal = canNative(portNum, CMD_SendFrame, 0, null, frame);

          if (retVal < 0)
          {
             if (retVal == ERROR_BUSOFF)
                throw new CanBusException("Send Failed: BusOff",CanBusException.PORT_DISABLED);
             else if (retVal == ERROR_PORT_DISABLED)
                throw new CanBusException("Send Failed: Controller Disabled",CanBusException.PORT_DISABLED);
             else if (retVal == ERROR_TRANSMIT_NO_ACK)
                throw new CanBusException("Send Failed: No Ack",CanBusException.TRANSMIT_NO_ACK);
             else if (retVal == ERROR_BIT_ONE)
                throw new CanBusException("Send Failed: Bit One",CanBusException.BIT_ONE);
             else if (retVal == ERROR_BIT_ZERO)
                throw new CanBusException("Send Failed: Bit Zero",CanBusException.BIT_ZERO);
             else
                throw new CanBusException("failed in send " + retVal,CanBusException.NONE);
          }
      }
      else
        throw new CanBusException("Controller not open",CanBusException.PORT_NOTOPENED);
   }

   /**
    * Sends a CAN Data Frame from first available message center without
    * using an explicit <code>CanFrame</code>.
    * If transmit queue size is one, this method will block until the
    * message has been successfully transmitted, or an error has occurred.
    * If transmit queue size is greater than one, messages will be queued
    * for transmit and method will return immediately.
    *
    * @param ID 11 or 29 bit ID for frame
    * @param extendedID <code>true</code> for extended ID (29 bit), or <code>false</code>
    * for standard ID (11 bit)
    * @param data to send in frame (max 8 bytes)
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void sendDataFrame (int ID, boolean extendedID, byte[] data)
      throws CanBusException
   {
      synchronized (transmitFrame)
      {
        CanFrame frame = transmitFrame;
        frame.ID = ID;
        frame.extendedID = extendedID;
        System.arraycopy(data,0,frame.data,0,data.length);
        frame.length = data.length;
        frame.messageCenter = 0;
        frame.remoteFrameRequest = false;

        sendFrame(frame);
      }
   }

   /**
    * Sends a CAN Remote Frame Request from first available message center without
    * using an explicit <code>CanFrame</code>.
    * If transmit queue size is one, this method will block until the
    * message has been successfully transmitted, or an error has occurred.
    * If transmit queue size is greater than one, messages will be queued
    * for transmit and method will return immediately.
    *
    * @param ID 11 or 29 bit ID for frame
    * @param extendedID <code>true</code> for extended ID (29 bit), or <code>false</code>
    * for standard ID (11 bit)
    * @param data to send in frame (max 8 bytes)
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void sendRemoteFrameRequest (int ID, boolean extendedID,
                                       byte[] data)
      throws CanBusException
   {
      synchronized (transmitFrame)
      {
        CanFrame frame = transmitFrame;
        frame.ID = ID;
        frame.extendedID = extendedID;
        System.arraycopy(data,0,frame.data,0,data.length);
        frame.length = data.length;
        frame.messageCenter = 0;
        frame.remoteFrameRequest = true;

        sendFrame(frame);
      }
   }

   /**
    * Sets up a message center to auto reply to remote frame requests. <br>
    * Note: This resets the state of a message center for this purpose only.
    *       Also, the message center becomes enabled during this method.
    *
    * @param messageCenter message center to use for this reply
    * @param ID 11 or 29 bit ID for frame
    * @param data to send in frame (max 8 bytes)
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void autoAnswerRemoteFrameRequest (int messageCenter, int ID,
                                             byte[] data)
      throws CanBusException
   {
      // This is not implemented at this time
      throw new CanBusException("Don't use CanBus:autoAnswerRemoteFrameRequest()",CanBusException.NONE);
   }

   /**
    * Sets the global media ID mask which determines what bits to match
    * in the first two bytes of the incoming frame data area.
    *
    * @param mask 16 bit media mask
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMediaIDMask (int mask)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (mask >>> 8);
      temparr [1] = ( byte ) mask;

      canNative(portNum, CMD_SetMediaIDMask, 0, temparr, null);
   }

   /**
    * Sets the global media ID arbitration value which matches bits
    * in the first two bytes of the incoming frame data area.
    *
    * @param value 16 bit data value
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMediaIDArbitration (int value)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (value >>> 8);
      temparr [1] = ( byte ) value;

      canNative(portNum, CMD_SetMediaIDArbitration, 0, temparr, null);
   }

   /**
    * Sets the 11 Bit (STANDARD) Global ID Mask.
    * The Global ID Mask is used to denote which bits to match in the incoming
    * frame ID.
    *
    * @param mask 11 bit mask
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void set11BitGlobalIDMask (int mask)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (mask >>> 24);
      temparr [1] = ( byte ) (mask >>> 16);
      temparr [2] = ( byte ) (mask >>> 8);
      temparr [3] = ( byte ) mask;

      canNative(portNum, CMD_Set11BitGlobalIDMask, 0, temparr, null);
   }

   /**
    * Sets the 29 Bit (EXTENDED) Global ID Mask.
    * The Global ID Mask is used to denote which bits to match in the incoming
    * frame ID.
    *
    * @param mask 29 bit mask
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void set29BitGlobalIDMask (int mask)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (mask >>> 24);
      temparr [1] = ( byte ) (mask >>> 16);
      temparr [2] = ( byte ) (mask >>> 8);
      temparr [3] = ( byte ) mask;

      canNative(portNum, CMD_Set29BitGlobalIDMask, 0, temparr, null);
   }

   /**
    * Sets the 29 Bit Message Center 15 ID Mask.
    * Message Center 15 has it's own ID Mask.
    * The Message Center 15 ID Mask is used to denote
    * which bits to match in the Message Center 15 ID.
    *
    * @param mask 29 bit mask
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void set29BitMessageCenter15IDMask (int mask)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (mask >>> 24);
      temparr [1] = ( byte ) (mask >>> 16);
      temparr [2] = ( byte ) (mask >>> 8);
      temparr [3] = ( byte ) mask;

      canNative(portNum, CMD_Set29BitMC15IDMask, 0, temparr, null);
   }

   /**
    * Sets an 11 Bit Arbitration ID for a message center.
    * When this value matches an incoming frame ID subject to
    * the Global ID Mask or Message Center 15 Mask, the incoming
    * frame will be received.<br>
    * Note: This will change the specified message center to only
    * respond to 11 bit messages.
    *
    * @param messageCenter message center to change
    * @param ID 11 bit ID
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void set11BitMessageCenterArbitrationID (int messageCenter, int ID)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (ID >>> 24);
      temparr [1] = ( byte ) (ID >>> 16);
      temparr [2] = ( byte ) (ID >>> 8);
      temparr [3] = ( byte ) ID;

      canNative(portNum, CMD_Set11BitMCArbitrationID, messageCenter, temparr, null);
   }

   /**
    * Sets a 29 Bit Arbitration ID for a message center.
    * When this value matches an incoming frame ID subject to
    * the Global ID Mask or Message Center 15 Mask, the incoming
    * frame will be received.<br>
    * Note: This will change the specified message center to only
    * respond to 29 bit messages.
    *
    * @param messageCenter message center to change
    * @param ID 29 bit ID
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void set29BitMessageCenterArbitrationID (int messageCenter, int ID)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      temparr [0] = ( byte ) (ID >>> 24);
      temparr [1] = ( byte ) (ID >>> 16);
      temparr [2] = ( byte ) (ID >>> 8);
      temparr [3] = ( byte ) ID;

      canNative(portNum, CMD_Set29BitMCArbitrationID, messageCenter, temparr, null);
   }

   /**
    * Sets the state of write over in a specific message center.
    * This allows incoming frames to overwrite received frames
    * that haven't been moved out of the message center.
    *
    * @param messageCenter message center to change
    * @param writeover set to <code>true</code> to allow overwrite.
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMessageCenterWriteOverEnable (int messageCenter,
           boolean writeover)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      if (writeover)
         temparr [0] = 1;
      else
         temparr [0] = 0;

      canNative(portNum, CMD_SetMCWriteOverEnable, messageCenter, temparr, null);
   }

   /**
    * Sets a specific message center as transmitter. <br>
    * Note: Message Center 15 is always RECEIVE only.
    *
    * @param messageCenter message center to change
    * @throws CanBusException if controller is not open, or if bus error occurs
    * @throws IllegalArgumentException if message center is out of range
    */
   public void setMessageCenterTXMode (int messageCenter)
      throws CanBusException
   {
      checkOpen();

      // Exclude MC 15
      if (messageCenter == 15)
          throw new IllegalArgumentException("Argument out of range.");

      byte[] temparr = new byte [4];

      temparr [0] = 1;

      canNative(portNum, CMD_SetMCRXTXMode, messageCenter, temparr, null);
   }

   /**
    * Sets a specific message center as receiver. <br>
    * Note: Message Center 15 is always RECEIVE only.
    *
    * @param messageCenter message center to change
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMessageCenterRXMode (int messageCenter)
      throws CanBusException
   {
      checkOpen();

      // Exclude MC 15
      byte[] temparr = new byte [4];

      temparr [0] = 0;

      canNative(portNum, CMD_SetMCRXTXMode, messageCenter, temparr, null);
   }

   /**
    * Enables or disables Message ID Masking for a specific message center. <br>
    * Note: If masking is disabled, this message center ID must match ALL bits of incoming ID.
    *
    * @param messageCenter message center to change
    * @param maskEnable <code>true</code> to allow Global Mask, <code>false</code> to require exact match of ID
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMessageCenterMessageIDMaskEnable (int messageCenter,
           boolean maskEnable)
      throws CanBusException
   {
      checkOpen();
      byte[] temparr = new byte [4];

      if (maskEnable)
         temparr [0] = 1;
      else
         temparr [0] = 0;

      canNative(portNum, CMD_SetMCIDMaskEnable, messageCenter, temparr, null);
   }

   /**
    * Enables or disables Media ID Masking for a specific message center. <br>
    * Note: If masking is disabled, no checks will occur on the first two data bytes.
    *
    * @param messageCenter message center to change
    * @param maskEnable <code>true</code> to allow Media ID masking
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setMessageCenterMediaIDMaskEnable (int messageCenter,
           boolean maskEnable)
      throws CanBusException
   {
      checkOpen();

      byte[] temparr = new byte [4];

      if (maskEnable)
         temparr [0] = 1;
      else
         temparr [0] = 0;

      canNative(portNum, CMD_SetMCMediaIDMaskEnable, messageCenter, temparr, null);
   }

   /**
    * Gets the transmitter error count.
    *
    * @return transmitter error count
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public int getTXErrorCount ()
      throws CanBusException
   {
      checkOpen();
      byte[] paramArray = new byte [4];

      return canNative(portNum, CMD_GetTXErrorCount, 0, paramArray, null);
   }

   /**
    * Gets the receiver error count.
    *
    * @return receiver error count
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public int getRXErrorCount ()
      throws CanBusException
   {
      checkOpen();
      byte[] paramArray = new byte [4];

      return canNative(portNum, CMD_GetRXErrorCount, 0, paramArray, null);
   }

   /**
    * Gets the last bus error.
    *
    * @return Last bus error
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   private int getLastBusError ()
      throws CanBusException
   {
      checkOpen();
      return 0;
   }

   /**
    * Puts the CAN Controller in SIESTA (low power) mode.
    * When Bus activity is detected, the controller will wake up
    * and participate on the bus.
    *
    * @throws CanBusException if controller is not open, or if bus error occurs
    */
   public void setSiestaMode ()
      throws CanBusException
   {
      checkOpen();
      byte[] paramArray = new byte [4];

      canNative(portNum, CMD_SetSiestaMode, 0, paramArray, null);
   }

   private void checkOpen() throws CanBusException
   {
      if (!isOpen) throw new CanBusException("Controller not open",CanBusException.PORT_NOTOPENED);
   }
}
