
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



/**
 * The <code>CANFrame</code> class is used with the <code>CanBus</code>
 * class to send and receive frames on the CAN Bus.
 *
 * @see CanBus
 */
public class CanFrame
{
   /*
   DANGER:  Do not add, remove, or modify any instance fields
   in CanFrame without understanding the consequences
   of doing so.  This means native code changes in several places including
   nat_can.a51.
   You have been warned.

   -Pinky
   */

   /**
    * CAN Controller Message Center.
    */
   public int messageCenter;

   /**
    * Remote Frame if <code>true</code> or Data Frame if <code>false</code> .
    */
   public boolean remoteFrameRequest = false;

   /**
    * Standard ID (11 bits) if <code>false</code> or Extended ID (29 bits) if <code>true</code>.
    */
   public boolean extendedID = false;

   /**
    * Standard ID (11 bits) or Extended ID (29 bits) based on the value of <code>extendedID</code>.
    */
   public int ID;

   /**
    * Length of data area.
    */
   public int length;

   /**
    * Frame data area.
    */
   public byte[] data = new byte [8];

   /**
    * Creates a new CAN Frame with default parameters.
    * Equivalent of the following code:
    * <pre>
    * ID = 0;
    * extendedID = false;
    * messageCenter = 0;
    * remoteFrameRequest = false;
    * data = new byte[8];
    * length = 0;
    * </pre>
    */
   public CanFrame ()
   {
   }

   /**
    * Creates a new CAN Frame with specified parameters.
    *
    * @param ID identifier on frame
    * @param extendedID <code>true</code> for extended ID (29 bit), or <code>false</code> for
    * standard ID (11 bit)
    * @param buf data for frame (maximum 8 bytes)
    * @param length length of data
    */
   public CanFrame (int ID, boolean extendedID, byte[] buf, int length)
   {
      this.ID         = ID;
      this.extendedID = extendedID;

      System.arraycopy(buf, 0, data, 0, length);

      this.length = length;
   }

   /**
    * Sets the ID field on a CAN Frame.
    *
    * @param ID identifier on frame
    */
   public void setID (int ID)
   {
      this.ID = ID;
   }

   /**
    * Gets the ID field on a CAN Frame to be sent
    * or already received.
    *
    * @return identifier on frame
    */
   public int getID ()
   {
      return ID;
   }

   /**
    * Sets the ID field type on a CAN Frame.
    *
    * @param extendedID <code>true</code> for extended ID (29 bit), or <code>false</code> for
    * standard ID (11 bit)
    */
   public void setExtendedID (boolean extendedID)
   {
      this.extendedID = extendedID;
   }

   /**
    * Gets the ID field type on a CAN Frame.
    *
    * @return <code>true</code> for extended ID (29 bit), or <code>false</code> for
    * standard ID (11 bit)
    */
   public boolean getExtendedID ()
   {
      return extendedID;
   }

   /**
    * Sets the message center from which to send this CAN Frame.
    *
    * @param MC message Center to use
    */
   public void setMessageCenter (int MC)
   {
      messageCenter = MC;
   }

   /**
    * Gets the message center where this frame was received, or where
    * this frame will be sent.
    *
    * @return message center
    */
   public int getMessageCenter ()
   {
      return messageCenter;
   }

   /**
    * Sets the RTR bit in the frame to change this from a Data Frame to a
    * Remote Frame.
    *
    * @param RTR <code>true</code> for Remote Frame, <code>false</code> for Data Frame
    */
   public void setRemoteFrameRequest (boolean RTR)
   {
      remoteFrameRequest = RTR;
   }

   /**
    * Gets the RTR bit in the frame to determine if it is a
    * Remote Frame.
    *
    * @return <code>true</code> for Remote Frame, <code>false</code> for Data Frame
    */
   public boolean getRemoteFrameRequest ()
   {
      return remoteFrameRequest;
   }

   /**
    * Sets the data for the outgoing frame.
    * Maximum data length is 8 bytes.
    *
    * @param buf data to put in frame
    */
   public void setData (byte[] buf)
   {
      synchronized (this)
      {
         System.arraycopy(buf, 0, data, 0, buf.length);

         length = buf.length;
      }
   }

   /**
    * Gets the data for the incoming or outgoing frame.
    *
    * @return data in frame
    */
   public byte[] getData ()
   {
      return data;
   }

   /**
    * Sets the length for the data in the outgoing frame.
    * Maximum data length is 8 bytes.
    *
    * @param len length of data in frame
    */
   public void setLength (int length)
   {
      if (length > 8)
         throw new ArrayIndexOutOfBoundsException();

      this.length = length;
   }

   /**
    * Gets the length of the data in the incoming or outgoing frame.
    *
    * @return length of data in frame
    */
   public int getLength ()
   {
      return length;
   }
}
