
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



import java.io.*;


/**
 * <CENTER><H2>LCDOutputStream</H2></CENTER>
 * <H3>Description</H3>
 * <dode>LCDOutputStream</code> is used for writing data to a LCD Display.
 * This class can be used for direct writing to the LCD display on the TINI.
 * It has been tested with LCDs that use the Samsung
 * KS0066U and KS0070B segment drivers.  Using this stream, you can create
 * a PrintStream and do println()s to the display.  Each println will show
 * up as the next available line on the display.
 *
 * <H3>Data Sheets</H3>
 * <A HREF="http://www.usa.samsungsemi.com/">Samsung</A><BR>
 * <A HREF="http://www.usa.samsungsemi.com/System_LSI/LCD_Driver_IC/Character/KS0070B/DS-KS0070B-V0.PDF">
 * KS0070B Datasheet</A><BR>
 * <A HREF="http://www.usa.samsungsemi.com/System_LSI/LCD_Driver_IC/Character/KS0066U/KS0066U.pdf">
 * KS0066U Datasheet</A><BR>
 *
 * <H3> Example Usuage </H3>
 * <PRE><CODE>
 *
 * PrintStream out = null;
 * try
 * {
 *    out = new PrintStream(lp.getOutputStream());
 * }
 * catch (IOException e)
 * {
 *    System.out.println(e);
 * }
 * LCDPort.setAddress(0x40);  // set the address
 * out.println("HELLO");      // write Hello to the LCD
 *
 * </CODE></PRE>
 */
public class LCDOutputStream
   extends java.io.OutputStream
{
   LCDPort LCDPortObject = null;

   /**
    * Private constructor
    *
    *
    */
   private LCDOutputStream ()
   {
   }

   /**
    * Creates a new data output stream to write data to the LCD port.
    * @param lcd the LCD port to use for communications
    *
    */
   public LCDOutputStream (LCDPort lcd)
   {
      LCDPortObject = lcd;
   }

   /**
    * Opens a LCD output stream for writing. Calls the LCDPorts open method.
    *
    * @throws  IOException if an unable to open the port
    */
   private void open ()
      throws IOException
   {
      LCDPortObject.open();
   }

   /**
    * Writes the given character to the LCD display.  This method blocks until
    * the byte has actually been sent.
    *
    * @param ch the byte to write to the display
    *
    * @throws IOException if port is not open
    *
    */
   public synchronized void write (int ch)
      throws IOException
   {
      LCDPortObject.write(ch);
   }

   /**
    * Displays the array on the LCD display.  This method blocks until all bytes
    * have been sent.
    *
    * @param barr the array to send
    * @param offset the offset into the array to begin sending
    * @param len the number of bytes to send
    *
    * @throws IOException if port is not open
    *
    */
   public synchronized void write (byte[] barr, int offset, int len)
      throws IOException
   {
      LCDPortObject.write(barr, offset, len);
   }

   /**
    * Closes this output stream and releases any system resources
    * associated with this stream.  Calls the underlying LCDPort close method.
    *
    * @exception  IOException  if port is not open
    */
   public void close ()
      throws IOException
   {
      LCDPortObject.close();
   }
}
