
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



import java.lang.*;
import java.io.*;


/**
 * <CENTER><H2>LCDPort</H2></CENTER>
 * <H3> Description </H3>
 * LCDPort is used to write to the LCD display.  This class was written and tested using the
 * KS0066U and KS0070B segment drivers/controller.  The class can be used
 * as part of a streams based protocol.  There are also methods for communicating
 * directly to the LCD display.  Using the sendControl and sendData commands you
 * can control the LCD directly without the need for streams.
 *
 * <H3> DataSheets </H3>
 * <A HREF="http://www.usa.samsungsemi.com/">Samsung</A><BR>
 * <A HREF="http://www.usa.samsungsemi.com/System_LSI/LCD_Driver_IC/Character/KS0070B/DS-KS0070B-V0.PDF">
 * KS0070B Datasheet</A><BR>
 * <A HREF="http://www.usa.samsungsemi.com/System_LSI/LCD_Driver_IC/Character/KS0066U/KS0066U.pdf">
 * KS0066U Datasheet</A><BR>
 *
 * <H3> Example Usuage </H3>
 * <PRE><CODE>
 * LCDPort.sendControl(0x38);  // Set up the display mode
 * LCDPort.sendControl(0x0C);  // Turn on display with no cursor
 * LCDPort.setAddress(0x20);
 * LCDPort.sendControl(1);     // Clear the screen
 * LCDPort.setAddress(8);
 * LCDPort.sendData((int)85);  // Write a U
 * </CODE></PRE>
 *
 * @author CW
 * @version 1.0
 */
public class LCDPort
{

   /**
    * Port Type for NativeComm
    */
   private static final int LCD_DRIVER = 2;

   /** COMMAND Stream           */
   //private static final int COMMAND = 0;

   /** STDOUT Stream           */
   private static final int STDOUT = 2;

   /** STDERR Stream           */
   private static final int STDERR = 3;

   /** GENERIC Stream          */
   //private static final int GENERIC = 4;

   /** Field LCD_DEFAULT_READ_TIMEOUT           */
   private static final int LCD_DEFAULT_READ_TIMEOUT = 0;

   //    public static final int PARAM_COMMAND_TABLE = 1;
   private int              LCDHandle      = 0;
   private int              PortType       = 0;
   private int              StreamType     = 0;
   private boolean          IsOpen         = false;
   private int              LCDReadTimeout = LCD_DEFAULT_READ_TIMEOUT;
   private int              streamType;
   private byte[]           LCDPortWriteArray;
   private static final int LCD_PORT_ARRAY_LENGTH = 32;

   /**
    * Method declaration
    *
    *
    * @param paramNum
    * @param b
    * @param length
    *
    *
    */
   private static native void SetLCDParams (int paramNum, byte[] b,
                                            int length);

   /**
    * Sends a parameter to the LCD display.  The only parameter currently
    * supported is number of lines.  Valid line numbers supported
    * are 1, 2, and 4.<BR>
    * <B>Example:</B> Set the number of lines to 1
    *  <PRE><CODE>
    * byte[] b1 = {0x01};
    * LCDPort.setLCDParams(1, b1, b1.length);
    *  </CODE></PRE>
    *
    * @param paramNum parameter number (only 1 is supported)
    * @param params the parameter values, currently only number of lines is supported, allowable values include 1, 2 or 4
    * @param length the length of the parameter array
    *
    *
    */
   public static void setLCDParams (int paramNum, byte[] params, int length)
   {
      SetLCDParams(paramNum, params, length);
   }

   /**
    * Default constructor.
    *
    *
    *
    */
   private LCDPort ()
   {
   }

   /**
    * Constructor declaration
    *
    *
    * @param portNum
    *
    *
    */
   private LCDPort (int portNum)
   {
   }

   /**
    * Constructs a new LCDPort.
    * <PRE><CODE>
    * LCDPort port = new LCDPort(NativeComm.PORT_LCD, NativeComm.STREAM_STDOUT);
    * </CODE></PRE>
    * @param portNum the port number of the driver, should be NativeComm.PORT_LCD
    * @param stream the stream to open, only stdout is supported
    * @throws IOException if portNum is not a valid port type or
    * stream type is not valid
    */
   public LCDPort (int portNum, int stream)
      throws IOException
   {
      switch (portNum)
      {

         case LCD_DRIVER :
            PortType = NativeComm.PORT_LCD;
            break;
      }

      if (stream == STDOUT)
      {
         streamType = NativeComm.STREAM_STDOUT;
      }
      else if (stream == STDERR)
      {
         streamType = NativeComm.STREAM_STDERR;
      }
      else
         throw new IOException("Invalid LCD Stream Type");

      LCDPortWriteArray = new byte [LCD_PORT_ARRAY_LENGTH];
   }

   /**
    * Opens the LCD port for communication.  This method is called from getInputStream
    * and getOutputStream.
    */
   public void open ()
   {
      if (!IsOpen)
      {
         LCDHandle = NativeComm.open(PortType, StreamType);
         IsOpen    = true;
      }
   }

   /**
    * Closes the port and frees up system resources.
    * @throws IOException if port is not already open.
    */
   public void close ()
      throws IOException
   {
      if (IsOpen)
      {
         NativeComm.close(LCDHandle);

         IsOpen = false;
      }
      else
         throw new IOException("Could not close LCD port");
   }

   /**
    * Writes the array to the LCD port.  Bytes are written sequentially.  This method 
    * blocks until all bytes have been sent.
    *
    * @param arr the array to be written to the LCD port
    * @throws IOException if port is not open
    */
   public void write (byte[] arr)
      throws IOException
   {
      if (IsOpen)
      {

         // com.dalsemi.jellybeans.Debug.debugDump(0x5b);
         NativeComm.write(LCDHandle, arr);
      }
      else
         throw new IOException("Could not write to LCD port");
   }

   /**
    * Writes the array to the LCD port.  Bytes are written sequentially.  This method 
    * blocks until all bytes have been sent.
    *
    * @param arr the array to be written
    * @param offset the offset in the array to begin sending
    * @param len the number of bytes to send
    * @throws IOException if port is not open
    */
   public void write (byte[] arr, int offset, int len)
      throws IOException
   {
      int bytesleft;
      int curoffset;
      int sendlength;

      if (IsOpen)
      {

         // com.dalsemi.jellybeans.Debug.debugDump(0x5c);
         // Native doesn't handle offsets other than 0.
         // So we need to check for this and format it correctly
         // in a new buffer.
         //if (offset == 0)
         NativeComm.write(LCDHandle, arr, offset, len);

         /*else
         {

             // Copy the array to a LCD buffer, and send each
             // one individually.
             bytesleft = len;
             curoffset = offset;

             while (bytesleft > 0)
             {
                 if (bytesleft > LCD_PORT_ARRAY_LENGTH)
                     sendlength = LCD_PORT_ARRAY_LENGTH;
                 else
                     sendlength = bytesleft;

                 synchronized (LCDPortWriteArray)
                 {
                    System.arraycopy(arr, curoffset, LCDPortWriteArray, 0,
                                     sendlength);
                    NativeComm.write(LCDHandle, LCDPortWriteArray,
                                     sendlength);

                    bytesleft -= sendlength;
                    curoffset += sendlength;
                 }
             }
         }*/
      }
      else
         throw new IOException("Could not write to LCD port");
   }

   /**
    * Writes a byte to the LCD port. This method blocks until the byte has 
    * actually been sent.
    * @param ch the data to be written
    * @throws IOException if port is not open
    */
   public void write (int ch)
      throws IOException
   {
      if (IsOpen)
      {
         synchronized (LCDPortWriteArray)
         {
            LCDPortWriteArray [0] = ( byte ) ch;

            // com.dalsemi.jellybeans.Debug.debugDump(0x5d);
            NativeComm.write(LCDHandle, LCDPortWriteArray, 0, 1);

            //  com.dalsemi.jellybeans.Debug.debugDump(0x5e);
         }
      }
      else
         throw new IOException("Could not write to LCD port");
   }

   /*
    *
    * @param arr - the array for the data read to be placed in
    * @param offset - the offset to begin saving data
    * @param len - the number of bytes to read
    *
    * @return The byte read.
    * @throws IOException
    */
/*   public int read (byte[] arr, int offset, int len)
      throws IOException
   {
      throw new IOException("Input unsupported");
   }
*/
   /*
    * read from the LCD port.
    * @return the byte read.
    * @exception IOException
    */
/*   public int read ()
      throws IOException
   {
      throw new IOException("Input unsupported");
   }
*/
   /*
    * gets a LCD port Input stream
    * @return InputStream
    * @exception throws IOException
    */
/*   public InputStream getInputStream ()
      throws IOException
   {
      throw new IOException("Input not currently supported");
   }
*/
   /**
    * Gets a LCD port Output stream
    * @return OutputStream
    * @exception throws IOException if unable to open the port
    */
   public OutputStream getOutputStream ()
      throws IOException
   {
      open();

      return new LCDOutputStream(this);
   }

   /*
    * gets the size of the input buffer.
    * @return the size of the input buffer.
    */
  /* public int getInputBufferSize ()
   {
      return 0;
   }*/

   /**
    * Gets the size of the output buffer.
    * @return the size of the output buffer
    */
   public int getOutputBufferSize ()
   {
      return 1;
   }

   /**
    * Returns the number of bytes available on the port.
    *
    *
    * @return The number of bytes available
    *
    *
    */
/*   public int available ()
   {

      //return NativeComm.available(LCDHandle);
      return 0;
   }
*/
   /**
    * Sets the number of lines for the LCD driver.
    *
    * @param num number of lines
    */
   public static void setNumberOfLines (int num)
   {
      byte[] param = new byte [1];

      param [0] = ( byte ) num;

      SetLCDParams(1, param, 1);
   }

   /**
    * Method declaration
    *
    *
    * @param value
    *
    *
    */
   private static native void sendControlNative (int value);

   /**
    * Sends a control byte to the LCD display.  Refer to the data sheets of the
    * KS0066U and KS0070B for information on control characters.  This can be
    * used to set the address, shift, clear, etc.
    *
    *
    * @param value the control value to be sent
    */
   public static void sendControl (int value)
   {
      sendControlNative(value);
   }

   /**
    * Sends a data byte to the display.  This can be used to directly read and
    * write data to the LCD display.
    *
    *
    * @param value the byte to be displayed
    */
   private static native void sendDataNative (int value);

   /**
    * Sends a data byte to the display.  This can be used to directly write
    * data to the LCD display.
    *
    * @param value the byte to be displayed
    */
   public static void sendData (int value)
   {
      sendDataNative(value);
   }

   /**
    * Method declaration
    *
    *
    * @param on
    *
    *
    */
   private static native void shiftModeNative (boolean on);

   /**
    * Turns automatic shifting on and off.
    * @param on set to true to turn shifting on
    */
   public static void setShiftMode (boolean on)
   {
      shiftModeNative(on);
   }

   /**
    * Method declaration
    *
    *
    * @param dir
    *
    *
    */
   private static native void setShiftDirectionNative (boolean dir);

   /**
    * Sets the direction of shifts.
    *
    *
    * @param dir set to true to shift characters to the left
    */
   public static void setShiftDirection (boolean dir)
   {
      setShiftDirectionNative(dir);
   }

   /**
    * Method declaration
    *
    *
    * @param count
    *
    *
    */
   private static native void setShiftIntervalNative (int count);

   /**
    * Sets the interval of the automatic shift in milliseconds.
    *
    *
    * @param num_ms number of milliseconds between shifts
    */
   public static void setShiftInterval (int num_ms)
   {
      int count = ( int ) (num_ms / 8);

      setShiftIntervalNative(count);
   }

   /**
    * Method declaration
    *
    *
    * @param address
    *
    *
    */
   private static native void setAddressNative (int address);

   /**
    * Sets the address for the next write.  Refer to the data sheets for the
    * KS0066U and KS0070B for information on control characters.
    *
    *
    * @param address the address for the next write
    */
   public static void setAddress (int address)
   {
      setAddressNative(address);
   }
}
