
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;
import com.dalsemi.system.Debug;



import java.io.*;


/**
 * An Serial output stream is an output stream for writing data to a serial port.
 */
public class SerialOutputStream
   extends java.io.OutputStream
{
   //
   // Handle to IO output stream
   //
   private int outputHandle;
   private int portNum;
   //
   // Referece to creating serial port. Don't use it, but can't 
   // hurt to hold onto it.
   //
   private TINISerialPort tsp;
   private byte[] writeArray;
   private int             SERIAL_PORT_ARRAY_LENGTH      = 1024;
   private int             SERIAL_PORT_READ_ARRAY_LENGTH = 1024;


   /**
    * Constructor declaration
    *
    * @see
    */
   private SerialOutputStream ()
   {
   }

   /**
    * Creates a new data output stream to write data to the serial port.
    * @param sp the serial port to use for communications.
    *
    */
   public SerialOutputStream (TINISerialPort t)
   {
      outputHandle = -1;
      portNum = t.portNum;//TINISerialPort.getPortNumByName(t.getName());
      //Debug.dump("Port number = " + portNum);
      //Debug.dump("Port name = " + t.getName());
      tsp = t;
      writeArray = new byte [SERIAL_PORT_ARRAY_LENGTH];
   }

   /**
    * Opens a serial output stream for writing.
    *
    * @exception  IOException  if an I/O error occurs.
    */
   protected void open ()
      throws IOException
   {

      if (!NativeComm.io_isHandleOpen(outputHandle))
     {

         //Debug.dump("Creating Handle");
         outputHandle = NativeComm.io_openNative(portNum,NativeComm.STREAM_STDOUT);
         NativeComm.setHandleNative(outputHandle);
     }
      else
     {
         //Debug.dump("Port Already Open");
         throw new IOException("Port Already Open");
     }

   }

   /**
    * Writes the given character to the serial port.  This method blocks until
    * the byte has actually be sent.
    *
    * @param ch the byte to send.
    *
    * @throws IOException
    *
    * @see
    */
   public synchronized void write (int ch)
      throws IOException
   {
      //Debug.dump("w");
      if (NativeComm.io_isHandleOpen(outputHandle))
     {
          //Debug.dump(ch);
          writeArray[0] = (byte) ch;
          NativeComm.io_ioctlNative(NativeComm.WRITE,outputHandle,writeArray,0,1,0x3FFFFFFF,true);
     }
      else
     {
     //Debug.dump("Ex");
     throw new IOException("Could not write to serial port");      
     }
     
   }

   /**
    * Sends the given array out the serial port.  This method blocks until all bytes
    * have been sent.
    *
    * @param barr the array to send.
    * @param offset the offset into the array to begin sending.
    * @param len the number of bytes to send.
    *
    * @throws IOException
    *
    * @see
    */
   public synchronized void write (byte[] barr, int offset, int len)
      throws IOException
   {
      ////Debug.dump("w");
      if (!NativeComm.io_isHandleOpen(outputHandle))
      {
         //Debug.dump("Could not write to serial port");
         throw new IOException("Could not write to serial port");
      }

      if ((len < 0) || (offset < 0) || (offset + len > barr.length)) 
        throw new ArrayIndexOutOfBoundsException();
      if (len == 0) return ; //DON'T GO NATIVE IN THiS CASE!!!!!!
	      

      NativeComm.io_ioctlNative(NativeComm.WRITE,outputHandle,barr, offset, len,0x3FFFFFFF,true);
   }

   /**
    * Closes this output stream and releases any system resources
    * associated with this stream.
    *
    * @exception  IOException  if an I/O error occurs.
    */
   public void close ()
      throws IOException
   {
     if (NativeComm.io_isHandleOpen(outputHandle))
     {
        NativeComm.io_closeNative(outputHandle);
     }

   }


   public void flush()
      throws IOException
   {
     if (NativeComm.io_isHandleOpen(outputHandle))
     {
        NativeComm.io_ioctlNative(NativeComm.WRITE,outputHandle,writeArray,0,0,0x3FFFFFFF,true);
        
     }

   }

}
