/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;
import com.dalsemi.system.Debug;
import java.io.*;
import java.util.TooManyListenersException;
import javax.comm.*;

/**
 * This class is the actual serial port implementation used by
 * <code>javax.comm.SerialPort</code> and should be used directly only
 * if portability is not required and if <code>javax.comm.SerialPort</code>
 * does not implement a method that is supported in this class. <p>One potentially
 * useful method in this class is <code>setDivisor</code>. This method allows
 * an application to set the divisor register on the 16552 DUART (serial2 and
 * serial3 only). This will allow the hardware designer to choose the crystal
 * frequency used by the DUART and giving full control over the baud rates (see
 * the data sheet for the 16552 for formulas to calculate the buad rate using
 * the divisor value and crystal frequency). <p> The serial port is opened
 * automatically when a <code>TINISerialPort</code> object is constructed.
 * <code>serial0</code>, by default, is owned by the shell's SerialServer.
 * The shell command "downserver -s" must be called if an application, running
 * from the shell, wishes to take ownership of the port. <code>serial1</code>,
 * by default, is owned by the external 1-Wire adapter. The DS2480 chip must be
 * removed and level shifters must be added to make this an RS-232 compatible
 * port. <code>com.dalsemi.system.TINIOS.enableSerialPort1</code> must be called
 * to enable this port after the system is booted. <code>serial2</code> and
 * <code>serial3</code> require external hardware support (16552 Dual UART)
 * on a socket board to make use of these ports.
 * The class <code>com.dalsemi.system.TINIOS</code> contains methods for
 * enabling and setting the address of the external serial ports. These methods
 * need to be called only one time after TINI's heap is cleared.
 *<p>
 * <A HREF="http://www.national.com/pf/PC/PC16552D.html">National Semiconductor PC16552 data sheet</A><BR>
 *
 * @see com.dalsemi.system.TINIOS
 *
 */
public class TINISerialPort
   extends javax.comm.SerialPort
{

   /**
    * Constructor - Attempts to open the requested serial port specified by the
    *  parameter <code>name</code>.
    *
    * @param name the name of the serial port that is to be opened
    *
    * @throws IllegalArgumentException if the specified port name is not
    *         supported on TINI. Valid port names include <code>serial0</code>
    *         through <code>serial3</code>
    */
   public TINISerialPort (String name)
   {
      super();

      String[] localPortNames = getPortNames();


      for (int i = 0; i < localPortNames.length; i++)
      {
         if (name.equals(localPortNames [i]))
         {
            this.name = name;
            portNum   = localPortNames [i].charAt(localPortNames [i].length() - 1)
                        - '0';
            open();

            return;
         }
      }
/*
      if (name.indexOf("COM")==0)
      {
            portNum = (name.charAt(name.length()) - '0');
            this.name = "serial" + portNum;
            open();
            return;
      }
*/
      throw new IllegalArgumentException("port " + name
                                         + " is not a valid port name");
   }

  /**
   * Sets the receive FIFO trigger level for the external serial ports. Valid
   * trigger levels are 14, 8, 4 and 1. Invalid levels will be rounded down to
   * the next valid level. A 0 level will be adjusted to 1. Attempts to set the
   * trigger level for the internal ports will fail silently.
   *
   * @param trigger number of bytes in the receive FIFO that will cause a
   * receive interrupt
   */
  public void setRcvFifoTrigger(int trigger)
  {
    SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETFIFOTRIGGER,
                                 (new byte []{
                                 (byte) (portNum & 0xFF),
                                 (byte)(trigger & 0xFF)}));

  }


   /**
    * Closes the serial port and associated streams.
    */
   public void close ()
   {
      //Debug.dump("Closing Serial Port");
      if (portState != PORT_STATE_OPEN)
         return;
      removeEventListener();
      try
      {

         super.close();

         portState = PORT_STATE_CLOSED;

         setOwned(portNum, false);

         inputStream.close();
         outputStream.close();
      }
      catch (Exception i){}
   }

   /**
    * Returns the port's current baud rate in bits per second.
    *
    * @return the current baud rate
    */
   public int getBaudRate ()
   {
      if (portNum < 2)
      {

        switch (internalBaudRate)
        {

           case SPEED_300 :
              return 300;

           case SPEED_600 :
              return 600;

           case SPEED_1200 :
              return 1200;

           case SPEED_2400 :
              return 2400;

           case SPEED_4800 :
              return 4800;

           case SPEED_9600 :
              return 9600;

           case SPEED_19200 :
              return 19200;

           case SPEED_38400 :
              return 38400;

           case SPEED_57600 :
              return 57600;

           case SPEED_115200 :
              return 115200;
        }
      }
      else
      {
        switch(SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_GETBAUDRATE,
                                            (new byte []{
                                               ( byte ) (portNum & 0x0ff) })))
        {
          case DIVISOR_300:
          return 300;

          case DIVISOR_600:
          return 600;

          case DIVISOR_1200:
          return 1200;

          case DIVISOR_2400:
          return 2400;

          case DIVISOR_4800:
          return 4800;

          case DIVISOR_9600:
          return 9600;

          case DIVISOR_19200:
          return 19200;

          case DIVISOR_38400:
          return 38400;

          case DIVISOR_57600:
          return 57600;

          case DIVISOR_115200:
          return 115200;

          case DIVISOR_230400:
          return 230400;
        }
      }

      return -1;
   }

   /**
    * Returns the port's current the data bit (word size) setting.
    *
    * @return the current word size
    */
   public int getDataBits ()
   {
      int retVal = 0;

      retVal = SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_GETBITS,
                                            (new byte []{
                                               ( byte ) (portNum & 0xFF) }));

      switch (retVal & 0x03)
      {

         case 0 :
            return javax.comm.SerialPort.DATABITS_5;

         case 1 :
            return javax.comm.SerialPort.DATABITS_6;

         case 2 :
            return javax.comm.SerialPort.DATABITS_7;

         case 3 :
            return javax.comm.SerialPort.DATABITS_8;
      }

      return javax.comm.SerialPort.DATABITS_8;
   }

   /**
    * Returns the port's current stop bits setting.
    *
    * @return the current stop bits
    */
   public int getStopBits ()
   {
      int retVal = 0;

      retVal = SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_GETBITS,
                                            (new byte []{
                                               ( byte ) (portNum & 0x0ff) }));

      if ((retVal & 0x04) == 0)
         return javax.comm.SerialPort.STOPBITS_1;

      switch (retVal & 0x03)
      {

         case 0 :
            return javax.comm.SerialPort.STOPBITS_1_5;

         default :
            return javax.comm.SerialPort.STOPBITS_2;
      }
   }

   /**
    * Returns the port's current parity setting.
    *
    * @return the current parity
    */
   public int getParity ()
   {
      int retVal = 0;

      retVal = SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_GETBITS,
                                            (new byte []{
                                               ( byte ) (portNum & 0x0ff) }));

      if ((retVal & 0x08) == 0)
         return javax.comm.SerialPort.PARITY_NONE;

      switch ((retVal & 0x30) >>> 4)
      {

         case 0 :
            return javax.comm.SerialPort.PARITY_ODD;

         case 1 :
            return javax.comm.SerialPort.PARITY_EVEN;

         case 2 :
            return javax.comm.SerialPort.PARITY_MARK;

         case 3 :
            return javax.comm.SerialPort.PARITY_SPACE;
      }

      return javax.comm.SerialPort.PARITY_NONE;
   }

   /**
    * Sets the serial port's word size, stop bits and parity parameters.
    *
    * @param dataBits desired word size
    * @param stopBits desired stop bits
    * @param parity desired parity
    *
    * @throws IllegalStateException if the port is not open
     */
   public void setBitParameters (int dataBits, int stopBits, int parity) throws UnsupportedCommOperationException
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      int bitComposite = 0;

      switch (dataBits)
      {
         case javax.comm.SerialPort.DATABITS_5 :
            if(portNum < 2) throw new UnsupportedCommOperationException("invalid dataBits");
            break;

         case javax.comm.SerialPort.DATABITS_6 :
            if(portNum < 2) throw new UnsupportedCommOperationException("invalid dataBits");
            bitComposite |= 1;
            break;

         case javax.comm.SerialPort.DATABITS_7 :
            bitComposite |= 2;
            break;

         case javax.comm.SerialPort.DATABITS_8 :
            bitComposite |= 3;
            break;

         default :
            throw new UnsupportedCommOperationException("invalid dataBits");
      }

      switch (stopBits)
      {
         case javax.comm.SerialPort.STOPBITS_1 :
            break;

         case javax.comm.SerialPort.STOPBITS_1_5 :
            if(portNum < 2) throw new UnsupportedCommOperationException("invalid stopBits");
         case javax.comm.SerialPort.STOPBITS_2 :
            if (dataBits == javax.comm.SerialPort.DATABITS_8 && portNum < 2)
            {
               throw new UnsupportedCommOperationException("invalid stopBits");
            }

            bitComposite |= 4;
            break;

         default :
            throw new UnsupportedCommOperationException("invalid stopBits");
      }

      switch (parity)
      {

         case javax.comm.SerialPort.PARITY_NONE :
            if(dataBits == javax.comm.SerialPort.DATABITS_7 && stopBits == javax.comm.SerialPort.STOPBITS_1 && portNum < 2)
            {
               throw new UnsupportedCommOperationException("7,1,NONE not supported");
            }
            break;

         case javax.comm.SerialPort.PARITY_ODD :
            if(dataBits == javax.comm.SerialPort.DATABITS_7 && stopBits == javax.comm.SerialPort.STOPBITS_2 && portNum < 2)
            {
               throw new UnsupportedCommOperationException("7,2,ODD not supported");
            }
            bitComposite |= 0x08;
            break;

         case javax.comm.SerialPort.PARITY_EVEN :
            if(dataBits == javax.comm.SerialPort.DATABITS_7 && stopBits == javax.comm.SerialPort.STOPBITS_2 && portNum < 2)
            {
               throw new UnsupportedCommOperationException("7,2,EVEN not supported");
            }
            bitComposite |= 0x18;
            break;

         case javax.comm.SerialPort.PARITY_MARK :
            if(portNum < 2) throw new UnsupportedCommOperationException("invalid parity");
            bitComposite |= 0x28;
            break;

         case javax.comm.SerialPort.PARITY_SPACE :
            if(portNum < 2) throw new UnsupportedCommOperationException("invalid parity");
            bitComposite |= 0x38;
            break;

         default :
            throw new UnsupportedCommOperationException("invalid parity");
      }

      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETBITS,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                                 ( byte ) (bitComposite
                                                           & 0x0ff) }));
   }

   /**
    * Sends a break on the serial port.
    *
    * @param millis desired duration of the break in milliseconds
    *
    * @throws IllegalStateException if the port is not open
    */
   public void sendBreak (int millis)
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      sendBreakNative((byte)(portNum & 0xFF), millis);
   }

   /**
    * Native method
    *
    * @param timeMS
    *
    * @see
    */
   private static native void sendBreakNative (byte portNum, int timeMS);
   /**
    * Sets the port's flow control mode. Valid modes include the
    * <code>javax.comm.SerialPort</code> values <code>FLOWCONTROL_NONE</code>,
    * <code>FLOWCONTROL_RTSCTS_IN</code>, <code>FLOWCONTROL_RTSCTS_OUT</code>,
    * <code>FLOWCONTROL_XONXOFF_IN</code> and <code>FLOWCONTROL_XONXOFF_OUT</code>
    *
    * @throws IllegalStateException if the port is not open
    */
   public void setFlowControlMode (int flowC)
      throws UnsupportedCommOperationException
   {

      /*
     FLOWCONTROL_NONE        = 0;
     FLOWCONTROL_RTSCTS_IN   = 1;
     FLOWCONTROL_RTSCTS_OUT  = 2;
     FLOWCONTROL_XONXOFF_IN  = 4;
     FLOWCONTROL_XONXOFF_OUT = 8;
      */
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (portNum > 1)
      {
         if (flowC > 4)
            throw new UnsupportedCommOperationException(
               "XONXOFF not supported");

         this.flowControl = flowC;
         setFlowControlNative(portNum, flowC);
         return;
      }
      else
      {
         this.flowControl = flowC;
         setFlowControlNative(portNum, flowC);
      }
   }

   /**
    * Returns the port's current flow control mode.
    *
    * @return the current flow control mode
    * @see #setFlowControlMode(int flowc)
    */
   public int getFlowControlMode ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      int retVal = getFlowControlNative(portNum);
      return retVal;
   }

   /**
    * Sets serial port parameters including baud rate, data bits, stop bits
    *  and parity.
    *
    * @param baudRate the desired baud rate
    * @param dataBits the desired word size
    * @param stopBits the desired stop bits
    * @param parity the desired parity
    *
    * @throws UnsupportedCommOperationException never
    * @throws IllegalStateException if the port is not open
    */
   public void setSerialPortParams (int baudRate, int dataBits, int stopBits,
                                    int parity)
      throws UnsupportedCommOperationException
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (portNum < 2)
      {

         setInternalBaudRate(baudRate);
      }
      else
      {
         setBaudRate(baudRate);
      }

      setBitParameters(dataBits, stopBits, parity);
   }

   /** Sets the baud rate divisor register on the 16552 DUART. NOTE: New
    *  behavior for TINI 1.02d and above. For serial ports 0 and 1, this
    *  method allows you to set the reload value for the serial port.
    *  For ports 2 and 3, it functions as normal. <br> <br>
    *
    *  If you plan to set the internal serial port divisor, we advise you
    *  to first use "setSerialPortParams" to initialize the baud rate to 
    *  the first supported baud rate below what you intend to move to.
    *  (If you plan to change the baud rate to 8192, initialize the port
    *  to 4800 baud first). This will help set the interrupt timing correctly
    *  on older models of the 80c390.
    *
    * @param divisor the value to be loaded into the port's divisor register
    */
   public void setDivisor (int divisor)
   {
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETBAUDRATE,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                                 ( byte ) (divisor & 0x0ff),
                                                 ( byte ) ((divisor >>> 8)
                                                           & 0x0ff) }));
   }

   /**
    * Sets the baud rate for the port.
    *
    * @param speed use the static final ints
    *
    * @throws UnsupportedCommOperationException
    *
    * @see
    */
   protected void setInternalBaudRate (int speed)
      throws UnsupportedCommOperationException
   {
      int tiniBaudRate = SPEED_115200;
      //Debug.dump("Setting internal baud rate");

      switch (speed)
      {

         case 300 :
            tiniBaudRate = SPEED_300;
            break;

         case 600 :
            tiniBaudRate = SPEED_600;
            break;

         case 1200 :
            tiniBaudRate = SPEED_1200;
            break;

         case 2400 :
            tiniBaudRate = SPEED_2400;
            break;

         case 4800 :
            tiniBaudRate = SPEED_4800;
            break;

         case 9600 :
            tiniBaudRate = SPEED_9600;
            break;

         case 19200 :
            tiniBaudRate = SPEED_19200;
            break;

         case 38400 :
            tiniBaudRate = SPEED_38400;
            break;

         case 57600 :
            tiniBaudRate = SPEED_57600;
            break;

         case 115200 :
            tiniBaudRate = SPEED_115200;
            break;

         case 230400 :
            tiniBaudRate = SPEED_230400;
            break;

         default :
            throw new UnsupportedCommOperationException(
               "unsupported baud rate");
      }

      switch (portNum)
      {

         case NativeComm.PORT_SERIAL0 :
            if ((tiniBaudRate >= SPEED_300) && (tiniBaudRate <= SPEED_115200))
            {
               SetSerBaud(NativeComm.PORT_SERIAL0, tiniBaudRate - 1);

               internalBaudRate = tiniBaudRate;

               break;
            }
            else
            {
               throw new UnsupportedCommOperationException(
                  "Invalid baud rate");
            }
         case NativeComm.PORT_SERIAL1 :
            if ((tiniBaudRate >= SPEED_2400)
                    && (tiniBaudRate <= SPEED_115200))
            {

               //com.dalsemi.system.//Debug.debugDump("Baud = ");
               //com.dalsemi.system.//Debug.debugDump(tiniBaudRate);
               SetSerBaud(NativeComm.PORT_SERIAL1, tiniBaudRate - 1);

               internalBaudRate = tiniBaudRate;

               break;
            }
            else
            {
               throw new UnsupportedCommOperationException(
                  "Invalid baud rate");
            }
         default :
         {
//            //Debug.debugDump("INVALID PORT TYPE IN INTERNAL SETBAUDRATE");
//            //Debug.debugDump(PortType);

            throw new UnsupportedCommOperationException();
         }
      }
   }



   /**
    * Sets the port's baud rate assuming the default crystal frequency.
    *
    * @param desired baud rate in bits per second
    */
   public void setBaudRate (int baudRate)
   {
      if (portNum < 2)
      {
         try
         {
            setInternalBaudRate(baudRate);

            return;
         }
         catch (Exception e)
         {
            return;
         }
      }

      //Debug.dump("Setting Baud Rate");
      switch (baudRate)
      {

         case 300 :
            setDivisor(DIVISOR_300);
            break;

         case 600 :
            setDivisor(DIVISOR_600);
            break;

         case 1200 :
            setDivisor(DIVISOR_1200);
            break;

         case 2400 :
            setDivisor(DIVISOR_2400);
            break;

         case 4800 :
            setDivisor(DIVISOR_4800);
            break;

         case 9600 :
            setDivisor(DIVISOR_9600);
            break;

         case 19200 :
            setDivisor(DIVISOR_19200);
            break;

         case 38400 :
            setDivisor(DIVISOR_38400);
            break;

         case 57600 :
            setDivisor(DIVISOR_57600);
            break;

         case 115200 :
            setDivisor(DIVISOR_115200);
            break;
      }
   }

   /**
    * Sets the state of DTR.
    *
    * @param dtr set to true to enable DTR
    *
    * @throws IllegalStateException if the port is not open
    */
   public void setDTR (boolean dtr)
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETDTR,
                                   (new byte []{
                                   ( byte ) (portNum & 0x0ff),
                                   ( byte ) (dtr ? 1 : 0) }));
   }

   /**
    * Returns the current state of DTR at the exact time it was sampled by the
    * native driver.  Same as getDTR.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if DTR is currently asserted
    * @throws IllegalStateException if the port is not open
    * @see #getDTR()
    */
   public boolean isDTR ()
   {
      return getDTR();
   }

   /**
    * Returns the current state of DTR at the exact time it was sampled by the
    * native driver.
    *
    * @return true if DTR is currently asserted
    *
    * @throws IllegalStateException if the port is not open
    */
   public boolean getDTR ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (portNum > 1)
         return (SerExt_deviceIOControlNative(
            DEVICE_IO_DISPATCH_GETDTR, (new byte []{ ( byte ) (portNum
               & 0x0ff) })) != 0);

      return false;
   }

   /**
    * Sets the state of RTS.
    *
    * @param desired set to true to enable RTS
    *
    * @throws IllegalStateException if the port is not open
    */
   public void setRTS (boolean rts)
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETRTS,
                                   (new byte []{
                                   ( byte ) (portNum & 0x0ff),
                                   ( byte ) (rts ? 1 : 0) }));
   }

   /**
    * Returns the current state of RTS at the exact time it was sampled by the
    * native driver.  Same as getRTS.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if RTS is asserted
    * @throws IllegalStateException if the port is not open
    * @see #getRTS()
    */
   public boolean isRTS ()
   {
      return getRTS();
   }

   /**
    * Returns the current state of RTS at the exact time it was sampled by the
    * native driver.
    *
    * @return true if RTS is asserted
    *
    * @throws IllegalStateException if the port is not open
    */
   public boolean getRTS ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      return (SerExt_deviceIOControlNative(
         DEVICE_IO_DISPATCH_GETRTS, (new byte []{ ( byte ) (portNum
            & 0x0ff) })) != 0);

   }

   /**
    * Returns the current state of CTS at the exact time it was sampled by the
    * native driver.  Same as getCTS.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if CTS is asserted
    * @throws IllegalStateException if the port is not open
    * @see #getCTS()
    */
   public boolean isCTS ()
   {
      return getCTS();
   }

   /**
    * Returns the current state of CTS at the exact time it was sampled by the
    * native driver.
    *
    * @return true if CTS is asserted
    *
    * @throws IllegalStateException if the port is not open
    */
   public boolean getCTS ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      return (SerExt_deviceIOControlNative(
              DEVICE_IO_DISPATCH_GETCTS, (new byte []{ ( byte ) (portNum
              & 0x0ff) })) != 0);

   }

   /**
    * Returns the current state of DSR at the exact time it was sampled by the
    * native driver.  Same as getDSR.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if DSR is asserted
    * @throws IllegalStateException if the port is not open
    * @see #getDSR()
    */
   public boolean isDSR ()
   {
      return getDSR();
   }

   /**
    * Returns the current state of DSR at the exact time it was sampled by the
    * native driver.
    *
    * @return true if DSR is asserted
    *
    * @throws IllegalStateException if the port is not open
    */
   public boolean getDSR ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (portNum > 1)
         return (SerExt_deviceIOControlNative(
            DEVICE_IO_DISPATCH_GETDSR, (new byte []{ ( byte ) (portNum
               & 0x0ff) })) != 0);

      return false;
   }

   /**
    * Returns the current state of RI at the exact time it was sampled by the
    * native driver.  Same as getRI.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if RI is asserted
    * @throws IllegalStateException if the port is not open
    * @see #getRI()
     */
   public boolean isRI ()
   {
      return getRI();
   }

   /**
    * Returns the current state of RI at the exact time it was sampled by the
    * native driver.
    *
    * @return true if RI is asserted
    *
    * @throws IllegalStateException if the port is not open
    *
    */
   public boolean getRI ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (portNum > 1)
         return (SerExt_deviceIOControlNative(
            DEVICE_IO_DISPATCH_GETRI, (new byte []{ ( byte ) (portNum
               & 0x0ff) })) != 0);

      return false;
   }

   /**
    * Returns the current state of CD at the exact time it was sampled by the
    * native driver. Same as getCD.
    * Used by <code>javax.comm.SerialPort</code>.
    *
    * @return true if RI is asserted
    * @throws IllegalStateException if the port is not open
    * @see #getCD()
    */
   public boolean isCD ()
   {
      return getCD();
   }

   /**
    * Returns the current state of CD at the exact time it was sampled by the
    * native driver.
    *
    * @return true if CD is asserted
    *
    * @throws IllegalStateException if the port is not open
    */
   public boolean getCD ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      return (SerExt_deviceIOControlNative(
              DEVICE_IO_DISPATCH_GETCD, (new byte []{ ( byte ) (portNum
              & 0x0ff) })) != 0);

   }

   /**
    * Adds an event listener for the port. A new Thread will be created to
    * manage events as soon as one of the notifyOn methods are called.
    *
    * @param eventListener listener class to be installed
    *
    * @throws IllegalStateException if the port is not open
    * @throws TooManyListenersException if the maximum number of listeners has
    *         been installed
    */
   public void addEventListener (SerialPortEventListener eventListener)
      throws TooManyListenersException
   {
      boolean activate;

      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      if (this.eventListener != null)
         throw new TooManyListenersException();

      this.eventListener = eventListener;

      if (tiniNativeThread == null)
      {
          activate = requestedNotifyOnOutputEmpty |
                     requestedNotifyOnDataAvailable |
                     requestedNotifyOnCTS |
                     requestedNotifyOnDSR |
                     requestedNotifyOnRingIndicator |
                     requestedNotifyOnCarrierDetect |
                     requestedNotifyOnOverrunError |
                     requestedNotifyOnFramingError |
                     requestedNotifyOnBreakInterrupt;

          if (activate)
          {
                tiniNativeThread = new TININativeThread(this);
          }
       }

   }

   /**
    * Removes an event listener. This method removes all threads associated
    * with event listeners when the last listener is removed.
    */
   public synchronized void removeEventListener ()
   {
//      if (tiniEventThread != null)
//         tiniEventThread.stopEventListener();

      if (tiniNativeThread != null)
         tiniNativeThread.stopEventListener();

//      tiniEventThread  = null;
      tiniNativeThread = null;
   }



   /**
    * Requests notification when data is available to read.
    * This method will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnDataAvailable (boolean enable)
   {
      requestedNotifyOnDataAvailable = enable;
      addNotification(enable, (byte)NOTIFY_ON_DATA_AVAILABLE);
   }

   private void addNotification(boolean enable, byte eventType)
   {
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte[] {(byte)(portNum & 0xff),
                                   (byte)(enable ? 1:0),
                                   (byte)(eventType)}));
      if (enable && (this.eventListener != null)
                 && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }

   }
   /**
    * Requests notification when output buffer is empty. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnOutputEmpty (boolean enable)
   {
      requestedNotifyOnOutputEmpty = enable;
/*      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_OUTPUT_EMPTY)}));
*/
      addNotification(enable, (byte)NOTIFY_ON_OUTPUT_EMPTY);
/*      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
*/
   }

   /**
    * Requests notification when CTS changes state. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnCTS (boolean enable)
   {
      requestedNotifyOnCTS = enable;
/*
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_CTS)}));

      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
*/
      addNotification(enable, (byte)NOTIFY_ON_CTS);

   }

   /**
    * Requests notification when DSR changes state. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnDSR (boolean enable)
   {
      requestedNotifyOnDSR = enable;
      /*SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_DSR)}));

      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
      */
      addNotification(enable, (byte)NOTIFY_ON_DSR);
   }

   /**
    * Requests notification when ring indicator changes state. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnRingIndicator (boolean enable)
   {
      requestedNotifyOnRingIndicator = enable;
      /*
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_RI)}));

      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
      */
      addNotification(enable, (byte)NOTIFY_ON_RI);
   }

   /**
    * Requests notification when carrier detect changes state. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnCarrierDetect (boolean enable)
   {
      requestedNotifyOnCarrierDetect = enable;
      /*
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_CARRIER_DETECT)}));

      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
      */
      addNotification(enable, (byte)NOTIFY_ON_CARRIER_DETECT);
   }

   /**
    * Requests notification when overrun error occurs. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnOverrunError (boolean enable)
   {
      requestedNotifyOnOverrunError = enable;
      /*
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_OVERRUN_ERROR)}));

      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
      */
      addNotification(enable, (byte)NOTIFY_ON_OVERRUN_ERROR);
   }

   /**
    * Requests notification when parity error occurs. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnParityError (boolean enable)
   {
      requestedNotifyOnParityError = enable;

      /*
      SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_ENABLENOTIFY,
                                   (new byte []{ ( byte ) (portNum & 0x0ff),
                                   ( byte ) (enable ? 1:0),
                                   ( byte ) (NOTIFY_ON_PARITY_ERROR)}));
      if (enable && (this.eventListener != null)
              && (tiniNativeThread == null))
      {
         tiniNativeThread = new TININativeThread(this);
      }
      */
      addNotification(enable, (byte)NOTIFY_ON_PARITY_ERROR);
   }

   /**
    * Requests notification when framing error occurs. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnFramingError (boolean enable)
   {
      requestedNotifyOnFramingError = enable;
      addNotification(enable, (byte)NOTIFY_ON_FRAMING_ERROR);
   }

   /**
    * Requests notification when break interrupt occurs. This method
    * will create a new Thread to manage events if one has not already been
    * created.
    *
    * @param enable set to true to enable
    */
   public void notifyOnBreakInterrupt (boolean enable)
   {
      requestedNotifyOnBreakInterrupt = enable;
      addNotification(enable, (byte)NOTIFY_ON_BREAK_INTERRUPT);
   }

   /** Returns the InputStream associated with this port.
    *
    * @return the InputStream associated with this port
    *
    * @throws IOException if the stream is not available
    * @throws IllegalStateException if the port is not open
    */
   public InputStream getInputStream ()
      throws IOException
   {
      if (portState == PORT_STATE_CLOSED)
      {
         throw new IllegalStateException("serial port is closed");
      }
      if (inputStream == null)
      {
         inputStream = new SerialInputStream(this);
      }

      return inputStream;
   }

   /** Returns the OutputStream associated with this port.
    *
    * @return the OutputStream associated with this port
    *
    * @throws IOException if the stream is not available
    * @throws IllegalStateException if the port is not open
    */
   public OutputStream getOutputStream ()
      throws IOException
   {
      if (outputStream == null)
      {
         outputStream = new SerialOutputStream(this);
      }

      return outputStream;
   }

   /**
    * Enables the port's receive threshold.
    *
    * @param thresh threshold in milliseconds
    *
    * @throws UnsupportedCommOperationException never
    * @throws IllegalStateException if the InputStream is not accessable
    */
   public void enableReceiveThreshold (int thresh)
      throws UnsupportedCommOperationException
   {
      try
      {
         if (inputStream == null)
            inputStream = getInputStream();
      }
      catch (IOException e)
      {
         throw new IllegalStateException(
            "Unable to get inputstream");
      }

      threshold = thresh;

      (( SerialInputStream ) inputStream).enableReceiveThreshold(thresh);
   }

   /**
    * Disables the port's receive threshold.
    */
   public void disableReceiveThreshold ()
   {
      try
      {
         if (inputStream == null)
            inputStream = getInputStream();
      }
      catch (IOException e)
      {
         return;
      }

      threshold = 1;

      (( SerialInputStream ) inputStream).disableReceiveThreshold();
   }

   /**
    * Returns the port's threshold enable state. Used by
    *  <code>javax.comm.SerialPort</code>.
    *
    * @return true if receive threshold is enabled
    */
   public boolean isReceiveThresholdEnabled ()
   {
      if (inputStream == null)
         return false;

      return (( SerialInputStream ) inputStream).isReceiveThresholdEnabled();
   }

   /**
    * Returns the port's threshold enable state.
    *
    * @return true if receive threshold is enabled
    */
   public int getReceiveThreshold ()
   {
      return threshold;
   }

   /**
    * Enables the port's receive timeout.
    *
    * @param rcvTimeout timeout in milliseconds
    *
    * @throws UnsupportedCommOperationException never
    * @throws IllegalStateException if the port is closed
    */
   public void enableReceiveTimeout (int rcvTimeout)
      throws UnsupportedCommOperationException
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      try 
      {
      if (inputStream == null)
         inputStream = getInputStream();
      }
      catch (Exception E)
      {
         return;
      }

      ((SerialInputStream)inputStream).enableReceiveTimeout(rcvTimeout);

      receiveTimeout = rcvTimeout;
   }

   /**
    * Disables the port's receive timeout.
    *
    * @throws IllegalStateException if the port is closed
    */
   public void disableReceiveTimeout ()
   {
      if (portState == PORT_STATE_CLOSED)
         throw new IllegalStateException("serial port is closed");

      receiveTimeout = DEFAULT_SERIAL_READ_TIMEOUT;

      if (inputStream == null) return;

      ((SerialInputStream)inputStream).disableReceiveTimeout();
   }

   /**
    * Returns the port's receive timeout enable state.
    *
    * @return true if receive timeout is enabled
     */
   public boolean isReceiveTimeoutEnabled ()
   {
      if (inputStream == null) return false;
      else return ((SerialInputStream)inputStream).isReceiveThresholdEnabled();
   }

   /**
    * Returns the port's current receive timeout.
    *
    * @return the current timeout value
    */
   public int getReceiveTimeout ()
   {
      return receiveTimeout;
   }

   /**
    * Enables the port's receive framing. This method is not implemented on
    * TINI.
    *
    * @throws UnsupportedCommOperationException always
    */
   public void enableReceiveFraming (int framingByte)
      throws UnsupportedCommOperationException
   {
      throw new UnsupportedCommOperationException();
   }

   /**
    * Disables the port's receive framing.
    */
   public void disableReceiveFraming ()
   {
   }

   /**
    * Returns the port's receive framing enable state.
    *
    *  @return false always
    */
   public boolean isReceiveFramingEnabled ()
   {
      return false;
   }

   /**
    * Returns the port's receive framing byte.
    *
    * @return -1 always
    */
   public int getReceiveFramingByte ()
   {
      return -1;
   }

   /**
    * Sets the port's input buffer size.
    *
    * @param size number of bytes to use for the input receive buffer
    */
   public void setInputBufferSize (int size)
   {
      int retVal = 0;

      retVal = SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_SETBUFFERSIZE,
                                            (new byte []{
                                            (byte) (portNum & 0xFF),
                                            (byte)(size&0xFF),
                                            (byte)((size >>> 8)&0xFF)}));
   }

   /**
    * Returns the port's input buffer size.
    *
    * @return the current input buffer size
    */
   public int getInputBufferSize ()
   {
      int retVal = 0;

      retVal = SerExt_deviceIOControlNative(DEVICE_IO_DISPATCH_GETBUFFERSIZE,
                                            (new byte []{
                                            (byte) (portNum & 0xFF) }));
      return retVal;
   }
   /**
    * Sets the port's output buffer size.
    *
    * @param size number of bytes to use for the output transmit buffer
    */
   public void setOutputBufferSize (int size)
   {
    //No TINI serial port supports setting the output buffer size.  Leave it at the default value.

//      try
//      {
//         serialPortOut.setOutputBufferSize(size);
//         outputBufferSize = size;
//      }
//      catch (UnsupportedCommOperationException u)
//      {
//         outputBufferSize = -1;
//  Just going to let the output buffer size revert back to what it was before this method was called.
//      }
   }

   /**
    * Returns the port's current output buffer size.
    *
    * @return the current output buffer size
    */
   public int getOutputBufferSize ()
   {
      return outputBufferSize;
   }

   /* --- package protected section ---*/

   final int                PORT_STATE_CLOSED           = 0;
   final int                PORT_STATE_OPEN             = 1;
   int                      portNum;
   boolean                  isOpen = false;
   int                      portState;
   SerialPortEventListener  eventListener;
//   TINIEventThread          tiniEventThread;
   TININativeThread         tiniNativeThread;
   InputStream              inputStream;
   OutputStream             outputStream;


   static String PortNames[];
   static final int DEFAULT_SERIAL_READ_TIMEOUT = 0x3FFFFFFF;

   static final int NOTIFY_ON_CARRIER_DETECT         = 1;
   static final int NOTIFY_ON_CTS                    = 2;
   static final int NOTIFY_ON_DSR                    = 4;
   static final int NOTIFY_ON_PARITY_ERROR           = 8;
   static final int NOTIFY_ON_RI                     = 16;
   static final int NOTIFY_ON_OVERRUN_ERROR          = 32;
   static final int NOTIFY_ON_OUTPUT_EMPTY           = 64;

   static final int NOTIFY_ON_FRAMING_ERROR          = 128;
   static final int NOTIFY_ON_BREAK_INTERRUPT        = 128;
   static final int NOTIFY_ON_DATA_AVAILABLE         = 128;

   static final int SPEED_300 = 1;
   static final int SPEED_600 = 2;
   static final int SPEED_1200 = 3;
   static final int SPEED_2400 = 4;
   static final int SPEED_4800 = 5;
   static final int SPEED_9600 = 6;
   static final int SPEED_19200 = 7;
   static final int SPEED_38400 = 8;
   static final int SPEED_57600 = 9;
   static final int SPEED_115200 = 10;

   private static final int SPEED_230400 = 11;



   int     inputBufferSize  = 1;
   int     outputBufferSize = 1024;
   int     internalBaudRate = SPEED_115200;
   boolean requestedNotifyOnDataAvailable;
   boolean requestedNotifyOnOutputEmpty;
   boolean requestedNotifyOnCTS;
   boolean requestedNotifyOnDSR;
   boolean requestedNotifyOnRingIndicator;
   boolean requestedNotifyOnCarrierDetect;
   boolean requestedNotifyOnOverrunError;
   boolean requestedNotifyOnParityError;
   boolean requestedNotifyOnFramingError;
   boolean requestedNotifyOnBreakInterrupt;

   /**
    * Returns the array of possible port names.  This was created
    * in order to avoid creating these strings during API class
    * initialization.  All functions should call this instad of
    * grabbing portNamesB[] directly.
    */
   static String[] getPortNames ()
   {
      if (portNamesB == null)
      {
         portNamesB = new String[4];
         portNamesB[0] = "serial0";
         portNamesB[1] = "serial1";
         portNamesB[2] = "serial2";
         portNamesB[3] = "serial3";
      }

      return portNamesB;
   }
   /**
     */
   static boolean isOwned (String portName)
   {
      String[] localPortNames = getPortNames();

      for (int i = 0; i < localPortNames.length; i++)
         if (portName.equals(localPortNames [i]))
            return (getOwned(localPortNames [i].charAt(localPortNames [i].length() - 1)
                             - '0') != 0);

      throw new IllegalArgumentException("port " + portName
                                         + " is not a valid port name");
   }




   /**
     */
   static native void setOwned (int portNum, boolean isOwned);

   /**
     */
   static native int getOwned (int portNum);

   /* --- private section --- */

   private int              threshold                   = 1;
   private int              receiveTimeout              =
      DEFAULT_SERIAL_READ_TIMEOUT;
   private int              flowControl                 =
      SerialPort.FLOWCONTROL_NONE;
   private static String[]  portNamesB;
   private static final int DEVICE_IO_DISPATCH_GETBAUDRATE    = 0;
   private static final int DEVICE_IO_DISPATCH_SETBAUDRATE    = 1;
   private static final int DEVICE_IO_DISPATCH_GETADDRESS     = 2;
   private static final int DEVICE_IO_DISPATCH_SETADDRESS     = 3;
   private static final int DEVICE_IO_DISPATCH_GETBUFFERSIZE  = 4;
   private static final int DEVICE_IO_DISPATCH_SETBUFFERSIZE  = 5;
   private static final int DEVICE_IO_DISPATCH_GETDTR         = 6;
   private static final int DEVICE_IO_DISPATCH_SETDTR         = 7;
   private static final int DEVICE_IO_DISPATCH_GETRTS         = 8;
   private static final int DEVICE_IO_DISPATCH_SETRTS         = 9;
   private static final int DEVICE_IO_DISPATCH_GETCTS         = 10;
   private static final int DEVICE_IO_DISPATCH_GETDSR         = 11;
   private static final int DEVICE_IO_DISPATCH_GETRI          = 12;
   private static final int DEVICE_IO_DISPATCH_GETCD          = 13;
   private static final int DEVICE_IO_DISPATCH_GETENABLED     = 14;
   private static final int DEVICE_IO_DISPATCH_SETENABLED     = 15;
   private static final int DEVICE_IO_DISPATCH_GETBITS        = 16;
   private static final int DEVICE_IO_DISPATCH_SETBITS        = 17;
   private static final int DEVICE_IO_DISPATCH_ENABLENOTIFY   = 18;
   private static final int DEVICE_IO_DISPATCH_SETFIFOTRIGGER = 19;

   private static final int DIVISOR_230400                   = 1;
   private static final int DIVISOR_115200                   = DIVISOR_230400
                                                                  << 1;
   private static final int DIVISOR_57600                    = DIVISOR_115200
                                                                  << 1;
   private static final int DIVISOR_38400                    = DIVISOR_115200
                                                                  * 3;
   private static final int DIVISOR_19200                    = DIVISOR_38400
                                                                  << 1;
   private static final int DIVISOR_9600                     = DIVISOR_19200
                                                                  << 1;
   private static final int DIVISOR_4800                     = DIVISOR_9600
                                                                  << 1;
   private static final int DIVISOR_2400                     = DIVISOR_4800
                                                                  << 1;
   private static final int DIVISOR_1200                     = DIVISOR_2400
                                                                  << 1;
   private static final int DIVISOR_600                      = DIVISOR_1200
                                                                  << 1;
   private static final int DIVISOR_300                      = DIVISOR_600
                                                                  << 1;
   /**
     */
   private TINISerialPort ()
   {

   }

   /**
     */
   private void open ()
   {
      try
      {
         //Debug.dump("Opening Serial Port");
         if (portNum > 1)
         {
            if (!com.dalsemi.system.TINIOS.getExternalSerialPortEnable(
                    portNum))
               throw new IllegalStateException("serial port is disabled");
         }

         if (isOwned(this.name))
         {
            throw new IllegalStateException("Port Owned by Unknown Application");
         }
         setOwned(portNum, true);

         portState = PORT_STATE_OPEN;

         inputStream = getInputStream();
         ((SerialInputStream)inputStream).open();
         outputStream = getOutputStream();
         ((SerialOutputStream)outputStream).open();
         
         ((SerialInputStream)inputStream).enableReceiveThreshold(1);
      }
      catch (IOException i){}
      catch (UnsupportedCommOperationException E){};
   }

   private static native void SetSerBaud (int portNum, int baudRateSpecifier);

   private static native int SerExt_deviceIOControlNative (int funcNum,
           byte[] paramBuff);
   private static native void setFlowControlNative (int portnum, int mode);

   private static native int getFlowControlNative (int portnum);
}

/**
  */
  /*
class TINIEventThread
   extends Thread
{
   private boolean         eventListenerEnabled;
   private SerialPortEvent serialPortEvent;
   private TINISerialPort  tiniSerialPort;

   private TINIEventThread ()
   {
   }

   public TINIEventThread (TINISerialPort tiniSerialPort)
   {
      this.tiniSerialPort  = tiniSerialPort;
      eventListenerEnabled = true;
      setDaemon(true);
      start();
   }

   void stopEventListener ()
   {
      eventListenerEnabled = false;

      this.stop();
   }

   public void run ()
   {
      byte[] readAByte      = new byte [1];
      byte[] returnArray    = null;
      int    bytesAvailable = 0;

      try
      {
         SerialInputStream in =
            ( SerialInputStream ) tiniSerialPort.getInputStream();

         while (eventListenerEnabled)
         {
            if (in.read(readAByte, 0, 1) > 0)
            {
               in.unread(readAByte [0]);

               serialPortEvent           = new SerialPortEvent(
                  ( javax.comm.SerialPort ) tiniSerialPort,
                  SerialPortEvent.DATA_AVAILABLE, false, false);
               serialPortEvent.eventType = SerialPortEvent.DATA_AVAILABLE;

               tiniSerialPort.eventListener.serialEvent(serialPortEvent);
            }
         }
      }
      catch (IOException i){}
   }
}
*/
class TININativeThread
   extends Thread
{
   private boolean         eventListenerEnabled;
   private SerialPortEvent serialPortEvent;
   private TINISerialPort  tiniSerialPort;

   private static native int waitOnSerialNative (int portNum);

   /**
     */
   private TININativeThread ()
   {
   }

   /**
     */
   public TININativeThread (TINISerialPort tiniSerialPort)
   {
      this.tiniSerialPort  = tiniSerialPort;
      eventListenerEnabled = true;
      setDaemon(true);
      start();
   }


   private static native void stopEventThread(int port);

   /**
     */
   void stopEventListener ()
   {
      eventListenerEnabled = false;
      stopEventThread(tiniSerialPort.portNum);
//      this.interrupt();
   }

   /**
     */
   public void run ()
   {
      while (eventListenerEnabled)
      {
         try
         {

            int reason = waitOnSerialNative(tiniSerialPort.portNum);

            if (eventListenerEnabled == false)
               return;
            switch (reason & 0xFF)
            {
              case SerialPortEvent.DATA_AVAILABLE:
                  if (tiniSerialPort.requestedNotifyOnDataAvailable)
                     fireEvent(SerialPortEvent.DATA_AVAILABLE, 0);
              break;

              case SerialPortEvent.PE:
                  if (tiniSerialPort.requestedNotifyOnParityError)
                  {
                    /*serialPortEvent           = new SerialPortEvent(
                      ( javax.comm.SerialPort ) tiniSerialPort, SerialPortEvent.PE, false,
                      false);
                    tiniSerialPort.eventListener.serialEvent(serialPortEvent);
                    */
                    fireEvent(SerialPortEvent.PE, 0);
                  }
              break;

              case SerialPortEvent.CD:
                if (tiniSerialPort.requestedNotifyOnCarrierDetect)
                {
                  /*serialPortEvent           = new SerialPortEvent(
                    ( javax.comm.SerialPort ) tiniSerialPort, SerialPortEvent.CD,
                    ((reason & 0x0000FF00) != 0),
                    ((reason & 0x00ff0000) != 0));
                  tiniSerialPort.eventListener.serialEvent(serialPortEvent);*/
                  fireEvent(SerialPortEvent.CD, reason);
                }
              break;

              case SerialPortEvent.CTS:
                if (tiniSerialPort.requestedNotifyOnCTS)
                {
                  /*serialPortEvent           = new SerialPortEvent(
                    ( javax.comm.SerialPort ) tiniSerialPort, SerialPortEvent.CTS,
                    ((reason & 0x0000FF00) != 0),
                    ((reason & 0x00ff0000) != 0));
                  tiniSerialPort.eventListener.serialEvent(serialPortEvent);*/
                  fireEvent(SerialPortEvent.CTS, reason);
                }
              break;

              case SerialPortEvent.DSR:
                if (tiniSerialPort.requestedNotifyOnDSR)
                {
                 /* serialPortEvent           = new SerialPortEvent(
                    ( javax.comm.SerialPort ) tiniSerialPort, SerialPortEvent.DSR,
                    ((reason & 0x0000FF00) != 0),
                    ((reason & 0x00ff0000) != 0));
                  tiniSerialPort.eventListener.serialEvent(serialPortEvent);*/
                  fireEvent(SerialPortEvent.DSR, reason);
                }
              break;

              case SerialPortEvent.RI:
                if (tiniSerialPort.requestedNotifyOnRingIndicator)
                {
                  fireEvent(SerialPortEvent.RI, reason);
                }
              break;
              case SerialPortEvent.OE:
                if (tiniSerialPort.requestedNotifyOnOverrunError)
                {
                 /* serialPortEvent           = new SerialPortEvent(
                    ( javax.comm.SerialPort ) tiniSerialPort,
                    SerialPortEvent.OE,
                    ((reason & 0x0000FF00) != 0),
                    ((reason & 0x00ff0000) != 0));
                  tiniSerialPort.eventListener.serialEvent(serialPortEvent);*/
                  fireEvent(SerialPortEvent.OE, reason);
                }
              break;

              case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
                if (tiniSerialPort.requestedNotifyOnOutputEmpty)
                {
                  /*serialPortEvent           = new SerialPortEvent(
                    ( javax.comm.SerialPort ) tiniSerialPort,
                    SerialPortEvent.OUTPUT_BUFFER_EMPTY,
                    ((reason & 0x0000FF00) != 0),
                    ((reason & 0x00ff0000) != 0));
                  tiniSerialPort.eventListener.serialEvent(serialPortEvent);*/
                  fireEvent(SerialPortEvent.OUTPUT_BUFFER_EMPTY, reason);
                }
              break;
              case SerialPortEvent.FE:
                if (tiniSerialPort.requestedNotifyOnFramingError)
                   fireEvent(SerialPortEvent.FE, reason);
                break;

              case SerialPortEvent.BI:
                if (tiniSerialPort.requestedNotifyOnBreakInterrupt)
                   fireEvent(SerialPortEvent.BI, reason);
                break;
              default:
                //com.dalsemi.system.//Debug.debugDump("Not a standard event");
                //com.dalsemi.system.//Debug.debugDump(reason & 0xFF);
                break;

            }

         }
         catch (Exception e)
         {

            // Drain
         }
      }



   }
   public void fireEvent(int event, int reason)
   {
      serialPortEvent = new SerialPortEvent(
                            ( javax.comm.SerialPort ) tiniSerialPort,
                              event,
                              ((reason & 0x0000FF00) != 0),
                              ((reason & 0x00ff0000) != 0));
      tiniSerialPort.eventListener.serialEvent(serialPortEvent);
   }

}
