
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

/*     7-12-00  KLA Added BufferedOutputStream to listLong method to
 *              increase transmission efficiency.
 *
 *
 *
 */
package com.dalsemi.fs;



import java.io.*;
import com.dalsemi.system.TINIOS;
import com.dalsemi.shell.server.*;


/**
 * An extension to the standard <code>java.io.File</code> class.  Methods are provided for accessing the
 * permissions and the owner ID.  Instances of this class can also be used to execute the
 * file it represents, although the preferred way to execute a file is through the
 * methods provided in <code>java.lang.Runtime</code>.
 */
public class DSFile
   extends File
{
   private static final String CURR_DIR_KEY = "current directory";
   private static final String EMPTY_STRING = "";
   private byte[]              nameBytes;
   private static final int    EXEC_BADAPI = 1;
   private static final int    EXEC_BADCRC = 2;

   /**
    * Creates a new <code>DSFile</code> instance by converting the given
    * pathname string into an abstract pathname.  If the given string is
    * the empty string, then the result is the empty abstract pathname.
    *
    * @param       path                    the file pathname
    * @exception   NullPointerException    if the file path is equal to <code>null</code>
    */
   public DSFile (String path)
   {
      super(path);

      //LS -- FIXME -- setup the nameBytes, nameStart, nameLength variables;
      nameBytes = buildAbsolutePath(null, path);
   }


   /**
    * Creates a new <code>DSFile</code> instance from a parent pathname string
    * and a child pathname string.
    *
    * <p> If <code>path</code> is <code>null</code> then the new
    * <code>DSFile</code> instance is created as if by invoking the
    * single-argument <code>DSFile</code> constructor on the given
    * <code>name</code> pathname string.
    *
    * <p> Otherwise the <code>path</code> pathname string is taken to denote
    * a directory, and the <code>name</code> pathname string is taken to
    * denote either a directory or a file.  If the <code>name</code> pathname
    * string is absolute then it is converted into a relative pathname in a
    * system-dependent way.  If <code>path</code> is the empty string then
    * the new <code>DSFile</code> instance is created by converting
    * <code>name</code> into an abstract pathname and resolving the result
    * against the current working directory.  Otherwise each pathname
    * string is converted into an abstract pathname and the child abstract
    * pathname is resolved against the parent.
    *
    * @param       path                    the directory pathname
    * @param       name                    the file pathname
    * @exception   NullPointerException    if the name is equal to <code>null</code>
    */
   public DSFile (String path, String name)
   {

      //LS -- FIXME -- setup the nameBytes, nameStart, nameLength variables;
      super(path, name);

      nameBytes = buildAbsolutePath(path, name);
   }

   /**
    * Creates a new <code>DSFile</code> instance from a parent abstract
    * pathname and a child pathname string.
    *
    * <p> If <code>dir</code> is <code>null</code> then the new
    * <code>DSFile</code> instance is created as if by invoking the
    * single-argument <code>DSFile</code> constructor on the given
    * <code>name</code> pathname string.
    *
    * <p> Otherwise the <code>dir</code> abstract pathname is taken to
    * denote a directory, and the <code>name</code> pathname string is taken
    * to denote either a directory or a file.  If the <code>name</code>
    * pathname string is absolute then it is converted into a relative
    * pathname in a system-dependent way.  If <code>dir</code> is the empty
    * abstract pathname then the new <code>DSFile</code> instance is created by
    * converting <code>name</code> into an abstract pathname and resolving
    * the result against the current working directory.  Otherwise each
    * pathname string is converted into an abstract pathname and the child
    * abstract pathname is resolved against the parent.
    *
    * @param       dir                     the directory
    * @param       name                    the file pathname
    * @exception   NullPointerException    if the name is equal to <code>null</code>
    */
   public DSFile (File dir, String name)
   {

      //LS -- FIXME -- setup the nameBytes, nameStart, nameLength variables;
      super(dir, name);

      if (dir == null)
         nameBytes = DSFile.buildAbsolutePath(null, name);
      else
         nameBytes = DSFile.buildAbsolutePath(dir.getPath(), name);
   }

   /**
    * Sets the user (owner) permissions for this file.  The three lowest order
    * bits are used for the permissions, all others are ignored.
    * <p><code>
    * Bit 0 = Execute<p>
    * Bit 1 = Write<p>
    * Bit 2 = Read
    * </code><p>
    * Set the bit to grant that permission, clear it to deny.
    *
    * @param       perms           the new user/owner permissions
    * @exception   IOException     if the file does not exist or the
    *                              caller does not own the file and is
    *                              not an administrator
    * @see com.dalsemi.fs.DSFile#getUserPermissions()
    */
   public void setUserPermissions (int perms)
      throws IOException
   {
      DSFile.setUserPermissionsNative(nameBytes, perms,
                                      TINIOS.getCurrentUID());
   }

   /**
    * Retrieves the user (owner) permissions for this file.
    * The three lowest order bits of the return are used,
    * all others should be ignored.
    * <p><code>
    * Bit 0 = Execute<p>
    * Bit 1 = Write<p>
    * Bit 2 = Read
    * </code><p>
    * If the bit is set, that permission is granted, otherwise it
    * is denied.
    *
    * @return      the permissions
    * @exception   FileNotFoundException   if the file does not exist
    * @see com.dalsemi.fs.DSFile#setUserPermissions(int)
    */
   public int getUserPermissions ()
      throws FileNotFoundException
   {
      return DSFile.getUserPermissionsNative(nameBytes);
   }

   /**
    * Sets the other (non-owner) permissions for this file.  The three
    * lowest order bits are used for the permissions, all others are ignored.
    * <p><code>
    * Bit 0 = Execute<p>
    * Bit 1 = Write<p>
    * Bit 2 = Read
    * </code><p>
    * Set the bit to grant that permission, clear it to deny.
    *
    * @param       perms           the new other permissions
    * @exception   IOException     if the file does not exist or the
    *                              caller does not own the file and is
    *                              not an administrator
    * @see com.dalsemi.fs.DSFile#getOtherPermissions()
    */
   public void setOtherPermissions (int perms)
      throws IOException
   {
      DSFile.setOtherPermissionsNative(nameBytes, perms,
                                       TINIOS.getCurrentUID());
   }

   /**
    * Retrieves the other (non-owner) permissions for this file.
    * The three lowest order bits of the return are used,
    * all others should be ignored.
    * <p><code>
    * Bit 0 = Execute<p>
    * Bit 1 = Write<p>
    * Bit 2 = Read
    * </code><p>
    * If the bit is set, that permission is granted, otherwise it
    * is denied.
    *
    * @return      the permissions
    * @exception   FileNotFoundException   if the file does not exist
    * @see com.dalsemi.fs.DSFile#setOtherPermissions(int)
    */
   public int getOtherPermissions ()
      throws FileNotFoundException
   {
      return DSFile.getOtherPermissionsNative(nameBytes);
   }

   /**
    * Sets the user (owner) of this file.  Only owners of a file are allowed
    * to make changes to its permissions.
    *
    * @param       uid             the ID of the new user
    * @exception   IOException     if the file does not exist or the
    *                              caller does not own the file and is
    *                              not an administrator
    * @see com.dalsemi.fs.DSFile#getUser()
    * @see com.dalsemi.system.TINIOS#getCurrentUID()
    * @see com.dalsemi.system.TINIOS#getUserNameByUID(byte)
    * @see com.dalsemi.system.TINIOS#getUIDByUserName(String)
    */
   public void setUser (byte uid)
      throws IOException
   {
      DSFile.setUserNative(nameBytes, uid, TINIOS.getCurrentUID());
   }

   /**
    * Gets the ID of the user (owner) of this file.
    *
    * @return      uid                         the ID of the user
    * @exception   FileNotFoundException       if the file does not exist
    * @see com.dalsemi.fs.DSFile#setUser(byte)
    * @see com.dalsemi.system.TINIOS#getCurrentUID()
    * @see com.dalsemi.system.TINIOS#getUserNameByUID(byte)
    * @see com.dalsemi.system.TINIOS#getUIDByUserName(String)
    */
   public int getUser ()
      throws FileNotFoundException
   {
      return DSFile.getUserNative(nameBytes);
   }

   /**
    * Determines if this file can be executed.  Will return false if
    * the file does not exist or if the caller does not have execute
    * permission for the file.
    *
    * @return  <code>true</code> if the application can execute this file,
    *          <code>false</code> otherwise
    *
    * @see #executeFile(InputStream,OutputStream,OutputStream,String[],boolean,String)
    * @see #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String)
    * @see #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String,Process)
    * @see #executeFile()
    */
   public boolean canExec ()
   {
      return DSFile.canExecNative(nameBytes, TINIOS.getCurrentUID());
   }

   /**
    * Executes this file as a new foreground process.  Applications should not invoke
    * executeFile directly, but should instead use the methods of the <code>java.lang.Runtime</code>
    * class to start new processes.  <code>System.in</code>, <code>System.out</code>, and
    * <code>System.err</code> of the calling process are used for the new process's streams.
    * This call will block until the new process finishes.
    *
    * @return      the new process's ID (pid)
    * @exception   IOException     if the <code>canExec()</code> method returns false
    * @see         com.dalsemi.fs.DSFile#canExec()
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],boolean,String)
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String)
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String,Process)
    *
    */
   public int executeFile ()
      throws IOException
   {
      return executeFile(System.in, System.out, System.err, new String [0],
                         true, getName());
   }

   /**
    * Sets the modified time of this file to the current time.  If the file
    * does not exist, a file of length 0 will be created.
    *
    * @exception   if the caller cannot write the file
    */
   public void touch ()
      throws IOException
   {
      DSFile.touchNative(nameBytes, TINIOS.getCurrentUID());
   }

   /**
    * Prints a verbose file listing to the given stream.  The listing will include
    * the files name, size, owner and user permissions, owner name, and last
    * modified time.  This method can also print the listing to match the format
    * of the Un*x 'ls -l' command.
    *
    *
    * @param out  the output stream the listing will be sent to
    * @param unixStyle  whether the Un*x format should be used
    * @return  false if the user does not have permissions to read the
    *          directory or if this file does not exist, true otherwise
    */
   public boolean listLong (OutputStream out, boolean unixStyle)
      throws IOException
   {
      if (exists())
      {
         try
         {
            BufferedOutputStream bout   = new BufferedOutputStream(out);
            byte[]               buffer = new byte [54];

            if (isFile())
            {
               getListingLine(buffer, this, unixStyle);
               bout.write(buffer);
               bout.write(getName().getBytes());
               bout.write("\r\n".getBytes());
               bout.flush();

               return true;
            }
            else
            {
               String[] listing = list();
               String   pnt     = getParent();

               bout.write(("total " + (listing.length + ((pnt == null) ? 1
                                                                       : 2)) + "\r\n").getBytes());
               getListingLine(buffer, this, unixStyle);
               bout.write(buffer);
               bout.write(".\r\n".getBytes());

               if (pnt != null)
               {
                  getListingLine(buffer, new DSFile(getParent()), unixStyle);
                  bout.write(buffer);
                  bout.write("..\r\n".getBytes());
               }

               for (int i = 0; i < listing.length; i++)
               {
                  getListingLine(buffer, new DSFile(this, listing [i]),
                                 unixStyle);
                  bout.write(buffer);
                  bout.write((listing [i] + "\r\n").getBytes());
               }

               bout.flush();

               return true;
            }
         }
         catch (Exception e)
         {
            return false;
         }
      }

      return false;
   }

   private void getListingLine (byte[] ba, DSFile f, boolean unixStyle)
      throws Exception
   {
      ba [0] = f.isDirectory() ? ( byte ) 'd'
                               : ( byte ) '-';

      int perms = f.getUserPermissions();

      ba [1] = ((perms & 4) == 0) ? ( byte ) '-'
                                  : ( byte ) 'r';
      ba [2] = ((perms & 2) == 0) ? ( byte ) '-'
                                  : ( byte ) 'w';
      ba [3] = ((perms & 1) == 0) ? ( byte ) '-'
                                  : ( byte ) 'x';

      if (!unixStyle)
      {
         perms  = f.getOtherPermissions();
         ba [4] = ((perms & 4) == 0) ? ( byte ) '-'
                                     : ( byte ) 'r';
         ba [5] = ((perms & 2) == 0) ? ( byte ) '-'
                                     : ( byte ) 'w';
         ba [6] = ((perms & 1) == 0) ? ( byte ) '-'
                                     : ( byte ) 'x';
         ba [7] = ( byte ) ' ';
         ba [8] = ( byte ) ' ';
         ba [9] = ( byte ) ' ';
      }
      else
      {
         ba [4] = ( byte ) '-';
         ba [5] = ( byte ) '-';
         ba [6] = ( byte ) '-';
         perms  = f.getOtherPermissions();
         ba [7] = ((perms & 4) == 0) ? ( byte ) '-'
                                     : ( byte ) 'r';
         ba [8] = ((perms & 2) == 0) ? ( byte ) '-'
                                     : ( byte ) 'w';
         ba [9] = ((perms & 1) == 0) ? ( byte ) '-'
                                     : ( byte ) 'x';
      }

      //System.arraycopy(SPACES, 0, ba, 10, 3);
      com.dalsemi.system.ArrayUtils.arrayFill(ba, 10, 13, (byte)0x20);

      ba [13] = ( byte ) '1';
      ba [14] = ( byte ) ' ';

      int id = f.getUser() & 0x0FF;

      if (lastname == null)
      {
         lastname = new byte [0];
      }

      synchronized (lastname)
      {
         if (id != lastuid)
         {
            String userName = TINIOS.getUserNameByUID(( byte ) id);

            if (userName == null)
            {
               ba [15] = ( byte ) ('0' + (id / 100));
               ba [16] = ( byte ) ('0' + ((id % 100) / 10));
               ba [17] = ( byte ) ('0' + (id % 10));

               //System.arraycopy(SPACES, 0, ba, 18, 5);
               com.dalsemi.system.ArrayUtils.arrayFill(ba,18,23,(byte)0x20);
            }
            else
            {
               lastname = userName.getBytes();

               int nameLen = lastname.length;

               System.arraycopy(lastname, 0, ba, 15, Math.min(nameLen, 8));

               if (nameLen < 8)
                  //System.arraycopy(SPACES, 0, ba, 15 + nameLen, 8 - nameLen);
                  com.dalsemi.system.ArrayUtils.arrayFill(ba,15 + nameLen, 23,(byte)0x20);
            }

            lastuid = id;
         }
         else
         {
            int nameLen = lastname.length;

            System.arraycopy(lastname, 0, ba, 15, Math.min(nameLen, 8));

            if (nameLen < 8)
               //System.arraycopy(SPACES, 0, ba, 15 + nameLen, 8 - nameLen);
                  com.dalsemi.system.ArrayUtils.arrayFill(ba,15 + nameLen, 23,(byte)0x20);
         }
      }

      ba [23] = ( byte ) ' ';

      /*if ((lastuid & 0x080) == 0)
      {
         //System.arraycopy(USERS, 0, ba, 24, 9);
         System.arraycopy("user     ".getBytes(), 0, ba, 24, 9);
      }
      else
      {
         //System.arraycopy(ADMINS, 0, ba, 24, 9);
         System.arraycopy("admin    ".getBytes(), 0, ba, 24, 9);
      }*/

      System.arraycopy( (((lastuid & 0x080) == 0) ? "user     " : "admin    ").getBytes(), 0,ba,24,9);

      intToAsciiBytes(( int ) f.length(), ba, 33);

      ba [40] = ( byte ) ' ';

      long   time = DSFile.origLastModifiedNative(f.nameBytes);
      byte[] regs = new byte [8];

      for (int i = 0; i < 8; i++)
      {
         regs [i] = ( byte ) time;
         time     = time >> 8;
      }

      try
      {
         int month = com.dalsemi.system.Clock.bcdToInt(regs [6]);

         if (month < 1)
            month = 1;

         //System.arraycopy(MONTH [month - 1], 0, ba, 41, 3);
         System.arraycopy("JanFebMarAprMayJunJulAugSepOctNovDec".getBytes(), 3*(month-1), ba, 41, 3);


         ba [44] = ( byte ) ' ';
         ba [45] = ( byte ) ((regs [5] >> 4) + 48);
         ba [46] = ( byte ) ((regs [5] & 0x0F) + 48);
         ba [47] = ( byte ) ' ';

         int hour = regs [3] & 0x0FF;   // Make sure we mask, so we don't go negative.

         if (hour > 127)   // are we in 12 hour mode???
         {                 // yes !!!
            boolean pm = ((hour & 0x20) == 0x20);

            hour &= 0x1f;
            hour = com.dalsemi.system.Clock.bcdToInt(hour);

            if ((hour < 1) || (hour > 12))
            {
               hour = 1;
            }

            hour %= 12;

            if (pm)
            {
               hour += 12;
            }
         }
         else              // no !!!
         {
            hour &= 0x3f;
            hour = com.dalsemi.system.Clock.bcdToInt(hour);

            if ((hour < 0) || (hour > 23))
               hour = 0;
         }

         ba [48] = ( byte ) ((hour / 10) + 48);
         ba [49] = ( byte ) ((hour % 10) + 48);
         ba [50] = ( byte ) ':';
         ba [51] = ( byte ) ((regs [2] >> 4) + 48);
         ba [52] = ( byte ) ((regs [2] & 0x0F) + 48);
      }
      catch (ArrayIndexOutOfBoundsException x)
      {
         /*ba [41] = ( byte ) 'M';
         ba [42] = ( byte ) Character.forDigit(((regs [6] & 0x0FF) >> 4), 16);
         ba [43] = ( byte ) Character.forDigit((regs [6] & 0x0F), 16);
         ba [44] = ( byte ) 'D';
         ba [45] = ( byte ) Character.forDigit(((regs [5] & 0x0FF) >> 4), 16);
         ba [46] = ( byte ) Character.forDigit((regs [5] & 0x0F), 16);
         ba [47] = ( byte ) 'h';
         ba [48] = ( byte ) Character.forDigit(((regs [3] & 0x0FF) >> 4), 16);
         ba [49] = ( byte ) Character.forDigit((regs [3] & 0x0F), 16);
         ba [50] = ( byte ) 'm';
         ba [51] = ( byte ) Character.forDigit(((regs [2] & 0x0FF) >> 4), 16);
         ba [52] = ( byte ) Character.forDigit((regs [2] & 0x0F), 16);*/

         int k=41;
         for (int i=0;i<4;i++)
         {
            ba[k++] = (byte) (0x6d68444d >> (8 * i));
            ba[k++] = (byte) Character.forDigit(((regs [6 - i] & 0x0FF) >> 4), 16);
            ba[k++] = (byte) Character.forDigit((regs [6 - i] & 0x0F), 16);
         }


/*         com.dalsemi.system.Debug.debugDump("EXCEPTION BUILDING DATE STRING: "
                                            + x);
         com.dalsemi.system.Debug.debugDump(
            "LAST MODIFIED TIME = "
            + DSFile.origLastModifiedNative(f.nameBytes));
*/
      }

      ba [53] = ( byte ) ' ';
   }

   byte[]                lastname;
   int                   lastuid = -1;
   //static final byte[]   SPACES  =
   //{
   //   32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32
   //};
   //static final byte[]   ADMINS  = "admin    ".getBytes();
   /*{
      ( byte ) 'a', ( byte ) 'd', ( byte ) 'm', ( byte ) 'i', ( byte ) 'n',
      32, 32, 32, 32
   };*/
   //static final byte[]   USERS   = "user     ".getBytes();
   /*{
      ( byte ) 'u', ( byte ) 's', ( byte ) 'e', ( byte ) 'r', 32, 32, 32, 32,
      32
   };*/
   /*static final byte[][] MONTH   =
   {
      { ( byte ) 'J', ( byte ) 'a', ( byte ) 'n' },
      { ( byte ) 'F', ( byte ) 'e', ( byte ) 'b' },
      { ( byte ) 'M', ( byte ) 'a', ( byte ) 'r' },
      { ( byte ) 'A', ( byte ) 'p', ( byte ) 'r' },
      { ( byte ) 'M', ( byte ) 'a', ( byte ) 'y' },
      { ( byte ) 'J', ( byte ) 'u', ( byte ) 'n' },
      { ( byte ) 'J', ( byte ) 'u', ( byte ) 'l' },
      { ( byte ) 'A', ( byte ) 'u', ( byte ) 'g' },
      { ( byte ) 'S', ( byte ) 'e', ( byte ) 'p' },
      { ( byte ) 'O', ( byte ) 'c', ( byte ) 't' },
      { ( byte ) 'N', ( byte ) 'o', ( byte ) 'v' },
      { ( byte ) 'D', ( byte ) 'e', ( byte ) 'c' }
   };*/

   //FIXME -- take this function out
   private void intToAsciiBytes (int value, byte[] data, int offset)
   {
      int     div        = 1000000;
      boolean firstFound = false;

      for (int i = 0; i < 7; i++)
      {
         int x = value / div;

         if (x == 0)
         {
            if (firstFound || i == 6)
            {
               data [offset + i] = ( byte ) '0';
            }
            else
            {
               data [offset + i] = ( byte ) ' ';
            }
         }
         else
         {
            firstFound        = true;
            data [offset + i] = ( byte ) (x + 48);
         }

         value = value % div;
         div   = div / 10;
      }
   }

   /**
    * Builds the fully qualified path name from the directory name and
    * the file name.  The current working directory is also used as needed.
    * The final result will have the leading and trailing separator
    * characters (/) removed.  The name parameter cannot be null. Both
    * parameters can, however, be the empty string. The resulting byte
    * array will have the following format.<p>
    * <p>
    * Byte 0-3 = The offset into the array where the name begins (LSB, xx, xx, MSB).<p>
    * Byte 4-7 = The length of the name stored in the array (LSB, xx, xx, MSB).<p>
    * Byte 8-x = The data holding the ASCII name of the file.  There may be extra
    *            bytes preceding or following the actual data.  Use the first eight
    *            bytes of the array to determine which bytes to use.
    * <p>
    *
    *
    * @param parent  the directory or path of the file
    * @param name  the file name
    *
    * @return byte array containing the fully qualified path name
    *
    */
   public static byte[] buildAbsolutePath (String parent, String name)
   {
      byte[] nameBuffer;
      int    bufferLength = 0;
      int    nameLength   = name.length();

      if (parent == null)                 // use current directory and file name
      {
         String currDir = null;

         if ((nameLength != 0 && name.charAt(0) == '/') || ((currDir =
                 TINIOS.getFromCurrentEnvironment(CURR_DIR_KEY)) == null))   // no current directory set, just use file name
         {
            bufferLength                  = nameLength;
            nameBuffer                    = new byte [bufferLength + 10];
            nameBuffer [8]                = ( byte ) '/';
            nameBuffer [bufferLength + 9] = ( byte ) '/';

            name.getBytes(0, bufferLength, nameBuffer, 9);

            bufferLength += 2;
         }
         else                             // must use current directory and file
         {
            bufferLength                                = currDir.length();
            nameBuffer                                  =
               new byte [bufferLength + nameLength + 11];
            nameBuffer [8]                              = ( byte ) '/';
            nameBuffer [bufferLength + 9]               = ( byte ) '/';
            nameBuffer [bufferLength + nameLength + 10] = ( byte ) '/';

            currDir.getBytes(0, bufferLength, nameBuffer, 9);
            name.getBytes(0, nameLength, nameBuffer, bufferLength + 10);

            bufferLength += (nameLength + 3);
         }
      }
      else if (parent.length() == 0)      // just use the file name
      {
         bufferLength                  = nameLength;
         nameBuffer                    = new byte [bufferLength + 10];
         nameBuffer [8]                = ( byte ) '/';
         nameBuffer [bufferLength + 9] = ( byte ) '/';

         name.getBytes(0, bufferLength, nameBuffer, 9);

         bufferLength += 2;
      }
      else if (parent.charAt(0) == '/')   // use parent and file name
      {
         bufferLength                                = parent.length();
         nameBuffer                                  =
            new byte [bufferLength + nameLength + 11];
         nameBuffer [8]                              = ( byte ) '/';
         nameBuffer [bufferLength + 9]               = ( byte ) '/';
         nameBuffer [bufferLength + nameLength + 10] = ( byte ) '/';

         parent.getBytes(0, bufferLength, nameBuffer, 9);
         name.getBytes(0, nameLength, nameBuffer, bufferLength + 10);

         bufferLength += (nameLength + 3);
      }
      else                                // use all three (current dir, parent, and file name)
      {
         String currDir = TINIOS.getFromCurrentEnvironment(CURR_DIR_KEY);

         if (currDir == null)
         {
            bufferLength                                = parent.length();
            nameBuffer                                  =
               new byte [bufferLength + nameLength + 11];
            nameBuffer [8]                              = ( byte ) '/';
            nameBuffer [bufferLength + 9]               = ( byte ) '/';
            nameBuffer [bufferLength + nameLength + 10] = ( byte ) '/';

            parent.getBytes(0, bufferLength, nameBuffer, 9);
            name.getBytes(0, nameLength, nameBuffer, bufferLength + 10);

            bufferLength += (nameLength + 3);
         }
         else
         {
            bufferLength = 8;

            int temp1 = currDir.length();
            int temp2 = parent.length();

            nameBuffer                  =
               new byte [temp1 + temp2 + nameLength + 12];
            nameBuffer [bufferLength++] = ( byte ) '/';

            currDir.getBytes(0, temp1, nameBuffer, bufferLength);

            bufferLength                += temp1;
            nameBuffer [bufferLength++] = ( byte ) '/';

            parent.getBytes(0, temp2, nameBuffer, bufferLength);

            bufferLength                += temp2;
            nameBuffer [bufferLength++] = ( byte ) '/';

            name.getBytes(0, nameLength, nameBuffer, bufferLength);

            bufferLength                += nameLength;
            nameBuffer [bufferLength++] = ( byte ) '/';
            bufferLength                -= 8;
         }
      }

      for (int i = 7 + bufferLength; i > 8; )
      {
         if (nameBuffer [i] == '/' && nameBuffer [--i] == '/')
         {
            if (bufferLength == 2)
            {
               bufferLength = 0;

               break;
            }
            else if (i != bufferLength + 6)
            {
               System.arraycopy(nameBuffer, i + 2, nameBuffer, i + 1,
                                bufferLength + 6 - i);
            }

            bufferLength--;
         }
         else
         {
            i--;
         }
      }

      boolean dotSlashFound = true;
      int     lastDotSlash  = 8;
      boolean breakToWhile  = false;
      int     nameEnd       = 8 + bufferLength;
      int     nameBegin     = 8;

      while (dotSlashFound)
      {
         dotSlashFound = false;

         int cond = nameEnd - 1;

         for (int i = lastDotSlash; i < cond; i++)
         {
            if (nameBuffer [i] == '.' && nameBuffer [i + 1] == '/')
            {
               dotSlashFound = true;

               int upDirCount = 0;
               int offset     = i;

               while (offset != lastDotSlash && nameBuffer [--offset] == '.')
               {
                  upDirCount++;
               }

               if (nameBuffer [offset] != '/' && nameBuffer [offset] != '.')
               {
                  lastDotSlash = i + 1;

                  break;
               }

               if (offset == nameBegin)
               {
                  nameBegin = i + 1;

                  if (nameBegin + 1 == nameEnd)
                  {
                     nameBegin++;
                  }

                  lastDotSlash = nameBegin;

                  break;
               }

               int parentStart = offset;

               for (int j = 0; j < upDirCount; j++)
               {
                  int k;

                  for (k = parentStart - 1; k >= nameBegin; k--)
                  {
                     if (nameBuffer [k] == '/')
                     {
                        parentStart = k;

                        break;
                     }
                  }

                  if (k <= nameBegin)
                  {
                     if (i + 2 == nameEnd)
                     {
                        nameBegin = i + 2;
                     }
                     else
                     {
                        nameBegin = i + 1;
                     }

                     lastDotSlash = nameBegin;
                     breakToWhile = true;

                     break;
                  }
               }

               if (breakToWhile)
               {
                  breakToWhile = false;

                  break;
               }

               int lengthRemaining = nameEnd - (i + 1);

               System.arraycopy(nameBuffer, i + 1, nameBuffer, parentStart,
                                lengthRemaining);

               nameEnd = parentStart + lengthRemaining;

               break;
            }
         }
      }

      if (nameEnd != nameBegin)
      {
         nameBegin++;
         nameEnd--;
      }

      intToByteArray(nameBuffer, 0, nameBegin);
      intToByteArray(nameBuffer, 4, nameEnd - nameBegin);

      return nameBuffer;
   }

   /*
       //This method should build the fully qualified path name from
       //the two parameters and the current working directory.  The
       //final result will have the leading and trailing separator
       //characters (/) removed.  The name parameter cannot be null.
       //Both parameters can be an empty string.
       public static byte[] buildAbsolutePath(String parent, String name)
       {
           //Use this character array to build the entire name.
           //Then copy the data into the 'nameBytes' byte array.
           //This saves us from creating a bunch of unneccesary Strings
           //and StringBuffers.
           String currDir = TINIOS.getFromCurrentEnvironment(CURR_DIR_KEY);
           if(currDir == null)
               currDir = EMPTY_STRING;
           int currDirLength = currDir.length();

           if(parent == null)
               parent = EMPTY_STRING;
           int parentLength = parent.length();

           int nameLength = name.length();

           byte[] nameBytes = new byte[currDirLength + parentLength + nameLength + 11];

           if(currDirLength != 0)
           {
               currDir.getBytes(0, currDirLength, nameBytes, 8);
               nameBytes[currDirLength++ + 8] = (byte)File.separatorChar;
           }

           currDirLength += 8;

           if(parentLength != 0)
           {
               parent.getBytes(0, parentLength, nameBytes, currDirLength);
               currDirLength += parentLength;
               if(!parent.endsWith(File.separator))
               {
                   nameBytes[currDirLength++] = (byte)File.separatorChar;
               }
           }

           name.getBytes(0, nameLength, nameBytes, currDirLength);
           nameLength += currDirLength;

           if(nameBytes[nameLength - 1] != File.separatorChar)
           {
               nameBytes[nameLength++] = (byte)File.separatorChar;
           }

           return normalizeName(nameBytes, nameLength);
       }

       private static byte[] normalizeName(byte[] name, int nameEnd)
       {
           int nameBegin = 8;
           for(int i = nameEnd - 1; i > nameBegin; i--)
           {
               if(name[i] == File.separatorChar &&
                  name[i - 1] == File.separatorChar)
               {
                   nameBegin = i + 1;
                   break;
               }
           }

           boolean dotSlashFound = true;

           int lastDotSlash = nameBegin;

           boolean breakToWhile = false;

           while(dotSlashFound)
           {
               dotSlashFound = false;
               for(int i = lastDotSlash; i < nameEnd - 1; i++)
               {
                   if(name[i] == '.' && name[i + 1] == File.separatorChar)
                   {
                       dotSlashFound = true;
                       int upDirCount = 0;
                       int offset = i;
                       while(offset != lastDotSlash && name[--offset] == '.')
                       {
                           upDirCount++;
                       }

                       if(name[offset] != File.separatorChar  && name[offset] != '.')
                       {
                           lastDotSlash = i + 1;
                           break;
                       }

                       if(offset == nameBegin)
                       {
                           nameBegin = i + 1;
                           lastDotSlash = nameBegin;
                           break;
                       }

                       int parentStart = offset;

                       for(int j = 0; j < upDirCount; j++)
                       {
                           int k;
                           for(k = parentStart - 1; k >= nameBegin; k--)
                           {
                               if(name[k] == File.separatorChar)
                               {
                                   parentStart = k;
                                   break;
                               }
                           }

                           if(k <= nameBegin)
                           {
                               nameBegin = i + 2;
                               lastDotSlash = nameBegin;
                               breakToWhile = true;
                               break;
                           }
                       }

                       if(breakToWhile)
                       {
                           breakToWhile = false;
                           break;
                       }

                       int lengthRemaining = nameEnd - (i + 1);
                       System.arraycopy(name, i + 1, name, parentStart, lengthRemaining);
                       nameEnd = parentStart + lengthRemaining;
                       break;
                   }
               }
           }

           if(nameEnd != nameBegin && name[nameBegin] == File.separatorChar)
               nameBegin++;

           if(nameEnd != nameBegin && name[nameEnd - 1] == File.separatorChar)
               nameEnd--;

   //FIXME -- only a temporary solution.
           if(nameBegin < 9)
           {
               System.arraycopy(name, nameBegin, name, nameBegin + 1, nameEnd - nameBegin);
               nameBegin++;
               nameEnd++;
           }
           name[nameBegin - 1] = (byte)'/';





           intToByteArray(name, 0, nameBegin);
           intToByteArray(name, 4, nameEnd - nameBegin);

           return name;
       }
   */
   private static void intToByteArray (byte[] b, int offset, int value)
   {
      b [offset]     = ( byte ) (value);
      b [offset + 1] = ( byte ) (value >> 8);
      b [offset + 2] = ( byte ) (value >> 16);
      b [offset + 3] = ( byte ) (value >> 24);
   }

   //Native Functions
   private static native int wos_getAPIVersion ();

   private static native void setUserPermissionsNative (byte[] nameBytes,
           int perms, byte uid);

   private static native int getUserPermissionsNative (byte[] nameBytes);

   private static native void setOtherPermissionsNative (byte[] nameBytes,
           int perms, byte uid);

   private static native int getOtherPermissionsNative (byte[] nameBytes);

   private static native void setUserNative (byte[] nameBytes, byte newUID,
                                             byte uid);

   private static native int getUserNative (byte[] nameBytes);

   private static native boolean canExecNative (byte[] nameBytes, byte uid);

   private static native void touchNative (byte[] nameBytes, byte uid);

   private static native long origLastModifiedNative (byte[] name);

   /*****
   *****/

   /**
    * Executes this file as a new process with the default environment.  Applications should not invoke
    * executeFile directly, but should instead use the methods of the <code>java.lang.Runtime</code>
    * class to start new processes.
    *
    * @param       stdin           the stream that will become the new
    *                              process's <code>System.in</code>
    * @param       stdout          the stream that will become the new
    *                              process's <code>System.out</code>
    * @param       stderr          the stream that will become the new
    *                              process's <code>System.err</code>
    * @param       args            the arguments that will be passed to the
    *                              applications <code>main(String[])</code> method
    * @param       foreground      whether this method should block until the new process finishes
    * @param       processName     the name of the new process
    *
    * @return      the new process's ID (pid)
    *
    * @exception   IOException     if the <code>canExec()</code> method returns false
    * @see         com.dalsemi.fs.DSFile#canExec()
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String)
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String,Process)
    * @see         #executeFile()
    *
    */
   public int executeFile (InputStream stdin, OutputStream stdout,
                           OutputStream stderr, String[] args,
                           boolean foreground, String processName)
      throws IOException
   {
      return executeFile(stdin, stdout, stderr, args, null, foreground,
                         processName, null);
   }

   /**
    * Executes this file.  Applications should not invoke
    * executeFile directly, but should instead use the methods of the <code>java.lang.Runtime</code>
    * class to start new processes.
    *
    * @param       stdin           the stream that will become the new
    *                              process's <code>System.in</code>
    * @param       stdout          the stream that will become the new
    *                              process's <code>System.out</code>
    * @param       stderr          the stream that will become the new
    *                              process's <code>System.err</code>
    * @param       args            the arguments that will be passed to the
    *                              applications <code>main(String[])</code> method
    * @param       env             the environment that will be added to the process's
    *                              System properties
    * @param       foreground      whether this method should block until the new process finishes
    * @param       processName     the name of the new process
    *
    * @return      the new process's ID (pid)
    *
    * @exception   IOException     if the <code>canExec()</code> method returns false
    * @see         com.dalsemi.fs.DSFile#canExec()
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],boolean,String)
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String,Process)
    * @see         #executeFile()
    *
    */
   public int executeFile (InputStream stdin, OutputStream stdout,
                           OutputStream stderr, String[] args, String[] env,
                           boolean foreground, String processName)
      throws IOException
   {
      return executeFile(stdin, stdout, stderr, args, env, foreground,
                         processName, null);
   }

   /**
    * Executes this file.  Applications should not invoke
    * executeFile directly, but should instead use the methods of the <code>java.lang.Runtime</code>
    * class to start new processes.
    *
    * @param       stdin           the stream that will become the new
    *                              process's <code>System.in</code>
    * @param       stdout          the stream that will become the new
    *                              process's <code>System.out</code>
    * @param       stderr          the stream that will become the new
    *                              process's <code>System.err</code>
    * @param       args            the arguments that will be passed to the
    *                              applications <code>main(String[])</code> method
    * @param       env             the environment that will be added to the process's
    *                              System properties
    * @param       foreground      whether this method should block until the new process finishes
    * @param       processName     the name of the new process
    * @param       procObj         the Process object to associate with the new process
    *
    * @return      the new process's ID (pid).
    *
    * @exception   IOException     if the <code>canExec()</code> method returns false
    * @see         com.dalsemi.fs.DSFile#canExec()
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],String[],boolean,String)
    * @see         #executeFile(InputStream,OutputStream,OutputStream,String[],boolean,String)
    * @see         #executeFile()
    */
   public int executeFile (InputStream stdin, OutputStream stdout,
                           OutputStream stderr, String[] args, String[] env,
                           boolean foreground, String processName,
                           Process procObj)
      throws IOException
   {

      if(stdin != null && !(stdin instanceof SystemInputStream))
      {
        stdin = new SystemInputStream(stdin, null);
      }
      if(stdout != null && !(stdout instanceof SystemPrintStream))
      {
        stdout = new SystemPrintStream(stdout);
      }
      if(stderr != null && !(stderr instanceof SystemPrintStream))
      {
        stderr = new SystemPrintStream(stderr);
      }

      //com.dalsemi.jellybeans.Debug.debugDump("TINIFS.execute()"); //DEBUG
      int retval;

      //We can only have one process pending execution at a time.
      //This means if we want to execute two new processes, we have to wait
      //for the first one to start before starting the second.
      while (!wos_processLockTestAndSet())
      {
/*         com.dalsemi.system.Debug.debugDump(
            "DEBUG: Waiting for mutex to clear executeFile()");
*/
         try
         {
            Thread.sleep(2000);
         }
         catch (InterruptedException ie)
         {

            //DRAIN
         }
      }

      try
      {

         // Create the new process, could throw an exception.
         retval = executeFileNative(nameBytes, processName.getBytes());

         if (retval > 0)
         {
            Object[] temparr = new Object [6];

            temparr [0] = (stdin instanceof SystemInputStream) ? stdin : new SystemInputStream(stdin, null);
            temparr [1] = (stdout instanceof SystemPrintStream) ? stdout : new SystemPrintStream(stdout);
            temparr [2] = (stderr instanceof SystemPrintStream) ? stderr : new SystemPrintStream(stderr);
            temparr [3] = System.getProperties();   //Will be deep copied in the new process.
            temparr [4] = TINIOS.getFromCurrentEnvironment(
               com.dalsemi.shell.server.Session.CURRENT_DIRECTORY);
            temparr [5] = env;   //Commandline "-Dkey=val" properties.  Will be deep copied in the new process.  Can be null.

            wos_setJavaProcessStateNative(retval, temparr);

            if (foreground)
               wos_setProcessRestartParentNative(retval, true);

            wos_setJavaProcessArgsNative(retval, args);
            wos_setTaskOwnerNative(retval, TINIOS.getCurrentUID());
            wos_setProcessObjectNative(retval, procObj);
            wos_wakeUpProcessNative(retval);

            //Now, wait for child process to signal that he is through with
            //the shared variables.
            while (!wos_hasProcessStartedNative(retval))
            {
/*               com.dalsemi.system.Debug.debugDump(
                  "DEBUG: Waiting for child process to start");
*/

               try
               {
                  Thread.sleep(2000);
               }
               catch (InterruptedException ie)
               {

                  //DRAIN
               }
            }

            // Release process creation lock.
            wos_processLockClear();

            // If foreground, block here.
            if (foreground)
            {
               /* This errorcode is only returned on foreground execution, and
                  may help in debugging problems with <clinit>.
               */
               int errorCode = wos_waitForProcessExitNative(retval);
//com.dalsemi.system.Debug.dump("errorCode = "+Integer.toHexString(errorCode));
               if (errorCode != 0)
               {
                 if (((errorCode >> 24 & 0xFF) == 0xA5) &&
                     ((errorCode >> 16 & 0xFF) == 0x5A))
                 {
                    if (((errorCode >> 8 & 0xFF) == 0x00) &&
                        ((errorCode & 0xFF) == 0x00))
                    throw new ExceptionInInitializerError();
                 else
                    throw new ExceptionInInitializerError("errorCode = "+(Integer.toHexString(errorCode)).toUpperCase());
                 }
               }
            }

            // Return the ID of the process created.
            return (retval);
         }
         else
         {

            // throw an exception with decoded value from native.
            switch (retval & 0xFF)
            {

               case EXEC_BADAPI :
                  throw new RuntimeException(
                     "TINI executable has wrong API version: "
                     + Integer.toHexString(((retval >>> 8) & 0xFFFF))
                     + ", TINI firmware API version: "
                     + Integer.toHexString(wos_getAPIVersion()));
               case EXEC_BADCRC :
                  throw new RuntimeException(
                     "Bad CRC.");
               default :
                  throw new RuntimeException(
                     "retval = " + retval);
            }
         }
      }
      catch (ExceptionInInitializerError initerr)
      {
         throw new RuntimeException("Could not execute file: " + initerr);
      }
      catch (Exception e)
      {
         throw new RuntimeException("Could not execute file: " + e);
      }
      catch (OutOfMemoryError m)
      {
         throw new RuntimeException("Could not execute file: Insufficient contiguous heap");
      }
      finally
      {
         // No matter what, we must release process creation lock.
         wos_processLockClear();
      }
   }

   private static native boolean wos_hasProcessStartedNative (int processID);

   private static native boolean wos_processLockTestAndSet ();

   private static native void wos_processLockClear ();

   private static native int executeFileNative (byte[] data,
           byte[] processName);

   private static native void wos_setJavaProcessStateNative (int processID,
           Object[] javaState);

   private static native void wos_setJavaProcessArgsNative (int processID,
           Object[] javaState);

   private static native void wos_setProcessRestartParentNative (
      int processID, boolean restart);

   private static native void wos_setTaskOwnerNative (int processID,
           int owner);

   private static native void wos_wakeUpProcessNative (int processID);

   private static native int wos_waitForProcessExitNative (int processID);

   private static native void wos_setProcessObjectNative (int processID,
           Process obj);

   //    private static native byte[] getListPointersNativeDS(byte[] nameBytes, byte uid);
   //    private static native void getFileNameNativeDS(int pointer, byte[] dest);
}
