
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.io;



import com.dalsemi.system.ArrayUtils;
import java.io.*;


/**
 *  12-8-00 KLA - added useModifiedUTF8 to do the wierd Java-UTF8, where the value 0
 *                is encoded as 2 bytes, rather than 1.  this allows java.io.RandomAccessFile
 *                and java.io.DataOutputStream to use this class.
 */
public class CharToByteUTF8
   extends CharToByteConverter
{

   private boolean useModifiedUTF8 = false;
   
   /**
    *
    */
   public int getMaxByteCount (char[] forThis, int start, int end)
   {
      int utfLength = 0;

      for (int i = start; i < end; i++)
      {
         int ch = forThis [i];

         if (ch < 0x80)
         {
            utfLength++;
            if (useModifiedUTF8 && (ch==0))
                utfLength++;
         }
         else if (ch < 0x0800)
         {
            utfLength += 2;
         }
         else
         {
            utfLength += 3;
         }
      }

      return utfLength;
   }

   /**
    *
    */
   public int convert (char[] src, int srcStart, int srcEnd, byte[] dst,
                       int dstStart, int dstEnd)
      throws CharConversionException
   {
      int pos = dstStart;

      for (int i = srcStart; i < srcEnd; i++)
      {
         int ch = src [i];

         if (ch < 0x80)
         {
            if (useModifiedUTF8 && (ch==0))
            {
                dst [pos++] = (byte)0xc0;
                dst [pos++] = (byte)0x80;
            }
            else
                dst [pos++] = ( byte ) ch;
         }
         else if (ch < 0x0800)
         {
            dst [pos++] = ( byte ) (0xC0 | ((ch >> 6) & 0x1F));
            dst [pos++] = ( byte ) (0x80 | (ch & 0x3F));
         }
         else
         {
            dst [pos++] = ( byte ) (0xE0 | ((ch >> 12) & 0x0F));
            dst [pos++] = ( byte ) (0x80 | ((ch >> 6) & 0x3F));
            dst [pos++] = ( byte ) (0x80 | (ch & 0x3F));
         }
      }

      return pos - dstStart;
   }

   /**
    *
    */
   public int flush (byte[] buff, int start, int length)
   {
      return 0;
   }

   /**
    *
    */
   public String getName ()
   {
      return "UTF8";
   }
   
   public void useModifiedUTF8(boolean useMod)
   {
      synchronized(this)
      {
        useModifiedUTF8 = useMod;
      }
   }
   
   public static int write(String str, DataOutput writer)
        throws IOException
   {
      CharToByteUTF8 utf8 = new CharToByteUTF8();
      utf8.useModifiedUTF8 = true;
      char[] string = str.toCharArray();
      int utfLength = utf8.getMaxByteCount(string, 0, string.length);
      int total = 0;

      if (utfLength > 65535)
      {
         throw new UTFDataFormatException();
      }

      byte[] buf = null;
      int pos = 0;
      int start_in_buf = 2;

      if (utfLength > 65530)
      {
         pos = string.length >> 1;
         int tempLength = utf8.getMaxByteCount(string, 0, pos);
         buf = new byte [tempLength+2];

         buf [0] = ( byte ) (utfLength >> 8);
         buf [1] = ( byte ) utfLength;
         
         utf8.convert(string, 0, pos, buf, 2, 0);
         writer.write(buf, 0, buf.length);
         total = buf.length;

         //add that 2 back in cause of the length
         utfLength = utfLength - buf.length + 2;
         buf = new byte [utfLength];
         start_in_buf = 0;
      }
      else
      {
         buf = new byte[utfLength + 2];
         buf [0] = ( byte ) (utfLength >> 8);
         buf [1] = ( byte ) utfLength;
      }
         
      utf8.convert(string, pos, string.length, buf, start_in_buf, 0);
      writer.write(buf, 0, buf.length);
      total += buf.length;
      return total;
      
   }
}
