
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.protocol.ftp;



import java.io.*;
import java.net.*;
import java.util.*;
import com.dalsemi.protocol.*;


/**
 * Connection.java
 * An extension of the abstract URLConnection class.
 *
 * @author KLA
 */
public class Connection
   extends URLConnection
{

   /* The following 6 variables form the meat of an FTP URL.
    * Some (maybe most) will be null or trivial do to no
    * specification. 'host' and 'port' are required.
    */
   String        host;
   String        user;
   String        pass;
   int           port;
   String        file;
   String        type;
   HeaderManager hm;
   InputStream   ftpin = null;

   /* the object that will actually perform
      the FTP protocol for us
    */
   FTPClient     ftp;           

   /* tells us we need to call the connect()
      method if this has not yet happened
    */
   boolean connected = false;   


   /**
    * Constructor Connection
    *
    *
    * @param url
    *
    */
   public Connection (URL url)
   {
      super(url);

      setFileNameMap(new DefaultFileNameMap());

      hm   = new HeaderManager();
      port = url.getPort();
      file = url.getFile();
      type = url.getRef();
      host = url.getHost();
      
      String type = guessContentTypeFromName(file);

      if (type != null)
         hm.add("content-type", type);
      int index = host.indexOf('@');
      if (index!=-1)
      {
        user = host.substring(0,index);
        int passindex = user.indexOf(':');
        if (passindex!=-1)
        {
            pass = user.substring(passindex+1);
            user = user.substring(0,passindex);
        }
      }

      int colon_index = host.indexOf(':', index);
      if (colon_index!=-1)
        host = host.substring(index+1, colon_index);
      else
        host = host.substring(index+1);

   }

   /**
    * Method connect
    *
    *
    * @throws IOException
    *
    */
   public void connect ()
      throws IOException
   {

      //here's where i need to set up a session and all
      //log in, set up the Sockets, Streams
      ftp = new FTPClient(host, port);

      ftp.analyze(ftp.userName(user));
      ftp.analyze(ftp.password(pass));

      connected = true;
   }

   /**
    * Method getInputStream
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public InputStream getInputStream ()
      throws IOException
   {
      if (ftpin == null)
         return ( InputStream ) getContent();
      else
         return ftpin;
   }

   /**
    * Method getContent
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public Object getContent ()
      throws IOException
   {
      if (ftpin != null)
         return ftpin;

      //returns an FTPInputStream
      if (!connected)
         connect();

      //first determine type - try to determine what we need
      int t_type = -1;

      if (type != null)
      {
         try
         {
            if (type.substring(0, 5).equalsIgnoreCase("type="))
            {
               char t = type.charAt(5);

               if ((t == 'a') || (t == 'A'))
                  t_type = 0;

               if ((t == 'i') || (t == 'I'))
                  t_type = 1;

               if ((t == 'd') || (t == 'D'))
                  t_type = 2;
            }
            else
               throw new Exception();
         }
         catch (Exception e)
         {   //means the URL is not properly formatted
            throw new IOException("Improper type declaration: " + type);
         }
      }
      else
      {

         //try to logically determine: html, txt are ascii
         if (file.endsWith(".htm") || file.endsWith(".html")
                 || file.endsWith(".txt"))
            t_type = 0;
         else if (file.endsWith("/"))
            t_type = 2;
         else
            t_type = 1;
      }

      //support for the funky ~ character
      if ((file != null) && (file.length() > 1))
      {
            if (file.charAt(1) == '~')   //want to change to home directory first
            {
               ftp.issueCommand("CWD ~\r\n");

               try
               {
                  file = new String("." + file.substring(2));
               }
               catch (Exception e)
               {   //means that file was just ~ directory
                  file = new String(".");
               }
            }
      }

      if (t_type == 1)
         ftp.binary();

      int     code = 0;
      boolean dir  = true;

      ftp.analyze(ftp.port());
      if (t_type == 2)
         code = ftp.list(file);
      else
      {
         code = ftp.retr(file);

         if (code == 550)   //might be a directory, lets try that
         {
            ftp.analyze(ftp.port());
            code = ftp.list(file);
         }
         else
         {
            dir = false;
            //look at the response string that has " bytes)" in it...
            //the ftp client already does that on a RETR, so just see what it found.
            if (ftp.getFileSize()!=-1)
               hm.add("content-length", String.valueOf(ftp.getFileSize()));
         }
      }

      ftp.analyze(code);

      ftpin = ftp.getDataStream();

      return ftpin;
   }

   /**
    * Method getHeaderField
    *
    *
    * @param s
    *
    * @return
    *
    */
   public String getHeaderField (String s)
   {
      return hm.get(s);
   }

   /**
    * Method getHeaderFieldKey
    *
    *
    * @param n
    *
    * @return
    *
    */
   public String getHeaderFieldKey (int n)
   {
      return hm.getKey(n);
   }

   /**
    * Method getHeaderField
    *
    *
    * @param n
    *
    * @return
    *
    */
   public String getHeaderField (int n)
   {
      return hm.get(n);
   }
}
