
/*---------------------------------------------------------------------------
 * Copyright (C) 1999-2001 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.protocol.ftp;



/*
 * FTPClient.java
 *
 * @author KLA
 *
 */
import java.net.*;
import java.io.*;
import java.util.*;
import com.dalsemi.protocol.*;


public class FTPClient
   extends BasicClient
{
   int    lastResponse   = -1;      //stores the last response integer
   String responseString = null;    //will store the last response string,
   String lastCommand = null;

   //functions will return the status int's
   InputStream reader   = null;     //for reading data and command responses
   int         fileSize = -1;       //for files, finds out from FTP response line
                                    //so that .available() is accurate

   ServerSocket sock = null;
   
   boolean passive = false;

   /** Field binary           */
   public boolean binary = false;   //are we in binary xmission mode

   /* Create a connection to an FTP server running at the
    * specified HOST on the default FTP port
    */

   /**
    * Constructor FTPClient
    *
    *
    * @param SERVER
    *
    * @throws IOException
    *
    */
   public FTPClient (String SERVER)
      throws IOException
   {
      this(SERVER, 21, DEFAULT_SOCKET_TIMEOUT);   //well known FTP port
   }

   /* Create a connection to an FTP server running at the
    * specified HOST:PORT
    */

   /**
    * Constructor FTPClient
    *
    *
    * @param SERVER
    * @param PORT
    *
    * @throws IOException
    *
    */
   public FTPClient (String SERVER, int PORT)
      throws IOException
   {
      this (SERVER, PORT, DEFAULT_SOCKET_TIMEOUT);
   }

   /**
    * Constructor FTPClient
    *
    *
    * @param SERVER
    * @param PORT
    * @param timeout
    *
    * @throws IOException
    *
    */
   public FTPClient (String SERVER, int PORT, int timeout)
      throws IOException
   {
      super(SERVER, PORT, timeout);

      issueCommand(null);
   }

   /* Send a user name to login to an FTP server.  Send
    * 'anonymous' if no user specified.
    */

   /**
    * Method userName
    *
    *
    * @param user
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int userName (String user)
      throws IOException
   {
      if (user == null)
         user = new String("anonymous");

      return issueCommand("USER "
                          + com.dalsemi.protocol.URLDecoder.decode(user)
                          + "\r\n");
   }

   /* Send a password to the FTP server.  If we are doing anonymous
    * login, send something that looks like an email address.
    */

   /**
    * Method password
    *
    *
    * @param password
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int password (String password)
      throws IOException
   {

      //get SYSTEM USER PROPERTY if possible for anonymous login
      if (password == null)
         password = new String("user@");

      return issueCommand("PASS "
                          + com.dalsemi.protocol.URLDecoder.decode(password)
                          + "\r\n");
   }

   /* Change to BINARY transfer mode */

   /**
    * Method binary
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int binary ()
      throws IOException
   {
      issueCommand("TYPE I\r\n");

      binary = true;

      return lastResponse;
   }

   /* Change to ASCII transfer mode */

   /**
    * Method ascii
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int ascii ()
      throws IOException
   {
      issueCommand("TYPE A\r\n");

      binary = false;

      return lastResponse;
   }

   /* Sets up a passive connection to the server. */

   /**
    * Method passiveConnection
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int passiveConnection ()
      throws IOException
   {
      issueCommand("PASV\r\n");

      String          r  =
         responseString.substring(responseString.indexOf(( int ) '(') + 1,
                                  responseString.indexOf(( int ) ')'));
      StringTokenizer st = new StringTokenizer(r, ",");

      if (st.countTokens() != 6)
         throw new IOException("Bad Response Format");

      StringBuffer ip = new StringBuffer();

      for (int i = 0; i < 4; i++)
      {
         ip.append(st.nextToken());

         if (i != 3)
            ip.append('.');
      }

      int p = (Integer.parseInt(st.nextToken()) << 8)
              | Integer.parseInt(st.nextToken());

      setupDataSocket(ip.toString(), p);
      synchronized (this)
      {
        passive = true;
      }
      return lastResponse;
   }


    public synchronized int port ()
        throws IOException
    {
        passive = false;
        int tries = 0;
        
        if (sock!=null)
        {
            sock.close();
            sock = null;
        }
        
        while (sock==null)
        {
            try
            {
                sock = new ServerSocket(1024+((int)(Math.random()*1000))%1000);
            }
            catch(Exception e)
            {
                sock = null;
                tries++;
                if (tries==5)
                {
                    throw new SocketException("Could not passive open a port!");
                }
            }
        }
        byte[] ip = control.getLocalAddress().getAddress();
        int port = sock.getLocalPort();
        String command = "PORT "+(ip[0] & 0x0ff)+","+(ip[1] & 0x0ff)+","+
                         (ip[2] & 0x0ff)+","+(ip[3] & 0x0ff)+","+(port/256)+","+(port%256)+"\r\n";
        issueCommand(command);
        return lastResponse;
    }
    


   /* Issues a command to get a file from an FTP server.  Also parses
    * the response string for the proper size of the file.  Filesize is
    * invalid for directories.
    */
   /**
    * Method retr
    *
    *
    * @param path
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int retr (String path)
      throws IOException
   {
      try
      {
        
         issueCommand("RETR " + com.dalsemi.protocol.URLDecoder.decode(path)
                      + "\r\n");

         //now get the length of the file
         String r =
            responseString.substring(responseString.lastIndexOf(( int ) '(')
                                     + 1);

         fileSize = Integer.parseInt(r.substring(0, r.indexOf(( int ) ' ')));

         //System.out.println("Filesize: "+fileSize);
         //                      lastResponse = issueCommand(null);
      }
      catch (NumberFormatException e)
      {
         fileSize = -1;
      }
      
      return lastResponse;
   }

   /* List the files in a directory pointed to by the URL.
    * Returns data as 'ls -l' would.
    */

   /**
    * Method list
    *
    *
    * @param path
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int list (String path)
      throws IOException
   {
      if (path != null)
        issueCommand("LIST " + com.dalsemi.protocol.URLDecoder.decode(path) + "\r\n");
      else
        issueCommand("LIST\r\n");

      fileSize = -1; 
      return lastResponse;
   }

   /* Get the size of a file we are downloading */

   /**
    * Method getFileSize
    *
    *
    * @return
    *
    */
   public int getFileSize ()
   {
      return fileSize;
   }

   /* Close the FTP session. */

   /**
    * Method quit
    *
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int quit ()
      throws IOException
   {
      return issueCommand("QUIT\r\n");
   }

   static final int DIGIT1         = 0;
   static final int DIGIT2         = 1;
   static final int DIGIT3         = 2;
   static final int LASTLINEMARK   = 3;
   static final int ALINE_CRLF     = 4;
   static final int LASTLINE_CRLF  = 5;
   static final int DONE           = 6;
   private int      globalBufIndex = 0;
   private int      globalBufLen   = 0;

   /* Sends a command to the FTP server and returns
    * the success code reported by the FTP server.
    */

   /**
    * Method issueCommand
    *
    *
    * @param command
    *
    * @return
    *
    * @throws IOException
    *
    */
   public int issueCommand (String command)
      throws IOException
   {
      boolean lastLine = false;
      int     state    = DIGIT1;

      reader = command(command);

      StringBuffer sb        = new StringBuffer();
      byte[]       buf       = new byte [256];
      int          index     = globalBufIndex;
      int          initindex = globalBufIndex;
      int          len       = globalBufLen;

      try
      {
         while ((state != DONE) && (len != -1))
         {
            if (index >= len)
            {
               index     = 0;
               initindex = 0;
               len       = reader.read(buf, 0, buf.length);   //single byte read OK - this is the status message
            }

            if (len != -1)
            {
               while ((state != DONE) && (index < len))
               {
                  char ch = ( char ) (buf [index] & 0xFF);

                  switch (state)
                  {

                     case DIGIT1 :
                        if (Character.isDigit(ch))
                           state = DIGIT2;
                        break;

                     case DIGIT2 :
                        if (Character.isDigit(ch))
                           state = DIGIT3;
                        break;

                     case DIGIT3 :
                        if (Character.isDigit(ch))
                           state = LASTLINEMARK;
                        break;

                     case LASTLINEMARK :
                        if (ch == '-')
                           state = ALINE_CRLF;
                        else
                           state = LASTLINE_CRLF;
                        break;

                     case ALINE_CRLF :
                        if (ch == 0x0A)
                           state = DIGIT1;
                        break;

                     case LASTLINE_CRLF :
                        if (ch == 0x0A)
                           state = DONE;
                        break;

                     case DONE :
                        break;
                  }

                  index++;
               }

               sb.append(new String(buf, initindex, index - initindex));
            }
         }
      }
      catch (IOException e)
      {
         globalBufLen = 0;

         // Make sure superclass' socket is closed on IOException.
         control.close();

         throw e;
      }

      /* Update global variables */
      if ((len < 0) || (index >= len))
      {
         globalBufLen   = 0;
         globalBufIndex = 0;
      }
      else
      {
         globalBufLen   = len;
         globalBufIndex = index;
      }

      String x = sb.toString();

      //com.dalsemi.jellybeans.Debug.debugDump(x.length());

      /*if (x.length() > 0)
          com.dalsemi.jellybeans.Debug.debugDump(x.getBytes(),x.length());*/

      synchronized(this)
      {
        lastCommand = command;
        if (x.length() > 4)
        {

            //com.dalsemi.jellybeans.Debug.debugDump(0x55);
            responseString = x.substring(4);

            //System.out.println(x.length()+":RESPONSE: "+x);
            //System.out.println("response num: "+x.substring(0,3));
            //com.dalsemi.jellybeans.Debug.debugDump((x.substring(0,3)).getBytes(),3);
            try
            {
                lastResponse = Integer.parseInt(x.substring(0, 3));
            }
            catch(NumberFormatException e)
            {
                responseString = x;
                lastResponse = 500;
            }
        }
        else
        {

            // What response should we fake if the string is too small
            // to parse for a 3 byte integer.
            responseString = "";
            lastResponse = 0;
        }
      }
      return lastResponse;
   }

   /* Gets the stream to the file/directory info we
    * have requested using retr()
    */
   public InputStream getDataStream ()
      throws IOException
   {
      if (passive)
        return readDataStream();
      else
      {
        InputStream din = null;
        synchronized(this)
        {
            Socket d = sock.accept();
            din = d.getInputStream();
            sock.close();
            int tries=0;
            while (din.available()==0)
            {
                tries++;
                try
                {
                    Thread.sleep(10);
                }
                catch(InterruptedException e)
                {
                    //drain...
                }
                if (tries==400)
                    throw new SocketException("Timed out waiting for input");
            }
        }
        return din;
      }
   }


   public OutputStream getOutputStream ()
      throws IOException
   {
      if (passive)
        return dout;
      else
      {
        OutputStream my_dout = null;
        synchronized(this)
        {
            Socket d = sock.accept();
            my_dout = d.getOutputStream();
            sock.close();
        }
        return my_dout;
      }
   }


   /* Outputs meaningful IOExceptions in the event of bad
    * FTP return codes.
    */

   /**
    * Method analyze
    *
    *
    * @param code
    *
    * @throws IOException
    *
    */
   public void analyze (int code)
      throws IOException
   {
      String temp = null;
      switch (code)
      {
         case 421 :
            temp = "FTP: Service not available";
            break;
         case 425 :
            temp = "FTP: Can't open data connection";
            break;
         case 500 :
            temp = "FTP: Syntax Error: Command";
            break;
         case 501 :
            temp = "FTP: Syntax Error: Parameters";
            break;
         case 530 :
            temp = "FTP: Not Logged In";
            break;
         case 503 :
            temp = "FTP: Bad Command Sequence";
            break;
         case 332 :
            temp = "FTP: Need Account for Login";
            break;
         case 502 :
            temp = "FTP: Command Not Implemented";
            break;
         case 550 :
            temp = "FTP: Action not taken (file unavailable)";
      }
      if (temp==null)
      {
        if (code > 499)
            temp = "FTP: Terminal Exception: Permanent Failure";
        else if (code > 399)
            temp = "FTP: Temporary Exception: Please try again";
      }
      if (temp!=null) throw new IOException(temp);
   }
   
   public String getResponseString()
   {
        return responseString;
   }
   
   public String getLastCommand()
   {
        return lastCommand;
   }
   
   public void close()
   {
        try
        {
            if (cin!=null) cin.close();
            if (din!=null) din.close();
            if (cout!=null) cout.close();
            if (dout!=null) dout.close();
        }
        catch(IOException e)
        {
            //drain...
        }
   }
}
