
/*---------------------------------------------------------------------------
 * Copyright (C) 1999-2001 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

/*
 * TINI
 * CIM
 *
 * Revised 8/17
 * 1. System properties
 * 2. InetAddress.getLocalHost().getHostName() not supported so fall back fromaddr not supported
 * 3. mailto class with eventually be mailtourl class when url class supported
 */
package com.dalsemi.protocol.mailto;



import java.io.*;
import java.net.*;
import com.dalsemi.tininet.*;
import com.dalsemi.system.*;
import com.dalsemi.protocol.*;
import com.dalsemi.tininet.*;


/**
 * Class to construct and deliver mail messages using the SMTP protocol.
 *
 *
 * @CIM
 */
public class Connection
   extends URLConnection
{
   MailToPrintStream           os;
   String                      recipient;
   String                      ccRcpt = null;
   Socket                      mailSocket;
   boolean                     isTo = true;
   String                      userName;
   String                      fromAddr = null;
   BufferedReader              socketReader;
   PrintWriter                 printer;
   String                      toLine       = null, ccLine = null;
   private static final String DEFAULT_USER = "TINI";

   /**
    * Constructs a new mailTo object addressed to the parameter Recipient.
    * Multiple recipients should be separated by commas.
    *
    * @param Recipient recipient or recipients to send the message to
    *
    * @param url
    *
    */
   public Connection (URL url)
   {
      super(url);

      toLine = "";
      ccLine = "";

      HeaderManager hm = new HeaderManager();

      hm.add("content-type", "text/html");

      //recipient = Recipient;
      userName = DEFAULT_USER;
   }

   /**
    * Method setUserName
    *
    *
    * @param toThis
    *
    */
   public void setUserName (String toThis)
   {
      userName = toThis;
   }

   /**
    * Not yet implemented
    *
    *
    * @throws IOException
    *
    */
   public void connect ()
      throws IOException
   {
   }

   /**
    * Start Mail Message and get output stream to socket
    *
    * @return OutputStream to socket
    *
    * @throws IOException
    *
    */
   public synchronized OutputStream getOutputStream ()
      throws IOException
   {
      if (os != null)
         return os;
      else
      {
         String s = url.getFile();

         //parse s for cc:
         int ccBegin = s.indexOf("cc:");

         if (ccBegin != -1)
         {
            if (ccBegin + 3 < s.length())
               ccRcpt = s.substring(ccBegin + 3);

            s = s.substring(0, ccBegin);
         }

         String user = null;

         //Get username.  If we are running under a shell, use the logged in name.
         user = TINIOS.getCurrentUserName();

         if ((user == null) || (user.length() == 0))
         {
            user = DEFAULT_USER;
         }

         user += "@" + TININet.getHostname() + "." + TININet.getDomainname();

         getMailSocket();

         //setSMTPOutputStream();
         os = new MailToPrintStream(mailSocket.getOutputStream(), this);

         if (fromAddr != null)
         {
            sendFrom(fromAddr);
         }
         else
         {
            sendFrom(user);
         }

         sendTo(s);

         if (ccRcpt != null)
            sendCC(ccRcpt);

         sendSMTPCommand("DATA", 354);

         // Add to and cc line to data show it shows up in header
         if (toLine.length() != 0)
            parseLine("To: ", toLine);

         if (ccLine.length() != 0)
            parseLine("cc: ", ccLine);

         return os;
      }
   }

   private void sendFrom (String sender)
      throws IOException
   {
      sendSMTPCommand("MAIL FROM: <" + sender + ">", 250);
   }

   /**
    * Method setFromAddr
    *
    *
    * @param sender
    *
    */
   public void setFromAddr (String sender)
   {
      fromAddr = sender;
   }

   /**
    * Method getFromAddr
    *
    *
    * @return
    *
    */
   public String getFromAddr ()
   {
      return (fromAddr);
   }

   void sendSMTPCommand (String Command, int response)
      throws IOException
   {
      int    sResp;
      String reply;

      printer.print(Command + "\r\n");
      printer.flush();

      do {
        reply = socketReader.readLine();
        if (reply.length() < 4)
          throw new IOException(reply); // too short, must be CCC <SP>
      } while (reply.charAt(3) == '-'); // handle multiline replies, RFC 821 appendix E


      while ((sResp = Integer.parseInt(reply.substring(0, 3))) != response)
      {
         if (sResp != 220)
            throw new IOException(reply);
         else
         {
           do {
             reply = socketReader.readLine();
             if (reply.length() < 4)
               throw new IOException(reply); // too short, must be CCC <SP>
           } while (reply.charAt(3) == '-'); // handle multiline replies, RFC 821 appendix E
         }
      }
   }

   private void sendTo (String whoTo)
      throws IOException
   {
      boolean done = false;
      String  user = null;
      int     start, end, last;

      end  = 0;
      last = 0;

      while (!done)
      {
         start = 0;

         if (whoTo.indexOf('<') == 0)
            start++;

         end = whoTo.indexOf(',');

         if (end == -1)
         {
            end = whoTo.indexOf('>');

            if (end == -1)
               end = whoTo.length();

            last = end;
         }
         else
         {
            last = end;
            end  = whoTo.indexOf('>');

            if (end == -1 || end > last)
               end = last;
         }

         user = whoTo.substring(start, end);

         sendSMTPCommand("RCPT TO: <" + user + ">", 250);

         if (isTo)
         {
            if (user.length() > 0)
               toLine += user + ",";
         }
         else
         {
            if (user.length() > 0)
               ccLine += user + ",";
         }

         int len = whoTo.length();

         last++;

         if (last >= len)
         {
            done = true;
         }
         else
         {
            whoTo = whoTo.substring(last, len);
         }
      }
   }

   private void sendCC (String whoTo)
      throws IOException
   {
      isTo = false;

      this.sendTo(whoTo);

      isTo = true;
   }

   public void closeConnection ()
      throws IOException
   {
      if (mailSocket == null)
         return;
      else
      {
         mailSocket.close();

         printer      = null;
         socketReader = null;
      }
   }

   void getMailSocket ()
      throws IOException
   {

      // open provided string, then try mail.host system property,
      // then localhost and mailhost
      int port = 25;

      try
      {
         String s1;

         if ((s1 = System.getProperty("mail.host")) == null)
            s1 = TININet.getMailhost();

         if (s1 != null)
         {
            openMailServer(s1, port);

            return;
         }
         else
         {
            openMailServer("mailhost", port);
         }
      }
      catch (Exception ex)
      {
         openMailServer("mailhost", port);
      }
   }

   void openMailServer (String server, int port)
      throws IOException, UnknownHostException
   {
      mailSocket = new Socket(InetAddress.getByName(server), port);

      if (mailSocket != null)
      {

         // Get input and output stream
         socketReader = new BufferedReader(
            new InputStreamReader(mailSocket.getInputStream()));
         printer      = new PrintWriter(mailSocket.getOutputStream());

         // Identify computer with mailhost
         sendSMTPCommand("HELO " + TININet.getHostname(), 250);
      }
   }

   void parseLine (String command, String line)
   {
      int    firstChar = 0;
      int    lastChar  = 0;
      String totalLine = "";

      // separate recipients by delimitting commas
      lastChar = line.indexOf(',', firstChar);

      // loop until no more recipients
      while (lastChar != -1)
      {
         totalLine += command + line.substring(firstChar, lastChar) + "\n";

         if (lastChar + 1 < line.length())
         {
            firstChar = lastChar + 1;
            lastChar  = line.indexOf(',', firstChar);
         }
         else
            lastChar = -1;
      }

      os.print(totalLine);
   }
}
