
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell;



import java.util.*;
import java.io.*;
import com.dalsemi.system.*;
import com.dalsemi.shell.server.*;


/**
 * The default implementation of a TINIShell.  In this shell, all users are assigned the
 * same UID and user name, and all are given adminstrator privileges.  This shell has no
 * commands associated with it.  For an example of a complete implementation of a TINIShell,
 * see the <code>slush</code> shell that is included with the TINI distribution files.
 */
public class DefaultTINIShell
   extends TINIShell
{

   /**
    * Creates the initial environment for a shell.  The environment
    * is initialized to hold the current working directory.  This value
    * is stored under the "current directory" key.
    */
   public DefaultTINIShell ()
   {
      super();
   }

   /**
    * Returns the name of the shell.  The name of the default TINI shell is simply "slushLite".
    *
    * @return  the name of the shell
    */
   public String getName ()
   {
      return "slushLite";
   }

   /**
    * Returns the version of the shell.
    *
    * @return   the shell version
    */
   public String getVersion ()
   {
      return "1.0";
   }

   /**
    * Logs a user into the system and sets
    * that user's privilege level correctly.  Most shells will have a list
    * of valid users and their corresponding passwords and user id's
    * to consult when this method is called (such as a passwd file).
    * This shell always returns the default adminstrator UID (0x80).
    *
    * @param  userName   the user name of the user logging in
    * @param  password   the password for this user
    *
    * @returns the user ID of the newly logged in user, or -1 if
    * the userName password combination was not correct
    */
   public int login (String userName, String password)
   {
      return (getCurrentUID() & 0x0FF);
   }

   /**
    * Does nothing in this implementation.
    *
    * @param  info   any extra information needed in the logout process
    */
   public void logout (Object info)
   {
   }

   /**
    * Returns the user ID of the current user.  This implementation will always
    * return the default adminstrator UID (0x80).
    *
    * @return  the current user's ID
    */
   public byte getCurrentUID ()
   {

      return (byte)TINIOS.getOwnerIDByTaskID(TINIOS.getTaskID());
      //DEBUG, BIG, BIG, DEBUG! Uncomment previous line when Bryan
      //fixes the default uid to be root instead of 0!!!
      //return adminUID;
   }

   /**
    * Returns the user name of the user with the given user ID.  This implementation will always
    * return the default adminstrator name ("root").
    *
    * @param  uid   The uid to check
    *
    * @return  the user name of the user with the given user ID
    */
   public String getUserNameByUID (byte uid)
   {
      return "root";
   }

   /**
    * Returns the ID of the user with the given user name.  This implementation will always
    * return the default adminstrator UID (0x80).
    *
    * @param  username   the username to check
    *
    * @return  the ID of the user with the given user name,
    * or -1 if that user does not exist
    */
   public int getUIDByUserName (String username)
   {
      return (adminUID & 0x0FF);
   }

   /**
    * Returns the name of the current user logged in.    This implementation will always
    * return the default adminstrator name ("root").
    *
    * @return  the name of the current user
    */
   public String getCurrentUserName ()
   {
      return "root";
   }

   /**
    * Returns a copy of the current environment.
    *
    * @return the environment
    */
   public Hashtable getCurrentEnvironment ()
   {
      return ( Hashtable ) environment.clone();
   }

   /**
    * Gets the value of the given key from the current environment.
    *
    * @param  key   the key to lookup in the current environment
    *
    * @return  the value of the key, or an empty String if the key
    * is not set in the current environment
    */
   public String getFromCurrentEnvironment (String key)
   {
      String val = ( String ) environment.get(key);

      if (val == null)
      {
         return "";
      }
      else
      {
         return val;
      }
   }

   /**
    * Executes a command in the shell.  This implementation has no commands.  A call to this
    * method will print "Command not found." to the given error stream.
    *
    * @param commandLine  an Object array containing the command in the first element,
    * followed by any parameters need for that command in a String[] in the second element
    * @param in  the stream the command will use to get input
    * @param out  the stream used to report non-critical messages
    * @param err  the stream used to report critical messages
    * @param env  a table of environment variables
    *
    * @throws Exception any exception raised by the command.
    */
   public void execute (Object[] commandLine, SystemInputStream in,
                        SystemPrintStream out, SystemPrintStream err,
                        Hashtable env)
      throws Exception
   {

      //DEBUG ==> Implement a small command interpreter?
      err.println("Command not found.");   //DEBUG
   }
}
