
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell;



import java.util.*;
import java.io.*;
import com.dalsemi.shell.server.*;


/**
 * This abstract class encapsulates all of the functionality of a
 * system shell for TINI.  It interacts between a process and the
 * system OS, and provides the current environment and the user ID of
 * the current process.  Methods are provided for converting to 
 * and from the numerical user ID and the text representation and for
 * determining whether a particular user is an administrator.  A user
 * will be considered an administrator if the most significant bit of
 * their user ID is set.
 */
public abstract class TINIShell
{

   /**
    * Standard administrator UID.  Also used as mask to determine if other UIDs are administrators.
    */
   public static final byte adminUID = ( byte ) 0x80;

   /**
    * Contains the current environment variables for the process.
    */
   protected Hashtable environment;

   /**
    * Creates the initial environment for a shell.  The environment
    * is initialized to hold the current working directory.  This value
    * is stored under the "current directory" key.
    */
   public TINIShell ()
   {

      //Set up the environment.  If this process was spawned from
      //another process, then we deep copy the environment from the
      //parent process.
      environment = new Hashtable(15, 10);

      environment.put(Session.CURRENT_DIRECTORY,
                      System.getProperty("user.dir"));
   }

   /**
    * Logs a user into the system and sets that user's
    * privilege level correctly.  Most shells will have a list
    * of valid users and their corresponding passwords and user id's
    * to consult when this method is called.  (Such as a passwd file.)
    *
    * @param  userName   the user name of the user logging in
    * @param  password   the password for this user
    *
    * @returns the user ID of the newly logged in user, or -1 if
    * the userName password combination was not correct
    */
   public abstract int login (String userName, String password);

   /**
    * Logs the current user out of the system, returning the
    * privilege level of the current process to the default state.
    *
    * @param  info   any extra information needed in the logout process
    */
   public abstract void logout (Object info);

   /**
    * Returns the name of the shell.
    *
    * @return  the name of the shell
    */
   public abstract String getName ();

   /**
    * Returns the version of the shell.
    *
    * @return   the shell version
    */
   public abstract String getVersion ();

   /**
    * Indicates whether or not the current user has admin privileges.
    *
    * @return  <code>true</code> if the user has admin rights,
    * <code>false</code> otherwise
    */
   public boolean isCurrentUserAdmin ()
   {
      return isAdmin(getCurrentUID());
   }

   /**
    * Returns the user ID of the current user.
    *
    * @return  the current user's ID
    */
   public abstract byte getCurrentUID ();

   /**
    * Determines if the given user ID has administrative rights.
    *
    * @param uid  the user ID
    *
    * @return  <code>true</code> if the user is an administrator,
    * <code>false</code> otherwise
    */
   public boolean isAdmin (byte uid)
   {
      return ((uid & adminUID) == adminUID);
   }

   /**
    * Returns the user name of the user with the given user ID.
    *
    * @param  uid   the uid to check
    *
    * @return  the user name of the user with the given user ID,
    * or null if that ID is not valid
    */
   public abstract String getUserNameByUID (byte uid);

   /**
    * Returns the ID of the user with the given user name.
    *
    * @param  username   the username to check
    *
    * @return  the ID of the user with the given user name,
    * or -1 if that user does not exist
    */
   public abstract int getUIDByUserName (String username);

   /**
    * Returns the name of the current user logged in.
    *
    * @return  the name of the current user
    */
   public abstract String getCurrentUserName ();

   /**
    * Returns the system environment (the environment that
    * all processes see when they start up).  The Hashtable
    * that is returned is a clone of the original.
    *
    * @return  a copy of the system environment
    */
   public Hashtable getSystemEnvironment ()
   {
      return ( Hashtable ) environment.clone();
   }

   /**
    * Returns a copy of the current environment.
    *
    * @return the environment
    */
   public abstract Hashtable getCurrentEnvironment ();

   /**
    * Gets the value of the given key from the current environment.
    *
    * @param  key  the key to lookup in the current environment
    *
    * @return  the value of the key, or <code>null</code> if the key
    * is not set in the current environment
    */
   public abstract String getFromCurrentEnvironment (String key);

   /**
    * Executes a command in the shell.
    *
    * @param commandLine  an Object array containing the command in the first element,
    * followed by any parameters need for that command in a String[] in the second element
    * @param in  the stream the command will use to get input
    * @param out  the stream used to report non-critical messages
    * @param err  the stream used to report critical messages
    * @param env  a table of environment variables
    *
    * @throws Exception any exception raised by the command
    */
   public abstract void execute (Object[] commandLine, SystemInputStream in,
                                 SystemPrintStream out,
                                 SystemPrintStream err, Hashtable env)
      throws Exception;
}
