
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell.server.serial;



import java.io.*;
import java.net.*;
import java.util.*;
import javax.comm.*;
import com.dalsemi.shell.server.*;
import com.dalsemi.system.TINIOS;


/**
 * A server that uses a serial port to listen for user connections.  For each connection made,
 * a Serial session is created.  All command processing is handled by the session, not this server.
 */
public class SerialServer
   extends Server
{
   SystemPrintStream serOut     = null;
   SerialInputStream serIn      = null;
   SerialPort        serialPort = null;

   /**
    * Prepares the server to listen on an arbitrary port.  The server will not
    * be started and no connections will be accepted until its <code>run()</code> method is executed.
    *
    *
    * @param portName the name of the serial port
    * @param speed the baud rate for the serial port communications
    * @param dataBits the data bits used by the serial port communications
    * @param stopBits the stop bits used by the serial port communications
    * @param parity the parity of the serial port communications
    *
    * @throws Exception if the serial port could not be opened or any of the
    * specified parameters are invalid
    *
    */
   public SerialServer (String portName, int speed, int dataBits,
                        int stopBits, int parity)
      throws Exception
   {
      if (com.dalsemi.system.Debug.defaultStreams)
      {
         com.dalsemi.system.Debug.resetDefaultStreams();
      }

      // Obtain a CommPortIdentifier object for the port to open.
      CommPortIdentifier portId =
         CommPortIdentifier.getPortIdentifier(portName);

      // Open the port represented by the CommPortIdentifier object.
      serialPort = ( SerialPort ) portId.open(TINIOS.getShellName(), 1000);

      // Set the parameters of the connection.
      serialPort.setSerialPortParams(speed, dataBits, stopBits, parity);

      serOut = new SystemPrintStream(serialPort.getOutputStream());
      serIn  = new SerialInputStream(serialPort.getInputStream(), serOut);

      try
      {
         while (serIn.available() > 0)
            serIn.read();
      }
      catch (IOException ioe)
      {
         closeAllPorts();

         throw ioe;
      }
   }

   /**
    * Listens on the serial port for connection requests.  Once a
    * request is made, it creates, initializes, and returns a new 
    * SerialSession to handle that request.  This method will block until
    * a connection is made.  Since only one connection is allowed at any
    * time on the serial server, if a connection already exists, this
    * method suspends the server.  The server will be resumed when the
    * current connection terminates.
    *
    * @return  a new <code>SerialSession</code>
    */
   protected Session acceptNewSession ()
   {
      SerialSession newSession = null;

      //Put SerialServer to sleep.  SerialSession will wake it up
      //when you log out or when the SerialServer is shutdown...
      if (sessions.size() > 0)
      {
         Thread.currentThread().suspend();
      }

      if (shutdown)
      {
         return null;
      }

      try
      {
         serOut.println("\r\n\r\nHit any key to login.");
         serIn.setRawMode(true);
         serIn.read();
         serIn.setRawMode(false);

         newSession = new SerialSession(serIn, serOut, serOut, this);

         try
         {
            newSession.start();
         }
         catch(Throwable t)
         {
            //aah! must be over the limit on Threads.
            //no more threads for you!  it's closing time.
            newSession = null;
            //i don't think we have anything to close, do we?
         }            
      }
      catch (IOException ioe)
      {
/*         com.dalsemi.system.Debug.debugDump("SS.acceptNewSession: "
                                            + ioe.toString());   //DEBUG
*/
         try
         {
            shutDown();
         }
         catch (Throwable t)
         {
//            com.dalsemi.system.Debug.debugDump(t.toString());    //DEBUG

            shutdown = true;
         }
      }

      return newSession;
   }

   /**
    * Closes and releases ownership of this server's serial port.
    */
   public void closeAllPorts ()
      throws IOException
   {
      serIn.close();
      serOut.close();
      serialPort.close();

      serialPort = null;
   }

   /**
    * Displays this message to all sessions initiated by this server.
    *
    * @param sendThis  the message to display
    */
   public void broadcast (String sendThis)
   {
      if (sessions.size() == 0)
      {

         //If there is no one logged in to the serial
         //server, we want to act as a console, so we
         //display it here.
         serOut.println(sendThis);
      }
      else
         super.broadcast(sendThis);
   }
}
