/********************************************************************************
 *  Copyright (C) 1999,2000 Dallas Semiconductor Corporation.                   *
 *  All rights Reserved. Printed in U.S.A.                                      *
 *  This software is protected by copyright laws of                             *
 *  the United States and of foreign countries.                                 *
 *  This material may also be protected by patent laws of the United States     *
 *  and of foreign countries.                                                   *
 *  This software is furnished under a license agreement and/or a               *
 *  nondisclosure agreement and may only be used or copied in accordance        *
 *  with the terms of those agreements.                                         *
 *  The mere transfer of this software does not imply any licenses              *
 *  of trade secrets, proprietary technology, copyrights, patents,              *
 *  trademarks, maskwork rights, or any other form of intellectual              *
 *  property whatsoever. Dallas Semiconductor retains all ownership rights.     *
 ********************************************************************************/
package com.dalsemi.slush.command.sled;

import java.io.*;

/**
 * General purpose byte buffer used as an alternative to
 * a StringBuffer.  This buffer also handles backspace events.
 * That is, if the byte 0x08 is inserted or appended, the
 * byte proceeding the current position is deleted.
 *
 * It is important to note that this all works on TINI because
 * there is no special character encoding on TINI.  If
 * this is ever ported to another platform, you will need to
 * take character encoding into account.
 *
 * @author Stephen Hess, Lorne Smith
 * @version 1.0
 */
public class ByteBuffer
{
    public static final int LEFT  = 1;
    public static final int RIGHT = 0;

    int                     BUFFER_LENGTH;

    int                     currentPosition;

    int                     endPosition;

    byte[]                  buffer;

    /**
     * Construct a buffer with the default size.
     */
    public ByteBuffer()
    {
        this(50);
    }

    /**
     * Construct a buffer with specified size.
     *
     * @param  size     Size of buffer.
     */
    public ByteBuffer(int size)
    {
        BUFFER_LENGTH = size;
        buffer = new byte[BUFFER_LENGTH];
        currentPosition = 0;
        endPosition = 0;
    }

    /**
     * Return the number of bytes in the buffer.
     *
     * @return  number of bytes currently held.
     */
    public int length()
    {
        return endPosition;
    }

    /**
     * Set the character (ok, byte) at the specified position
     * in the buffer.
     *
     * @param  here     Position in the buffer.
     * @param  toThis   New character.
     */
    public void setCharAt(int here, char toThis)
    {
        buffer[here] = (byte)toThis;
    }

    /**
     * Return the index of the first occurrence of the given character.
     *
     * @param  ch     The character for which to search.
     *
     * @return  The index of the first occurrence of the given character,
     * or -1 if the character is not in the buffer.
     */
    public int indexOf(char ch)
    {
        for(int i = 0; i < endPosition; i++)
        {
            if(buffer[i] == ch)
            {
                return i;
            }
        }
        return -1;
    }

    /**
     * Move the buffer pointer in the specified direction.
     *
     * @param direction     <code>LEFT</code> or <code>RIGHT</code>
     */
    public void movePosition(int direction)
    {
        if (direction == LEFT)
            currentPosition--;
        else if (direction == RIGHT)
            currentPosition++;

        if (currentPosition > endPosition)
        {
            currentPosition = endPosition;
        }
        else if (currentPosition < 0)
        {
            currentPosition = 0;
        }
    }

    /**
     * Insert the character in the buffer at the current position.
     *
     * @param insertThis     Character to insert.
     */
    public void insert(char insertThis)
    {
        if (insertThis == 0x08)
        {
            backspace();

            return;
        }

        increaseBuffer(1);

        if (currentPosition < endPosition)
        {
            System.arraycopy(buffer, currentPosition, buffer,
                             currentPosition + 1,
                             endPosition - currentPosition);
        }

        buffer[currentPosition] = (byte)insertThis;
        currentPosition++;
        endPosition++;
    }

    /**
     * Insert the character at the specified position in the buffer.
     *
     * @param  here        Position in the buffer.
     * @param  insertThis  Character to insert.
     */
    public void insert(int here, char insertThis)
    {
        currentPosition = here + 1;
        if (insertThis == 0x08)
        {
            backspace();

            return;
        }

        increaseBuffer(1);

        if (here < endPosition)
        {
            System.arraycopy(buffer, here, buffer,
                             here + 1,
                             endPosition - here);
        }

        buffer[here] = (byte)insertThis;
        endPosition++;
    }

    /**
     * Append the character to the buffer.  This will move the current
     * pointer to the end.
     *
     * @param  appendThis     Character to append to the buffer.
     */
    public void append(char appendThis)
    {
        currentPosition = endPosition;

        if (appendThis == 0x08)
        {
            backspace();

            return;
        }

        increaseBuffer(1);

        buffer[currentPosition] = (byte)appendThis;
        currentPosition++;
        endPosition++;
    }

    /**
     * Append the <code>String</code> to the buffer.  This will move the current
     * pointer to the end.
     *
     * @param  appendThis    <code>String</code> to append to the buffer.
     */
    public void append(String appendThis)
    {
        int length = appendThis.length();
        increaseBuffer(length);

        appendThis.getBytes(0, length, buffer, currentPosition);
        endPosition = endPosition + length;
        currentPosition = endPosition;
    }

    /**
     * Insert the <code>String</code> at the current position in the buffer.
     *
     * @param  str     <code>String</code> to insert.
     */
    public void insert(String str)
    {
        int len = str.length();
        increaseBuffer(len);
        if(currentPosition < endPosition)
        {
            System.arraycopy(buffer, currentPosition, buffer, currentPosition + len, endPosition - currentPosition);
        }
        str.getBytes(0, len, buffer, currentPosition);
        endPosition = endPosition + len;
        currentPosition = currentPosition + len;
    }

    /**
     * Delete the last character prior to the current position.
     */
    public void backspace()
    {
        movePosition(LEFT);
        delete();
    }

    /**
     * Delete the character at the current position.
     */
    public void delete()
    {
        if (currentPosition == endPosition)
            return;

        System.arraycopy(buffer, currentPosition + 1, buffer,
                         currentPosition, endPosition - currentPosition - 1);

        endPosition--;
    }

    /**
     * Return the current buffer contents in a <code>String</code>.
     *
     * @return  The current buffer contents.
     */
    public String toString()
    {
        return new String(buffer, 0, endPosition);
    }

    /**
     * Return the current buffer contents with no leading or trailing
     * white space.
     *
     * @return  A <code>String</code> containing the current buffer
     * contents with no leading or trailing white space.
     */
    public String trimmedString()
    {
        if(buffer.length > 0)
        {
            int first = 0;
            while(buffer[first] <= ' ' && first != endPosition)
                first++;

            int last = endPosition - 1;
            if(endPosition > first)
            {
                while(buffer[last] <= ' ')
                    last--;
            }

            return new String(buffer, first, last + 1 - first);
        }
        return "";
    }

    /**
     * Ensure that the buffer has room for the specified new amount.
     */
    private void increaseBuffer(int i)
    {

        // Only increase if needed...
        if (endPosition + i > buffer.length)
        {
            byte[]  temp = new byte[buffer.length + BUFFER_LENGTH + i];

            System.arraycopy(buffer, 0, temp, 0, buffer.length);

            buffer = temp;
        }
    }

    /**
     * Set the length of the buffer.
     *
     * @param len     The new length.
     */
    public void setLength(int len)
    {
        endPosition = len;

        if (currentPosition > len)
            currentPosition = len;
    }

    /**
     * Delete just past the first newline in the buffer.
     */
    public void delFirstLine()
    {
        String  lines = toString();
        int     index = lines.indexOf('\n');

        if (index != -1)
        {
            if (endPosition > index)
            {
                endPosition = endPosition - index - 1;

                System.arraycopy(buffer, index + 1, buffer, 0, endPosition);

                if (currentPosition > index)
                {
                    currentPosition = currentPosition - index - 1;
                }
                else
                {
                    currentPosition = 0;
                }
            }
            else
            {
                setLength(0);
            }
        }
    }

    /**
     * Write from the specified position to the end of the buffer
     * to the given stream.
     *
     * @param  out     The stream to which to write.
     * @param  off     The offset into the buffer.
     */
    public void writeTo(OutputStream out, int off)
    {
        writeTo(out, off, endPosition - off);
    }

    /**
     * Write the specified portions of the buffer to the given stream.
     *
     * @param  out     The stream to which to write.
     * @param  off     The offset into the buffer.
     * @param  len     The number of bytes to send.
     */
    public void writeTo(OutputStream out, int off, int len)
    {
        try
        {
            out.write(buffer, off, len);
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    /**
     * Return the current buffer contents in a byte array.
     */
    public byte[] toByteArray()
    {
        byte[] ret = new byte[endPosition];
        for(int i=0; i<ret.length; i++)
        {
            ret[i] = (byte)buffer[i];
        }

        return ret;
    }
}