/****************************************************************************
*
*  Copyright (C) 1999,2000 Dallas Semiconductor Corporation. 
*  All rights Reserved. Printed in U.S.A.
*  This software is protected by copyright laws of
*  the United States and of foreign countries.
*  This material may also be protected by patent laws of the United States 
*  and of foreign countries.
*  This software is furnished under a license agreement and/or a
*  nondisclosure agreement and may only be used or copied in accordance
*  with the terms of those agreements.
*  The mere transfer of this software does not imply any licenses
*  of trade secrets, proprietary technology, copyrights, patents,
*  trademarks, maskwork rights, or any other form of intellectual
*  property whatsoever. Dallas Semiconductor retains all ownership rights.
*
*     Module Name: Sled.java
*/
package com.dalsemi.slush.command.sled;

import com.dalsemi.shell.server.*;
import com.dalsemi.slush.command.sled.*;
import java.util.*;
import java.io.*;
import com.dalsemi.slush.*;
import com.dalsemi.fs.*;

/**
 * Basic vi like text editor.  I stress the "like" portion of that last 
 * sentence. 
 *
 * Sled currently has what I believe to be most of the minimum or core amount 
 * of vi functionality (plus a few features that aren't really core, and 
 * arguably not really needed.) This is just my opinion of base function, 
 * every vi user has his/her own ideas. 
 *
 * This implementation is already far chunkier than I would like, so I 
 * think that future versions will be optimizations to slim this down 
 * for memory and code size, and of course, bug fixes.  I gladly welcome 
 * suggestions (and coded implementations!!) for new features or
 * optimization of current functionality.
 *
 * Would take much, much too long to list vi features that sled does not 
 * support.  So, here is a rundown of most of the ones it does: 
 *
 * Insert/Normal mode:
 *      Inserting: i, I, o, O, a, A, s, S, C, cw, p, ., r, R, &, 
 *      Deleting : dw, D, x
 *      Movement : w, b, ^, $, G, H, L, M, ^l, ^d, ^u, ^b, ^f,
 *                 h, j, k, l, m, ', %
 *      Searching: ?, /, n, N
 *      Misc     : J, yy, ~, ZZ
 *
 * Command mode:
 *      Search & Replace: :[range]s/oldStr/newStr/[options]
 *                        where [range] can be '%' (file global),
 *                        'start,end' to limit to between start and end lines inclusive,
 *                        or nothing for current line only.
 *                        [options] can be any combo of 'c' (confirm), 'g' (line
 *                        global), or 'i' (case insensitive search)
 *      Text Movement   : :[range] move [addr]  (move lines in range to addr)
 *                        :[range] d            (delete lines in range)
 *                        where range can be line numbers or marks separated my a ','
 *                        (e.g. :'a,'b move .)
 *                        addr can be '.', '$', or a line number
 *      Set             : Only set option currently supported is IC (make searches
 *                        case insensitive.)
 *      Saving/Exiting  : wq, w!, w [new filename]
 *      Misc            : e (edit new file), r [filename] (insert this file)
 *                        :lineNumber (goto that line)
 *
 * Sled supports an unlimited undo buffer.
 *
 * Hint: use refresh (CTRL-l) if you feel what you see on the screen isn't
 *       what really is there.  This will repaint the screen and correctly set
 *       the cursor in the correct position.
 *
 *
 * Some todo's/notable things NOT supported by sled:
 *      o Support tabs (yeah, yeah, this is a big one, I know, I forgot.)  Right now
 *        I replace any tabs with 5 spaces.  This slows reading a file since I have to
 *        parse each line looking for tabs.  Plus, it changes your file to have spaces
 *        instead of tabs.  This is so very pathetic, but such is life.
 *      o Support ctrl-v control characters.  (i.e. the user types ctrl-v, the next thing
 *        typed such as a ctrl-M should be taken literally.)
 *      o Desperately need to fix all of the completely unnecessary repainting.
 *        For example, the user types 'dd' to delete a line, the WHOLE screen
 *        currently will be repainted.  Really, it should only delete the current line
 *        (remember that a line can span multiple screen lines - there's the rub), and
 *        calculate how many lines not being shown (if any) can be brought up from the
 *        bottom of the screen.
 *  **  o Need to use the bufOut (BufferedOutputStream) for ALL of the prints - control
 *        sequences included, not just in the refresh.
 *      o 'e', 'E', 'W', 'B'
 *      o Support coefficient in 'j', 'k', 'H', 'L'
 *      o Support updating marks after inserts, deletes, etc.
 *      o Deleting into other buffers 
 *      o Support large files by reading only part of the file and file 
 *        paging as needed.  This one would take a total gutting of the
 *        system so I really don't think it will ever happen, mainly because
 *        someday I'd like to investigate this "having a life" thing.
 *      o :g/X/cmd      (Run cmd on all lines that match X.)
 *      o :g/X/s//Y/opt (Global change X to Y)
 *      o :set number   (Show line numbers in far left columns.)
 *      o Support single line editor (ed).
 *    So, so many more...
 *
 *
 * @author Stephen Hess
 * @version 0.01, 03/17/00  ELVIS IS IN THE BUILDING!
 */
public class Sled
{
    SledInputStream in;
    SledPrintStream out;

    //Various definitions
    static final char        UP_ARROW       = 'k';
    static final char        DOWN_ARROW     = 'j';
    static final char        RIGHT_ARROW    = 'l';
    static final char        LEFT_ARROW     = 'h';
    static final char        ESC            = 0x1B;
    static final String      EMPTY          = "";
    static final String      TAB_REPLACE    = "     ";  //See todo list.

    //Possible states for the editor engine
    static final int         FINISHED       = -1;
    static final int         NORMAL         =  0;
    static final int         COMMAND        =  1;
    static final int         INSERT         =  2;
    //Possible substates
    static final int         NONE           =  3;
    static final int         REPLACE        =  4;
    static final int         CHANGING       =  5;
    static final int         DELETING       =  6;
    
    //Common command Messages
    static final String      CMD_INSERT     = "-- INSERT --";
    static final String      CMD_REPLACE    = "-- REPLACE --";
    static final String      CMD_NO_WRITE   = "No write"; //"No write since last change.";
    static final String      CMD_NO_MARK    = "Mark not set";
    static final String      CMD_NOT_FOUND  = "Not found";
    static final String      CMD_NO_EXPR    = "No previous expression";
    static final String      CMD_BAD_RANGE  = "Invalid range";
    static final String      CMD_BAD_ADDR   = "Invalid address";
    
    //Set variables
    private boolean          IC;

    //State and global variables
    private int              state          = NORMAL;
    private int              substate       = NONE;
    private boolean          modified;
    private Vector           copyBuffer;
    private Vector           undoBuffer;
    private ByteBuffer       insertRedoBuff;         //There must be a better way to do this...
    private SledEvent        currentEvent;
    private String           line;
    private int[]            marks;
    private Vector           buffer;                 //Text display buffer
    private ByteBuffer       lineBuffer;
    private String           searchString   = EMPTY;
    private int[]            searchInfo;
    
    private byte[]           displayBuffer;
    
    private String           coefficientStr;         //Hold a number (e.x. 55dd -> 55 is coeff.)

    //DEBUG DOWNSIZE TEST private String           lastReplace;
    private boolean          searchBack;
    private String           searchCMDStr   = EMPTY;

    private int              currentLine    = 0;     //Y
    private int              cursor         = 0;     //X
    private int              lineAtTopOfScr = 0;     //First line displayed on the page
    private int              cursorStart    = 0;     //First column displayed on the page
    
    private int              SCREEN_WIDTH   = 80;
    private int              SCREEN_HEIGHT  = 25;
    
    private RandomAccessFile file;
    private String           fileName;
    
    private boolean          readOnlyMode;
    
    private String           CURRENT_NEWLINE;
    
    public Sled(InputStream in, PrintStream out, String fileName)
    {
        this.out = new SledPrintStream(out);
        this.in  = new SledInputStream(in, out);

        try
        {
            this.fileName  = fileName;
            displayBuffer  = new byte[SCREEN_HEIGHT*(SCREEN_WIDTH+2)+10];
            lineBuffer     = new ByteBuffer(SCREEN_WIDTH);
            copyBuffer     = new Vector(10, 10);
            buffer         = new Vector(SCREEN_HEIGHT*2, SCREEN_HEIGHT);
            undoBuffer     = new Vector(10, 10);
            marks          = new int[52];
            searchInfo     = new int[2];
            insertRedoBuff = new ByteBuffer(SCREEN_WIDTH);
            
            readData();

            runEditor();
        }
        catch(Throwable t)
        {
            out.println(t.toString());
        }
    }
    
    /**
     * Runs the sled editor runtime engine.
     */
    private void runEditor()
    {
        in.setEcho(false);
        
        while (state != FINISHED)
        {
            try
            {
                switch(state)
                {
                    case COMMAND:
                        handleCommandMode(null);
                    break;

                    case INSERT:
                        handleInsertMode(in);
                    break;

                    default:
                        handleNormalMode(in, (char)in.read());
                    break;
                }
            }
            catch(Throwable t)
            {
                repaint(false,t.toString());
                state    = NORMAL;
                substate = NONE;
                in.setEcho(false);
            }
        }

        //Finished, clean up...
        try
        {
            if (file != null)
                file.close();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }

        out.clearDisplay();
        out.setCursorPosition(0,0);
        in.setEcho(true);
    }
    
    /**
     * This handles events when the editor is not inserting or in command
     * mode.  (You get to those modes from this mode.)
     */
    void handleNormalMode(SledInputStream in, char input) throws IOException
    {
        int coefficient;
        
	    //Look for up/down/left/right arrows and translate
	    if ((input == ESC) && ((input = parseControlSequence()) == 0xFF))
		{
			return;
		}
	    
	    //Handle command
	    switch(input)
	    {
		    case 0x02: //PAGE_UP
		    case 0x04: //HALF_PAGE_DOWN
		    case 0x06: //PAGE_DOWN
		    case 0x15: //HALF_PAGE_UP
		        if (input == 0x02)
		            currentLine = Math.max(currentLine-SCREEN_HEIGHT, 0);
		        else if (input == 0x04)
		            currentLine = Math.min(currentLine+SCREEN_HEIGHT/2-2, buffer.size()-1);
		        else if (input == 0x06)
		            currentLine = Math.min(currentLine+SCREEN_HEIGHT-2, buffer.size()-1);
		        else
		            currentLine = Math.max(currentLine-SCREEN_HEIGHT/2, 0);
		        cursor = 0;
		        line = (String)buffer.elementAt(currentLine);
		        gotoCurrentLine();
		    break;
		    
            case 0x0A:
                //DEBUG!!  Man, I don't get this at all!  For some reason, at least
                //on Windows Telnet, the cursor will be moved down one on a \n, even
                //when echo is turned off!  This makes no sense.  This is universe folding
                //in on itself sort of stuff.  Throwing this in for now, should really
                //figure this out later.
                out.moveCursor(SledPrintStream.UP, 1); //DEBUG TEST!
            break;
		    
		    case 0x0C: //Refresh
		        line = (String)buffer.elementAt(currentLine); //DEBUG --> Needed?!?
		        if ((cursor < 0) || (cursor >= line.length()))
		            cursor = 0;
		        if (cursor < cursorStart)
		            cursorStart = cursor;
			    repaint();
		    break;

		    case ESC:   //Escape out of current operation
                state = NORMAL;
                substate = NONE;
                coefficientStr = null;
                in.setEcho(false);
		    break;

		    case '$':   //Goto end of line
		        cursor = (line.length() == 0) ? 0 : line.length() - 1;
			    gotoCurrentLine();
		    break;
		    
/*DEBUG DOWNSIZE TEST
		    case '%':   //Bracket search
		        boolean backwards = false;
		        char bracket = line.charAt(cursor);
		        switch(bracket)
		        {
		            case '(':
		                searchString = ")";
		                backwards = false;
		            break;
		            
		            case ')':
		                searchString = "(";
		                backwards = true;
		            break;
		            
		            case '[':
		                searchString = "]";
		                backwards = false;
		            break;
		            
		            case ']':
		                searchString = "[";
		                backwards = true;
		            break;
		            
		            default:
		                //Ignore
		            return;
		        }
		        
		        search(backwards, true);
		        
		        if (searchInfo[0] != -1)
		        {
		            currentLine = searchInfo[0];
		            cursor = searchInfo[1];
		            gotoCurrentLine();
		        }
		    break;
DEBUG DOWNSIZE TEST*/
		    
/*DEBUG DOWNSIZE TEST
		    case '&':   //Repeat last substitution
		        if (lastReplace == null)
		            repaintCommandMessage(CMD_NO_EXPR, true, false);
		        else
		            handleCommandMode(lastReplace);
		    break;
DEBUG DOWNSIZE TEST*/

		    case '\'':  //Set mark
		        input = (char)in.read();
		        if (Character.isLetter(input))
		        {
		            if (Character.isUpperCase(input))
		                coefficient = (input & 0x0FF) - 65;
		            else
		                coefficient = (input & 0x0FF) - 71;

		            int tLine = marks[coefficient];
		            if ( (tLine == 0) || (tLine >= buffer.size()))
		            {
		                repaintCommandMessage(CMD_NO_MARK, true, false);
		            }
		            else
		            {
		                currentLine = tLine - 1;
		                cursor = 0;
		                line = (String)buffer.elementAt(currentLine);
        		        		    
		                gotoCurrentLine();
                    }
                }
		    break;
		    
		    case '.': //Redo
		        if (currentEvent != null)
		        {
		            SledInputStream bIn;
		            modified = true;
                    switch(currentEvent.type)
                    {
                        case SledEvent.DELETING_CHARS:
                            coefficientStr = Integer.toString(currentEvent.miscInfo);
                            handleNormalMode(in, currentEvent.command);
                        break;
                        
                        case SledEvent.DELETING_LINES:
                            if (buffer.size() > 0)
                            {
                                substate = DELETING;
                                coefficientStr = Integer.toString(currentEvent.miscInfo);
                                handleNormalMode(in, 'd');
                            }
                        break;

                        case SledEvent.DELETING_WORD:
                            substate = DELETING;
                            handleNormalMode(in, 'w');
                        break;
                        
                        case SledEvent.REPLACING_CHAR:
                            if (currentEvent.command == 'r')
                            {
                                insertRedoBuff.setLength(0);
                                insertRedoBuff.insert(currentEvent.miscInfoB);
                                bIn = new SledInputStream(new ByteArrayInputStream(insertRedoBuff.toByteArray()), out);
                            }
                            else
                                bIn = in;
                                
                            handleNormalMode(bIn, currentEvent.command);
                        break;

                        case SledEvent.INSERTING:
                            bIn = new SledInputStream(new ByteArrayInputStream(insertRedoBuff.toByteArray()), out);
                            bIn.setEcho(true);
                            handleNormalMode(bIn, currentEvent.command);
                            while (state != NORMAL)
                                handleInsertMode(bIn);
                            in.setEcho(false);
                        break;
                        
                        case SledEvent.PASTE:
                            handleNormalMode(in, 'p');
                        break;
                        
                        case SledEvent.JOINING_LINES:
                            handleNormalMode(in, 'J');
                        break;
		            }
		        }
		    break;
		    
		    case ':':   //Enter command mode
			    state = COMMAND;
			    in.setEcho(true);
			    out.saveCursor();
			    repaintCommandMessage(":", false, false);
		    break;
		    
		    case '^':   //Goto start of line
			    cursor = findStartOfWord(line, 0);//, false);
			    gotoCurrentLine();
		    break;
		    
		    case '?':   //Seach backwards
		    case '/':   //Search forwards
                lineBuffer.setLength(0);
                char currChar = 0;
                boolean finished = false;
                in.setEcho(true);
                out.saveCursor();   //Save in case the pattern does not exist...
                repaintCommandMessage((input == '?' ? "?" : "/"), false, false);
                while (!finished)
                {
                    currChar = (char)in.read();
                    if ( (currChar != '\r') && (currChar != '\n') && (currChar != ESC) )
                        lineBuffer.insert(currChar);
                    else
                        finished = true;
                }
                in.setEcho(false);
                searchString = (IC ? lineBuffer.toString().toLowerCase() : lineBuffer.toString());
                searchBack = (input == '?');
                search(searchBack, false);
                if (searchInfo[0] == -1)
                {
                    repaintCommandMessage(CMD_NOT_FOUND, false, false);
                    out.restoreCursor();
                }
                else
                {
                    currentLine = searchInfo[0];
                    cursor      = searchInfo[1];
                    line = (String)buffer.elementAt(currentLine);
                    gotoCurrentLine();
                }
		    break;
		    
		    case 'A':   //Start appending to end of line.
			    state = INSERT;
			    //Need to put cursor at end of line...
			    cursor = line.length();
			    if (cursor >= (cursorStart + SCREEN_WIDTH))
			    {
			        //Need to horizontally page
			        cursorStart = cursor;
			        repaint(true, CMD_INSERT);
			    }
			    else
			        repaintCommandMessage(CMD_INSERT, true, true);

			    modified = true;
			    in.setEcho(true);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
			                                 0, ' ', input);
			    currentEvent.changed.addElement(line);
			    insertRedoBuff.setLength(0);
		    break;

		    case 'C' :  //Change from the cursor to the end of the line
		        currentEvent = new SledEvent(SledEvent.INSERTING, 10, currentLine, cursor, 
		                                     0, ' ', input);
		        currentEvent.changed.addElement(line);
		        insertRedoBuff.setLength(0);
		        
			    modified = true;
			    line = line.substring(0, cursor);
			    state = INSERT;
			    repaintCommandMessage(CMD_INSERT, true, true);
			    in.setEcho(true);
			    lineBuffer.setLength(0);
			    lineBuffer.append(line);
			    out.clearToEOL();
		    break;
		    
		    case 'D':   //Delete from cursor to the end of the line
			    if (line.length() > 0)
			    {
		            currentEvent = new SledEvent(SledEvent.DELETING_WORD, 1, currentLine, cursor,
		                                        0, ' ', input);
		            currentEvent.changed.addElement(line);
		            undoBuffer.insertElementAt(currentEvent, 0);

			        modified = true;
			        line = line.substring(0, cursor);
			        buffer.setElementAt(line, currentLine);
			        if (cursor > 0)
			            cursor--;
			        if (cursor == (cursorStart + SCREEN_WIDTH))
			        {
			            cursorStart -= SCREEN_WIDTH;
			            if (cursorStart < 0)
			                cursorStart = 0;
			            repaint();
			        }
			        else
			        {
			            out.clearToEOL();
			            out.moveCursor(SledPrintStream.LEFT, 1);
			        }
			    }
		    break;
		    
		    case 'G':
		        currentLine = buffer.size() - 1;
			    cursor = 0;
			    line = (String)buffer.elementAt(currentLine);
			    gotoCurrentLine();
		    break;
		    
/*DEBUG DOWNSIZE TEST
		    case 'G':
		        if (coefficientStr == null)
		        {
			        currentLine = buffer.size() - 1;
			    }
			    else
			    {
			        int temp = getCoefficient() - 1;
			        if (temp < buffer.size())
			            currentLine = temp;
			        else
			            currentLine = buffer.size() - 1;
			    }
			    cursor = 0;
			    line = (String)buffer.elementAt(currentLine);
			    gotoCurrentLine();
		    break;
		    
		    case 'H':
		        coefficient = getCoefficient();
		        //DEBUG ==> MUST check and use coefficient here!!
		        currentLine = lineAtTopOfScr;
		        line = (String)buffer.elementAt(currentLine);
		        cursor = 0;
		        out.setCursorPosition(0,0);
		    break;
DEBUG DOWNSIZE TEST*/

		    case 'I':
			    modified = true;
			    state = INSERT;
			    //Move the cursor to the beginning of the line.
			    cursor = findStartOfWord(line, 0);//, false);
			    gotoCurrentLine();
			    repaintCommandMessage(CMD_INSERT, false, true);
			    //Page if needed.
			    gotoCurrentLine();
			    in.setEcho(true);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
			                                 0, ' ', input);
			    currentEvent.changed.addElement(line);
			    insertRedoBuff.setLength(0);
		    break;
		    
		    case 'J':
		        //Join the next line with the current one.
		        if (currentLine < (buffer.size() - 1))
		        {
		            String nextLine = (String)buffer.elementAt(currentLine+1);
		            currentEvent = new SledEvent(SledEvent.JOINING_LINES, 2, currentLine, cursor,
		                                         0, ' ', input);
			        currentEvent.changed.addElement(line);
			        currentEvent.changed.addElement(nextLine);
			        undoBuffer.insertElementAt(currentEvent, 0);

		            line += nextLine;
		            buffer.setElementAt(line, currentLine);
		            buffer.removeElementAt(currentLine+1);
		            //DEBUG ==> Don't really need to repaint here!
		            repaint();
		        }
		    break;
		    
/*DEBUG DOWNSIZE TEST
		    case 'L':
		        coefficient = getCoefficient();
		        //DEBUG ==> MUST check and use coefficient here!!
		        currentLine = Math.min(buffer.size()-1, lineAtTopOfScr+SCREEN_HEIGHT-2);
		        line = (String)buffer.elementAt(currentLine);
		        cursor = 0;
		        gotoCurrentLine();
		    break;
		    
		    case 'M':
		        currentLine += (SCREEN_HEIGHT-2-lineAtTopOfScr)/2;
		        line = (String)buffer.elementAt(currentLine);
		        cursor = 0;
		        gotoCurrentLine();
		    break;
DEBUG DOWNSIZE TEST*/

		    case 'O':
                //Clear the command line
                out.setCursorPosition(SCREEN_HEIGHT-1, 0);
                out.clearToEOL();
                out.deleteLines(SCREEN_HEIGHT-2, 1);
                out.insertLines(currentLine, 1);
			    line = EMPTY;
			    cursor = 0;
			    modified = true;
			    buffer.insertElementAt(line, currentLine);
                lineBuffer.setLength(0);
			    state = INSERT;
			    repaintCommandMessage(CMD_INSERT, true, true);
			    in.setEcho(true);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, -1, currentLine, cursor,
			                                 0, ' ', input);
			    insertRedoBuff.setLength(0);
		    break; //ELVIS??
		    
		    case 'R':
			    in.setEcho(true);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
			                                 0, ' ', input);
			    currentEvent.changed.addElement(line);
			    lineBuffer.setLength(0);
			    lineBuffer.append(line);
			    insertRedoBuff.setLength(0);
			    
			    modified = true;
			    substate = REPLACE;
			    state = INSERT;
			    repaintCommandMessage(CMD_REPLACE, true, true);
		    break;
		    
/*DEBUG DOWNSIZE TEST
		    case 'S':
		        //Substitute the entire line
		        currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
		                                     0, ' ', input);
			    currentEvent.changed.addElement(line);
		        line = EMPTY;
		        cursor = 0;
                buffer.setElementAt(line, currentLine);
                state = INSERT;
                in.setEcho(true);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
                out.setCursorPosition(cursor - cursorStart, 0);
                out.clearToEOL();
		    break; //ELVIS??
DEBUG DOWNSIZE TEST*/

		    case 'Z':
		        input = (char)in.read();
		        if (input == 'Z')
		        {
		            if (modified)
					    saveFile();
					    
		            state = FINISHED;
		        }
		    break;

		    case 'a':
			    cursor++;
			    state = INSERT;
			    modified = true;
			    in.setEcho(true);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
                //Need to put the cursor at the end of the line...
                if (cursor >= (cursorStart + SCREEN_WIDTH))
                {
                    cursorStart = cursor;
                    repaint(true, CMD_INSERT);
                }
			    else
			    {
			        out.moveCursor(SledPrintStream.RIGHT, 1);
			        repaintCommandMessage(CMD_INSERT, true, true);
			    }
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
			                                 0, ' ', input);
			    currentEvent.changed.addElement(line);
			    insertRedoBuff.setLength(0);
		    break;

            case 'b': //ELVIS Savings of 19
                //Goto prev word
                coefficient = getCoefficient();
                int lastWord = cursor;
                while (coefficient > 0)
                {
                    lastWord = findPrevWord(line, lastWord);
                    if (lastWord == -1)
                    {
                        if (currentLine == 0)
                        {
                            //EOF, stop the search
                            break;
                        }
                        else
                        {
                            //Grab the next line.
                            currentLine--;
                            line = (String)buffer.elementAt(currentLine);
                            cursor = 0;
                            cursorStart = -1;
                            lastWord = 0;
                        }
                    }
                    coefficient--;
                }
                
                if (lastWord >= 0)
                {
                    cursor = lastWord;
                    //Now, move to that new position
                    if ((cursorStart == -1) || (cursor < cursorStart))
                    {
                        cursorStart = cursor - 10;
                        if (cursorStart < 0)
                            cursorStart = 0;
                        repaint();
                    }
                    else
                    {
                        out.setCursorPosition(currentLine - lineAtTopOfScr, cursor - cursorStart);
                    }
                }
            break;

		    case 'c':
			    substate = CHANGING;
		    break;
		    
		    case 'd':
		        //Delete lines
			    if (substate == DELETING)
			    {
				    modified = true;
				    copyBuffer.removeAllElements();
				    
			        coefficient = getCoefficient();
				    
				    if ((coefficient+currentLine) > (buffer.size()-1))
				        coefficient = buffer.size() - currentLine ;

                    //Wow, this is going to be **SERIOUSLY** slow!!
                    //Must fix this.  Hmm.
                    String temp;
                    currentEvent = new SledEvent(SledEvent.DELETING_LINES, coefficient, currentLine, cursor,
                                                 coefficient, ' ', input);
				    for (int i=0; i<coefficient; i++)
				    {
				        temp = (String)buffer.elementAt(currentLine);
				        copyBuffer.addElement(temp);
				        currentEvent.changed.addElement(temp);
				        buffer.removeElementAt(currentLine);
				    }
				    undoBuffer.insertElementAt(currentEvent, 0);
				    
				    currentLine--;
				    if (currentLine < 0)
				        currentLine = 0;
				    
				    line = (buffer.size() == 0) ? EMPTY : (String)buffer.elementAt(currentLine);
				    
				    cursor = (line.length() == 0) ? 0 : line.length() - 1;

				    substate = NONE;

				    repaint(); //DEBUG ==> only after current line?				    
			    }
			    else
			    {
				    substate = DELETING;
			    }
		    break;
		    
            case 0x08:  //ELVIS Savings of -2
		    case LEFT_ARROW:
		        coefficient = getCoefficient();
		        if ((cursor-coefficient) >= 0)
		        {
		            cursor -= coefficient;
		            if (cursor < cursorStart)
		            {
		                cursorStart = cursor - 10;
		                if (cursorStart < 0)
		                    cursorStart = 0;
		                repaint();
		            }
		            else
		            {
		                out.moveCursor(SledPrintStream.LEFT, coefficient);
		            }
                }
            break;
		    
		    case 'i':
			    modified = true;
			    state = INSERT;
			    repaintCommandMessage(CMD_INSERT, true, true);
			    in.setEcho(true);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
			                                 0, ' ', input);
			    currentEvent.changed.addElement(line);
			    insertRedoBuff.setLength(0);
		    break;
		    
		    case 0x0d:  //ELVIS Savings of 26
		    case DOWN_ARROW:
		        coefficient = getCoefficient();
                if (buffer.size() > (currentLine+coefficient))
                {
                    currentLine += coefficient;
                    line = (String)buffer.elementAt(currentLine);
                    int normalized = currentLine - lineAtTopOfScr;  //Screen line we are on
                    
                    //We moved to a new line, adjust the cursor back if needed.
                    if (cursor >= line.length())
                    {
                        cursor = line.length() - 1;
                        if (cursor < 0)
                            cursor = 0;
                    }
                    
                    //Now, move the cursor, scrolling screen if necessary.
                    if (coefficient >= SCREEN_HEIGHT)
                    {
                        //We are scrolling more than what is currently displayed.
                        //Just give up and repaint...
                        lineAtTopOfScr = currentLine;
                        repaint();
                        return;
                    }
                    else if (cursor < cursorStart)
                    {
                        //Horizontal page needed.
                        cursorStart = cursor; //Yes?
                        repaint();
                        return;
                    }
                    else if (normalized > (SCREEN_HEIGHT-2))
                    {
                        //The screen needs to scroll a little.
                        int offScr = normalized - (SCREEN_HEIGHT - 2);
                        lineAtTopOfScr += offScr;
                        //Delete the command line.
                        out.setCursorPosition(SCREEN_HEIGHT-1, 0);
                        out.clearToEOL();
                        //Now, delete the first lines...
                        out.deleteLines(0, offScr);
                        out.setCursorPosition(SCREEN_HEIGHT-1-offScr, 0);

                        byte[] currBytes;
                        for (int i=0; i<offScr; i++)
                        {
                            currBytes = ((String)buffer.elementAt(currentLine+i)).getBytes();
                            if (currBytes.length > cursorStart)
                                out.write(currBytes, cursorStart, Math.min(currBytes.length-cursorStart, SCREEN_WIDTH));
                            out.write('\r');
                            out.write('\n');
                        }
                        normalized -= offScr;
                    }
                    
                    out.setCursorPosition(normalized, (cursor-cursorStart) % SCREEN_WIDTH);
                }
		    break;
		    
		    case UP_ARROW:  //ELVIS Savings of 10
		        coefficient = getCoefficient();
                if ((currentLine-coefficient) >= 0)
                {
                    currentLine -= coefficient;
                    line = (String)buffer.elementAt(currentLine);
                    int normalized = currentLine - lineAtTopOfScr;  //Screen line we are on
                    
                    //We moved to a new line, adjust the cursor back if needed.
                    if (cursor >= line.length())
                    {
                        cursor = line.length() - 1;
                        if (cursor < 0)
                            cursor = 0;
                    }
                    
                    //Now, move the cursor, scrolling screen if necessary.
                    if (coefficient >= SCREEN_HEIGHT)
                    {
                        //We are scrolling more than what is currently displayed.
                        //Just give up and repaint...
                        lineAtTopOfScr = currentLine;
                        repaint();
                        return;
                    }
                    else if (cursor < cursorStart)
                    {
                        //Horizontal page needed.
                        cursorStart = cursor; //Yes?
                        repaint();
                        return;
                    }
                    else if (currentLine < lineAtTopOfScr)
                    {
                        //The screen needs to scroll a little.
                        lineAtTopOfScr = currentLine;
                        //Delete the command line.
                        out.setCursorPosition(SCREEN_HEIGHT-1, 0);
                        out.clearToEOL();
                        //Now, delete the last lines...
                        normalized = Math.abs(normalized);
                        out.deleteLines(SCREEN_HEIGHT-2-normalized, normalized);
                        out.insertLines(0, normalized);
                        
                        out.setCursorPosition(0, 0);

                        byte[] currBytes;
                        for (int i=0; i<normalized; i++)
                        {
                            currBytes = ((String)buffer.elementAt(currentLine+i)).getBytes();
                            if (currBytes.length > cursorStart)
                                out.write(currBytes, cursorStart, Math.min(currBytes.length-cursorStart, SCREEN_WIDTH));
                            out.write('\r');
                            out.write('\n');
                        }
                        normalized = 0;
                    }
                    
                    out.setCursorPosition(normalized, (cursor-cursorStart) % SCREEN_WIDTH);
                }
		    break;
		
            case ' ':
		    case RIGHT_ARROW: //ELVIS Savings of -1
		        coefficient = getCoefficient();
		        if (line.length() > (cursor + coefficient))
		        {
		            cursor += coefficient;
		            if (cursor >= (cursorStart + SCREEN_WIDTH))
		            {
		                cursorStart = cursor - 10;
		                if (cursorStart < 0)
		                    cursorStart = 0;
		                repaint();
		            }
		            else
		            {
		                out.moveCursor(SledPrintStream.RIGHT, coefficient);
		            }
                }
            break;
            
		    case 'm':
		        input = (char)in.read();
		        if (Character.isLetter(input))
		        {
		            if (Character.isUpperCase(input))
		                coefficient = (input & 0x0FF) - 65;
		            else
		                coefficient = (input & 0x0FF) - 71;
		            marks[coefficient] = currentLine + 1;
		        }
		    break;
		    
		    case 'N':
		    case 'n':
		        if (searchString.length() == 0)
		        {
		            repaintCommandMessage(CMD_NO_EXPR, true, false);
		            break;
		        }
		        out.saveCursor();
		        repaintCommandMessage(EMPTY, false, false);
                boolean backwards;
                if (input == 'N')
                {
                    backwards = !searchBack;
                }
                else
                {
                    backwards = searchBack;
                }   
                if (!backwards)
                    cursor++;
                else
                    cursor--;
                search(backwards, false);
                if (searchInfo[0] == -1)
                {
                    //Reset the cursor
                    if (!backwards)
                        cursor--;
                    else
                        cursor++;
                    repaintCommandMessage(CMD_NOT_FOUND, false, false);
                    out.restoreCursor();
                }
                else
                {
                    currentLine = searchInfo[0];
                    cursor      = searchInfo[1];
                    line        = (String)buffer.elementAt(currentLine);
                    gotoCurrentLine();
                    if (searchCMDStr.length() != 0)
                    {
                        //This is a hack.  Since gotoCurrentLine tends to repaint
                        //the screen, if a message occurred during search (hit end, hit
                        //top), we'll display it again.  Yes, yes, this is inefficient.
                        repaintCommandMessage(searchCMDStr, true, false);
                        searchCMDStr = EMPTY;
                    }
                }
		    break;
		    
		    case 'o':
			    currentLine++;
			    modified = true;
			    line = EMPTY;
			    cursor = 0;
			    buffer.insertElementAt(line, currentLine);
                lineBuffer.setLength(0);
                lineBuffer.append(line);
			    state = INSERT;
			    in.setEcho(true);
			    
			    currentEvent = new SledEvent(SledEvent.INSERTING, -1, currentLine, cursor,
			                                 0, ' ', input);
			    insertRedoBuff.setLength(0);

				//Now, open up a new line.  If we are on the last line of the page,
				//we need to only delete the first line.  If in the middle of the screen,
				//we also need to open a new one under us.
				out.saveCursor();
				out.setCursorPosition(SCREEN_HEIGHT-1, 0);
				out.clearToEOL();
				if ((currentLine-lineAtTopOfScr) != (SCREEN_HEIGHT-1))
				{
				    //In the middle somewhere
				    out.deleteLines(SCREEN_HEIGHT-2, 1);
				    out.restoreCursor();
				    out.moveCursor(SledPrintStream.DOWN, 1);
				    out.insertLines(currentLine, 1);
				}
				else
				{
				    //At the last line.
				    lineAtTopOfScr++;
				    out.deleteLines(0, 1);
				    out.setCursorPosition(SCREEN_HEIGHT-2, 0);
				}
		    break;
		    
		    case 'p':
		        currentEvent = new SledEvent(SledEvent.PASTE, -1, currentLine+1, cursor,
		                                     copyBuffer.size(), ' ', input);
		        undoBuffer.insertElementAt(currentEvent, 0);

		        pasteText(copyBuffer, false);
		    break;
		    
		    case 'r':
				in.setEcho(true);
				input = (char)in.read();
			    in.setEcho(false);
			    
			    if (input != ESC)
			    {
			        currentEvent = new SledEvent(SledEvent.REPLACING_CHAR, 1, currentLine, cursor,
			                                     0, input, 'r');
			        currentEvent.changed.addElement(line);
			        undoBuffer.insertElementAt(currentEvent, 0);

				    modified = true;
				    line = line.substring(0,cursor) + input + line.substring(cursor+1);
				    buffer.setElementAt(line, currentLine);
				    if (((cursor+1) % SCREEN_WIDTH) == 0)
				    {
				        out.moveCursor(SledPrintStream.UP, 1);
				        out.moveCursor(SledPrintStream.RIGHT, SCREEN_WIDTH);
				    }
				    else
				    {
				        out.moveCursor(SledPrintStream.LEFT, 1);
				    }
			    }
		    break;
			
/*DEBUG DOWNSIZE TEST
		    case 's':
				in.setEcho(true);
				input = (char)in.read();
			    in.setEcho(false);
			    if (input != ESC)
			    {
			        modified = true;
			        state = INSERT;
			        repaintCommandMessage(CMD_INSERT, true, true);
			        in.setEcho(true);
    			    
    			    currentEvent = new SledEvent(SledEvent.INSERTING, 1, currentLine, cursor,
    			                                 0, input, 's');
			        currentEvent.changed.addElement(line);
			        insertRedoBuff.setLength(0);
			        insertRedoBuff.insert(input);

				    line = line.substring(0,cursor) + input + line.substring(cursor+1);
			        lineBuffer.setLength(0);
			        lineBuffer.append(line);
				    cursor++;
			    }
		    break;
DEBUG DOWNSIZE TEST*/

		    case 'u':
		        if (undoBuffer.size() != 0)
		        {
		            SledEvent event = (SledEvent)undoBuffer.elementAt(0);
		            
                    switch(event.type)
                    {
                        case SledEvent.DELETING_WORD:
                        case SledEvent.DELETING_CHARS:
                        case SledEvent.REPLACING_CHAR:
		                    currentLine = event.startLine;
		                    line = (String)event.changed.elementAt(0);
		                    buffer.setElementAt(line,event.startLine);
		                    
		                    if ((cursor >= line.length()) && (line.length() != 0))
		                    {
		                        cursor = line.length()-1;
		                    }
		                    
		                    repaint();
                        break;

                        case SledEvent.DELETING_LINES:
		                    currentLine = event.startLine;    		                    
		                    pasteText(event.changed, true);
                        break;
                        
                        case SledEvent.INSERTING:
                            currentLine = event.startLine;
                            cursor = event.startCursor;
		                    for (int i=0; i<event.miscInfo; i++)
		                    {
		                        buffer.removeElementAt(i+currentLine+1);
		                    }
		                    
		                    //'o' and 'O' events won't have an original line to restore
                            if (event.changed != null)
		                        buffer.setElementAt(event.changed.elementAt(0),event.startLine);
		                    else
		                        buffer.removeElementAt(currentLine);
		                        
		                    line = (String)buffer.elementAt(currentLine);

		                    if ((cursor >= line.length()) && (line.length() != 0))
		                    {
		                        cursor = line.length()-1;
		                    }
		                        
		                    repaint();
                        break;
                        
                        case SledEvent.PASTE:
                            currentLine = event.startLine-1;
		                    for (int i=0; i<event.miscInfo; i++)
		                    {
		                        buffer.removeElementAt(event.startLine);
		                    }

		                    line = (String)buffer.elementAt(currentLine);

		                    if ((cursor >= line.length()) && (line.length() != 0))
		                    {
		                        cursor = line.length()-1;
		                    }

                            if (currentEvent.startCursor == -1)
                            {
                                //startCursor will be set to -1 if this paste is
                                //part of a :[range] move [addr] command.
                                //So, we've undone the paste, now need to also
                                //undo the delete.
                                undoBuffer.removeElementAt(0);
                                handleNormalMode(null, 'u');
                                return;
                            }
                            else
		                        repaint();
                        break;
                        
                        case SledEvent.JOINING_LINES:
		                    currentLine = event.startLine;
		                    line = (String)event.changed.elementAt(0);
		                    buffer.setElementAt(line, currentLine);
		                    buffer.insertElementAt(event.changed.elementAt(1), currentLine+1);
		                    repaint(); //DEBUG ==> Should do only after current line
                        break;
                        
                        case SledEvent.INSERTING_FILE:
		                    currentLine = event.startLine;
		                    int sizeDiff = buffer.size() - event.miscInfo;
		                    for (int i = 0; i<sizeDiff; i++)
		                        buffer.removeElementAt(currentLine);
		                    line = (String)buffer.elementAt(--currentLine);
		                    cursor = 0;
		                    repaint();
                        break;

                        case SledEvent.SEARCH_REPLACE:
                            for (int i=0; i<event.changed.size(); i++)
                            {
                                buffer.setElementAt(event.changed.elementAt(i), ((Integer)event.misc.elementAt(i)).intValue());
                            }
                            currentLine = event.startLine;
                            cursor = event.miscInfo;
                            line = (String)buffer.elementAt(currentLine);
                            repaint();
                        break;
		            }
		            undoBuffer.removeElementAt(0);
		        }
		    break;
		    
		    case 'w': //Elvis Savings of 41
			    if ((substate == CHANGING) || (substate == DELETING))
			    {
                    String backup = line;
			        int nextWord = findNextWord(line, cursor);
			        line = line.substring(0, cursor);
			        if (nextWord != -1)
			        {
			            line += backup.substring(nextWord);
			        }
			            
                    buffer.setElementAt(line, currentLine);
                    
                    currentEvent = new SledEvent(0, 10, currentLine, cursor,
                                                 0, ' ', input);
			        currentEvent.changed.addElement(backup);

                    out.saveCursor();
                    out.clearToEOL();
                    out.write(line.getBytes(), cursor, Math.min(line.length()-cursor, SCREEN_WIDTH-(cursor%SCREEN_WIDTH)));
                    out.restoreCursor();
				    if (substate == CHANGING)
				    {
				        currentEvent.type = SledEvent.INSERTING;
                        state = INSERT;
                        in.setEcho(true);
                        lineBuffer.setLength(0);
                        lineBuffer.append(line);
                        repaintCommandMessage(CMD_INSERT, true, true);
                    }
                    else
                    {
			            currentEvent.type = SledEvent.DELETING_WORD;
                        undoBuffer.insertElementAt(currentEvent, 0);
                    }
                    
                    modified = true;
                    substate = NONE;
                }
                else
                {
                    //Goto next word
                    coefficient = getCoefficient();
                    int nextWord = cursor;
                    while (coefficient > 0)
                    {
                        nextWord = findNextWord(line, nextWord);
                        if (nextWord == -1)
                        {
                            if (currentLine >= (buffer.size() - 1))
                            {
                                //EOF, stop the search
                                break;
                            }
                            else
                            {
                                //Grab the next line.
                                currentLine++;
                                line = (String)buffer.elementAt(currentLine);
                                cursor = 0;
                                cursorStart = -1;
                                nextWord = 0;
                            }
                        }
                        coefficient--;
                    }
                    
                    if (nextWord >= 0)
                    {
                        cursor = nextWord;
                        //Now, move to that new position
                        if ((cursorStart == -1) || (cursor >= (cursorStart+SCREEN_WIDTH)))
                        {
                            cursorStart = cursor - 10;
                            if (cursorStart < 0)
                                cursorStart = 0;
                            repaint();
                        }
                        else
                        {
                            out.setCursorPosition(currentLine - lineAtTopOfScr, cursor - cursorStart);
                        }
                    }
                }
		    break;

		    case 'x':
		        //Delete the character under the  cursor
		        //Wow, this'll all be slow.
		        //First handle undo
		        currentEvent = new SledEvent(SledEvent.DELETING_CHARS, 1, currentLine, cursor,
		                                     0, ' ', input);
		        currentEvent.changed.addElement(line);
		        undoBuffer.insertElementAt(currentEvent, 0);
		            
		        //Now handle deletion
		        coefficient = getCoefficient();
		        if ((coefficient+cursor) > line.length())
		            coefficient = line.length() - cursor;
		        currentEvent.startCursor = coefficient;
		        line = line.substring(0, cursor) + line.substring(cursor+coefficient);
		        buffer.setElementAt(line, currentLine);
		        if ((cursor >= line.length()) && (line.length() != 0))
		        {
		            cursor = line.length()-1;
		        }
		        //DEBUG ==> Sigh, ok, I'm cheating yet again.  Shouldn't need to
		        //repaint the whole screen, just the current line, but life's rough.
		        repaint();
		    break;

		    case 'y':
			    if (((char)in.read()) == 'y')
			    {
			        //Again, really slow...
			        coefficient = getCoefficient();
				    copyBuffer.removeAllElements();
				    if ((coefficient+currentLine) > buffer.size())
				        coefficient = buffer.size() - currentLine;
				    for (int i=0; i<coefficient; i++)
				    {
				        copyBuffer.addElement(buffer.elementAt(i+currentLine));
				    }
				    if (coefficient > 1)
				    {
				        repaintCommandMessage(coefficient + " lines yanked", true, false);
				    }
				    else
				    {
				        repaintCommandMessage(EMPTY, true, false);
				    }
			    }
		    break;

/*DEBUG DOWNSIZE TEST
		    case '~':
		        if ((line.length() > 0) && (Character.isLetter(line.charAt(cursor))))
		        {
		            char curr = line.charAt(cursor);
		            char changed;
		            
		            if (Character.isUpperCase(curr))
		                changed = Character.toLowerCase(curr);
		            else
		                changed = Character.toUpperCase(curr);
		                
		            currentEvent = new SledEvent(SledEvent.REPLACING_CHAR, 1, currentLine, cursor,
		                                         0, ' ', input);
			        currentEvent.changed.addElement(line);
			        undoBuffer.insertElementAt(currentEvent, 0);

				    modified = true;
				    line = line.substring(0,cursor) + changed + line.substring(cursor+1);
				    buffer.setElementAt(line, currentLine);
				    out.print(changed);
				    if (((cursor+1) % SCREEN_WIDTH) == 0)
				    {
				        out.moveCursor(SledPrintStream.UP, 1);
				        out.moveCursor(SledPrintStream.RIGHT, SCREEN_WIDTH);
				    }
				    else
				    {
				        out.moveCursor(SledPrintStream.LEFT, 1);
				    }
		        }
		    break;
DEBUG DOWNSIZE TEST*/
		    
		    default:
		        if (Character.isDigit(input))
		        {
		            if (coefficientStr == null)
		                coefficientStr = EMPTY + input;
		            else
		                coefficientStr += input;
		        }
		        else
		        {
		            substate = NONE;
			        repaintCommandMessage(("Unknown cmd: " + input), true, false);
			    }
		    break;
	    }
    }
    
    /**
     * Handles the mode in which the user is actively entering text
     * to the currently open file.  This includes inserting ('i','I'),
     * appending ('a', 'A'), and replacing ('R').  The new text is kept
     * in a buffer until the user ends the insert session by hitting
     * ESC, then added to the buffer as part of the file.
     * ELVIS
     */
    void handleInsertMode(SledInputStream tin) throws IOException
    {
        if (readOnlyMode)
        {
            repaintCommandMessage("Warning: changing a readonly file.", true, true);
        }
		char input = (char)tin.read();
		if (input == '\t')
		{
		    //DEBUG ==> Oh, man, you just had to enter a tab, didn't you?  DIDN'T YOU?!?  Well,
		    //I don't handle it yet.  I threw this in to cover.  I replace tabs with a
		    //predetermined number of spaces.  Yeah, yeah, need to fix.  See the todo list for
		    //more info.  In the meantime, just don't enter tabs, do you see how much time
		    //and memory you are gonna burn?
            SledInputStream tabIn = new SledInputStream(new ByteArrayInputStream(TAB_REPLACE.getBytes()), out);
            tabIn.setEcho(true);
            for (int i=0; i<TAB_REPLACE.length(); i++)
            {
		        handleInsertMode(tabIn);
		    }
		    return;
		}
		else
            insertRedoBuff.insert(input);
		
		if (input != ESC)
		{
			if ((input == '\r') || (input == '\n') )
			{
			    //Handle new line...
			    if (input == '\r')
			    {
			        if (tin.available() > 0)
			            tin.read();  //Take care of '\n' ==> Is this the best way to handle this?
			    }
			    currentEvent.miscInfo++;    //Keep up with the number of new lines in this insert
                //User just hit enter, so set the line he was editing in the buffer.
                String fullLine = lineBuffer.toString();
				buffer.setElementAt(fullLine.substring(0,cursor), currentLine);
			    //Now, set up the next line for continued editing...
				line = fullLine.substring(cursor);
				cursor = 0;
				currentLine++;
				lineBuffer.setLength(0);
				lineBuffer.append(line);
				if (currentLine == buffer.size())
                    buffer.addElement(line);
                else
				    buffer.insertElementAt(line, currentLine);
				if (cursorStart != 0)
				{
				    cursorStart = 0;
				    repaint();
				}
				else
				{
				    //Now, open up a new line.  If we are on the last line of the page,
				    //we need to only delete the first line.  If in the middle of the screen,
				    //we also need to open a new one under us.
				    out.saveCursor();
				    out.setCursorPosition(SCREEN_HEIGHT-1, 0);
				    out.clearToEOL();
				    if ((currentLine-lineAtTopOfScr) != (SCREEN_HEIGHT-1))
				    {
				        //In the middle somewhere
				        out.deleteLines(SCREEN_HEIGHT-2, 1);
				        out.restoreCursor();
				        out.moveCursor(SledPrintStream.DOWN, 1);
				        out.insertLines(currentLine, 1);
				    }
				    else
				    {
				        //At the last line.
				        lineAtTopOfScr++;
				        out.deleteLines(0, 1);
				        out.setCursorPosition(SCREEN_HEIGHT-2, 0);
				    }
				}
			}
			else if ((input == 0x08) && (cursor > currentEvent.startCursor))
			{
			    //Handle backspace event.
			    lineBuffer.backspace();
			    cursor--;
			    if ( cursor < cursorStart )
			    {
			        cursorStart = cursor - 10;
			        if (cursorStart < 0)
			            cursorStart = 0;
			        repaint();
			    }
			    else
			    {
			        out.write(' ');
			        out.write(0x08);
			    }
			}
			else
			{
			    //Normal character input, add (or insert) into line...
				if (cursor >= lineBuffer.length())
				{
                    //Appending...
					lineBuffer.append(input);
					cursor++;
				}
				else if (substate == REPLACE)
				{
				    //Replacing
					lineBuffer.setCharAt(cursor++, input);
				}
				else
				{
				    //Inserting...
				    //Need to print out the rest of the line since this is
				    //insert mode.
			        //Great wobbling plates of custard, this is going to be
			        //slow.  Should I think about buffering this up somehow?
				    out.saveCursor();
				    lineBuffer.writeTo(out, cursor, Math.min(lineBuffer.length()-cursor, SCREEN_WIDTH-(cursor-cursorStart)-1));
				    out.restoreCursor();
					lineBuffer.insert(cursor++, input);
				}
				
				if (cursor == (cursorStart + SCREEN_WIDTH))
				{
				    //Need to horizontal page...
				    cursorStart = cursor;
				    //DEBUG ==> This could be handled better.  Really only need to do a
				    //lineBuffer.writeTo after the repaint to handle printing out the
				    //current line that we are editing, but such is life.
				    buffer.setElementAt(lineBuffer.toString(), currentLine);
				    repaint();
				}
			}
		}
		else
		{
		    //ESC hit, user is done inserting, cleanup...
            line = lineBuffer.toString();

            if (currentLine >= buffer.size())
                buffer.addElement(line);
            else
                buffer.setElementAt(line, currentLine);
            state = NORMAL;
            substate = NONE;
            in.setEcho(false);
            if (cursor != 0)
            {
                cursor--;
                out.moveCursor(SledPrintStream.LEFT, 1);
            }
            undoBuffer.insertElementAt(currentEvent, 0);
            repaintCommandMessage(EMPTY, true, false);
            System.gc();
		}
    }
    
    /**
     * Handles commands.  (i.e. the user types ':' while in normal
     * mode.)  User can set options, save, exit, etc. from
     * command mode.
     */
    void handleCommandMode(String command) throws IOException
    {
        lineBuffer.setLength(0);
        char    currChar   = 0;
        boolean fileGlobal = false;
        
        String newLine;
        if (command == null)
        {
            //Read the line the user enters in...
            while ( (currChar != '\r') && (currChar != '\n') )
            {
                currChar = (char)in.read();
                if ((currChar == ESC) || ((currChar == 0x08) && lineBuffer.length() == 0))
                {
                    //Abort the command entry if the user hits ESC, or backspaces past the
                    //start of the currently entered command
			        state = NORMAL;
			        in.setEcho(false);
			        out.restoreCursor();
			        return;
                }
                else if ( (currChar != '\r') && (currChar != '\n') )
                    lineBuffer.insert(currChar);
            }
            newLine = lineBuffer.toString();
        }
        else
            newLine = command;
        		
        int newLineLength = newLine.length();
        if (newLineLength == 0)
            return;
		
        //First, look to see if this is a :set request
        if (newLine.startsWith("set "))
        {
            boolean set = true;
            newLine = newLine.substring(4);
            //Not very optimized...
            if (newLine.startsWith("no "))
            {
                set = false;
                newLine = newLine.substring(2);
            }
            
            //Available options:
            if (newLine.equals("ic"))
                IC = set;
            //else
            //{
            //    repaintCommandMessage("Unsupported.", false, false);
            //}
        }
        else
        {
		    //Not a set command, so parse it and handle accordingly...
            int lowRange  = -1;
            int highRange = -1;
            
            //Hack support for :x
            if (newLine.charAt(0) == 'x')
            {
                newLine = "wq";
                newLineLength = 2;
            }
            
		    for (int i=0; i<newLineLength; i++)
		    {
			    switch(newLine.charAt(i))
			    {
                    case '.':
                        if (lowRange == -1)
                            lowRange = currentLine;
                        else
                            highRange = currentLine;
                    break;
                    
                    case '$':
                        if (lowRange == -1)
                            lowRange = buffer.size() - 1;
                        else
                            highRange = buffer.size() - 1;
                    break;
                    
				    case '%':
				        fileGlobal = true;
				    break;
                    
                    case ' ':
                        //Skip past the space
                    break;
                    
                    case '\'':
                        //Look for a range setting using marks such as:
                        //:'a,'b move .
                        if (i+1 < newLineLength)
                        {
                            char mark = newLine.charAt(i+1);
                            int num;
		                    if (Character.isUpperCase(mark))
		                        num = marks[(mark & 0x0FF) - 65];
		                    else
		                        num = marks[(mark & 0x0FF) - 71];
		                    if ((num == 0) || (num >= buffer.size()))
		                    {
		                        repaintCommandMessage(CMD_NO_MARK, false, false);
		                        i += newLineLength;
		                        break;
		                    }
		                    num--;
                            if (lowRange == -1)
                            {
                                lowRange = num;
                                if ( ((i+2) < newLineLength) &&
                                     (newLine.charAt(i+2) == ',') )
                                {
                                    //Looks good so far...
                                    i += 2;    //Skip past comma and mark
                                }
                                else
                                {
                                    repaintCommandMessage(CMD_BAD_RANGE, false, false);
                                    i += newLineLength;
                                }
                            }
                            else
                            {
                                highRange = num;
                                i++; //Skip past mark;
                            }
                        }
                    break;
                    
                    case 'd':
                        //Delete lines
                        if ((lowRange == -1) || (highRange == -1) || (lowRange > highRange) ||
                            (lowRange < 0) || (highRange >= buffer.size()))
                        {
                            repaintCommandMessage(CMD_BAD_RANGE, false, false);
                            i += newLineLength;
                            break;
                        }
                        else
                        {
                            currentLine = lowRange;
                            coefficientStr = Integer.toString(highRange - lowRange + 1);
                            substate = DELETING;
                            handleNormalMode(null, 'd');
                        }
                    break;

				    case 'e':
				        if (modified)
				        {
					        repaintCommandMessage(CMD_NO_WRITE, false, false);
					    }
					    else
					    {
					        out.setCursorPosition(0,0);
					        out.saveCursor();
				            fileName = newLine.substring(2);
					        readData();
					    }
				        i+=newLineLength; //Kick out of the loop
				    break;
				    
				    case 'r':
				        insertFile(newLine.substring(2));
				        out.saveCursor();
				        i += newLineLength; //Kick out of the loop
				    break;
				    
                    case 'm':
                        //Move lines
                        //Look for address...
                        int addrPos = newLine.indexOf(' ', i);
                        if (addrPos == -1)
                        {
                            repaintCommandMessage(CMD_BAD_ADDR, false, false);
                            i += newLineLength;
                        }
                        else if ((lowRange == -1) || (highRange == -1) || (lowRange > highRange) ||
                                 (lowRange < 0) || (highRange >= buffer.size()))
                        {
                            repaintCommandMessage(CMD_BAD_RANGE, false, false);
                            i += newLineLength;
                        }
                        else
                        {
                            String temp = newLine.substring(addrPos+1);
                            addrPos = 0;
                            if (temp.length() == 1)
                            {
                                if (temp.charAt(0) == '.')
                                    addrPos = currentLine;
                                else if (temp.charAt(0) == '$')
                                    addrPos = buffer.size() - 1;
                                else
                                {
                                    repaintCommandMessage(CMD_BAD_ADDR, false, false);
                                    i += newLineLength;
                                    break;
                                }
                            }
                            else
                            {
                                try
                                {
                                    addrPos = Integer.parseInt(temp) - 1;
                                }
                                catch(NumberFormatException nfe)
                                {
                                    addrPos = -1;
                                }
                                if (((addrPos < 0)  || (addrPos >= buffer.size())) ||
                                    ((addrPos > lowRange) && (addrPos < highRange)))
                                {
                                    repaintCommandMessage(CMD_BAD_ADDR, false, false);
                                    i += newLineLength;
                                    break;
                                }
                            }
                            highRange = highRange - lowRange + 1;
                            coefficientStr = Integer.toString(highRange);
                            if (addrPos > lowRange)
                            {
                                addrPos -= highRange;
                            }
                            currentLine = lowRange;
                            handleNormalMode(null, 'd');
                            handleNormalMode(null, 'd');
                            currentLine = addrPos;
                            handleNormalMode(null, 'p');
                            currentEvent.startCursor = -1;
                            i += newLineLength;
                            out.saveCursor();
                        }
                    break;
				    
				    case 's':
				        //Valid replace command line should look like:
				        //:[range]s/oldStr/newStr/CMD
				        //Where [range] can consist of:
				        //      '%'   File global, or
				        //      'start,end' to limit between start and end inclusive
				        //      If no range is given, assume current line only
				        //Where CMD can consist of:
				        //      g   Line global (replace every occurrence on the line)
				        //      c   Confirm changes
				        //      i   Case insensitive search
				        //      (DEBUG ==> Should include: p      Print results)
						if ((i+2) < newLineLength)
						{
							if (newLine.charAt(i+1) == '/')
							{
							    int firstEndPos = newLine.indexOf('/',i+2);
							    if (firstEndPos != -1)
							    {
							        String oldStr = newLine.substring(i+2, firstEndPos);
							        int endPos = newLine.indexOf('/', firstEndPos+1);
							        if (endPos == -1)
							            endPos = newLineLength;
							        String newStr = newLine.substring(firstEndPos+1, endPos);
							        //Ok, we're finally ready to replace.  See if there are any
							        //options...
							        boolean lineGlobal = false;
							        boolean confirm    = false;
							        boolean ic         = false;
							        boolean backIC     = false;
							        if (newLineLength > endPos)
							        {
							            String opts = newLine.substring(endPos+1);
							                   
							            lineGlobal = (opts.indexOf('g') != -1);
							            confirm = (opts.indexOf('c') != -1);
							            ic = (opts.indexOf('i') != -1);
							            if (ic)
							            {
							                backIC = IC;
							                IC = true;
							            }
							        }
                                    in.setEcho(false);
                                    if ((lowRange != -1) && (highRange != -1))
                                    {
                                        if ((lowRange < highRange) && (lowRange > 0) &&
                                            (highRange < buffer.size()))
                                        {
                                            currentLine = lowRange;
                                            line = (String)buffer.elementAt(currentLine);
                                            cursor = 0;
                                            fileGlobal = true;
                                            highRange++;
                                        }
                                        else
                                        {
                                            repaintCommandMessage(CMD_BAD_RANGE, false, false);
                                            i += newLineLength;
                                            break;
                                        }
                                    }
							        replace(oldStr, newStr, lineGlobal, fileGlobal, confirm, highRange);
							        out.saveCursor();
							        //DEBUG DOWNSIZE TEST lastReplace = newLine;
							        if (ic)
							            IC = backIC;
							        i += newLineLength; //Kick out of the loop
							    }
							}
						}
                    break;
                    
				    case 'q':
					    if (modified)
					    {
							if (newLine.indexOf('!') != -1)
							{
								state = FINISHED;
							}
							else
							{
								repaintCommandMessage(CMD_NO_WRITE, false, false);
							}
							i+=newLineLength; //Kick out of the loop
					    }
					    else
					    {
						    state = FINISHED;
					    }
				    break;

				    case 'w':
				        boolean force = false;
				        if ((i+1) < newLineLength)
						{
						    //Look for force character
						    int pos = i+1;
							if (newLine.charAt(pos) == '!')
							{
							    force = true;
							    pos++;
							}
							    
							//Look for writing to another filename
							if ((pos < newLineLength) && (newLine.charAt(pos) == ' '))
							{
							    String newName = newLine.substring(++pos);
							    i += newLineLength;
							    if (newName.length() > 0)
							    {
							        File newFile = new File(newName);
							        if (!force && (newFile.exists()))
							        {
							            repaintCommandMessage("File exists (w! to override)", false, false);
							        }
							        else
							        {
							            saveAsFile(newName);
							            modified = false;
							        }
							        break;
							    }
							}
							else
							{
							    if (force && readOnlyMode)
							    {
							        if (file != null)
							            file.close();
							        if (changeFilePermission())
							        {
							            file = new RandomAccessFile(fileName, "rw");
							            readOnlyMode = false;
							        }
							        else
							        {
							            repaintCommandMessage(Slush.PERM_DENIED, false, false);
							            break;
							        }
							    }
							}
						}

                        //Nothing special, just save file.
						if (fileName != null)
						{
					        saveFile();
					        modified = false;
					    }
					    else
					    {
					        repaintCommandMessage("No file name", false, false);
					    }
					    //i += newLineLength;
				    break;
    				
				    default:
				        if (Character.isDigit(newLine.charAt(0)))
				        {
		                    try
		                    {
				                int commaPos;
				                if ((commaPos = newLine.indexOf(',', i)) == -1)
				                {
			                            int gotoLine = Integer.parseInt(newLine, 10);
			                            if (gotoLine > buffer.size())
			                            {
				                            gotoLine = buffer.size();
			                            }
			                            else if (gotoLine < 1)
			                            {
				                            gotoLine = 1;
			                            }
			                            currentLine = gotoLine - 1; //gotoLine is 1 based
			                            line = (String)buffer.elementAt(currentLine);
			                            cursor = 0;
			                            state = NORMAL;
                                        in.setEcho(false);
                                        
                                        gotoCurrentLine();
                                        
			                            return;
				                }
				                else
				                {
				                    String temp = newLine.substring(i, commaPos);
				                    lowRange = Integer.parseInt(temp, 10) - 1;
				                    int endPos = -1;
				                    for (int j = commaPos+1; j<newLineLength; j++)
				                    {
				                        if (!Character.isDigit(newLine.charAt(j)))
				                        {
				                            endPos = j;
				                            break;
				                        }
				                    }
				                    if (endPos != -1)
				                    {
				                        temp = newLine.substring(commaPos+1, endPos);
				                        highRange = Integer.parseInt(temp, 10) - 1;
				                        i = endPos - 1;
				                        break;
				                    }
				                }
		                    }
		                    catch(Exception e)
		                    {
			                    //Means it wasn't a good goto command
		                    }
				        }

				        repaintCommandMessage(("Unrecognized: " + newLine), false, false);
				        i += newLineLength; //Kick out of the loop
				    break;
			    }
		    }
		}

		if (state != FINISHED)
		{
			state = NORMAL;
			in.setEcho(false);
			out.restoreCursor();
		}
    }
    
    /**
     * Replace the old string with the new string.
     */
    private void replace(String oldStr, String newStr, 
                         boolean lineGlobal, boolean fileGlobal,
                         boolean confirm, int stopLine) throws IOException
    {
        if (confirm)
        {
            //Flush the input buffer first...
            while(in.available() > 0)
                in.read();
        }
        
        currentEvent     = null;
        searchString     = oldStr;
        int startLine    = currentLine;
        int startCursor  = cursor;
        int backupLine   = currentLine;
        int backupCursor = cursor;
        boolean replace  = true;
        boolean wrapped  = false;
        boolean finished = false;
        
        while(!finished)
        {
            replace = true;
            boolean thisOneWrapped = search(false, false);
            if (thisOneWrapped)
                wrapped = true;
                
            if (searchInfo[0] != -1)
            {
                //Look for end criteria.  We need to end when a) there are no
                //more matches in the file, or b) we wrap around to where we
                //started.
                if (wrapped)
                {
                    if ( ((searchInfo[0] == startLine) && (searchInfo[1] >= startCursor)) ||
                         (searchInfo[0] > startLine))
                    {
                        finished = true;
                        continue;
                    }
                }

                //Found one.
                if (!fileGlobal)
                {
                    if (currentLine != searchInfo[0])
                    {
                        finished = true;
                        continue;
                    }
                }
                if (confirm)
                {
                    gotoLine(searchInfo[0], searchInfo[1]);
                    //DEBUG DOWNSIZE TEST:
                    //repaintCommandMessage("replace with " + newStr + " (y/n/a/q/^E/^Y)?",
                    //                      true, false);
                    repaintCommandMessage("replace with " + newStr + " (y/n/a/q)?",
                                          true, false);
                    char response = (char)in.read();
                    switch(response)
                    {
/*DEBUG DOWNSIZE TEST
                        case 0x05:    //ctrl-e
                            if (currentLine != lineAtTopOfScr)
                                handleNormalMode(null, response);
                        break;
                        
                        case 0x13:    //ctrl-y
                            if (currentLine != lineAtEndOfScr)
                                handleNormalMode(null, response);
                        break;
DEBUG DOWNSIZE TEST*/
                        
                        case 'a':
                            confirm = false;
                        break;

                        case 'n':
                            replace = false;
                        break;
                            
                        case 'q':
                            finished = true;
                            replace  = false;
                        break;
                    }
                }

                currentLine = searchInfo[0];
                cursor = searchInfo[1];
                line = (String)buffer.elementAt(currentLine);
                if (replace)
                {
                    if (currentEvent == null)
                    {
                        currentEvent = new SledEvent(SledEvent.SEARCH_REPLACE, 10, startLine, startCursor,
                                                     startCursor, ' ', ' ');
                        currentEvent.misc = new Vector(10,10);
                    }
                    currentEvent.changed.addElement(line);
                    currentEvent.misc.addElement(new Integer(currentLine));

                    line = line.substring(0, cursor) + newStr +
                            line.substring(cursor+oldStr.length());
                    buffer.setElementAt(line, currentLine);
                    if (confirm)
                        repaint();
                    backupLine = currentLine;
                    backupCursor = cursor;
                }
                
                if (!lineGlobal)
                {
                    if (!fileGlobal)
                    {
                        finished = true;
                    }
                    else
                    {
                        if ((stopLine != -1) && ((currentLine + 1) == stopLine))
                        {
                            finished = true;
                        }
                        else
                        {
                            currentLine++;
                            if (currentLine == buffer.size())
                            {
                                currentLine = 0;
                            }
                            cursor = 0;
                            line = (String)buffer.elementAt(currentLine);
                        }
                    }
                }
                else
                {
                    cursor++;
                    if (cursor == line.length())
                    {
                        if ((stopLine != -1) && ((currentLine + 1) == stopLine))
                        {
                            finished = true;
                        }
                        else
                        {
                            currentLine++;
                            if (currentLine == buffer.size())
                            {
                                currentLine = 0;
                            }
                            cursor = 0;
                            line = (String)buffer.elementAt(currentLine);
                        }
                    }
                }
            }
            else
            {
                finished = true;
            }
        }

        currentLine = backupLine;
        cursor = backupCursor;
        line = (String)buffer.elementAt(currentLine);
        undoBuffer.insertElementAt(currentEvent, 0);
        currentEvent = null;
        if (confirm)
        {
            repaintCommandMessage(EMPTY, false, false);
            gotoCurrentLine();
        }
        else
        {
            if ((currentLine < lineAtTopOfScr) ||
                (currentLine > (lineAtTopOfScr + SCREEN_WIDTH - 2)))
            {
                //Need to page the screen
                lineAtTopOfScr = currentLine;
            }
            repaint();
        }
    }
  
    /**
     * Search for the searchStr String in the buffer.  If we find it, 
     * we set the instance variables searchInfo[0] to the line and 
     * searchInfo[1] to the cursor of the position of the found string. 
     * If not found, searchInfo[0] and [1] will contain -1;
     *
     * @return  <code>true>if search wrapped to beginning of file.
     */
    private boolean search(boolean backward, boolean stopOnRollOver)
    {
        boolean wrapped = false;
        String temp;
        
        if (IC)
            temp = line.toLowerCase();
        else
            temp = line;
            
        //Look on current line first        
        if ((searchInfo[1] = searchOnLine(temp, cursor, backward)) != -1)
        {
            searchInfo[0] = currentLine;
            return wrapped;
        }
        else if (buffer.size() != 1)
        {
            //Not in current line, look in rest of file
            int incrementVal = (backward ? -1 : 1);
            int startHere = (backward ? 5000 : 0);  //DEBUG ==> Can you say hack?
            int currLine = currentLine;
            do
            {
                currLine = currLine + incrementVal;
                if (currLine < 0)
                {
                    wrapped = true;
                    currLine = buffer.size()-1;
                    if (stopOnRollOver)
                        break;
                    searchCMDStr = "search hit TOP, continuing at BOTTOM";
                    repaintCommandMessage(searchCMDStr, false, false);
                }
                else if (currLine == buffer.size())
                {
                    wrapped = true;
                    currLine = 0;
                    if (stopOnRollOver)
                        break;
                    searchCMDStr = "search hit BOTTOM, continuing at TOP";                        
                    repaintCommandMessage(searchCMDStr, false, false);
                }
                temp = (IC ? ((String)buffer.elementAt(currLine)).toLowerCase() : (String)buffer.elementAt(currLine));
                if ((searchInfo[1] = searchOnLine(temp, startHere, backward)) != -1)
                {
                    searchInfo[0] = currLine;
                    return wrapped;
                }
            }
            while(currLine != currentLine);
        }
        
        searchInfo[0] = -1;
        searchInfo[1] = -1;
        return wrapped;
    }
    
    /**
     * Find the search string in the given line starting from the specified
     * position.
     */
    private int searchOnLine(String thisLine, int startHere, boolean backward)
    {
        int findPos;
        
        if (backward)
        {
            findPos = thisLine.lastIndexOf(searchString, startHere);
        }
        else
        {
            findPos = thisLine.indexOf(searchString, startHere);
        }
        
        return findPos;
    }
    
    /**
     * Inserts a file into the current file, after the line containing
     * the cursor.
     */
    private void insertFile(String fileName)
    {
        boolean appending = (buffer.size() == ++currentLine);
        int charCount = 0;
    	String currLine;
    	String commandMessage;
    	try
    	{
	        DataInputStream inStream = new DataInputStream(new BufferedInputStream(new FileInputStream(fileName)));
		    //DataInputStream doesn't return the newline from the readLine, so use it
		    //for the bulk reading of the file.
		        
		    //Handle the undo event.
		    SledEvent newEvent = new SledEvent(SledEvent.INSERTING_FILE, -1, currentLine, cursor,
		                                       buffer.size(), ' ', ' ');
			undoBuffer.insertElementAt(newEvent, 0);
		        
		    int newLineLength = CURRENT_NEWLINE.length();
		    int i=0;
		    int tabPos;
		    while ((currLine = inStream.readLine()) != null)
            {
                //DEBUG ==> Ok, here's the big cop out for tabs.
                //I'm going to replace all tabs with a predetermined
                //number of spaces.  I really need to fix this.  Sigh.
                while ((tabPos = currLine.indexOf('\t')) != -1)
                {
                    if (tabPos == (currLine.length()-1))
                        currLine = currLine.substring(0, tabPos) +
                                    TAB_REPLACE;
                    else
                        currLine = currLine.substring(0, tabPos) +
                                    TAB_REPLACE + currLine.substring(tabPos+1);
                }
                //DEBUG ==> What if there is a
                //control character?  I won't give an
                //accurate char count, will I?
			    charCount += currLine.length() + newLineLength;
			    if (appending)
			        buffer.addElement(currLine);
			    else
			        buffer.insertElementAt(currLine, currentLine + i);
			    i++;
		    }
		    inStream.close();

            commandMessage = "\"" + fileName + "\" " + buffer.size() +
                                " lines, " + charCount + " characters";
            line = (String)buffer.elementAt(currentLine);
            cursor = 0;
        }
        catch(FileNotFoundException fnfe)
        {
            commandMessage = "File not found";
        }
        catch(IOException ioe)
        {
            commandMessage = ioe.toString();
        }
        repaint(false, commandMessage);
    }
    
    /**
     * Pastes the text in the copy buffer above or below the current
     * line, based on the prepend parameter.
     *
     * @param  fromHere  Vector of Strings to paste
     * @param  prepend   paste the text before the current line or after
     */
    private void pasteText(Vector fromHere, boolean prepend)
    {
		if (fromHere.size() != 0)
		{
			modified = true;
			int temp;
			if (prepend)
			    temp = currentLine;
			else
			    temp = ++currentLine;

			for (int i=0; i<fromHere.size(); i++)
			{
				buffer.insertElementAt(fromHere.elementAt(i), temp++);
			}
			cursor = 0;
			line = (String)buffer.elementAt(currentLine);

			repaint();
		}
		else
		{
			repaintCommandMessage("Nothing in register", true, false);
		}
    }

    /**
     * Puts the cursor on the specified line, scrolling and repainting 
     * as needed.
     * ELVIS
     */
    private void gotoLine(int currentLine, int cursor)
    {
        if ((currentLine < lineAtTopOfScr) ||
            (currentLine > (lineAtTopOfScr+SCREEN_HEIGHT-2)))
        {
            //Need to page the screen, force the repaint
            lineAtTopOfScr = currentLine;
            repaint();
        }
        else if ((cursor < cursorStart) || (cursor >= (cursorStart + SCREEN_WIDTH)))
        {
            cursorStart = cursor;
            repaint();
        }
        else
        {
            out.setCursorPosition(currentLine - lineAtTopOfScr, cursor%SCREEN_WIDTH);
        }
    }
    
    /**
     * Puts the cursor on the current line, scrolling and repainting 
     * as needed.
     * ELVIS
     */
    private void gotoCurrentLine()
    {
        gotoLine(currentLine, cursor);
    }
    
    /**
     * Gets the user entered multiplier for the command.  For example, 
     * if the user types '12dd', the command is delete, the coefficient 
     * is 12.  This function would return 12 in this case.
     *
     * @return  the coefficient entered by the user, or 1 if none entered.
     */
    private int getCoefficient()
    {
        int retVal;
		if (coefficientStr != null)
		{
		    retVal = Integer.parseInt(coefficientStr);
		    coefficientStr = null;
		}
		else
		    retVal = 1;
		    
		return retVal;
    }

    /**
     * If the editor receives an ESC (0x1B), there could be an
     * escape control sequence following.  The next byte should be
     * 0x5B, followed by the control sequence.  Control sequences
     * only include arrows for now.
     */
    private char parseControlSequence()
    {
	    char input = 0x1B;
	    try
	    {
		    if (in.available() > 0)
		    {
			    if (in.read() == 0x5B)
			    {
				    input = (char)in.read();

				    switch(input)
				    {
					    case 'A':
						    input = UP_ARROW;
					    break;

					    case 'B':
						    input = DOWN_ARROW;
					    break;
    	
					    case 'C':
						    input = RIGHT_ARROW;
					    break;

					    case 'D':
						    input = LEFT_ARROW;
					    break;

					    default:
					        //Umm, some unknown control sequence.
						    input = 0xFF;
					    break;
				    }
			    }
		    }
	    }
	    catch(IOException ioe)
	    {
		    //DRAIN
	    }

	    return input;
    }

    /**
     * Search for the first non space character from the fromHere position.
     */
    int findStartOfWord(String inHere, int fromHere)//, boolean backward)
    {
	    //if (!backward)
	    //{
	        for (int i=fromHere; i<inHere.length(); i++)
	        {
		        if (inHere.charAt(i) != ' ')
		        {
		            return i;
		        }
	        }
	        return inHere.length();
	    //}
/*
	    else
	    {
	        //Dear heavens, this is **ultra** inefficient.  Hanging head in shame.
	        //I was tired, I promise.
	        boolean wordFound = (inHere.charAt(fromHere) != ' ');
	        if (fromHere > 0)
	        {
	            if (inHere.charAt(fromHere-1) == ' ')
	            {
	                fromHere--;
	                wordFound = false;
	            }
	            else if (wordFound && !Character.isLetterOrDigit(inHere.charAt(fromHere)))
	            {
	                fromHere--;
	            }
	        }   
	        for (int i=fromHere; i>=0; i--)
	        {
	            if (!wordFound)
	            {
		            if (inHere.charAt(i) != ' ')
		            {
		                if (!Character.isLetterOrDigit(inHere.charAt(i)))
		                    return i;
		                else
		                {
		                    wordFound = true;
		                }
		            }
		        }
		        else
		        {
		            if (!Character.isLetterOrDigit(inHere.charAt(i)))
		            {
		                return (i+1);
		            }
		        }
	        }
	        return 0;
	    }
*/
    }
    
    int findNextWord(String inHere, int fromHere)
    {
        boolean lookingForNextNonSpace = false;
        
        if (!Character.isLetterOrDigit(inHere.charAt(fromHere)))
        {
            //We are starting at a stop character, so just look for
            //first non space.
            for (int i=fromHere+1; i<inHere.length(); i++)
            {
                if (inHere.charAt(i) != ' ')
                    return i;
            }
        }
        else
        {
            for (int i=fromHere; i<inHere.length(); i++)
            {
                char currChar = inHere.charAt(i);
                if (currChar == ' ')
                {
                    lookingForNextNonSpace = true;
                }
		        else if ((!Character.isLetterOrDigit(currChar)) || lookingForNextNonSpace)
		        {
		            //Found a stop character.
		            return i;
		        }
            }
        }
        
        return -1;
    }
    
    //DEBUG ELVIS HERE:
    int findPrevWord(String inHere, int fromHere)
    {
        if (fromHere == 0)
            return -1;
        else if ((!Character.isLetterOrDigit(inHere.charAt(fromHere-1))))
        {
            int pos = -1;
            for (int i=fromHere-1; i>=0; i--)
            {
                if (inHere.charAt(i) != ' ')
                {
                    pos = i;
                    break;
                }
            }
            if ((pos == -1) || (!Character.isLetterOrDigit(inHere.charAt(pos))))
                return pos;
            else
                fromHere = pos;
        }
        
        if (!Character.isLetterOrDigit(inHere.charAt(fromHere)))
        {
            int pos = -1;
            //We are starting at a stop character, so first look for
            //first non space.
            for (int i=fromHere-1; i>=0; i--)
            {
                if (inHere.charAt(i) != ' ')
                {
                    pos = i;
                    break;
                }
            }
            if ((pos == -1) || (!Character.isLetterOrDigit(inHere.charAt(pos))))
            {
                return pos;
            }
        }
        
        //goto prev special char. return +1
        int newPos = -1;
        for (int i=fromHere-1; i>=0; i--)
        {
            if (!Character.isLetterOrDigit(inHere.charAt(i)))
            {
                newPos = i;
                break;
            }
        }
        
        if (newPos == -1)
            return (fromHere == 0) ? -1 : 0;
        else
            return (newPos+1);
    }
    
    /**
     * Finds the position of the next word starting from a certain position.
     * The end of a word is defined as the next non-space character after the next
     * space encountered, or the position of one of the stop characters ('.', 
     * etc.)
     *
     * As an aside, nextSpacePos is being passed in because I noticed that every
     * where I called this function, I already had that information.
     */
/*
    int findNextWord(String inHere, int fromHere, int nextSpacePos, boolean backward)
    {
        int length = inHere.length();
        if (nextSpacePos == -1)
        {
            //There isn't a space later in the string.
            nextSpacePos = (backward) ? 0 : length;
        }
        
        if ((fromHere > (length - 1)) ||
            (nextSpacePos > length))
        {
            return (backward) ? -1 : length;
        }
                       
        //First, search for an embedded "stop" character before the next
        //space.  For example, if the string was: "this.is a test", with
        //the cursor on the 't' in "this", then the '.' would be a stop
        //character.
	    int cursor = -1;
	    if (!backward)
	    {
	        for (int i=fromHere+1; i<nextSpacePos; i++)
	        {
		        if (!Character.isLetterOrDigit(inHere.charAt(i)))
		        {
			        cursor = i;
			        break;
		        }
	        }
	    }
	    else
	    {
	        for (int i=fromHere-1; i>nextSpacePos; i--)
	        {
		        if (!Character.isLetterOrDigit(inHere.charAt(i)))
		        {
			        cursor = i;
			        break;
		        }
	        }
	    }
	    
	    if (cursor == -1)
	    {
	        //Didn't find a stop character, so look for first non-whitespace.
	        //This is important because if the character at
	        //fromHere pos is a space, we need to find the
	        //next non space.
	        cursor = findStartOfWord(inHere, backward ? fromHere : nextSpacePos, backward);
	        if (cursor == fromHere)
	            cursor = backward ? -1 : length;
	    }

	    return cursor;
    }
*/
    
    /**
     * Display the given message on the command line, first clearing
     * what was there.  The user chooses whether or not to display the
     * text in bold, and whether or not to reset the cursor after the
     * message is displayed.
     */
    private void repaintCommandMessage(String commandMessage, boolean resetCursor,
                                       boolean bold)
    {
        if (resetCursor)
            out.saveCursor();

        out.setCursorPosition(SCREEN_HEIGHT-1, 0);
        out.clearLine();

        if (bold)
            out.setBold(true);

        out.print(commandMessage);
        out.flush();

        if (bold)
            out.setBold(false);

        if (resetCursor)
        {
            out.restoreCursor();
        }
    }

    /**
     * Clear and repaint the entire screen.  The command line message will
     * be cleared, and the cursor will be placed correctly on the screen.
     */
    void repaint()
    {
        repaint(false, EMPTY);
    }

    /**
     * Clear and repaint the entire screen.  The caller can specify a command
     * message to display.
     * ELVIS
     */
    void repaint(boolean boldCmdMsg, String commandMessage)
    {
        String currLine;
        byte[] currBytes;
        int    copyBytes;
        int    index = 0;
        int    emptyLines = SCREEN_HEIGHT - 1;
        int    linesToDisp = Math.min(buffer.size(), SCREEN_HEIGHT - 1);
        for (int i=0; i<linesToDisp; i++)
        {
            currLine = (String)buffer.elementAt(i + lineAtTopOfScr);
            if (currLine.length() > cursorStart)
            {
                currBytes = currLine.getBytes();
                copyBytes = Math.min(currBytes.length - cursorStart, SCREEN_WIDTH);
                System.arraycopy(currBytes, cursorStart, displayBuffer, index, copyBytes);
                index += copyBytes;
            }
            displayBuffer[index++] = (byte)'\r';
            displayBuffer[index++] = (byte)'\n';
            emptyLines--;
        }
        
        //Now, show the empty lines with '~'
        for (int i=0; i<emptyLines; i++)
        {
            displayBuffer[index++] = (byte)'~';
            displayBuffer[index++] = (byte)'\r';
            displayBuffer[index++] = (byte)'\n';
        }
        
        //Now show the command message, if any.
        if (commandMessage.length() > 0)
        {
            currBytes = commandMessage.getBytes();
            System.arraycopy(currBytes, 0, displayBuffer, index, currBytes.length);
            index += currBytes.length;
        }
        
        out.clearDisplay();
        out.setCursorPosition(0,0);
        out.write(displayBuffer, 0, index);

        out.setCursorPosition(currentLine - lineAtTopOfScr, cursor - cursorStart);
    }
    
    /**
     * TINI specific function to change the file permissions on the
     * current file to give user write permissions.
     *
     * @return <code>false</code> if unable to give write permission.
     */
    boolean changeFilePermission() throws IOException
    {
        DSFile file = new DSFile(fileName);
        file.setUserPermissions(6);     //R-W
        return file.canWrite();
    }

    /**
     * Open the file (specified by the instance variable fileName),
     * determine the type of end of line the file is using, then
     * read in that file and initialize the text buffer.
     */
    void readData()
    {
        String commandMessage = EMPTY;
        //First, reset some variables
        buffer.removeAllElements();
        clearMarks();
		modified = false;
		currentLine = 0;
		cursor = 0;
		lineAtTopOfScr = 0;
		CURRENT_NEWLINE = "\r\n";
		
		//Take care of closing the currently open file
		if (file != null)
		{
		    try
		    {
		        file.close();
		    }
		    catch(IOException ioe)
		    {
		        //Wake up the squirrels, Zelda, there's
		        //gonna be some trouble.
		    }
		    file = null;
		}
        
	    if (fileName == null)
	    {
	        repaint(false, "Buffer Empty");
	        out.setCursorPosition(SCREEN_HEIGHT/2-2, SCREEN_WIDTH/2-12);
	        out.print("Sled - TINI SLush EDitor");
	        out.setCursorPosition(SCREEN_HEIGHT/2, SCREEN_WIDTH/2-6);
	        out.print("Version 0.01");
	        out.setCursorPosition(SCREEN_HEIGHT/2+2, SCREEN_WIDTH/2-10);
	        out.print("Dallas Semiconductor Corp.");
	        out.setCursorPosition(0,0);
	        line = EMPTY;
	        buffer.addElement(line);
		    return;
	    }

	    File inputFile = new File(fileName);
	    if (inputFile.exists())
	    {
	        //Read the data...
	        try
	        {
		        int charCount = 0;
	            if (!inputFile.canWrite())
	            {
	                readOnlyMode = true;
	                file = new RandomAccessFile(inputFile, "r");
	            }
	            else
		            file = new RandomAccessFile(inputFile, "rw");
		        
		        //Find out the EOL type this file is using.
		        boolean modeFound = false;
		        char temp;
		        long avail = file.length();
		        while (!modeFound && (avail-- > 0))
		        {
		            temp = (char)file.read();
		            if (temp == '\r')
		            {
		                if (avail > 0)
		                {
		                    temp = (char)file.read();
		                    avail--;
		                    if (temp == '\n')
		                    {
		                        CURRENT_NEWLINE = "\r\n";
		                    }
		                    else
		                    {
		                        commandMessage = "[incomplete last line]";
		                        CURRENT_NEWLINE = "\r";
		                    }
		                }
		                else
		                {
		                    commandMessage = "[incomplete last line]";
		                    CURRENT_NEWLINE = "\r";
		                }
		                modeFound = true;
		            }
		            else if (temp == '\n')
		            {
		                commandMessage = "[unix format]";
		                modeFound = true;
		                CURRENT_NEWLINE = "\n";
		            }
		        }
		            
		        if (!modeFound)
		        {
		            commandMessage = "[freaky format]";
		            CURRENT_NEWLINE = "\r\n";
		        }
		        file.seek(0);		            

    	        String currLine;
	            DataInputStream inStream = new DataInputStream(new BufferedInputStream(new FileInputStream(inputFile)));
		        //DataInputStream doesn't return the newline from the readLine, so use it
		        //for the bulk reading of the file.
		        int newLineLength = CURRENT_NEWLINE.length();
		        int tabPos;
		        while ((currLine = inStream.readLine()) != null)
                {
                    //DEBUG ==> Ok, here's the big cop out for tabs.
                    //I'm going to replace all tabs with a predetermined
                    //number of spaces.  I really need to fix this.  Sigh.
                    while ((tabPos = currLine.indexOf('\t')) != -1)
                    {
                        if (tabPos == currLine.length())
                            currLine = currLine.substring(0, tabPos) +
                                       TAB_REPLACE;
                        else
                            currLine = currLine.substring(0, tabPos) +
                                       TAB_REPLACE + currLine.substring(tabPos+1);
                    }
                    //DEBUG add some for the newline character
                    //What if there is some other control character?
                    //I won't give an accurate char count, will I?
			        charCount += currLine.length() + newLineLength;
			        buffer.addElement(currLine);
		        }
		        inStream.close();

                commandMessage = "\"" + fileName + "\" " + (readOnlyMode ? "[readonly] " : EMPTY) 
                                 + commandMessage + " " + buffer.size() +
                                 " lines, " + charCount + " characters";
                line = (String)buffer.elementAt(0);
	        }
	        catch(Exception e)
	        {
	            commandMessage = e.toString();
	        }
	    }
	    else
	    {
	        line = EMPTY;
	        buffer.addElement(line);
	        commandMessage = fileName + " [New File]";
	    }
	    
	    repaint(false, commandMessage);
	    System.gc();
    }
    
    /**
     * Calculates PI to 8 billion decimal places.
     */
    void clearMarks()
    {
        for (int i=0; i<marks.length; i++)
            marks[i] = 0;
    }
    
    /**
     * Save the currently open file.
     */
    void saveFile()
    {
        if (readOnlyMode)
        {
            repaintCommandMessage("readonly is set. (! to override)", false, true);
        }
        else
        {
            String line;
            int charCount = 0;
	        try
	        {
	            if (file == null)
	                file = new RandomAccessFile(fileName, "rw");
	            else
	            {
	                //First, kill what is there...
	                //DEBUG FIXME DEBUG FIXME
	                //file.setLength(0);
	                //DEBUG FIXME DEBUG FIXME
	                //DEBUG! We are going to close the file and
	                //open it again until RandomAccessFile.setLength() is
	                //implemented again.
	                file.close();
	                //BEGIN DEBUG THIS ALSO will go away when set length works.
	                File temp = new File(fileName);
	                temp.delete();
	                //END DEBUG
	                file = new RandomAccessFile(fileName, "rw");
	            }
	            //Now, write the file...
	            for (int i=0; i<buffer.size(); i++)
	            {
	                line = (String)buffer.elementAt(i) + CURRENT_NEWLINE;
	                charCount += line.length();
	                file.write(line.getBytes());
	            }
	        }
	        catch(IOException ioe)
	        {
	            //DEBUG Umm, what now?!?
	        }
            line = "\"" + fileName + "\" " + buffer.size() +
                " lines, " + charCount + " characters";
	        repaintCommandMessage(line, false, false);
	    }
    }

    /**
     * Save the current text data to a new file.
     */
    void saveAsFile(String name)
    {
        String line;
        int    charCount = 0;
	    FileOutputStream fout = null;
	    try
	    {
	        fout = new FileOutputStream(new File(name));
	    }
	    catch(IOException ioe)
	    {
	        //WHOA, bad biscuits, baby
	        repaintCommandMessage(ioe.toString(), false, false);
	        return;
	    }
	    
	    try
	    {
	        //Now, write the file...
	        for (int i=0; i<buffer.size(); i++)
	        {
	            line = (String)buffer.elementAt(i) + CURRENT_NEWLINE;
	            charCount += line.length();
	            fout.write(line.getBytes());
	        }
	        
	        fout.close();
	    }
	    catch(IOException ioe)
	    {
	        //DEBUG Umm, what now?!?
	    }
        line = "\"" + name + "\" " + buffer.size() +
              " lines, " + charCount + " characters";
	    repaintCommandMessage(line, false, false);
    }
}