/****************************************************************************
*
*  Copyright (C) 1999,2000 Dallas Semiconductor Corporation. 
*  All rights Reserved. Printed in U.S.A.
*  This software is protected by copyright laws of
*  the United States and of foreign countries.
*  This material may also be protected by patent laws of the United States 
*  and of foreign countries.
*  This software is furnished under a license agreement and/or a
*  nondisclosure agreement and may only be used or copied in accordance
*  with the terms of those agreements.
*  The mere transfer of this software does not imply any licenses
*  of trade secrets, proprietary technology, copyrights, patents,
*  trademarks, maskwork rights, or any other form of intellectual
*  property whatsoever. Dallas Semiconductor retains all ownership rights.
*
*     Module Name: SledPrintStream.java
*/
package com.dalsemi.slush.command.sled;

import java.io.*;

/**
 * A PrintStream for general Sled use.  It provides VT100 command
 * sequences to control a VT100 terminal.  It only has partial VT100
 * support, but more will be added over time.
 *
 * @author  Stephen Hess
 * @version 1.0
 */
public class SledPrintStream extends PrintStream
{
    public static char UP       = 'A';
    public static char DOWN     = 'B';
    public static char RIGHT    = 'C';
    public static char LEFT     = 'D';
    
    OutputStream root;
    public static final byte[] CLEAR_DISPLAY = {0x1B, 0x5B, (byte)'2', (byte)'J'};
    static byte[] CONTROL_SEQ   = {0x1B, 0x5B};
    static byte   SEMICOLON     = 0x3B;
    static byte[] CLEAR_LINE    = {0x1B, 0x5B, 0x32, (byte)'K'};
    static byte[] CLEAR_TO_EOL  = {0x1B, 0x5B, (byte)'K'};
    static byte[] BOLD_ON       = {0x1B, 0x5B, 0x31, (byte)'m'};
    static byte[] BOLD_OFF      = {0x1B, 0x5B, 0x30, (byte)'m'};
    static byte[] DELETE_CHAR   = {0x1B, 0x5B, (byte)'P'};
    //DEBUG ==> These are just not used that often, so I'm commenting them out
    //for now to save space:
    //static byte[] BLINK_ON      = {0x1B, 0x5B, 0x35, (byte)'m'};
    //static byte[] BLINK_OFF     = {0x1B, 0x5B, 0x32, 0x35,(byte)'m'};
    //static byte[] INVERSE_ON    = {0x1B, 0x5B, 0x37, (byte)'m'};
    //static byte[] INVERSE_OFF   = {0x1B, 0x5B, 0x32, 0x37,(byte)'m'};
    //static byte[] CURSOR_ON     = {0x1B, 0x5B, (byte)'?', 0x32, 0x35, (byte)'h'};
    //static byte[] CURSOR_OFF    = {0x1B, 0x5B, (byte)'?', 0x32, 0x35, (byte)'l'};
    //static byte[] DELETE_LINE   = {0x1B, 0x5B, (byte)'M'};
    //static byte[] INSERT_LINE   = {0x1B, 0x5B, (byte)'L'};
    
    /**
     * Initialize the stream.
     */
    public SledPrintStream(OutputStream root)
    {
        super(root);
        this.root = root;
    }
    
    /**
     * This method clears the remote terminal display without moving the cursor.
     */
	public void clearDisplay()
	{
        try
        {
            root.write(CLEAR_DISPLAY);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
	}
	
	/*
	 * This method moves the cursor in a given direction the specified
	 * number of units.
	 *
	 * @param  direction  either <code>UP</code>, <code>down</code>,
	 * <code>RIGHT</code>, or <code>LEFT</code>
	 * @param  thisMany   number of units to move.
	 */
	public void moveCursor(char direction, int thisMany)
	{
	    try
	    {
	        root.write(CONTROL_SEQ);
	        int a;
            if (thisMany > 9)
            {
                a = (thisMany/10) + 0x30;
                root.write(a);
                a = (thisMany%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write(thisMany + 0x30);
            }
            
            root.write(direction);
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
	}

    /**
     * This method places the cursor based on the absolute coordinates
     * passed in.
     *
     * @param  line   Line number (0 indexed)
     * @param  column Column number (0 indexed)
     */
    public void setCursorPosition(int line, int column)
    {
        line++;
        column++;
        int a;
        try
        {
            root.write(CONTROL_SEQ);
            if (line > 9)
            {
                a = (line/10) + 0x30;
                root.write(a);
                a = (line%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write(line + 0x30);
            }
            root.write(SEMICOLON);

            if (column > 9)
            {
                a = (column/10) + 0x30;
                root.write(a);
                a = (column%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write(column + 0x30);
            }
            root.write(0x48);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
	
	/**
	 * This method deletes the indicated number of characters from
	 * the current cursor position.
	 *
	 * @param  thisMany   number of characters to delete.
	 */
    public void deleteChars(int thisMany)
    {
        try
        {
            //DEBUG ==> Why doesn't this work?!?
            //It works on Linux, but not Windows Telnet.
            //Oh, just answered my own question.
            /*
            root.write(CONTROL_SEQ);
            if (thisMany > 9)
            {
                a = (thisMany/10) + 0x30;
                root.write(a);
                a = (thisMany%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write((thisMany + 0x30)&0x0FF);
            }
            root.write(0x50); //Capital 'P'
            root.flush();
            */
            
            //DEBUG ==> Oh well, do it the slow way:
            for (int i=0; i<thisMany; i++)
            {
                root.write(DELETE_CHAR);
                root.flush();
            }
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
	
	/**
	 * This method clears the current line.
	 */
    public void clearLine()
    {
        try
        {
            root.write(CLEAR_LINE);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    /**
     * This method clears from the current cursor position to the end
     * of the line.
     */
    public void clearToEOL()
    {
        try
        {
            root.write(CLEAR_TO_EOL);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
	
	/**
	 *
	 */
    public void write(int b)
    {
        try
        {
            root.write(b);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
    
    /**
     *
     */
    public void write(byte[] b)
    {
        try
        {
            root.write(b);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
    
    /**
     *
     */
    public void write(byte[] b, int off, int len)
    {
        try
        {
            root.write(b, off, len);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    /**
     *
     */
    public void flush()
    {
        try
        {
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    /**
     * This method sets the bold mode on the remote terminal.
     * Once in bold mode, any new characters written will be
     * shown in bold until bold mode is turned off.
     *
     * @param  on   Indicates bold mode to set.
     */
    public void setBold(boolean on)
    {
        try
        {
            if (on)
                root.write(BOLD_ON);
            else
                root.write(BOLD_OFF);

            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

/*
    public void setBlink(boolean on)
    {
        try
        {
            if (on)
                root.write(BLINK_ON);
            else
                root.write(BLINK_OFF);

            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    public void setInverse(boolean on)
    {
        try
        {
            if (on)
                root.write(INVERSE_ON);
            else
                root.write(INVERSE_OFF);

            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    public void showCursor(boolean on)
    {
        try
        {
            if (on)
                root.write(CURSOR_ON);
            else
                root.write(CURSOR_OFF);

            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

    public void deleteLine(int thisLine)
    {
        try
        {
            setCursorPosition(thisLine, 0);
            root.write(DELETE_LINE);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
*/

    /**
     * This method deletes the specified number of lines from a
     * given start position.
     *
     * @param  start    line number to start (0 indexed)
     * @param  thisMany number of lines to delete
     */
	public void deleteLines(int start, int thisMany)
	{
	    try
	    {
	        setCursorPosition(start, 0);
	        root.write(CONTROL_SEQ);
	        int a;
            if (thisMany > 9)
            {
                a = (thisMany/10) + 0x30;
                root.write(a);
                a = (thisMany%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write(thisMany + 0x30);
            }
            
            root.write('M');
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
	}
    
    /**
     * This method tells the remote terminal to save the current
     * coordinates of the cursor.  You can have the terminal set
     * the cursor back at this spot by calling <code>restoreCursor()</code>.
     */
    public void saveCursor()
    {
        try
        {
            root.write(CONTROL_SEQ);
            root.write('s');
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
    
    /**
     * This method tells the remote terminal to place the cursor
     * to the position the cursor was in when <code>saveCursor()</code>
     * was called.
     */
    public void restoreCursor()
    {
        try
        {
            root.write(CONTROL_SEQ);
            root.write('u');
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }

/*
    public void insertLine(int beforeThis)
    {
        try
        {
            setCursorPosition(beforeThis, 0);
            root.write(INSERT_LINE);
            root.flush();
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
    }
*/
    
    /**
     * This method inserts the specified number of blank lines
     * from the given start line.
     *
     * @param  start    Line on which to start the insert (0 indexed).
     * @param  thisMany Number of blank lines to insert.
     */
	public void insertLines(int start, int thisMany)
	{
	    try
	    {
	        setCursorPosition(start, 0);
	        root.write(CONTROL_SEQ);
	        int a;
            if (thisMany > 9)
            {
                a = (thisMany/10) + 0x30;
                root.write(a);
                a = (thisMany%10) + 0x30;
                root.write(a);
            }
            else
            {
                root.write(thisMany + 0x30);
            }
            
            root.write('L');
        }
        catch(IOException ioe)
        {
            //DRAIN
        }
	}
}