
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;

/**
 * This class implements methods to manage arrays. This includes comparison,
 * copying, filling and inserting and removing simple data types into byte arrays.
 * Simple data types are stored in byte arrays in big-endian format.
 */

public class ArrayUtils
{

   /**
    * Copies <code>length</code> elements from the <code>src</code> array
    * at index <code>src_position</code> to the <code>dst</code> array
    * at index <code>dst_position</code>.  The <code>src</code> and
    * <code>dst</code> arrays must be of the same type.  This is similar to  
    * <code>java.lang.System.arraycopy()</code>, but more restrictive.
    * This allows it to be slightly faster, and calls one less method than
    * the same array copy performed through the <code>java.lang.System</code>
    * class.
    *
    * @param src source array
    * @param src_position starting index in source array
    * @param dst target array
    * @param dst_position starting index in target array
    * @param length number of elements to copy
    *
    * @return 0 under normal operation.  If both <code>dst</code> and <code>src</code>
    * are arrays of references, and are not the same type, returns -1.
    */
   public static native int arraycopy (Object src, int src_position,
                                       Object dst, int dst_position,
                                       int length);

   /**
    * Copies <code>length</code> common lower element bytes from
    * <code>fromArray</code> at <code>fromOffset</code> into
    * <code>toArray</code> at <code>toOffset</code>. <br>
    * If the destination element has more bytes than the source element, the
    * upper bytes of each element are set to zero. <br>
    * If the destination element has less bytes than the source element, the
    * upper bytes are truncated. <br>
    *
    * @param fromArray source array
    * @param fromIndex starting index in source array
    * @param toArray target array
    * @param toIndex starting index in target array
    * @param length number of elements to copy
    */
   public static native void arrayCopyUnequal (Object fromArray, int fromOffset,
                                               Object toArray,   int toOffset,
                                               int length);

   /**
    * Compares <code>length</code> elements in <code>array1</code> at
    * <code>offset1</code> with the elements in <code>array2</code>
    * at <code>offset2</code>.
    *
    * @param array1 source array
    * @param offset1 starting index in source array
    * @param array2 target array
    * @param offset2 starting index in target array
    * @param length number of elements to compare
    *
    * @return false if all elements are not the same
    */
   public static native boolean arrayComp (Object array1, int offset1,
                                           Object arayr2, int offset2,
                                           int length);

   /**
    * Fills <code>thisArray</code> from <code>fromIndex</code> to
    * <code>toIndex</code> with <code>fillValue</code>.
    *
    * @param thisArray target array
    * @param fromIndex starting index for fill
    * @param toIndex ending index for fill
    * @param fillValue value to fill array
    */
   public static native void arrayFill (byte[] thisArray, int fromIndex,
                                        int toIndex, byte fillValue);

   /**
    * Writes a short value <code>value</code> to byte array
    * <code>thisArray</code> starting at offset <code>offset</code>. The values
    * are stored in the array in Big Endian form.
    *
    * @param thisArray array that will be modified
    * @param offset starting offset
    * @param value value to be written
    *
    * @return array reference to <code>thisArray</code>
    */
   public static native byte[] setShort (byte[] thisArray, int offset,
                                         short value);

   /**
    * Writes an int value <code>value</code> to byte array
    * <code>thisArray</code> starting at offset <code>offset</code>. The values
    * are stored in the array in Big Endian form.
    *
    * @param thisArray array that will be modified
    * @param offset starting offset
    * @param value value to be written
    *
    * @return array reference to <code>thisArray</code>
    */
   public static native byte[] setInt (byte[] thisArray, int offset,
                                       int value);

   /**
    * Writes a long value <code>value</code> to byte array
    * <code>thisArray</code> starting at offset <code>offset</code>. The values
    * are stored in the array in Big Endian form.
    *
    * @param thisArray array that will be modified
    * @param offset starting offset
    * @param value value to be written
    *
    * @return array reference to <code>thisArray</code>
    */
   public static native byte[] setLong (byte[] thisArray, int offset,
                                        long value);

   /**
    * Reads a short value from a byte array. The values extracted are assumed
    * to be in Big Endian Form.
    *
    * @param thisArray array containing the value
    * @param offset starting offset
    *
    * @return value read
    */
   public static native short getShort (byte[] thisArray, int offset);

   /**
    * Reads an int value from a byte array. The values extracted are assumed
    * to be in Big Endian Form.
    *
    * @param thisArray array containing the value
    * @param offset starting offset
    *
    * @return value read
    */
   public static native int getInt (byte[] thisArray, int offset);

   /**
    * Reads a long value from a byte array. The values extracted are assumed
    * to be in Big Endian Form.
    *
    * @param thisArray array containing the value
    * @param offset starting offset
    *
    * @return value read
    */
   public static native long getLong (byte[] thisArray, int offset);

   /* --- private section --- */

   private ArrayUtils ()
   {
   }

}
