
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;


/**
 * This class allows bit manipulation of available processor port pins, SMC Ethernet
 * controller port pins and provides abstracted methods of manipulating bits
 * in memory mapped IO addresses through the <code>DataPort</code> class.
 *
 * @see com.dalsemi.system.DataPort
 */
public class BitPort
{

   /**
    * P5.0, C0TX, I2C SCL, pin 21.
    */
   public static final byte Port5Bit0 = 0;         // C0TX

   /**
    * P5.1, C0RX, I2C SDA, pin 20.
    */
   public static final byte Port5Bit1 = 1;         // C0RX

   /**
    * P5.2, C1RX, RXD1, DS2480 RX, "serial 1", pin 19.
    */
   public static final byte Port5Bit2 = 2;         // C1RX RXD1

   /**
    * P5.3, C1TX, TXD1, DS2480 TX, "serial 1", pin 18.
    */
   public static final byte Port5Bit3 = 3;         // C1TX TXD1

   /**
    * P5.4, /PCE0, pin 17.
    */
   public static final byte Port5Bit4 = 4;         // C1TX TXD1

   /**
    * P5.5, /PCE1, pin 16.
    */
   public static final byte Port5Bit5 = 5;         // C1TX TXD1

   /**
    * P5.6, /PCE2, pin 15.
    */
   public static final byte Port5Bit6 = 6;         // C1TX TXD1

   /**
    * P5.7, /PCE3, pin 14.
    */
   public static final byte Port5Bit7 = 7;         // C1TX TXD1

   /**
    * P3.0, RXD0, "serial 0", pin 4.
    */
   public static final byte Port3Bit0 = 8;         // RXD0

   /**
    * P3.1, TXD0, "serial 0", pin 5.
    */
   public static final byte Port3Bit1 = 9;         // TXD0

   /**
    * P3.2, /INT0, /SMCINT (Ethernet interrupt), pin 6.
    */
   public static final byte Port3Bit2 = 10;        // /INT0

   /**
    * P3.3, /INT1, /EXTINT (External interrupt), pin 7.
    */
   public static final byte Port3Bit3 = 11;        // /INT1

   /**
    * P3.4, T0, /SMCRST (Ethernet Reset), pin 10.
    */
   public static final byte Port3Bit4 = 12;        // T0

   /**
    * P3.5, T1, INTOW (Internal 1-Wire), pin 11.
    */
   public static final byte Port3Bit5 = 13;        // T1

   /**
    * SMC EESK, output ONLY, pin 7.
    * WARNING: This pin may also be used as serial flow control by the serial driver.
    * Be aware of resource usage.  You have been warned!
    */
   public static final byte ETH_EESK = 14;

   /**
    * SMC EEDO, output ONLY, pin 4.
    * WARNING: This pin may also be used as serial flow control by the serial driver.
    * Be aware of resource usage.  You have been warned!
    */
   public static final byte ETH_EEDO = 15;

   /**
    * SMC IOS0, input ONLY, pin 98.
    * WARNING: This pin may also be used as serial flow control by the serial driver.
    * Be aware of resource usage.  You have been warned!
    */
   public static final byte ETH_IOS0 = 16;

   /**
    * SMC IOS1, input ONLY, pin 99.
    * WARNING: This pin may also be used as serial flow control by the serial driver.
    * Be aware of resource usage.  You have been warned!
    */
   public static final byte ETH_IOS1 = 17;

   /**
    * SMC IOS2, input ONLY, pin 1.
    * WARNING: This pin may also be used as serial flow control by the serial driver.
    * Be aware of resource usage.  You have been warned!
    */
   public static final byte ETH_IOS2 = 18;

   private static final byte FUNC_SETBIT    = 0;
   private static final byte FUNC_CLEARBIT  = 1;
   private static final byte FUNC_READBIT   = 2;
   private DataPort          dataPortObject = null;
   private byte              bitPortName    = 0;
   private byte[]            bitPortArray   = new byte [2];

   /**
    * The last written value.
    */
   public int latchValue = 0;

   private static native int bitPortNative (byte[] arr);

   /*
    * No one is allowed here.
    */
   private BitPort ()
   {
   }

   /**
    * Creates a new instance of <code>BitPort</code> using the bit specified.
    *
    * @param bitname descriptor of bit to twiddle (e.g. Port5Bit0)
    *
    */
   public BitPort (byte bitname)
   {
      bitPortName = bitname;
   }

   /**
    * Creates a new instance of <code>BitPort</code> using the <code>DataPort</code> object specified.
    *
    * @param port An eight bit wide field of bits to twiddle
    *
    */
   public BitPort (DataPort port)
   {
      dataPortObject = port;
      latchValue = dataPortObject.readLatch();
   }

   /**
    * Sets the pin high.
    */
   public void set ()
   {

      // Make bitPortArray a local variable to speed up accesses
      byte[] tempPortArray = bitPortArray;

      synchronized (tempPortArray)
      {
         latchValue        = 1;
         tempPortArray [0] = FUNC_SETBIT;
         tempPortArray [1] = bitPortName;

         bitPortNative(tempPortArray);
      }
   }

   /**
    * Clears the port pin low.
    */
   public void clear ()
   {

      // Make bitPortArray a local variable to speed up accesses
      byte[] tempPortArray = bitPortArray;

      synchronized (tempPortArray)
      {
         latchValue        = 0;
         tempPortArray [0] = FUNC_CLEARBIT;
         tempPortArray [1] = bitPortName;

         bitPortNative(tempPortArray);
      }
   }

   /**
    * Reads the port pin.
    *
    * @return pin value
    */
   public int read ()
   {

      // Make bitPortArray a local variable to speed up accesses
      byte[] tempPortArray = bitPortArray;

      synchronized (tempPortArray)
      {
         tempPortArray [0] = FUNC_READBIT;
         tempPortArray [1] = bitPortName;

         return bitPortNative(tempPortArray);
      }
   }

   /**
    * Reads the current latch value.  This is always the last
    * value written.  To read the current value of the pin,
    * call the <code>read</code> method.
    *
    * @return latch value
    */
   public int readLatch ()
   {
      return latchValue;
   }

   /**
    * Sets the specified bit high.
    *
    * @param bitpos value 0-7 of bit to set
    * @throws IllegalAddressException if <code>DataPort</code> reports an exception during the operation
    *
    */
   public void set (int bitpos)
      throws IllegalAddressException
   {
      latchValue |= (0x01 << bitpos);

      dataPortObject.write(latchValue);
   }

   /**
    * Clears the specified bit low.
    *
    * @param bitpos value 0-7 of bit to clear
    * @throws IllegalAddressException if <code>DataPort</code> reports an exception during the operation
    *
    */
   public void clear (int bitpos)
      throws IllegalAddressException
   {
      latchValue &= ~(0x01 << bitpos);

      dataPortObject.write(latchValue);
   }

   /**
    * Reads the specified pin.
    * <br>Side effect: Will update all bits in the byte.
    *
    * @param bitpos value 0-7 of bit to read.
    *
    * @return port pin value
    * @throws IllegalAddressException if <code>DataPort</code> reports an exception during the operation
    */
   public int readBit (int bitpos)
      throws IllegalAddressException
   {
      latchValue = dataPortObject.read();

      return (latchValue >>> bitpos) & 0x01;
   }

   /**
    * Reads the latch value of the specified pin.
    * This is always the last value written.
    * To read the current value of the pin,
    * call the <code>read</code> method.
    *
    * @param bitpos value 0-7 of bit to read.
    *
    * @return latch value
    */
   public int readLatch (int bitpos)
   {
      return (latchValue >>> bitpos) & 0x01;
   }
}
