
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;

import java.util.*;

/**
 * This class allows an application to manage external interrupts received by
 * TINI. External interrupts can be used to respond to external events such as
 * a push button being pressed. The interrupt can be configured to be falling
 * edge triggered or low level triggered. Resources for external interrupts are
 * shared with TINI's external serial ports and the parallel port interrupt
 * circuitry that may exist on some socket boards.
 *
 * @see ExternalInterruptEvent
 * @see ExternalInterruptEventListener
 * @see ExternalInterruptException
 */
public class ExternalInterrupt
{

   /**
    * Adds an event listener to TINI's external interrupt event chain.
    *
    * @param externalEventListener reference to the class that will be added to
    *        the listener chain
    *
    * @throws TooManyListenersException if no more listiners can be added
    *
    */
   public void addEventListener (
           ExternalInterruptEventListener externalEventListener)
      throws TooManyListenersException
   {
      if (externalEventListener == null)
         throw new IllegalArgumentException();

      externalInterruptListeners.addElement(externalEventListener);

      if (externalInterruptThread == null)
         externalInterruptThread = new ExternalInterruptThread(this,
                 externalInterruptListeners);
   }

   /**
    * Removes an event listener from TINI's external interrupt event chain.
    *
    * @param externalEventListener reference to the class that will be removed
    *        from the listener chain
    *
    */
   public void removeEventListener (
           ExternalInterruptEventListener externalEventListener)
   {
      externalInterruptListeners.removeElement(externalEventListener);

      if(triggerOwner == externalEventListener)
        triggerOwner = null;
   }

   /**
    * Sets the external interrupt trigger state to edge or level triggered.
    *
    * @param edgeTrigger set to true for edge triggering or false for level
    *        triggering
    * @param owner reference to ExternalInterruptEventListener that wants to
    *        change the trigger. Once the trigger is owned the owner is the only
    *        object that can change the trigger.
    *
    * @throws ExternalInterruptException if triggering is owned by another
    *         listener in the current process, by another process or by the
    *         external serial ports
    */
   public static void setTrigger(boolean edgeTrigger,//int trigger,
       ExternalInterruptEventListener owner)
     throws ExternalInterruptException
   {
     if(triggerOwner != null)
       if(triggerOwner != owner)
         throw new ExternalInterruptException("Trigger is owned by another listener");

     int triggerVal = (edgeTrigger? 2: 1);

     switch(manageTrigger(triggerVal))
     {
       case 0:
         triggerOwner = owner;
       break;

       case 1:
         throw new ExternalInterruptException("Trigger is owned by another process");

       case 2:
         throw new ExternalInterruptException("Trigger is owned by an external serial port");

       default:
         throw new ExternalInterruptException("Trigger is owned");
     }

   }
   /**
    * Gets the current trigger state.
    *
    * @return true if edge triggered and false if level triggered
    */
   public static /*int*/ boolean getTrigger()
   {
     return (manageTrigger(0) == 2);
   }

   /* --- package section --- */

   Vector externalInterruptListeners =  new Vector(3);

   /* --- private section --- */

   private static native int manageTrigger(int triggerValue);
   private static ExternalInterruptThread externalInterruptThread;
   private static ExternalInterruptEventListener triggerOwner;
}

class ExternalInterruptThread
   extends Thread
{
   Vector            listeners;
   ExternalInterrupt externalInterrupt;

   /**
     */
   ExternalInterruptThread (ExternalInterrupt externalInterrupt,
                            Vector listeners)
   {
      this.listeners         = listeners;
      this.externalInterrupt = externalInterrupt;

      setDaemon(true);
      start();
   }

   /**
     */
   public void run ()
   {
      if (SerExt_RegisterExternalInterrupt() > 0)
      {
         return;
      }

      while (true)
      {
         //suspend();
         ExtIntWaitForEvent();

         for (int i = 0; i < listeners.size(); i++)
         {

            // for now the event is an airball
            // we will need to fill in the event fields
            // once we decide what they should be
            (( ExternalInterruptEventListener ) listeners.elementAt(
               i)).externalInterruptEvent(
                  new ExternalInterruptEvent(externalInterrupt));
         }
      }
   }

   private native int SerExt_RegisterExternalInterrupt ();
   private native void ExtIntWaitForEvent();
}
