
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;


/**
 * A SHA-1 Hash and a "Better than Java" random number generator.
 *
 */
public class Security
{
   /**
    * Don't allow instantiation.
    */
   private Security ()
   {
   }

   private static native int getSystemRandomNative ();

   /**
    * Return a 32 bit random number from the system
    * random number generator.  This value is constructed
    * from four calls into the system generator each which
    * returns a byte.  Each byte is the LSB of a 16
    * bit LFSR which is mixed using XOR with
    * clock/timer values from throughout the sytem.
    * This is definitely not cryptographically secure,
    * but it's better than pure pseudo random.
    *
    * @return one 32 bit pseudorandom doubleplusgood number
    */
   public static int getRandom ()
   {
      return getSystemRandomNative();
   }

   /**
    * Method declaration
    *
    *
    * @param x
    * @param n
    *
    * @return
    */
   private static int S (int x, int n)   // apply left rotate to x by n
   {
      return ((x << n) | (x >>> (32 - n)));
   }

   /**
    * Method declaration
    *
    *
    * @param MB
    * @param index
    * @param H
    */
   private static void HashBlock (byte[] MB, int index, int[] H)
   {
      int   A, B, C, D, E, Kt, Temp;
      int[] W = new int [80];
      int   t;

      for (t = 0; t < 16; t++)
      {
         W [t] = ArrayUtils.getInt(MB, (t<<2) + index);
      }

      // this line packs the data into W correctly
      for (t = 16; t < 80; t++)
      {
         W [t] = S(W [t - 3] ^ W [t - 8] ^ W [t - 14] ^ W [t - 16], 1);
      }

      A = H [0];
      B = H [1];
      C = H [2];
      D = H [3];
      E = H [4];

      for (t = 0; t < 80; t++)
      {
         //Temp is the old Ft() method just inlined and then placed here
         if (t < 20)
         {
            Kt = 0x5A827999;
            Temp = ((B & C) | ((B ^ 0xFFFFFFFF) & D));
         }
         else if (t < 40)
         {
            Kt = 0x6ED9EBA1;
            Temp = (B ^ C ^ D);
         }
         else if (t < 60)
         {
            Kt = 0x8F1BBCDC;
            Temp = ((B & C) | (B & D) | (C & D));
         }
         else
         {
            Kt = 0xCA62C1D6;
            Temp = (B ^ C ^ D);
         }
         
         //use '+=' because Temp already has the old Ft value
         Temp += S(A, 5) + /*Ft(t, B, C, D)*/ E + W [t] + Kt;
         E    = D;
         D    = C;
         C    = S(B, 30);
         B    = A;
         A    = Temp;
      }

      H [0] += A;
      H [1] += B;
      H [2] += C;
      H [3] += D;
      H [4] += E;
   }   // end HashBlock

   /**
    * Method declaration
    *
    *
    * @param Msg
    * @param MsgLen
    *
    * @return
    */
   private static int PadMessage (byte[] Msg, int MsgLen)

   // returns the BlockNum
   {
      int  BlockNum;
      int  index;
      byte LeftOver;

      BlockNum     = (MsgLen >>> 6) + 1;
      LeftOver     = ( byte ) (64 - (MsgLen & 0x3F));
      Msg [MsgLen] = ( byte ) 0x80;

      if (( int ) LeftOver > 8)
          BlockNum--;
      index = ( int ) (BlockNum << 6);
      BlockNum++;

      Msg [(index) + 56] = ( byte ) 0x00;
      
      ArrayUtils.setInt(Msg,index+60,MsgLen<<3);
      
      return BlockNum;
   }

   /**
    * Computes a SHA-1 hash on the given message.
    * <br>See FIPS 180-1 for more information on SHA-1.
    *
    * @param message the message to hash
    *
    * @return the result of the hash (20 bytes)
    */
   public static byte[] hashMessage (byte[] MsgStr)
   {
      byte[] PdMsg;
      int    BlockNum;
      int    i;
      int[]  H = new int [5];

      H [0] = 0x67452301;
      H [1] = 0xEFCDAB89;
      H [2] = 0x98BADCFE;
      H [3] = 0x10325476;
      H [4] = 0xC3D2E1F0;
      PdMsg = new byte [MsgStr.length + 128];

      System.arraycopy(MsgStr,0,PdMsg,0,MsgStr.length);

      BlockNum = PadMessage(PdMsg, MsgStr.length);

      for (i = 0; i < BlockNum; i++)
      {
         HashBlock(PdMsg, i << 6, H);
      }

      byte[] hash = new byte [20];

/*
      // no we don't want the Dallas 180-1 hash
      for (int j = 0; j < 5; j++)
      {
         hash [j << 2]       = ( byte ) (H [j] & 0xFF);
         hash [(j << 2) + 1] = ( byte ) ((H [j] >>> 8) & 0xFF);
         hash [(j << 2) + 2] = ( byte ) ((H [j] >>> 16) & 0xFF);
         hash [(j << 2) + 3] = ( byte ) ((H [j] >>> 24) & 0xFF);
      }
*/
      // let's go with the FIPS 180-1 hash
      for (i = 0; i < 5; i++)
         ArrayUtils.setInt(hash,i<<2,H[i]);

      return hash;
   }
}
