/*---------------------------------------------------------------------------
 * Copyright (C) 1999-2001 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.dhcp;

import java.io.*;
import java.net.*;
import com.dalsemi.system.*;
import com.dalsemi.tininet.*;
import java.util.*;


/**
 * The DHCPClient supports dynamically obtaining a new IP Address or reclaiming an
 * IP Address lost on reboot from a DHCP server.
 * Both constructors are passed a DHCPListener so that DHCP status messages can be
 * relayed. After the DHCPClient object is constructed, the process in invoked through
 * the start() method. When the user no longer needs the IP Address it is recommended
 * that the stopDHCP method is called to release the IP to the server and
 * close the socket.
 * <p>
 * Refer to the following RFCs:
 * <pre>           951 - BOOTSTRAP PROTOCOL (BOOTP)
 *          1048 - BOOTP Vendor Information Extensions
 *          2131 - Dynamic Host Configuration Protocol
 *          2132 - DHCP Options and BOOTP Vendor Extentions</pre>
 * RFC repository <A HREF="http://www.ietf.org/rfc.html">http://www.ietf.org/rfc.html</A>
 * @author CM, RM
 * @version 2.0
 */
public class DHCPClient extends Thread
{
   int             state;
   private int     retries    = 0;
   private int     count      = 0;

   // The following are set in parseOption / parsePacket
   private long    dhcpT1, dhcpT2;
   private long    dhcpLease;
   private byte[]  dhcpIP         = new byte [4];
   private byte[]  dhcpServer     = new byte [4];
   private byte[]  dhcpDomainname = null;
   private byte[]  dhcpGateway    = new byte [4];
   private byte[]  dhcpNetmask    = new byte [4];
   private byte[]  dhcpDNS1       = new byte [4];
   private byte[]  dhcpDNS2       = new byte [4];
   private byte[]  dhcpMailhost   = new byte [4];

   private byte[]  data           = new byte [548];
   private byte[]  recvdata       = new byte [548];
   private byte[]  xid            = new byte [4];
   private short   index          = 0;
   private boolean dhcprun        = true;
   private boolean resend         = false;
   private static Object lock     = new Object();   // lock for network synchonization
   private DatagramSocket sock;
   private Vector  dhcpListeners  = new Vector(3);   //Pick a small initial vector size

   // BOOTP message types
   private static final int BOOT_REQUEST     = 1;
   private static final int BOOT_REPLY       = 2;

   // BOOTP message offsets
   private static final int BOOTP_XID        = 4;
   private static final int BOOTP_CIADDR     = 12;
   private static final int BOOTP_YIADDR     = 16;
   private static final int BOOTP_SNAME	     = 44;
   private static final int BOOTP_FILE	     = 108;
   private static final int BOOTP_OPTIONS    = 236;

   // DHCP magic tag
   private static final int DHCP_TAG0        = 0x63;
   private static final int DHCP_TAG1	     = 0x82;
   private static final int DHCP_TAG2	     = 0x53;
   private static final int DHCP_TAG3	     = 0x63;

   // DHCP message types
   private static final int MSG_DHCPDISCOVER = 0x01;
   private static final int MSG_DHCPOFFER    = 0x02;
   private static final int MSG_DHCPREQUEST  = 0x03;
   private static final int MSG_DHCPACK      = 0x05;
   private static final int MSG_DHCPNAK      = 0x06;
   private static final int MSG_DHCPRELEASE  = 0x07;

   // RFC 2132
   private static final int OPTION_SUBNETMASK   = 1;
   private static final int OPTION_ROUTER       = 3;
   private static final int OPTION_DNSSERVER    = 6;
   private static final int OPTION_HOSTNAME     = 12;
   private static final int OPTION_DOMAINNAME   = 15;
   private static final int OPTION_REQUESTEDIP  = 50;
   private static final int OPTION_LEASETIME    = 51;
   private static final int OPTION_OVERLOAD     = 52;
   private static final int OPTION_MESSAGETYPE  = 53;
   private static final int OPTION_SERVERIDENT  = 54;
   private static final int OPTION_PARAMETERREQ	= 55;
   private static final int OPTION_T1RENEWAL    = 58;
   private static final int OPTION_T2REBINDING  = 59;
   private static final int OPTION_CLIENTIDENT  = 61;
   private static final int OPTION_SMTPSERVER   = 69;

   private static final int STATE_INIT       = 1;
   private static final int STATE_SELECTING  = 2;
   private static final int STATE_REQUESTING = 3;
   private static final int STATE_BOUND      = 4;
   private static final int STATE_RENEWING   = 5;
   private static final int STATE_REBINDING  = 6;
   private static final int STATE_INITREBOOT = 7;
   private static final int STATE_REBOOTING  = 8;

   private static final String ERROR_FATAL   =  "Fatal Error";
   private static final String ERROR_NOREPLY =  "No reply";
   private static final String ERROR_UNKNOWN =  "Unknown Host Exception";
   private static final String ERROR_IOEX    =  "I/O Exception";

   private native static boolean isDHCPRunning();

   private native static void setDHCPRunning(boolean set);

   private static final String ipv4Broadcast = "255.255.255.255";

   private static final int SOCK_TIMEOUT = 8000;

   /**
    * Constructs a dhcp object and intializes for obtaining a new IP.
    *
    *
    * @param DL DHCPListener to report DHCP status messages
    * @throws IllegalStateException
    */
   public DHCPClient (DHCPListener DL)
      throws IllegalStateException
   {
      this(DL, null, null);
   }

   /**
    * Constructs a dhcp object and intializes to attempt to rebind to previous
    * owned IP that was lost in a reboot.
    *
    *
    * @param DL DHCPListener
    * @param serverIP IP of DHCP server where previous IP was obtained
    * @param localIP IP to obtain
    * @throws IllegalStateException  */
   public DHCPClient (DHCPListener DL, byte[] serverIP, byte[] localIP)
      throws IllegalStateException
   {
      if (isDHCPRunning())
      {
         throw new IllegalStateException("DHCP is already running.");
      }

      TININet.setIPAddress(new byte[4]);

      // init reboot constructor (previous ip, want to try to regain)
      try
      {
         if ((serverIP == null) || (localIP == null))
         {
              state = STATE_INIT;

              // Only notify if we previously did DHCP
              if ((serverIP != null) || (localIP != null))
                notifyListenersIPLost();
         }
         else
         {
            System.arraycopy(serverIP, 0, dhcpServer, 0, 4);
            System.arraycopy(localIP, 0, dhcpIP, 0, 4);

            state = STATE_INITREBOOT;
         }

         // open a socket to port 68
         if (sock == null)
            sock = new DatagramSocket(68);

         dhcpListeners.addElement(DL);
         setDHCPRunning(true);
      }
      catch (SocketException e)
      {
        //drain
      }
   }

   /**
    * Starts dhcp thread.
    *
    *
    */
   public void run ()
   {
      while (dhcprun)
      {
         switch (state)
         {

            case STATE_INIT :
               state = init();
               break;

            case STATE_SELECTING :
               state = selecting();
               break;

            case STATE_REQUESTING :
               state = requesting();
               break;

            case STATE_BOUND :
               state = bound();
               break;

            case STATE_RENEWING :
               state = renewing();
               break;

            case STATE_REBINDING :
               state = rebinding();
               break;

            case STATE_INITREBOOT :
               state = initReboot();
               break;

            case STATE_REBOOTING :
               state = rebooting();
               break;

            case -1 :
               // call a shutdown to release ip if we have one
               retries = 0;
               state = STATE_INIT;
               break;

            default :
               notifyListenersIPError(ERROR_FATAL);
               dhcprun = false;
         }
      }

      // close socket
      if (sock != null)
         sock.close();
   }

   /**
    * Stops dhcp thread, releases ip, and closes any open sockets.
    *
    */
   public void stopDHCPThread ()
   {
      try
      {
         // protect so we don't step on the bound code toes
         synchronized (lock)
         {

            // can't tell if the socket is open or closed
            // so must close it and then open a new one to
            // guarantee we have a open socket
            dhcprun = false;

            if (sock != null)
            {
               sock.close();
            }

            sock = new DatagramSocket(68);

            // Send the message out to release current ip if we have one
            releaseIP();
            sock.close();
            setDHCPRunning(false);
            this.interrupt();
         }
      }
      catch (SocketException e){}
   }

   private int init ()
   {
      // Select random xid
      ArrayUtils.setInt(xid, 0, Security.getRandom());

      // Zero our IP
      ArrayUtils.arrayFill(dhcpIP, 0, 4, (byte) 0);
     
      // Zero all times 
      dhcpT1 = dhcpT2 = count = 0;

      backOff(1);

      DatagramPacket dp;

      // If it gets kicked back to init from other states >5x send error and continue
      if (retries >= 5)
      {
         notifyListenersIPError(ERROR_NOREPLY);
         return -1;
      }

      try
      {
         buildBasicMsg(MSG_DHCPDISCOVER);

         dp = new DatagramPacket(data, data.length, InetAddress.getByName(ipv4Broadcast), 67);

         sock.send(dp);

         return STATE_SELECTING;   // goto Selecting state*/
      }
      catch (UnknownHostException e)
      {
         notifyListenersIPError(ERROR_UNKNOWN);
         return -1;
      }
      catch (InterruptedIOException e)
      {
         return STATE_INIT;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private int selecting ()
   {
      byte[] data = this.data;  //local cache it!
      int i, j;
      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);   // recv
      DatagramPacket dp;      // Field offer, if timeout resend packet

      try
      {
         if (count >= 5)
         {
            notifyListenersIPError(ERROR_NOREPLY);
            return -1;
         }

         sock.setSoTimeout(SOCK_TIMEOUT);
         sock.receive(rdp);

         // if xid does not match or not an offer
         if (!ArrayUtils.arrayComp(recvdata, BOOTP_XID, xid, 0, 4) ||
             (messageType(recvdata) != MSG_DHCPOFFER))
         {
            count++;
            return STATE_SELECTING;                 // selecting
         }

         data[242] = MSG_DHCPREQUEST;

         // get gateway, lease times, ip, server ip
         parsePacket(recvdata);

         // accept first offer unless fields are bad, continue to field more offers
         if ((dhcpServer[0] == 0) || (dhcpT1 == 0) || (dhcpT2 == 0) || (dhcpIP[0] == 0))
         {
            count++;
            return STATE_SELECTING;
         }

         index--;                     // write over ending 0xff

         // copy requested ip from yiaddr into option field
         // request ip offered
         data [index++] = OPTION_REQUESTEDIP;
         data [index++] = 0x04;
         System.arraycopy(recvdata, BOOTP_YIADDR, data, index, 4);
         index += 4;

         // copy server ip into options
         data [index++] = OPTION_SERVERIDENT;
         data [index++] = 0x04;
         System.arraycopy(dhcpServer, 0, data, index, 4);
         index += 4;

         // finish packet
         data [index++] = ( byte ) 0xff;

         // point to beginning of option
         index -= 14;

         // clear yiaddr and siaddr 
         ArrayUtils.arrayFill(data, BOOTP_YIADDR, 23, (byte) 0);

         // construct packet
         dp = new DatagramPacket(data, data.length, InetAddress.getByName(ipv4Broadcast), 67);

         sock.send(dp);

         count = 0;

         return STATE_REQUESTING;
      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(ERROR_UNKNOWN);

         return -1;
      }
      catch (InterruptedIOException e)
      {

         // go back to init stage to resend UDP
         retries++;
         backOff(retries);
         return 1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private int requesting ()
   {
      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
      DatagramPacket dp;

      try
      {
         if (count >= 5)
         {
            notifyListenersIPError(ERROR_NOREPLY);
            return -1;
         }
         if (resend == true)
         {
            // dp = new DatagramPacket(data, data.length, InetAddress.getByName(TININet.createIPString(dhcpServer)), 67);
            dp = new DatagramPacket(data, data.length, InetAddress.getByName(ipv4Broadcast), 67);
            sock.send(dp);
            resend = false;
         }
         // time out in 3 secs, assume package lost
         sock.setSoTimeout(SOCK_TIMEOUT);
         sock.receive(rdp);

         int len = rdp.getLength();

         // xid matches and ACK
         if (ArrayUtils.arrayComp(recvdata, BOOTP_XID, xid, 0, 4) &&
             (messageType(recvdata) == MSG_DHCPACK))
         {
            parsePacket(recvdata);

            TININet.setOptions(dhcpServer, dhcpIP, dhcpGateway, dhcpNetmask,
                               dhcpDNS1, dhcpDNS2, dhcpDomainname, dhcpMailhost);

            notifyListenersIPLeased();
         }
         else
         {
            count++;
            return STATE_REQUESTING;
         }

         // Arp ip to make sure no one owns it.
         // T1 and T2 are already set go and spin
         return STATE_BOUND;                    // go to bound state

         /*
          * }
          * else
          * {
          * //send decline
          * data[242] = 0x04;
          * dp.setAddress(InetAddress.getByAddr(dhcpServer));
          * sock.send(dp);
          * retries++;
          * return 1; //go to init state
          * }
          */

      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(ERROR_UNKNOWN);
         return -1;
      }
      catch (InterruptedIOException e)
      {
         backOff(count);
         count++;
         resend = true;
         return STATE_REQUESTING;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private int bound ()
   {
      DatagramPacket dp;
      try
      {
         sock.close();
         count = retries = 0;

         this.sleep(dhcpT1 * 1000);   // need to be *1000 for release

         if (!dhcprun)            // we might get interrupted by DHCPStop
           return STATE_BOUND;

         dhcpT1 = 0;
         synchronized (lock)
         {

            // open socket again
            sock = new DatagramSocket(68);

            data [242] = MSG_DHCPREQUEST;

            // set our address
            System.arraycopy(dhcpIP, 0, data, BOOTP_CIADDR, 4);

            // remove requested ip and server ip from options(!!)
            data [index] = (byte) 0xff; // finish packet

            ArrayUtils.arrayFill(data, (int) index, index+14, (byte) 0);

            dp = new DatagramPacket(data, data.length,InetAddress.getByName(TININet.createIPString(dhcpServer)), 67);

            sock.send(dp);

            return STATE_RENEWING;                // go to renew state
         }
      }
      catch (InterruptedException e)
      {
         dhcpT1 = 0;
         return STATE_BOUND;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private int renewing ()
   {
      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
      DatagramPacket dp;

      try
      {
         if (dhcpT2 - System.currentTimeMillis()/1000 < 4)
         {
            // Send out renew request
            dp = new DatagramPacket(data, data.length, InetAddress.getByName(ipv4Broadcast), 67);
            sock.send(dp);
            count = 0;
            return STATE_REBINDING;
         }

         if (resend == true)
         {
            dp = new DatagramPacket(data, data.length, InetAddress.getByName(TININet.createIPString(dhcpServer)), 67);
            sock.send(dp);
            resend = false;
         }
         // time out in 3 secs, assume package lost
         sock.setSoTimeout(SOCK_TIMEOUT);
         sock.receive(rdp);

         // xid matches
         if (ArrayUtils.arrayComp(recvdata, BOOTP_XID, xid, 0, 4))
         {
            int type = messageType(recvdata);

            if (type == MSG_DHCPACK)
            {
                // Reset T1 and T2
                resetTimes(recvdata);

                if ((dhcpT1 > 0) && (dhcpT2 > 0))
                {
                    notifyListenersIPRenewed();
                    return STATE_BOUND;
                }
                else
                {
                    // times are hosed - see if there is another offer
                    count++;
                    backOff(1);
                    resend = true;
                    return STATE_REBINDING;
                }
            }
            else if (type == MSG_DHCPNAK)
            {
                notifyListenersIPLost();
                return STATE_INIT;
            }
         }
         count++;
         backOff(1);
         resend = true;
         return STATE_RENEWING;


      }
      catch (UnknownHostException e)
      {
         notifyListenersIPError(ERROR_UNKNOWN);
         return -1;
      }
      catch (InterruptedIOException e)
      {

         count++;
         backOff(1);
         resend = true;
         return STATE_RENEWING;

      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }
   // Timer 2 has expired, one last valient effort to get renew ip, by broadcast sent from renew
   private int rebinding ()
   {
       if (dhcpLease - System.currentTimeMillis()/1000 < 4)
       {
          // Lease has expired. Duh!
          notifyListenersIPLost();
          return STATE_INIT;
       }

       // try to broadcast the rebind
       DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
       try
       {
           sock.setSoTimeout(SOCK_TIMEOUT);
           sock.receive(rdp);

           // xid matches
           if (ArrayUtils.arrayComp(recvdata, BOOTP_XID, xid, 0, 4))
           {
               int type = messageType(recvdata);

               if (type == MSG_DHCPACK) { // ACK
                 resetTimes(recvdata);

                 if ((dhcpT1 > 0) && (dhcpT2 > 0))
                 {
                   notifyListenersIPRenewed();
                   return STATE_BOUND;
                 }
                 else { // too bad...
                   notifyListenersIPLost();
                   return STATE_INIT;
                 }
               }
               else if (type == MSG_DHCPNAK) { // NAK
                 // notify and start back in init
                 notifyListenersIPLost();
                 return STATE_INIT;
               }
            }
      }
      catch (InterruptedIOException e)
      {
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         notifyListenersIPLost();
         return -1;
      }

      count++;
      backOff(1);
      return STATE_REBINDING;
   }

   private int initReboot ()
   {
      // Select random xid
      ArrayUtils.setInt(xid, 0, Security.getRandom());

      // Save IP
      byte[] localIP = new byte[4];
      System.arraycopy(dhcpIP, 0, localIP, 0, 4);

      // Zero our IP
      ArrayUtils.arrayFill(dhcpIP, 0, 4, (byte) 0);

      // Zero all times 
      dhcpT1 = dhcpT2 = 0;

      DatagramPacket dp;

      backOff(1);

      try
      {
         buildBasicMsg(MSG_DHCPREQUEST);

         index--;    // write over ending 0xff

         // copy ip from ipconfig
         data [index++] = OPTION_REQUESTEDIP;
         data [index++] = 0x04;
         System.arraycopy(localIP, 0, data, index, 4);
         index += 4;

         data [index++] = ( byte ) 0xff;

         // create datagram packet with stored data and send to dhcpserver
         dp = new DatagramPacket(data, data.length, InetAddress.getByName(ipv4Broadcast), 67);

         sock.send(dp);

         return STATE_REBOOTING;   //rebooting
      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(ERROR_UNKNOWN);
         return -1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private int rebooting ()
   {
      try
      {
         DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);

         if (count > 1)
         {
            notifyListenersIPError(ERROR_NOREPLY);
            return -1;
         }

         sock.setSoTimeout(SOCK_TIMEOUT);
         sock.receive(rdp);

         // xid matches
         if (ArrayUtils.arrayComp(recvdata, BOOTP_XID, xid, 0, 4))
         {
            int type = messageType(recvdata);

            if (type == MSG_DHCPACK)
            {
                parsePacket(recvdata);

                TININet.setOptions(dhcpServer, dhcpIP, dhcpGateway, dhcpNetmask,
                                   dhcpDNS1, dhcpDNS2, dhcpDomainname, dhcpMailhost);

                notifyListenersIPLeased();

                return STATE_BOUND;
            }
            else if (type == MSG_DHCPNAK)
            {
                notifyListenersIPLost();
                count = 0;
                return STATE_INIT;
            }
        }
        count++;
        return STATE_INITREBOOT;
      }
      catch (InterruptedIOException e)
      {
         count++;
         backOff(count);
         return STATE_INITREBOOT;
      }
      catch (IOException e)
      {
         notifyListenersIPError(ERROR_IOEX);
         return -1;
      }
   }

   private void buildBasicMsg (int type)
   {
      byte[] data = this.data; //speed this up a little, local cache

      index = 0;
      byte[] temp = new byte [6];
      TININet.getEthernetAddress(temp);
      String hostname = TININet.getHostname();
      int len = hostname.length();

      // build dhcp request
      data [index++] = BOOT_REQUEST; // op
      data [index++] = 1;            // htype
      data [index++] = 0x06;         // hlen
      data [index++] = 0x00;         // hops

      // 4 bytes of xid
      System.arraycopy(xid, 0, data, index, 4);
      index += 4;

      // set first byte of flags (broadcast bit)
      ArrayUtils.arrayFill(data, index, index+20, (byte)0);
      data [index+2] = (byte) 0x80;
      index += 20;

      // set mac addr
      System.arraycopy(temp,0,data,index,6);
      index += 6;

      // zero fill
      ArrayUtils.arrayFill(data, index, index+202, (byte)0);

      index = BOOTP_OPTIONS;

      // options field - DHCP magic signature
      data [index++] = DHCP_TAG0;
      data [index++] = ( byte ) DHCP_TAG1;
      data [index++] = DHCP_TAG2;
      data [index++] = DHCP_TAG3;

      // DHCP message type DHCPDiscover/DHCPRelease/...
      data [index++] = OPTION_MESSAGETYPE;
      data [index++] = 0x01;
      data [index++] = ( byte ) type;

      // set client hardware identifier: type Ethernet, our MAC
      data [index++] = OPTION_CLIENTIDENT;       // option type
      data [index++] = 0x07;                     // length
      data [index++] = 0x01;                     // ETHERNET
      System.arraycopy(temp, 0, data, index, 6); // MAC
      index += 6;

      if (type != MSG_DHCPRELEASE)                    // not a release message
      {
         // add hostname
         data [index++] = OPTION_HOSTNAME;
         data [index++] = (byte) len;
         System.arraycopy(hostname.getBytes(),0,data,index,len);
         index += len;


         // subnet mask and gateway request
         data [index++] = OPTION_PARAMETERREQ;
         data [index++] = 0x05;			// 5 options requested
         data [index++] = OPTION_SUBNETMASK;
         data [index++] = OPTION_ROUTER;
         data [index++] = OPTION_SMTPSERVER;
         data [index++] = OPTION_DNSSERVER;
         data [index++] = OPTION_DOMAINNAME;
      }
      else {
         // set our address
         System.arraycopy(dhcpIP, 0, data, BOOTP_CIADDR, 4);

         // copy server ip into options
         data [index++] = OPTION_SERVERIDENT;
         data [index++] = 0x04;
         System.arraycopy(dhcpServer, 0, data, index, 4);
         index += 4;
      }

      data [index++] = ( byte ) 0xFF;   // end message

      // clear rest of data area
      ArrayUtils.arrayFill(data, index, data.length-1, (byte)0);
   }

   // method to get dhcp message type and to make sure message is ours
   private int messageType (byte[] recvdata)
   {
        if (recvdata[0] != BOOT_REPLY)
          return -1;

        // options field - DHCP magic signature
        if ((recvdata[BOOTP_OPTIONS]   != DHCP_TAG0) ||
            (recvdata[BOOTP_OPTIONS+1] != (byte) DHCP_TAG1) ||
            (recvdata[BOOTP_OPTIONS+2] != DHCP_TAG2) ||
            (recvdata[BOOTP_OPTIONS+3] != DHCP_TAG3))
          return -1;

        int i = BOOTP_OPTIONS+4; // start the search after dhcp tag

        while (recvdata[i] != ( byte ) 0xff)
        {
            if (recvdata [i] == OPTION_MESSAGETYPE)
                return (recvdata [i + 2]);

            i += (recvdata[i + 1] & 0xff) + 2;
       }

       return -1;
   }


   // examine single DHCP option and set variable - user hook
   /** This method is called at least once for each received DHCP
    * packet; it is also called for all DHCP options that are not
    * interpreted by the <CODE>DHCPClient</CODE> class. A subclass
    * could override this method and implement its own option
    * handling.
    * @param recvdata the byte array contains the complete packet
    * received from the DHCP server.
    * @param position position of the option inside the byte array.
    */   
   public void parseOption(byte[] recvdata, int position)
   {
   }


   // loop through and find lease times, gateway, subnet mask, serverip
   private void parsePacket(byte[] recvdata)
   {
      int optionOverload = 0;
      int i = BOOTP_OPTIONS+4;   // start the search after standard dhcp options

      do {
        while (recvdata[i] != ( byte ) 0xff)
        {
          int len = recvdata[i+1] & 0xff;

          switch (recvdata[i]) {
            // Times are set in resetTimes

            // server ip
            case OPTION_SERVERIDENT:
               System.arraycopy(recvdata, i + 2, dhcpServer, 0, 4);
               break;

            // gateway
            case OPTION_ROUTER:
               System.arraycopy(recvdata, i + 2, dhcpGateway, 0, 4);
               break;

            // subnet mask
            case OPTION_SUBNETMASK:
               System.arraycopy(recvdata, i + 2, dhcpNetmask, 0, 4);
               break;

            // domain name
            case OPTION_DOMAINNAME:
               // Strip off trailing 0 if there is one
               if (recvdata[i + len + 1] == 0) {
                 dhcpDomainname = new byte[len-1];
                 System.arraycopy(recvdata, i + 2, dhcpDomainname, 0, len - 1);
               }
               else {
                 dhcpDomainname = new byte[len];
                 System.arraycopy(recvdata, i + 2, dhcpDomainname, 0, len);
               }
               break;

            // smtp servers
            case OPTION_SMTPSERVER:
               System.arraycopy(recvdata, i + 2, dhcpMailhost, 0, 4);
               break;

            // dns servers
            case OPTION_DNSSERVER:
               // primary dns
               System.arraycopy(recvdata, i + 2, dhcpDNS1, 0, 4);

               // secondary dns
               if (len >= 8)
                    System.arraycopy(recvdata, i + 2 + 4, dhcpDNS2, 0, 4);

               break;

            // option overload
            case OPTION_OVERLOAD:
               optionOverload = recvdata[i + 2];
               break;

            // all others
            default:
               parseOption(recvdata, i);
               break;
          }

          i += len + 2;
        }

        if ((optionOverload == 1) || (optionOverload == 3)) { // the file field is used to hold options
          optionOverload--;
          i = BOOTP_FILE;
        }
        else
          if (optionOverload == 2) {    // the sname field is used to hold options
            optionOverload = 0;
            i = BOOTP_SNAME;
          }

      } while (optionOverload != 0);

      // set the times
      resetTimes(recvdata);

      // copy client ip
      System.arraycopy(recvdata, BOOTP_YIADDR, dhcpIP, 0, 4);
   }


   /* Function used to reset timers after a lease renewal*/
   private void resetTimes(byte[] recvdata)
   {
      dhcpLease = dhcpT1 = dhcpT2 = 0; // Reset T1 and T2

      int optionOverload = 0;
      int i = BOOTP_OPTIONS+4;   // start the search after standard dhcp options

      do {
        while (recvdata [i] != ( byte ) 0xff)
        {
          switch (recvdata [i])
          {
            // Times
            case OPTION_LEASETIME:
            case OPTION_T1RENEWAL:
            case OPTION_T2REBINDING:
            {
               long temp = 0;
               for (int j = 0; j < 4; j++)
               {
                  temp = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j))) | temp;
               }
               if (recvdata[i] == OPTION_LEASETIME)
                    dhcpLease = temp;
               else
               {
                    temp &= 0x7FFFFFFF;
                    if (recvdata[i] == OPTION_T1RENEWAL)
                        dhcpT1 = temp;
                    else
                        dhcpT2 = temp;
               }
               break;
            }

            // option overload
            case OPTION_OVERLOAD:
               optionOverload = recvdata[i + 2];
               break;

            default :
               break;
          }

          i += (recvdata[i + 1] & 0xff) + 2;
        }

        if ((optionOverload == 1) || (optionOverload == 3))  { // the file field is used to hold options
          optionOverload--;
          i = BOOTP_FILE;
        }
        else
          if (optionOverload == 2) {    // the sname field is used to hold options
            optionOverload = 0;
            i = BOOTP_SNAME;
          }
      } while (optionOverload != 0);

      if (dhcpT2 > dhcpLease)
        dhcpT2 = 0;
      if (dhcpT1 > dhcpT2)
        dhcpT1 = 0;
      if (((dhcpT1 == 0) || (dhcpT2 == 0)) && (dhcpLease > 0))
      {
         dhcpT1 = ( int ) ((dhcpLease / 2) & 0x7FFFFFFF);
         dhcpT2 = ( int ) (((dhcpLease * 7) / 8) & 0x7FFFFFFF);
      }
      if (dhcpT2 > 0)
        dhcpT2 += System.currentTimeMillis()/1000;
      dhcpLease += System.currentTimeMillis()/1000;
   }

   private void backOff (int bcount)
   {
      if (!dhcprun)
        return;

      try
      {
        int    timeout = 0;
        try {
          Random r       = new Random();

          timeout = r.nextInt();
          timeout &= 0x3F;

          bcount &= 0x07;

          if (bcount < 4)
            timeout %= (bcount * 7);
          else                     // keep timeouts around 1 minute
          {
            timeout %= (60);

            if (timeout < 40)
               timeout += 40;
          }
        }
        catch (ArithmeticException e) {
          timeout = 60; // I don't trust this code!!!
        }
        this.sleep(timeout * 1000);
      }
      catch (InterruptedException e){}
   }

   private void releaseIP ()
   {
      DatagramPacket dp;

      try
      {
         buildBasicMsg(MSG_DHCPRELEASE);

         dp = new DatagramPacket(data, data.length, InetAddress.getByName(TININet.createIPString(dhcpServer)), 67);
         sock.send(dp);
      }
      catch (InterruptedIOException e)
      {
      }
      catch (IOException e)
      {
      }
      finally
      {
         try {
           sock.close();
         }
         catch (Exception _)
         {
         }
      }

      notifyListenersIPLost();
   }

   /**
    * Adds listener for accepting dhcp status messages.
    *
    *
    * @param newListener DHCPListener for accepting status messages
    *
    */
   public void addDHCPListener (DHCPListener newListener)
   {
      dhcpListeners.addElement(newListener);
   }

   /**
    * Removes listener, to discontinue acceptance of status messages.
    *
    *
    * @param thisListener DHCPListener to be removed from list
    *
    */
   public void removeDHCPListener (DHCPListener thisListener)
   {
      dhcpListeners.removeElement(thisListener);
   }

   private void notifyListenersIPError (String message)
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipError(message);
      }
   }

   private void notifyListenersIPLost ()
   {
      // set TININet IP to undefined
      TININet.setIPAddress(new byte[4]);
      // HACK ALARM!!
      // We store IP to subnet mask so we can remember it
      // after a reboot and still have the current IP set
      // to 0.
      // HACK ALARM!!
      TININet.setSubnetMask(dhcpIP);
      count = 0;

      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipLost();
      }
   }

   private void notifyListenersIPRenewed ()
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipRenewed();
      }
   }

   private void notifyListenersIPLeased ()
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipLeased();
      }
   }

   /** This method returns the DHCP lease expiration timeout.
    *
    *
    * @return the number of seconds relative to system start, i.e.
    * subtract <CODE>System.currentTimeMillis()/1000</CODE> from this
    * value to get the remainder of the DHCP lease life in seconds.
    */   
   public long getLeaseTimeout()
   {
     return dhcpLease;
   }
}

