
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.dns;

import java.net.*;
import java.net.InetAddress;
import java.io.*;
import java.util.*;

/**
 * The <code>DNSClient</code> class allows DNS (Domain Name System)
 * lookups per RFC 1035.  Can perform forward or reverse
 * lookups.
 */
public class DNSClient
{
   private byte[]         HEADER  =
   {
      ( byte ) 0x55, ( byte ) 0x44, ( byte ) 0x01, ( byte ) 0x00,
      ( byte ) 0x00, ( byte ) 0x01, ( byte ) 0x00, ( byte ) 0x00,
      ( byte ) 0x00, ( byte ) 0x00, ( byte ) 0x00, ( byte ) 0x00
   };

   private byte[]         QUERY   = { ( byte ) 0x00, ( byte ) 0x01,
                              ( byte ) 0x00, ( byte ) 0x01 };

   private InetAddress DNS1    = null;
   private InetAddress DNS2    = null;
//   private InetAddress gateway = null;
   private int    DNSTimeout = 0;

   /**
    * Create an instance of <code>DNSClient</code>.
    */
   public DNSClient()
   {
   }

   /**
    * Sets the DNS timeout.
    *
    * @param timeout Set to 0 for exponential backoff timeout or
    *                non-zero value for single timeout in milliseconds.
    */
   public void setDNSTimeout(int timeout)
   {
     DNSTimeout = timeout;
   }

   // Need to check for a real IP address, not a name before calling this.
   private boolean verifyDNSIP(String str)
   {
     if ((str != null) && (str.length() > 0))
     {
        /* Make sure we have an IP address here, or we get into an
           infinite loop trying to resolve string addresses with
           InetAddress
        */
        int len = str.length();

        for (int i = 0; i < len; i++)
        {
           char ch = str.charAt(i);

           if (!((ch >= '0' && ch <= '9') || (ch == '.')))
           {
              //com.dalsemi.jellybeans.Debug.debugDump("DNS server string is not an address");
              //com.dalsemi.jellybeans.Debug.debugDump(dnsserver);
              return false;
           }
        }

       return true;
     }

     return false;
   }

   /**
    * Sets the primary DNS server to use for resolution.
    * <br>Note: Address must be in numerical IP format
    *
    * @param dns1 IP address of DNS Server
    */
   public void setPrimaryDNS (String dns1)
   {
      try
      {
        if (verifyDNSIP(dns1))
        {
          DNS1 = InetAddress.getByName(dns1);
        }
        else
          DNS1 = null;
      }
      catch (UnknownHostException e)
      {
        // DRAIN!
      }
   }

   /**
    * Sets the secondary DNS server to use for resolution.
    * <br>Note: Address must be in numerical IP format
    *
    * @param dns2 IP address of DNS Server
    */
   public void setSecondaryDNS (String dns2)
   {
      try
      {
        if (verifyDNSIP(dns2))
          DNS2 = InetAddress.getByName(dns2);
        else
          DNS2 = null;
      }
      catch (UnknownHostException e)
      {
        // DRAIN!
      }
   }

   /**
    * Method setProxy
    *
    *
    * @param d
    *
    */
/*   public void setProxy (String d)
   {
      try
      {
        if (verifyDNSIP(d))
          gateway = InetAddress.getByName(d);
        else
          gateway = null;
      }
      catch (UnknownHostException e)
      {
        // DRAIN!
      }
   }
*/

   /**
    * Performs a reverse DNS lookup.
    *
    * @param ip Address to look up
    * @return DNS response, or <code>null</code> for failed lookup
    */
   public String[] getByIP (String ip)
   {
     // return getByIP(toIPBytes(ip));
     return getByIP(com.dalsemi.tininet.TININet.createIPFromString(ip));
   }

   /**
    * Performs a reverse DNS lookup.
    *
    * @param ip Address to look up
    * @return DNS response, or <code>null</code> for failed lookup
    */
   public String[] getByIP (byte[] ip)
   {
      byte[] QUESTION = makeQuery(Integer.toString(ip [3] & 0x0ff) + "."
                                  + Integer.toString(ip [2] & 0x0ff) + "."
                                  + Integer.toString(ip [1] & 0x0ff) + "."
                                  + Integer.toString(ip [0] & 0x0ff)
                                  + ".in-addr.arpa");

      QUESTION [QUESTION.length - 3] = 0x0c;   //get a PTR

      return transmit(QUESTION);
   }

   /**
    * Performs a DNS lookup.
    *
    * @param name FQDN to look up
    * @return DNS response, or <code>null</code> for failed lookup
    */
   public String[] getByName (String name)
   {

      //System.out.println("DNSClient.getByName: "+name);
      byte[] QUESTION = makeQuery(name);

      return transmit(QUESTION);
   }

   /**
    * Build query for specified host
    *
    *
    * @param host
    *
    * @return query
    *
    */
   private byte[] makeQuery (String host)
   {
      int    d1       = 0;   //the location of the last dot
      byte[] QUESTION = null;

      QUESTION = new byte [host.length() + 6];

      int    i = 0;
      byte[] s = host.getBytes();

      for (i = 0; i < s.length; i++)
      {
         if (s [i] == ( byte ) '.')
         {
            QUESTION [d1] = ( byte ) (i - d1);
            d1            = i + 1;
         }
         else
            QUESTION [i + 1] = s [i];
      }

      QUESTION [d1]    = ( byte ) (i - d1);
      QUESTION [i + 1] = 0;

      System.arraycopy(QUERY, 0, QUESTION, i + 2, 4);

      return QUESTION;
   }

   /**
    * Convert a string such as "191.168.1.50" to the
    * four byte array representation.
    *
    * @param x string to convert
    *
    * @return array representation of the IP address
    *
    */
   /*private byte[] toIPBytes (String x)
   {
      byte[] b   = x.getBytes();
      int    i   = 0;
      byte[] ret = new byte [4];

      for (int index = 0; index < 4; index++)
      {
         while ((i < b.length) && (b [i] != ( byte ) '.'))
         {
            ret [index] = ( byte ) (ret [index] * 10 + b [i] - 0x30);

            i++;
         }

         i++;
      }

      return ret;
   }*/

   /**
    * Transmit data query.
    *
    *
    * @param data
    *
    * @return strings returned from DNS lookup
    *
    */
   private String[] transmit (byte[] data)
   {
      byte[] senddata = new byte [data.length + HEADER.length];

      System.arraycopy(HEADER, 0, senddata, 0, HEADER.length);
      System.arraycopy(data, 0, senddata, HEADER.length, data.length);

      String[] reply = null;

      if (DNS1 != null)
         reply = send(senddata, DNS1);

      if (reply == null)
         if (DNS2 != null)
            reply = send(senddata, DNS2);

/*      if (reply == null)
         if (gateway != null)
            reply = send(senddata, gateway);
*/

      return reply;
   }

   /**
    * Method send
    *
    *
    * @param outbuf
    * @param dnsserver
    *
    * @return
    *
    */
   private String[] send (byte[] outbuf, InetAddress dnsserver)
   {
      DatagramSocket sock = null;
      DatagramPacket pack;
      int            maxRetry;
      int            currentTimeout;

      if (DNSTimeout == 0)
      {

         /* Start at 2000ms, and double four times */
         currentTimeout = 2000;
         maxRetry       = 4;
      }
      else
      {
         currentTimeout = DNSTimeout;
         maxRetry       = 1;
      }

      for (int retry = 0; retry < maxRetry; retry++, currentTimeout *= 2)
      {
         try
         {
            sock = new DatagramSocket();

            sock.setSoTimeout(currentTimeout);

            //System.out.println("after getbyname dnsserver");
            pack = new DatagramPacket(outbuf, outbuf.length,
                                      dnsserver, 53);

            sock.send(pack);

            //                  System.out.print("new-sent...");
            DatagramPacket rec = new DatagramPacket(new byte [512], 512);

            sock.receive(rec);
            sock.close();

            //                  System.out.println("got it\n\n");
            byte[] data = rec.getData();

            //System.out.println("length "+rec.getLength());

            /*            for (int i = 0;i < rec.getLength();i++)
                                    {
                                        if (i % 16 == 0) System.out.println();
                                        int b = data[i] & 0xFF;
                                        int nibble;

                                        if (b < 16)
                                            System.out.write('0');
                                        else
                                        {
                                            nibble = (b >>> 4) & 0x0F;
                                            System.out.write((nibble > 9)?((nibble - 10) + 'A'):(nibble + '0'));
                                        }
                            nibble = b & 0x0F;
                            System.out.write((nibble > 9)?((nibble - 10) + 'A'):(nibble + '0'));
                            System.out.write(' ');
                                    }
            System.out.println();
            */

            // Check for RCODE. (errors)
            int x = data [3] & 0x0f;

            if (x != 0)
               return null;

            // Get the question count
            int qn = ((data [4] << 8) | (data [5] & 0x0ff)) & 0x0ffff;

            //                  System.out.println("QUESTIONS  : "+Integer.toHexString(x));
            // Get the answer count
            int an = ((data [6] << 8) | (data [7] & 0x0ff)) & 0x0ffff;

            //                  System.out.println("ANSWERS    : "+Integer.toHexString(x));
            // may need to handle this for recursion not implemented
            if (an == 0)
               return null;

            // Index past the question we sent.  We get copied back on the
            // question exactly, so we should just be able to index past the
            // length of outbuf.
            int index = outbuf.length;

            //                  System.out.println("\n\nANSWERS:--------------------");
            // Parse the answer and return an array of strings back to
            // the caller.
            return parse(an, index, data);
         }
         catch (IOException e)
         {
            if (sock != null)
               sock.close();
         }
         catch (Exception e)
         {

            //                  System.out.println("Caught an exception: "+e.toString());
            //                  e.printStackTrace();
         }
         finally
         {
            if (sock != null)
               sock.close();
         }
      }

      return null;
   }

   /**
    * Method parse
    *
    *
    * @param num
    * @param index
    * @param data
    *
    * @return
    *
    */
   private String[] parse (int num, int index, byte[] data)
   {

      // Array of strings to return.
      String[] ret = new String [num];

      for (int i = 0; i < num; i++)
      {

         // Skip the domain name data, it's one of the following
         // 1. A single compressed ID offset
         // 2. A domain name string with a single compresed ID offset at the end.
         // 3. A list of names with no compressed ID offset fields.b
         int count = data [index] & 0x0ff;

         index++;

         if ((count & 0xc0) == 0xc0)
         {
            index++;
         }
         else
            while (count != 0)
            {

               // If the byte is a compression byte, skip the next byte also.
               int y = count;

               for (int z = 0; z < y; z++)
               {

                  //                                      System.out.print(Integer.toHexString(data[index]&0x0ff)+" ");
                  index++;
               }

               count = data [index] & 0x0ff;

               index++;

               //                              System.out.print(" ");
            }

         // Get the query type
         int type = ((data [index] << 8) | (data [index + 1] & 0x0ff))
                    & 0x0ffff;

         // Get the query class
         int clss = ((data [index + 2] << 8) | (data [index + 3] & 0x0ff))
                    & 0x0ffff;

         index = index + 4;

         // Get time to live
         int ttl = ((data [index] << 8) | (data [index + 1] & 0x0ff))
                   & 0x0ffff;

         ttl = ttl
               | (((data [index + 2] << 8) | (data [index + 3] & 0x0ff))
                  & 0x0ffff);

         // Get the resource data length.
         int rdl = ((data [index + 4] << 8) | (data [index + 5] & 0x0ff))
                   & 0x0ffff;

         index = index + 6;

         StringBuffer sb = new StringBuffer();

         // If this is a PTR query (IP address to name)
         if (type == 0x0C)
         {
            int strlen   = data [index] & 0x0ff;
            int rdl_left = rdl;

            index++;
            rdl_left--;

            while ((strlen != 0) || (rdl_left > 0))
            {

               // If the byte is a compression byte, skip the next byte also.
               if ((strlen & 0xc0) == 0xc0)
               {
                  index++;
                  rdl_left--;

                  // Call function to get string pointed to by the length
               }

               // Append the next segment of the string to the buffer
               int y = strlen;

               for (int z = 0; z < y; z++)
               {

                  //                                      System.out.print(Integer.toHexString(data[index]&0x0ff)+" ");
                  sb.append(( char ) data [index]);

                  index++;
                  rdl_left--;
               }

               strlen = data [index] & 0x0ff;

               index++;
               rdl_left--;

               if (rdl_left != 0)
                  sb.append('.');
            }

            ret [i] = sb.toString();
         }
         else if (type == 0x01)
         {

            // Easy, just get the 4 byte ip address
            /*for (int k = 0; k < rdl; k++)
            {
               sb.append(Integer.toString(0x0ff & data [index]));

               if (k + 1 < rdl)
                  sb.append('.');

               index++;
            }*/

            //ret [i] = sb.toString();
            byte[] tempdata = new byte[4];
            System.arraycopy(data, index, tempdata, 0, 4);
            ret [i] = com.dalsemi.tininet.TININet.createIPString(tempdata);
            index += 4;
         }
         else
         {

            // Skip this resource, we don't want it.
            index += rdl;
         }
      }

      return ret;
   }

   /*      public static void followpointer(int ptr, byte[] data)
           {
                   ptr = ptr & 0x03f;  //get rid of 0xc0 part//

                   System.out.println("Following pointer "+ptr);
                   //try{Thread.sleep(2000);}catch(Exception e){}
                   int count=data[ptr] & 0x0ff;
                   ptr++;
                   System.out.print("( ");
                   while (count!=0)
                   {
                           int y=count;
                           for (int z=0;z<y;z++)
                           {
                                   System.out.print(Integer.toHexString(data[ptr]&0x0ff)+" ");
                                   ptr++;
                           }
                           count = data[ptr] & 0x0ff;
                           ptr++;
                           System.out.print(" ");
                   }
                   System.out.print(")");

           }
   */
}
