
/*---------------------------------------------------------------------------
 * Copyright (C) 1999-2001 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.http;

import java.net.*;
import java.io.*;
import java.util.*;
import com.dalsemi.system.*;

/**
 * This class implements an HTTP server. This class currently only supports the
 * GET method. The <code>serviceRequests</code> method blocks on a ServerSocket
 * accept call. A new thread is spawned to service each new connection.
 * The method <code>serviceRequest(Object lock)</code> can be used to
 * synchronize access, using the lock object, to a requested web page. This might be useful if the
 * server dynamically updates a page.
 * <H3> Usage </H3>
 * <H4> Example</H4><pre><CODE>
 * boolean    webServerEnabled = true;
 * Object     lock = new Object();
 * HTTPServer httpServer = new HTTPServer();
 * boolean    loggingFailed = false;
 *
 * //...
 *
 * try
 * {
 *   // create an instance of HTTPServer on port httpPort
 *   httpServer  = new HTTPServer(httpPort);
 *   // override the default index page
 *   httpServer.setIndexPage(webIndex);
 *   // override the default HTTP root
 *   httpServer.setHTTPRoot(webRoot);
 *   // override the default log file name
 *   httpServer.setLogFilename(webLog);
 * }
 * catch(HTTPServerException h)
 * {
 *   System.out.println(h.toString());
 * }
 *
 * try
 * {
 *   // enable logging
 *   // NOTE: the log file is always appended and
 *   //       will eventually consume all free memory
 *   //       if it is not managed.
 *   httpServer.setLogging(true);
 * }
 * catch(HTTPServerException h)
 * {
 *   // problem with log file
 *   loggingFailed = true;
 *
 *   if(debugOn)
 *   {
 *     System.out.println(h.toString());
 *   }
 * }
 *
 * while(webServerEnabled)
 * {
 *   httpServer.serviceRequests(lock);
 *   //...
 * }
 * </pre></CODE>
 */
public class HTTPServer
{

   /* 2XX: generally "OK" */

   /** Standard HTTP_OK response */
   public static final int HTTP_OK = 200;

   /** Standard HTTP_CREATED response */
   public static final int HTTP_CREATED = 201;

   /*
    * public static final int HTTP_ACCEPTED       = 202;
    * public static final int HTTP_NOT_AUTHORITATIVE  = 203;
    * public static final int HTTP_NO_CONTENT     = 204;
    * public static final int HTTP_RESET        = 205;
    * public static final int HTTP_PARTIAL      = 206;
    */

   /* 3XX: relocation/redirect */

   /*
    * public static final int HTTP_MULT_CHOICE     = 300;
    * public static final int HTTP_MOVED_PERM      = 301;
    * public static final int HTTP_MOVED_TEMP      = 302;
    * public static final int HTTP_SEE_OTHER       = 303;
    * public static final int HTTP_NOT_MODIFIED    = 304;
    * public static final int HTTP_USE_PROXY       = 305;
    */

   /* 4XX: client error */

   /** Standard HTTP_BAD_REQUEST response */
   public static final int HTTP_BAD_REQUEST = 400;

   /** Standard HTTP_UNAUTHORIZED response */
   public static final int HTTP_UNAUTHORIZED = 401;

   // public static final int HTTP_PAYMENT_REQUIRED   = 402;

   /** Standard HTTP_FORBIDDEN response */
   public static final int HTTP_FORBIDDEN = 403;

   /** Standard HTTP_NOT_FOUND response */
   public static final int HTTP_NOT_FOUND = 404;

   /*
    * public static final int HTTP_BAD_METHOD        = 405;
    * public static final int HTTP_NOT_ACCEPTABLE    = 406;
    * public static final int HTTP_PROXY_AUTH        = 407;
    * public static final int HTTP_CLIENT_TIMEOUT    = 408;
    * public static final int HTTP_CONFLICT          = 409;
    * public static final int HTTP_GONE              = 410;
    * public static final int HTTP_LENGTH_REQUIRED   = 411;
    * public static final int HTTP_PRECON_FAILED     = 412;
    * public static final int HTTP_ENTITY_TOO_LARGE  = 413;
    * public static final int HTTP_REQ_TOO_LONG      = 414;
    */

   /** Standard HTTP_UNSUPPORTED_TYPE response */
   public static final int HTTP_UNSUPPORTED_TYPE = 415;

   /* 5XX: server error */

   /** Standard HTTP_SERVER_ERROR response */
   public static final int HTTP_SERVER_ERROR = 500;

   /** Standard HTTP_INTERNAL_ERROR response */
   public static final int HTTP_INTERNAL_ERROR = 501;

   /*
    * public static final int HTTP_BAD_GATEWAY    = 502;
    * public static final int HTTP_UNAVAILABLE    = 503;
    * public static final int HTTP_GATEWAY_TIMEOUT  = 504;
    * public static final int HTTP_VERSION      = 505;
    */

   /** Standard type TYPE_SIMPLE_REQUEST */
   public static final int TYPE_SIMPLE_REQUEST = 1;

   /** Standard type TYPE_FULL_REQUEST */
   public static final int TYPE_FULL_REQUEST = 2;

   /** Standard type TYPE_FULL_RESPONSE */
   public static final int TYPE_FULL_RESPONSE = 3;

   /** Default port number (port 80)  */
   public static final int DEFAULT_HTTP_PORT = 80;
   // HTTP request types

   /** Request unsupported */
   public final static int UNSUPPORTED = 0;

   /** GET request */
   public final static int GET = 1;

   /** POST request */
   public final static int POST = 2;

   /** HEAD request */
   public final static int HEAD = 3;

   /** OPTIONS request */
   public final static int OPTIONS = 4;

   /** PUT request */
   public final static int PUT = 5;

   /** DELETE request */
   public final static int DELETE = 6;

   /** TRACE request */
   public final static int TRACE = 7;

   /**
    * Default constructor. Uses the default port and disables logging.
    *
    * @throws HTTPServerException if server instantiation error occurs
    */
   public HTTPServer ()
      throws HTTPServerException
   {
      this(DEFAULT_HTTP_PORT, false);
   }

   /**
    * Creates an <code>HTTPServer</code> using port <code>httpPort</code>. Logging is disabled.
    *
    * @param httpPort port number for server
    *
    * @throws HTTPServerException if server instantiation error occurs
    */
   public HTTPServer (int httpPort)
      throws HTTPServerException
   {
      this(httpPort, false);
   }

   /**
    * Creates an <code>HTTPServer</code> using port <code>httpPort</code>. Logging is enabled
    * if <code>logEnabled</code> is <code>true</code>.
    *
    * @param httpPort port number for server
    * @param logEnabled set logging option
    *
    * @throws HTTPServerException if server instantiation error occurs
    */
   public HTTPServer (int httpPort, boolean logEnabled)
      throws HTTPServerException
   {
      setPortNumber(httpPort);
      setLogging(logEnabled);
   }

   /**
    * Allows user to change the mime type for bitmaps. There
    * seem to be several mime-types for bmp files for different
    * browsers/servers.
    */
   public static void setBitmapMimeType (String newMimeType)
   {
      MIME_IMAGE_BMP = newMimeType;
   }

   /**
    * Returns the logging status.
    *
    * @return <code>true</code> if logging is enabled
    */
   public boolean getLogging ()
   {
      return logEnabled;
   }

   /**
    * Sets the logging status. If logging is enabled the server will attempt to
    * open a new log file if one does not exist. If logging is disabled the server
    * will attempt to close the log file. Log files are always appended.<BR><BR>
    * <B>NOTE: Logging is discouraged since the memory on TINI is limited</B><BR><BR>
    *
    * @param logEnabled <code>true</code> if the server is to write to a default log file
    *
    * @exception HTTPServerException if a log file error occurs
    */
   public void setLogging (boolean logEnabled)
      throws HTTPServerException
   {
      try
      {
        if(this.logEnabled)
        {
          if(!logEnabled)
          {
            if(logFileWriter != null)
              logFileWriter.close();
          }
        }
      }
      catch (IOException i)
      {
         throw new HTTPServerException("Failed log file access");
      }

      this.logEnabled = logEnabled;
   }

   /**
    * Returns the name of the log file. This represents the file name including path.
    *
    * @return log file name
    */
   public String getLogFilename ()
   {
      return logFileName;
   }

   /**
    * Sets the log file name. If the path is not given, the file will be written to
    * the current directory.
    *
    * @param logFileName name of the log file
    */
   public void setLogFilename (String logFileName)
   {
      this.logFileName = logFileName;

      // clean up the old log file
      try
      {
        if(logFileWriter != null)
          logFileWriter.close();
      }
      catch(Throwable t)
      {
        //drain
      }

      if(logEnabled)
      {
        setLogging(false);
        setLogging(true);
      }
   }

   /**
    * Returns the HTTP root of the server.
    *
    * @return server root
    */
   public String getHTTPRoot ()
   {
      return httpRoot;
   }

   /**
    * Sets the http root. All files from <code>httpRoot</code> to the leaves of
    * the directory tree can be accessed by HTTP requests. Files between the
    * root of the file system and <code>httpRoot</code> cannot be accessed with
    * HTTP requests.
    *
    * @param httpRoot path indicating the root of the server
    *
    */
   public void setHTTPRoot (String httpRoot)
   {
      if (httpRoot.charAt(httpRoot.length() - 1) != '/')
         httpRoot += "/";

      this.httpRoot = httpRoot;
   }

   /**
    * Returns the server's default index page.
    *
    * @return current index page
    */
   public String getIndexPage ()
   {
      return indexPage;
   }

   /**
    * Sets the server's index page.
    *
    * @param indexPage page to be used by the server as the default index page
    */
   public void setIndexPage (String indexPage)
   {
      this.indexPage = indexPage;
   }

   /**
    * Returns the server's current port number.
    *
    * @return current port number
    */
   public int getPortNumber ()
   {
      return httpPort;
   }

   /**
    * Sets the server's port number. <B>Note:</B> The server will close any
    * open ports and opens the new port.
    *
    * @param httpPort number of the port to be opened
    *
    * @throws HTTPServerException if the port is opened by another server
    */
   public void setPortNumber (int httpPort)
      throws HTTPServerException
   {
      this.httpPort = httpPort;

      try
      {

         // close any open sockets
         if (httpServerSocket != null)
            httpServerSocket.close();

         // Create a socket for our HTTP server
         httpServerSocket = new ServerSocket(httpPort);
      }
      catch (BindException b)
      {
         throw new HTTPServerException("Failed to bind on port " + httpPort);
      }
      catch (IOException i)
      {
         throw new HTTPServerException("Failed to create a ServerSocket");
      }
   }

   /**
    * Checks for incoming client HTTP request and services supported requests
    * that are detected.
    *
    * @return 0 is returned (return value is legacy from pre-Thread TINI APIs)
    *
    * @throws HTTPServerException if a server error occurs
    */
   public int serviceRequests ()
      throws HTTPServerException
   {
      return serviceRequests(new Object());
   }

   /**
    * Checks for incoming client HTTP request and services supported requests
    * that are detected. The lock is used to syncronize access to a web page.
    *
    * @param lock lock for exclusive access to web page
    *
    * @return 0 is returned (return value is legacy from pre-Thread TINI APIs)
    *
    * @throws HTTPServerException if a server error occurs
    */
   public int serviceRequests (Object lock)
      throws HTTPServerException
   {
     Socket       socket = null;

      try
      {
            socket = httpServerSocket.accept();

            (new Thread(new HTTPWorker(socket, this, lock))).start();
      }
      catch (IOException i)
      {
         throw new HTTPServerException("ServerSocket accept failure");
      }
      catch(Throwable t) // clean up socket if anything bad happens
      {        
        try
        {
          socket.close();
        }
        catch(Throwable _){}// drain
      }

      return 0;
   }

   /* --- package protected --- */

   final static String     DEFAULT_ROOT      = "/webroot/";
   final static String     DEFAULT_PAGE      = "index.html";
   final static String     DEFAULT_LOG_FILE  = "/default.log";

   // more strings
   final static String VERSION = "1.0";
   final static String CRLF    = "\r\n";

   // mime types
   final static String MIME_TEXT_PLAIN = "text/plain";
   final static String MIME_TEXT_HTML  = "text/html";
   final static String MIME_IMAGE_GIF  = "image/gif";
   final static String MIME_IMAGE_JPG  = "image/jpeg";

   // there seem to be multiple mime types for this
   // so we cannot make this final
   static String       MIME_IMAGE_BMP = "image/bmp";
   final static String MIME_APP_OS    = "application/octet-stream";

      // request strings
   final static String FIELD_USER_AGENT      = "User-Agent: ";
   final static String FIELD_HOST            = "Host: ";
   final static String FIELD_PRAGMA          = "Pragma: ";
   final static String FIELD_ACCEPT          = "Accept: ";
   final static String FIELD_ACCEPT_ENCODING = "Accept-Encoding: ";
   final static String FIELD_ACCEPT_LANGUAGE = "Accept-Language: ";
   final static String FIELD_ACCEPT_CHARSET  = "Accept-Charset: ";
   final static String FIELD_CONTENT_LENGTH  = "Content-length: ";

   // Socket for accepting incoming client connection requests
   ServerSocket httpServerSocket;
//   Socket       socket;

   // log file stuff
   BufferedWriter  logFileWriter;
   String          logFileName = DEFAULT_LOG_FILE;

   /*
    * Following are a bunch of static arrays and lengths that are used in a
    * desperate attempt to avoid the creation of a bunch of String/StringBuffer
    * obects.
    */
   static final byte[] h1_CRLF       = ("</h1>\r\n").getBytes();
   static final byte[] body_CRLF     = ("</body>" + CRLF).getBytes();
   static final byte[] notFound      = ("Error, file not found:").getBytes();
   static final byte[] httpVersion   = ("HTTP/1.0 ").getBytes();
   static final byte[] dateEtc       = ("Date: 00000000" + CRLF
                                  + "Server: Java/0.0" + CRLF
                                  + "Content-type: ").getBytes();
   static final byte[] contentLength = ("Content-length: ").getBytes();
   ;

   // pre-allocated buffers
   static final byte[] errorBufferPrefix = ("<body>" + CRLF + "<h1>").getBytes();

   /**
    * Writes an entry to the log file.
    *
    * @param message entry to be written to the log file
    * @param newLine prints newLine character if true
    */
   void logMessage (String message, boolean newLine)
      throws HTTPServerException
   {
      logMessage(message);

      if (logEnabled)   // bypass the String stuff if we are not logging
      {
         try
         {
            if (newLine)
            {
               logFileWriter.newLine();
               logFileWriter.flush();
            }
         }
         catch (IOException i)
         {
            throw new HTTPServerException("Log file write failed");
         }
      }
   }

   /**
    * Writes an entry to the log file.
    *
    * @param message entry to be written to the log file
    */
   void logMessage (byte[] ipAddress)
      throws HTTPServerException
   {
      if (logEnabled)   // bypass the String stuff if we are not logging
      {
         if (ipAddress != null && ipAddress.length >= 4)
         {
            /*logMessage(Integer.toString(ipAddress [0] & 0x0ff));
            logMessage(".");
            logMessage(Integer.toString(ipAddress [1] & 0x0ff));
            logMessage(".");
            logMessage(Integer.toString(ipAddress [2] & 0x0ff));
            logMessage(".");
            logMessage(Integer.toString(ipAddress [3] & 0x0ff));*/
            
            logMessage(com.dalsemi.tininet.TININet.createIPString(ipAddress));
         }
      }
   }

   /**
    * Writes an entry to the log file.
    *
    * @param message entry to be written to the log file
    */
   void logMessage (String message)
      throws HTTPServerException
   {
      if (logEnabled)
      {
         try
         {
           if(logFileWriter == null)
             logFileWriter = new BufferedWriter(new FileWriter(logFileName,true));

           logFileWriter.write(message);
         }
         catch (IOException i)
         {
            throw new HTTPServerException("Log file write failed");
         }
      }
   }

   /* --- private --- */

      // server port number
   private int httpPort;
   // log file stuff
   private boolean logEnabled;

   // server home page stuff
   private String httpRoot  = DEFAULT_ROOT;
   private String indexPage = DEFAULT_PAGE;
   // TODO: implement if performance is not affected

   /*
    * private String  requestHTTPVersion;
    * private String  fieldUserAgent;
    * private String  fieldHost;
    * private String  fieldPragma;
    * private String  fieldAccept;
    * private String  fieldAcceptEncoding;
    * private String  fieldAcceptLanguage;
    * private String  fieldAcceptCharset;
    */

}

/** HTTPWorker thread class
  */
class HTTPWorker
   implements Runnable
{
   Socket     socket;
   HTTPServer httpServer;
   byte[]     clientAddress;
   Object     lock;

   // our socket streams
   OutputStream socketOutputStream;
   InputStream  socketInputStream;

   // request fields
   private int      requestMethod;
   private String   requestURI;
   private String   requestParameters;
   private String   postContent;
   final static int MAX_SEND         = 1000;
   final static int BUFFER_SIZE      = 1000;
   byte[]           buffer           = new byte [BUFFER_SIZE];
   byte[]           errorBuffer      = new byte [512];
   byte[]           mimeHeaderBuffer = new byte [384];
   static byte[]    crlf             = { 0x0d, 0x0a };

   // incoming request string
   String request;

   HTTPWorker (Socket socket, HTTPServer httpServer, Object lock)
   {
      this.socket     = socket;
      this.httpServer = httpServer;
      this.lock       = lock;
   }

   /**
    * Encodes the content type into a byte array.
    *
    * @param contentType  string representing the content type.
    * @param size     size of the content type string.
    *
    * @return returns the byte encoded buffer
    */
   byte[] encodeContentType (String contentType, int size)
   {
      return encodeBody(HTTPServer.HTTP_OK, "OK", contentType, size);
   }

   /**
    * Encodes the body into a byte array.
    *
    * @param body  string representing the body.
    * @param size  size of the body string.
    *
    * @return returns the byte encoded buffer
    */
   byte[] encodeBody (int statusCode, String body, String contentType,
                      int size)
   {
      // "HTTP/1.0 "
      System.arraycopy(HTTPServer.httpVersion, 0, mimeHeaderBuffer, 0,
                       HTTPServer.httpVersion.length);

      int index = HTTPServer.httpVersion.length;

      // Add the status code
      insertStatusCode(statusCode, mimeHeaderBuffer, index);

      index                      += 3;
      mimeHeaderBuffer [index++] = ( byte ) ' ';

      // Add the body
      byte[] temp = body.getBytes();

      System.arraycopy(temp, 0, mimeHeaderBuffer, index, temp.length);

      index                      += temp.length;
      mimeHeaderBuffer [index++] = 0x0d;
      mimeHeaderBuffer [index++] = 0x0a;

      // "Date: 00000000"+CRLF+"Server: Java/0.0"+CRLF+"Content-type: "
      System.arraycopy(HTTPServer.dateEtc, 0, mimeHeaderBuffer, index,
                       HTTPServer.dateEtc.length);

      index += HTTPServer.dateEtc.length;

      // Add content type and CRLF
      temp = contentType.getBytes();

      System.arraycopy(temp, 0, mimeHeaderBuffer, index, temp.length);

      index                      += temp.length;
      mimeHeaderBuffer [index++] = 0x0d;
      mimeHeaderBuffer [index++] = 0x0a;

      if (size > 0)
      {

         // "Content-length: " + size + CRLF
         System.arraycopy(HTTPServer.contentLength, 0, mimeHeaderBuffer,
                          index, HTTPServer.contentLength.length);

         index                      += HTTPServer.contentLength.length;
         index                      += insertInteger(size, mimeHeaderBuffer,
                                                     index);
         mimeHeaderBuffer [index++] = 0x0d;
         mimeHeaderBuffer [index++] = 0x0a;
      }

      // Final CRLF
      mimeHeaderBuffer [index++] = 0x0d;
      mimeHeaderBuffer [index++] = 0x0a;

      // Create a new byte[] and copy mime header bytes
      byte[] mimeHeader = new byte [index];

      System.arraycopy(mimeHeaderBuffer, 0, mimeHeader, 0, index);

      return mimeHeader;
   }

   /**
    * Convert integer to 3 digit ascii HTTP status code. Note that I'm not
    * doing any check for a good value.
    *
    * @param statusCode status code to be inserted.
    * @param b      buffer that will contain the status code.
    * @param offset   offset of the status code in the buffer.
    */
   private static void insertStatusCode (int statusCode, byte[] b, int offset)
   {
      int temp = statusCode / 100;

      b [offset++] = ( byte ) (temp + '0');
      statusCode   = (statusCode - temp * 100);
      temp         = statusCode / 10;
      b [offset++] = ( byte ) (temp + '0');
      statusCode   -= temp * 10;
      b [offset]   = ( byte ) (statusCode + '0');
   }

   /**
    * Convert integer to ascii bytes in b at offset.
    *
    * @param statusCode status code to be inserted.
    * @param b      buffer that will contain the status code.
    * @param offset   offset of the status code in the buffer.
    */
   private static int insertInteger (int integer, byte[] b, int offset)
   {
      int temp, length = 0,

      // max integer is 2147483647
      loopMask = 1000000000;
      boolean foundFirstDigit = false;

      // no need to get messy for zero
      if (integer == 0)
      {
         b [offset++] = ( byte ) '0';

         return 1;
      }

      if (integer < 0)
      {
         b [offset++] = ( byte ) '-';
         integer      = -integer;

         length++;
      }

      while (loopMask > 0)
      {

         // get the bytes with no leading zeros
         if (foundFirstDigit || (integer >= loopMask))
         {

            // grind out the bytes
            foundFirstDigit = true;
            temp            = integer / loopMask;
            integer         -= temp * loopMask;
            b [offset++]    = ( byte ) (temp + '0');

            length++;
         }

         loopMask /= 10;
      }

      return length;
   }

   /**
    * Separates the URI from any parameters.
    */
   private void processCommandLine ()
   {
      int startIndex = requestURI.indexOf('?');

      if (startIndex > -1)
      {
         requestParameters = requestURI.substring(startIndex + 1);

         if (requestParameters != null)
            if (requestParameters.length() == 0)
               requestParameters = null;

         requestURI = requestURI.substring(0, startIndex);
      }
   }

   /**
    * returns the mime type string.
    *
    * @param filename name of the file which determines the mime type
    */
   private static String getContentType (String filename)
   {
      if (filename.endsWith(".html") || filename.endsWith(".htm"))
         return HTTPServer.MIME_TEXT_HTML;

      if (filename.endsWith(".class"))
         return HTTPServer.MIME_APP_OS;

      if (filename.endsWith(".gif"))
         return HTTPServer.MIME_IMAGE_GIF;

      if (filename.endsWith(".jpg") || filename.endsWith(".jpeg"))
         return HTTPServer.MIME_IMAGE_JPG;

      if (filename.endsWith(".bmp"))
         return HTTPServer.MIME_IMAGE_BMP;

      return HTTPServer.MIME_TEXT_PLAIN;
   }

   /**
    * returns the mime type string.
    *
    * @param filename name of the file which determines the mime type
    */
   private int getContentLength ()
   {
      int startIndex = request.indexOf(HTTPServer.FIELD_CONTENT_LENGTH);

      postContent = null;

      if (startIndex == -1)
         return -1;   // throw exception???

      int    stopIndex           = request.indexOf("\r\n", startIndex);
      String contentLengthString = request.substring(startIndex + 16,
                                      stopIndex);

      if ((contentLengthString == null)
              || (contentLengthString.length() == 0))
         return -1;

      postContent = request.substring(stopIndex + 4);

      return Integer.parseInt(contentLengthString);
   }

   /**
    * Nasty and cryptic way to cobble together an HTTP error message. This
    * is however pretty fast and doesn't consume too much memory.
    *
    * @param code http error code to be returned.
    *
    * @return byte array containing the error message.
    */
   private byte[] processError (int code)
      throws HTTPServerException
   {
      // copy <body> + CLRF + <h1> to buffer
      System.arraycopy(HTTPServer.errorBufferPrefix, 0, errorBuffer, 0,
                       HTTPServer.errorBufferPrefix.length);

      int index = HTTPServer.errorBufferPrefix.length;

      // Add the error code
      insertStatusCode(code, errorBuffer, index);

      index                 += 3;
      errorBuffer [index++] = ( byte ) ' ';

      byte[] temp;
      String errorMessage;

      switch (code)
      {

         case HTTPServer.HTTP_BAD_REQUEST :
            errorMessage = "Bad Request";
            break;

         case HTTPServer.HTTP_UNAUTHORIZED :
            errorMessage = "Unathorized";
            break;

         case HTTPServer.HTTP_FORBIDDEN :
            errorMessage = "Forbidden";
            break;

         case HTTPServer.HTTP_NOT_FOUND :
            errorMessage = "Not Found";
            break;

         default :
            errorMessage = "Unknown Error";
            break;
      }

      httpServer.logMessage(errorMessage, true);

      // Add the error message
      temp = errorMessage.getBytes();

      System.arraycopy(temp, 0, errorBuffer, index, temp.length);

      index += temp.length;

      // Add </h1> + CRLF
      System.arraycopy(HTTPServer.h1_CRLF, 0, errorBuffer, index,
                       HTTPServer.h1_CRLF.length);

      index += HTTPServer.h1_CRLF.length;

      // Add the name
      if (requestURI != null)
      {
         System.arraycopy(HTTPServer.notFound, 0, errorBuffer, index,
                          HTTPServer.notFound.length);

         index += HTTPServer.notFound.length;
         temp  = requestURI.getBytes();

         System.arraycopy(temp, 0, errorBuffer, index, temp.length);

         index                 += temp.length;
         errorBuffer [index++] = 0x0d;
         errorBuffer [index++] = 0x0a;
      }

      // Add </body> + CRLF
      System.arraycopy(HTTPServer.body_CRLF, 0, errorBuffer, index,
                       HTTPServer.body_CRLF.length);

      index += HTTPServer.body_CRLF.length;
      temp  = encodeBody(code, errorMessage, HTTPServer.MIME_TEXT_HTML, 0);

      // Finally create a new byte array to hold our errorMessage copy errorBuffer
      byte[] errorBytes = new byte [index + temp.length];

      System.arraycopy(temp, 0, errorBytes, 0, temp.length);
      System.arraycopy(errorBuffer, 0, errorBytes, temp.length, index);

      return errorBytes;
   }

   /**
    * Process a GET request.
    *
    * @return the HTTP error type.
    */
   private int processGetRequest ()
      throws HTTPServerException
   {
      String pageName = httpServer.getHTTPRoot() + requestURI;

      try
      {
         httpServer.logMessage(clientAddress);
         httpServer.logMessage(", GET, ");
         httpServer.logMessage(requestURI, true);

         File webPage = new File(pageName);

         if (!webPage.exists())
         {
            httpServer.logMessage(pageName);
            httpServer.logMessage(" ");
            socketOutputStream.write(processError(HTTPServer.HTTP_NOT_FOUND));

            return HTTPServer.HTTP_NOT_FOUND;
         }

         if (!webPage.getCanonicalPath().startsWith(httpServer.getHTTPRoot()))
         {
            httpServer.logMessage(pageName);
            httpServer.logMessage(" ");
            socketOutputStream.write(processError(HTTPServer.HTTP_FORBIDDEN));

            return HTTPServer.HTTP_FORBIDDEN;
         }

         if (!webPage.canRead())
         {
            httpServer.logMessage(pageName);
            httpServer.logMessage(" ");
            socketOutputStream.write(processError(HTTPServer.HTTP_FORBIDDEN));

            return HTTPServer.HTTP_FORBIDDEN;
         }

         InputStream fileInputStream = new FileInputStream(webPage);
         int         n               = fileInputStream.available();

         socketOutputStream.write(encodeContentType(getContentType(pageName),
                                                    n));

         synchronized (lock)
         {
            while ((n = fileInputStream.read(buffer, 0, MAX_SEND)) >= 0)
            {
               socketOutputStream.write(buffer, 0, n);
            }

            fileInputStream.close();
         }

         return HTTPServer.HTTP_OK;
      }
      catch (IOException i)
      {
         throw new HTTPServerException("Socket failure");
      }
   }

   /**
    * Process a POST request.
    *
    * @return the HTTP error type.
    */

   /*
     private int processPostRequest() throws HTTPServerException
     {
       String  pageName = httpServer.getHTTPRoot() + requestURI;
       int   contentLength = getContentLength();

       try
       {
         httpServer.logMessage(clientAddress);
         httpServer.logMessage(", POST, ");
         httpServer.logMessage(requestURI, true);

         File  webPage = new File(pageName);

         if(!webPage.exists())
         {
           httpServer.logMessage(pageName);
           httpServer.logMessage(" ");
           socketOutputStream.write(processError(HTTPServer.HTTP_NOT_FOUND));
           return HTTPServer.HTTP_NOT_FOUND;
         }

         if(!webPage.canRead())
         {
           httpServer.logMessage(pageName);
           httpServer.logMessage(" ");
           socketOutputStream.write(processError(HTTPServer.HTTP_UNAUTHORIZED));
           return HTTPServer.HTTP_UNAUTHORIZED;
         }

         // kludge until we get out executive/shell
         socketOutputStream.write(encodeContentType(HTTPServer.MIME_TEXT_PLAIN, 0));
         socketOutputStream.write("POST received:\n".getBytes());
         socketOutputStream.write(postContent.getBytes());
         return HTTPServer.HTTP_OK;
       }
       catch (IOException i)
       {
         throw new HTTPServerException("Socket failure");
       }
     }
   */

   /**
    * Process an HTTP request.
    */
   private int processRequest ()
      throws HTTPServerException
   {
      int response = HTTPServer.HTTP_SERVER_ERROR;

      try
      {
         socketOutputStream = socket.getOutputStream();
         socketInputStream  = socket.getInputStream();
         clientAddress      = socket.getInetAddress().getAddress();

         if (readRequest() != null)
         {
            switch (requestMethod)
            {

               case HTTPServer.GET :
                  response = processGetRequest();
                  break;

               /*
                         case HTTPServer.POST:
                           response = processPostRequest();
                         break;
               */
               default :
                  response = HTTPServer.HTTP_UNSUPPORTED_TYPE;
                  break;
            }
         }

         socketInputStream.close();
         socketOutputStream.close();
      }
      catch (IOException i)
      {
         throw new HTTPServerException("Socket failure");
      }

      return response;
   }

   /**
    * Returns a String representing the incomming request.
    *
    * @exception HTTPServerException if an error occurs while processing a request.
    */
   private String readRequest ()
      throws HTTPServerException
   {
      try
      {
         int     retry             = 0;
         int     index             = 0;
         int     available         = 0;
         int     bytesRead         = 0;
         int     postContentLength = -1;
         boolean done              = false;
         boolean post              = false;
         boolean get               = false;

         while (true)
         {
            available = socketInputStream.available();

            if (available == 0)
            {
               if (++retry > 100)
               {
                  throw new HTTPServerException("Incomplete request");
               }

               try
               {
                  Thread.sleep(10);
               }
               catch (InterruptedException i){}
            }
            else
            {
               socketInputStream.read(buffer, index, available);

               index += available;

               if (!get &&!post)
               {
                  if ((buffer [0] == ( byte ) 'g')
                          || (buffer [0] == ( byte ) 'G'))
                     get = true;
                  else if ((buffer [0] == ( byte ) 'p')
                           || (buffer [0] == ( byte ) 'P'))
                     post = true;
               }

               // see if we have all the data
               if (get)
               {
                  if ((buffer [index - 4] == 0x0d)
                          && (buffer [index - 3] == 0x0a)
                          && (buffer [index - 2] == 0x0d)
                          && (buffer [index - 1] == 0x0a))
                     done = true;
               }
               else if (post)
               {
                  request = new String(buffer, 0, index);

                  if (postContentLength == -1)
                     postContentLength = getContentLength();

                  if (postContent != null)
                     done = (postContentLength == postContent.length());
                  else
                     continue;
               }

               if (done)
               {
                  String temp         = (request = new String(buffer, 0,
                                                              index));
                  int    requestIndex = 0;

                  if (temp.regionMatches(true, 0, "get ", 0, 4))
                  {
                     requestMethod = HTTPServer.GET;
                     requestIndex  = 4;
                  }
                  else if (temp.regionMatches(true, 0, "post ", 0, 5))
                  {
                     requestMethod = HTTPServer.POST;
                     requestIndex  = 5;
                  }

                  /*
                   * else if(temp.regionMatches(true, 0, "head ", 0, 4))
                   * requestMethod = HEAD;
                   */
                  else
                  {
                     requestMethod = HTTPServer.UNSUPPORTED;

                     throw new HTTPServerException("Unsupported request");
                  }

                  requestURI = request.substring(requestIndex,
                                                 request.indexOf(' ',
                                                    requestIndex));

                  processCommandLine();

                  if (requestURI.endsWith("/"))
                     requestURI = requestURI + httpServer.getIndexPage();

                  if (requestURI.charAt(0) == '/')
                     requestURI = requestURI.substring(1);

                  return request;
               }
            }
         }
      }
      catch (IOException i)
      {
         throw new HTTPServerException("Socket failure");
      }
   }

   /** Entry point for this worker thread
     */
   public void run ()
   {
      request = null;

      try
      {
         int responseCode = processRequest();
      }
      catch (Throwable t)
      {
         throw new HTTPServerException("ServerSocket accept failure");
      }
      finally
      {
         try
         {
            socket.close();
         }
         catch (Throwable t){}
      }
   }
}
