
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.ppp;



import java.io.*;
import java.util.*;
import javax.comm.*;
import com.dalsemi.tininet.TININet;


/**
 * <P>
 * This class implements the Point-to-Point Protocol (PPP), which provides IP packet
 * transport over a serial link. PPP exists as a layer between the IP network-layer
 * and a physical interface. Being a point-to-point protocol, PPP does not distinguish
 * between a client and a server operation. For the purposes of this documentation, a peer which
 * requires a remote peer to authenticate and provides an IP address to the remote peer
 * will be known as a server. A peer which does not require a remote peer to authenticate
 * and uses a server assigned IP address will be know as a client. Many variations exist
 * between these configurations and are not covered in this documentation.
 * </P>
 * <P>
 * To satisfy the physical interface requirement, user application code must open,
 * initialize and pass a serial port object to an instance of PPP. After passing
 * the serial port object to PPP, the user application should not attempt to read
 * or write operations on the serial port until PPP has signaled a closed event. This
 * does not preclude the application code from implementing the <CODE>SerialPortEventListener</CODE>
 * interface or from polling the serial port status signals.
 * </P>
 * <P>
 * An application that creates a PPP object must provide a listener to receive notification
 * of PPP events. The listener is an instance of a class that implements the <CODE>PPPEventListener</CODE>
 * interface.
 * </P>
 *
 * <H3> Usage </H3>
 *
 * <DL>
 * <DD> <H4> Example 1</H4>
 * Setup a PPP connection using the already configured serial object 'so'.
 * <PRE> <CODE>
 *  (code example)
 * </CODE> </PRE>
 * </DL>
 *
 * <DL>
 * <DD> <H4> Example 2</H4>
 * Use the ppp instance 'pppConnection' that was configured in Example 1.
 * <PRE> <CODE>
 *  (code example)
 * </CODE> </PRE>
 * </DL>
 *
 * <DL>
 * <DD> <H4> Example 3</H4>
 * Close the ppp connection 'pppConnection'.
 * <PRE> <CODE>
 *  (code example)
 * </CODE> </PRE>
 * </DL>
 *
 * @see PPPEventListener
 * @see PPPEvent
 * @see javax.comm.SerialPort
 */
public class PPP
   extends Thread
{

   // Native interface
   private static native byte PPPInitNative ();

   private static native void FreeInterfaceNative (byte index);

   private static native void setPortNumberNative (int port, byte index);

   private static native void PPPUpNative (byte index);

   private static native void PPPDownNative (byte index);

   private static native void PPPOpenNative (byte index);

   private static native void PPPCloseNative (byte index);

   private static native void PPPAuthenticateNative (boolean value, byte index);

   private static native void PPPPortControlNative (boolean value, byte index);

   private static native void PPPIOPollNative (boolean active);

   private static native void setPPPOptionNative (byte option,
                                               boolean boolValue,
                                                   int intValue,
                                                byte[] arrayValue,
                                                  byte interfaceIndex);

   private static native int getPPPOptionNative (byte option,
                                               byte[] arrayValue,
                                                 byte interfaceIndex);

   private static native void stopPPPThreadNative(byte index);

    /**
     *  Gets the transmitter routine.
     *
     *  @param index interface index
     */
   public static native int getTransmitter (byte index);

   private static native int PPPWaitForEventNative (byte index);

   private static native int getPeerIDPasswordNative (int option,
                                                   byte[] value,
                                                     byte interfaceIndex);

   private Vector  eventListeners     = new Vector();

   private boolean eventThreadEnabled = false;

   // Native operation selectors
   private static final byte ACCM           = 0;
   private static final byte AUTHENTICATE   = 1;
   private static final byte DEFAULT_ROUTE  = 2;
   private static final byte LOCAL_ADDRESS  = 3;
   private static final byte MRU            = 4;
   private static final byte PASSIVE        = 5;
   private static final byte PASSWORD       = 6;
   private static final byte USERNAME       = 7;
   private static final byte PROXY_ARP      = 8;
   private static final byte REMOTE_ADDRESS = 9;
   private static final byte RESTART_PERIOD = 10;
   private static final byte MAX_CONFIG     = 11;
   private static final byte MAX_TERMINATE  = 12;
   private static final byte MAX_FAIL       = 13;
   private static final byte DIAL_ON_DEMAND = 14;
   private static final byte IDLE_TIMEOUT   = 15;
   private static final byte XON_XOFF       = 16;

   /**
    * Peer ID parameter to getUsernamePassword
    */
   public static final byte PEER_ID = 0;

   /**
    * Peer password parameter to getUsernamePassword
    */
   public static final byte PEER_PASSWORD  = 1;

   /**
    *
    */
   private byte             interfaceIndex = 99;

   /**
    * Default constructor
    */
   public PPP ()
   {

      // Allocate and intilize native PPP interface
      interfaceIndex = PPPInitNative();

      //com.dalsemi.system.Debug.debugDump("Using interface " + interfaceIndex);

      // Enable dispatch thread
      eventThreadEnabled = true;

      // Set thread for deamon operation
      this.setDaemon(true);

      // Start dispatch thread
      start();
   }

   /*
    * Configuration methods
    */


   /**
    * Sets the value of the Asynchronous Control Character Map (ACCM). If a bit in a specified position
    * is 1, the corresponding character is escaped otherwise it is transmitted normaly. For example
    * an ACCM value of 0x80000001 will escape only characters 0 and 31. By default the ACCM value
    * is 0xffffffff, escape all characters in the range 0x00 - 0x1f. Usefull ACCM values are
    * 0x00000000, escape no characters, and 0x000a0000, escape 0x11 and 0x13 for channels using
    * XON/XOFF flow control.
    *
    * @param newACCM Asynchronous Control Character Map (ACCM)
    */
   public void setACCM (int newACCM)
   {
      setPPPOptionNative(ACCM, false, newACCM, null, interfaceIndex);
   }

   /**
    * Gets the current Asynchronous Control Character Map (ACCM).
    *
    * @return Asynchronous Control Character Map (ACCM)
    *
    * @see com.dalsemi.tininet.ppp.PPP#setACCM setACCM()
    */
   public int getACCM ()
   {
      return getPPPOptionNative(ACCM, null, interfaceIndex);
   }



   /**
    * Sets the remote peer authentication option. When this option is set <CODE>true</CODE>
    * this peer will require the remote peer to authenticate before a connection is established.
    *
    * @param value <CODE>true</CODE> if remote peer authentication is required
    */
   public void setAuthenticate (boolean value)
   {
      setPPPOptionNative(AUTHENTICATE, value, 0, null, interfaceIndex);
   }

   /**
    * Gets current value of remote peer authenticate flag.
    *
    * @return current value of authentication flag
    *
    * @see com.dalsemi.tininet.ppp.PPP#setAuthenticate setAuthenticate()
    */
   public boolean getAuthenticate ()
   {
      if (getPPPOptionNative(AUTHENTICATE, null, interfaceIndex) == 0)
         return false;
      else
         return true;
   }




   /**
    * Sets the default interface flag. When this option is set <CODE>true</CODE>
    * this interface will be set to the default network interface once a connection
    * is established.
    *
    * @param value <CODE>true</CODE> if this interface should be set a default network interface
    */
   public void setDefaultInterface (boolean value)
   {
      setPPPOptionNative(DEFAULT_ROUTE, value, 0, null, interfaceIndex);
   }

   /**
    * Gets the current value of the default interface flag.
    *
    * @return current value of default interface flag
    *
    * @see com.dalsemi.tininet.ppp.PPP#setDefaultInterface setDefaultInterface
    */
   public boolean getDefaultInterface ()
   {
      if (getPPPOptionNative(DEFAULT_ROUTE, null, interfaceIndex) == 0)
         return false;
      else
         return true;
   }




   /**
    * Sets the local IP address for this interface. If the remote peer is vending an address
    * this address must match the vended address or PPP will fail to negotiate the link.
    *
    * @param address local IP address, index 0 is MSB of four byte network address
    *
    */
   public void setLocalAddress (byte[] address)
   {
      setPPPOptionNative(LOCAL_ADDRESS, false, 0, address, interfaceIndex);
   }

   /**
    * Gets the local IP address for this interface.
    *
    * @return local IP address, index 0 is MSB of four byte network address
    *
    * @see com.dalsemi.tininet.ppp.PPP#setLocalAddress setLocalAddress()
    */
   public byte[] getLocalAddress ()
   {
      byte[] address = new byte [4];

      getPPPOptionNative(LOCAL_ADDRESS, address, interfaceIndex);

      return address;
   }


   /**
    * Sets passive mode option. Passive mode effects the earliest phase of
    * of connection establishment, Line Control Protocol (LCP) negotiation. By default
    * passive mode is set to false and LCP attempts to initiate a connection by sending
    * configuration request messages to the remote peer. If no response is received LCP
    * times out and generates a stopped event. When passive mode is true LCP will time out
    * but not generate as stopped event and will wait indefinitly for the remote peer to
    * transmit a configuration request.
    *
    * @param value <CODE>true</CODE> for passive mode operation
    *
    */
   public void setPassive (boolean value)
   {
      setPPPOptionNative(PASSIVE, value, 0, null, interfaceIndex);
   }

   /**
    * Gets the current state of the passive mode option.
    *
    * @return current state of passive mode option
    *
    * @see com.dalsemi.tininet.ppp.PPP#setPassive setPassive()
    */
   public boolean getPassive ()
   {
      if (getPPPOptionNative(PASSIVE, null, interfaceIndex) == 0)
         return false;
      else
         return true;
   }




   /**
    * Sets the password sent to a remote peer during authentication.
    *
    * @param password user password
    *
    * @exception com.dalsemi.tininet.ppp.PPPException Password exceeds maximum length
    */
   public void setPassword (String password)
      throws PPPException
   {
      byte[] passwordAsByteArray;

      passwordAsByteArray = password.getBytes();

      setPPPOptionNative(PASSWORD, false, 0, passwordAsByteArray,
                         interfaceIndex);
   }


   /**
    * Sets the user name sent to a remote peer during authentication.
    *
    * @param userName user name
    *
    * @exception com.dalsemi.tininet.ppp.PPPException Username exceeds maximum length
    */
   public void setUsername (String userName)
      throws PPPException
   {
      byte[] userNameAsByteArray;

      userNameAsByteArray = userName.getBytes();

      setPPPOptionNative(USERNAME, false, 0, userNameAsByteArray,
                         interfaceIndex);
   }

   /**
    * Sets the IP address of the remote peer. Used when this peer is operating as a server
    * and needs to vend an IP address to the remote peer.
    *
    * @param address remote IP address, index 0 is MSB of four byte network address
    */
   public void setRemoteAddress (byte[] address)
   {
      setPPPOptionNative(REMOTE_ADDRESS, false, 0, address, interfaceIndex);
   }

   /**
    * Gets the IP address of the remote peer.
    *
    * @return remote IP address, index 0 is MSB of four byte network address
    *
    * @see com.dalsemi.tininet.ppp.PPP#setRemoteAddress setRemoteAddress()
    */
   public byte[] getRemoteAddress ()
   {
      byte[] address = new byte [4];

      getPPPOptionNative(REMOTE_ADDRESS, address, interfaceIndex);

      return address;
   }




   /**
    * Sets the restart time period in seconds (1 - 30). Expiration of the restart time is used
    * by the underlying state machine to retransmit configuration-request and terminate-request
    * packets. The default restart time is three seconds.
    *
    * @param timeout restart time in seconds
    * @exception com.dalsemi.tininet.ppp.PPPException Maximum time exceeded, Minimum time exceeded
    */
   public void setRestartPeriod (int timeout)
      throws PPPException
   {
      if ((timeout > 0) & (timeout < 31))
      {
          // Convert user timout in seconds to milliseconds for native
          int timeoutMillis = timeout * 1000;

          setPPPOptionNative(RESTART_PERIOD, false, timeoutMillis, null, interfaceIndex);
      }
      else
      {
         // Throw illegal argument exception
      }
   }

   /**
    * Gets the timeout period.
    *
    * @return timeout period in seconds
    *
    * @see com.dalsemi.tininet.ppp.PPP#setRestartPeriod setRestartPeriod()
    */
   public int getRestartPeriod ()
   {
      int timeoutSeconds = (getPPPOptionNative(RESTART_PERIOD, null, interfaceIndex) / 1000);

      return timeoutSeconds;
   }




   /**
    * Sets the maximum number of unanswered configuration requests that will be sent (1 - 20).
    * The default value is ten.
    *
    * @param count the maximum number of configuration retries
    *
    * @exception com.dalsemi.tininet.ppp.PPPException Invalid maximum/minimum retry count
    *
    */
   public void setMaxConfig (int count)
      throws PPPException
   {
      if ((count > 0) & (count < 21))
          setPPPOptionNative(MAX_CONFIG, false, count, null, interfaceIndex);
   }

   /**
    * Gets the maximum number of configuration retries allowed.
    *
    * @return maximum retries
    *
    * @see com.dalsemi.tininet.ppp.PPP#setMaxConfig setMaxConfig()
    */
   public int getMaxConfig ()
   {
      return getPPPOptionNative(MAX_CONFIG, null, interfaceIndex);
   }




   /**
    * Sets the maximum number of termination requests that will be sent (1 - 20).
    * The default value is two.
    *
    * @param count the maximum number of termination retries
    *
    * @exception com.dalsemi.tininet.ppp.PPPException Maximum count exceeded
    */
   public void setMaxTerminate (int count)
      throws PPPException
   {
      if ((count > 0) & (count < 21))
        setPPPOptionNative(MAX_TERMINATE, false, count, null, interfaceIndex);
   }

   /**
    * Gets the maximum number of termination retries allowed.
    *
    * @return maximum retries
    *
    * @see com.dalsemi.tininet.ppp.PPP#setMaxTerminate setMaxTerminate()
    */
   public int getMaxTerminate ()
   {
      return getPPPOptionNative(MAX_TERMINATE, null, interfaceIndex);
   }




   /**
    * Sets the maximum number of configuration failures allowed (1 - 20).
    * The default value is five.
    *
    * @param count the maximum number of failures allowed
    *
    * @exception com.dalsemi.tininet.ppp.PPPException Maximum count exceeded
    */
   public void setMaxFailure (int count)
      throws PPPException
   {
      if ((count > 0) & (count < 21))
        setPPPOptionNative(MAX_FAIL, false, count, null, interfaceIndex);
   }

   /**
    * Gets the maximum number of configuration failures allowed.
    *
    * @return maximum failures
    *
    *  @see com.dalsemi.tininet.ppp.PPP#setMaxFailure setMaxFailure()
    */
   public int getMaxFailure ()
   {
      return getPPPOptionNative(MAX_FAIL, null, interfaceIndex);
   }



   /**
    * Sets XON/XOFF flows control character escaping.
    * Default operation is not to escape XON/XOFF characters.
    *
    * @param value <CODE>true</CODE> for XON/XOFF character escaping
    */
    public void setXonXoffEscape(boolean value)
    {
         setPPPOptionNative(XON_XOFF, value, 0, null, interfaceIndex);
    }

    /**
     * Gets the current mode of XON/XOFF escaping.
     *
     * @return escaping mode
     */
     public boolean getXonXoffEscape()
     {
        if (getPPPOptionNative(XON_XOFF, null, interfaceIndex) == 0)
            return false;
        else
            return true;
     }


   /*
    * Operational methods
    */

   /**
    * Adds a PPP event listener.
    *
    * @param eventListener PPPEventListener object
    * @exception java.util.TooManyListenersException
    */
   public void addEventListener (PPPEventListener eventListener)
      throws TooManyListenersException
   {
      if (eventListener == null)
      {
         throw new IllegalArgumentException();
      }

      // If the event listener list is empty register the PPP native code
      // with the IO POLL structure and enable the event thread
      if (eventListeners.isEmpty())
      {
         register();

         eventThreadEnabled = true;
      }

      eventListeners.addElement(eventListener);
   }

   /**
    * Removes a PPP event listener.
    *
    * @param eventListener PPPEventListener object
    */
   public void removeEventListener (PPPEventListener eventListener)
   {
      if (eventListener == null)
      {
         throw new IllegalArgumentException();
      }

      eventListeners.removeElement(eventListener);

      // If the event listener list is empty unregister the PPP native code
      // with the IO POLL structure.
      if (eventListeners.isEmpty())
      {
         unregister();
      }
   }

   /**
    * Implements PPP deamon thread which propagates pppEvents to event listeners.
    * Overrides java.lang.Thread.run method.
    */
   public final void run ()
   {
      int event;
      int error;
      int evtErr;

      //com.dalsemi.system.Debug.debugDump("Event thread running");

      while (eventThreadEnabled)
      {
         // The return value for PPPWaitForEventNative now contains the event type and an error code where applicable
         // The event code is in the lower 16 bit word of the int and the error code is in the upper 16 bit word.
         //com.dalsemi.system.Debug.debugDump("Event thread going to sleep");
         evtErr = PPPWaitForEventNative(interfaceIndex);
         //com.dalsemi.system.Debug.debugDump("Event thread awake");
         // Extract event code
         event = evtErr & 0x00ff;
         // Extract error code
         error = (evtErr >> 16) & 0x00ff;

         if (event == 0xff)
         {
           eventThreadEnabled = false;
           continue;
         }

         //PPPEvent pppEvent = new PPPEvent(this, event, false, false);

         PPPEvent pppEvent = new PPPEvent(this, event, error);

         // Set error code, The PPPEvent constructor needs to change or another constructor with error as a paramter needs to be added.
         //pppEvent.setError(error);

         //pppEvent.eventType = event;
         for (int i = 0; i < eventListeners.size(); ++i)
         {
            (( PPPEventListener ) eventListeners.elementAt(i)).pppEvent(
               pppEvent);
         }
      }

      //com.dalsemi.system.Debug.debugDump("Event thread exit");
   }

   /**
    * Invoked when serial channel is ready for PPP traffic.
    *
    * @param sp SerialPort object
    */
   public void up(SerialPort sp)
   {
      String portName = sp.getName();
      int    portNum  = portName.charAt(portName.length() - 1) - 0x30;

      setPortNumberNative(portNum, interfaceIndex);

      // Take control of serial port
      PPPPortControlNative(true, interfaceIndex);

      // Issue up event
      //com.dalsemi.system.Debug.debugDump("Up on interface " + interfaceIndex);
      PPPUpNative(interfaceIndex);
   }

   /**
    * Invoked when serial channel is no longer available for PPP traffic.
    */
   public void down ()
   {

      // Release control of serial port
      PPPPortControlNative(false, interfaceIndex);
      PPPDownNative(interfaceIndex);
   }

   /**
    * Invoked to open PPP.
    */
   public void open ()
   {
      //com.dalsemi.system.Debug.debugDump("Open on interface "
      //                                   + interfaceIndex);
      PPPOpenNative(interfaceIndex);
   }

   /**
    * Invoked to close PPP.
    */
   public void close ()
   {
      PPPCloseNative(interfaceIndex);
   }

   /**
    * Method register
    *
    *
    */
   private void register ()
   {

      //com.dalsemi.system.Debug.debugDump("register");
      //PPPInitNative();
      PPPIOPollNative(true);
   }

   /**
    * Method unregister
    *
    *
    */
   private void unregister ()
   {

      //com.dalsemi.system.Debug.debugDump("unregister");
      PPPIOPollNative(false);

   }

   /**
    * @deprecated
    *
    * Gets remote peers username/password. Used during authentication phase
    * to retreive remote peers ID and password. Invoking this method with value
    * as null returns length of option.
    *
    * @param option indicates which value to retreive, PEER_ID or PEER_PASSWORD
    * @param value receives option value
    * @return length of username/password
    *
    * @see com.dalsemi.tininet.ppp.PPP#getPeerID getPeerID()
    * @see com.dalsemi.tininet.ppp.PPP#getPeerPassword getPeerPassword()
    */
   public int getUsernamePassword (int option, byte[] value)
   {
      return getPeerIDPasswordNative(option, value, interfaceIndex);
   }

   /**
    * Indicates if access should be granted to remote peer.
    *
    * @param value True to allows access, False to deny accesss
    */
   public void authenticate (boolean value)
   {
      PPPAuthenticateNative(value, interfaceIndex);
   }

   /**
    * Adds a PPP interface to the network interface table.
    *
    * @param name interface name
    * @param mss maximum TCP segment size sent
    * @param timeout initial TCP timeout
    */
   public void addInterface (String name, int mss, int timeout)
   {
      // The calculated subnet mask
      byte[] subnetMask = new byte []{ ( byte ) 255, ( byte ) 0, ( byte ) 0, ( byte ) 0 };

      // Default gateway address, required by addInterfaceEntry
      byte[] gateway    = new byte []{ 0, 0, 0, 0 };

      // Calculate default netmask for local interface address
      byte[] la = getLocalAddress();

      if ((la [0] & 0xc0) == 0x80)
      {
         subnetMask [1] = ( byte ) 255;
      }
      else if ((la [0] & 0xc0) == 0xc0)
      {
         subnetMask [1] = ( byte ) 255;
         subnetMask [2] = ( byte ) 255;
      }

      try
      {
         // Add the interface
         com.dalsemi.tininet.TININet.addInterfaceEntry(name.getBytes(),
                                                       getLocalAddress(),
                                                       subnetMask,
                                                       gateway,
                                                       ( byte ) 0x43/*71(base ten) */,
                                                       ( byte ) 0x02,
                                                       getTransmitter(interfaceIndex),
                                                       mss,
                                                       timeout);
      }
      catch (Exception e)
      {
         //com.dalsemi.system.Debug.debugDump("addInterfaceEntry");
         com.dalsemi.system.Debug.debugDump(e.toString());
      }

      // If this connection is marked as the default interface make it so
      if (getDefaultInterface())
      {
         try
         {
            com.dalsemi.tininet.TININet.setDefaultInterface(name.getBytes());
         }
         catch (Exception e)
         {
            //com.dalsemi.system.Debug.debugDump("setDefaultInterface");
            com.dalsemi.system.Debug.debugDump(e.toString());
         }
      }
   }

   /**
    * Adds a PPP interface to the network interface table.
    * The MSS and timeout are set to default value of 256 and 128
    *
    * @param name interface name
    */
   public void addInterface (String name)
   {
        addInterface(name, 256, TININet.TCP_TIMEOUT_128);
   }

   /**
    * Removes the named PPP interface from the network interface table.
    *
    * @param name interface name
    */
   public void removeInterface (String name)
   {
      try
      {
         com.dalsemi.tininet.TININet.removeInterfaceEntry(name.getBytes());
      }
      catch (Exception e)
      {
         com.dalsemi.system.Debug.debugDump(e.toString());
      }
   }

   /**
    * Frees native PPP interface.
    * If an application exits or is killed the interface will freed
    * automatically. If ana application does not exit but removes all
    * references to a PPP object the native interface will not be freed.
    * This method allow an application to fre the native interface cleanly.
    *
    */
   public void freeInterfaceWrapper ()
   {
      FreeInterfaceNative(interfaceIndex);
   }

   /**
    * Gets the remote peers authentication ID.
    *
    * @return remote peers authentication ID
    */
   public String getPeerID()
   {

        byte[] optionData = new byte[128];

        getUsernamePassword(PPP.PEER_ID, optionData);

        int len = getUsernamePassword(PPP.PEER_ID, null);

        String name = new String(optionData, 0, len);

        return name;
   }

   /**
    * Gets the remote peers authentication password.
    *
    * @return remote peers authentication password
    */
   public String getPeerPassword()
   {
        byte[] optionData = new byte[128];

        getUsernamePassword(PPP.PEER_PASSWORD, optionData);

        int len = getUsernamePassword(PPP.PEER_PASSWORD, null);

        String password = new String(optionData, 0, len);

        return password;
   }

   /**
    *  Terminates internal PPP event notification thread.
    *  The PPP event notification thread is marked as a daemon thread and is destroyed
    *  on application exit. If an application does not exit but removes all references to
    *  a PPP object the event thread will not terminate. This method allows an
    *  application to destroy the daemon thread prior to releasing the PPP object.
    *
    */
   public void stopPPPThread()
   {
        stopPPPThreadNative(interfaceIndex);
   }

}
