
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.ppp;



import java.util.EventObject;


/**
 * <P>
 * When a PPP object is created it starts a deamon thread to listen for events originating from the native PPP module.
 * This thread generates the events that notify PPP event listeners of important changes in the state
 * of a PPP connection. The events generated by the deamon thread are defined as integer constants and
 * shown below:
 * <UL>
 *  <LI><CODE><A HREF=#STARTING>STARTING</A></CODE>
 *  <LI><CODE><A HREF=#AUTHENTICATION_REQUEST>AUTHENTICATION_REQUEST</A></CODE>
 *  <LI><CODE><A HREF=#UP>UP</A></CODE>
 *  <LI><CODE><A HREF=#STOPPED>STOPPED</A></CODE>
 *  <LI><CODE><A HREF=#CLOSED>CLOSED</A></CODE>
 * </UL>
 * </P>
 *
 * <P>
 * Each event object also includes an error code. The error codes generated are defined as integer constants
 * and shown below:
 * <UL>
 *  <LI><CODE><A HREF=#NONE>NONE</A></CODE>
 *  <LI><CODE><A HREF=#ADDR>ADDR</A></CODE>
 *  <LI><CODE><A HREF=#AUTH>AUTH</A></CODE>
 *  <LI><CODE><A HREF=#TIME>TIME</A></CODE>
 *  <LI><CODE><A HREF=#REJECT>REJECT</A></CODE>
 * </UL>
 * </P>
 *
 * @see PPP
 * @see PPPEventListener
 */

public class PPPEvent
   extends EventObject
{

   /**
    * Event UP, PPP has entered the up state.
    *
    * <P>
    * An <CODE>UP</CODE> event is generated if the connection is successfully established. At this point the PPP
    * connection is established and is read for IP network traffic. However the application must invoke the <CODE>
    * PPP.addInterface</CODE> method before the TCP/IP stack will recognize the new PPP connection as a valid
    * network interface. The interface name passed to <CODE>PPP.addInterface</CODE> is typically formed by appending the
    * number of the serial port being used for PPP traffic to the lower case string 'ppp'.
    * </P>
    *
    */
   public static final int UP = 1;

   /**
    * Event CLOSED, PPP has entered the closed state.
    *
    * <P>
    * The <CODE>CLOSED</CODE> event occurs when the PPP link is no longer available. The link can become
    * unavailable because the link was shut down normally or as the result of an error. After receiving the
    * <CODE>CLOSED</CODE> event the PPP listener frees any resources that were consumed establishing the
    * connection. The listener should also call the <CODE>PPP.removeInterface</CODE> method to remove the
    * PPP network interface.
    * </P>
    *
    */
   public static final int CLOSED = 2;

   /**
    * Event AUTHENTICATION, PPP has received an authentication request from the remote peer.
    *
    * <P>
    * An <CODE>AUTHENTICATION_REQESTED</CODE> event is generated if <CODE>PPP.setAuthenticate</CODE> is invoked
    * with a parameters value of true. This will cause PPP to request authentication information from the
    * remote peer during its negotiation of the connection. This event gives the application the chance to
    * verify the login information. The <CODE>PPP.getRemoteUserName</CODE> and <CODE>PPP.getRemotePassword</CODE>
    * methods are used to retreive the remote peer's login information. After the application examines the
    * login information it invokes the <CODE>PPP.authenticate</CODE> method with a value of true if the remote
    * peer's login information is acceptable. In this case PPP continues its negotiation of the connection.
    * If the login information is invalid calling <CODE>PPP.authenticate</CODE> with a value of false will
    * cause PPP to reject the clients connection request and generate a <CODE>STOPPED</CODE> event.
    * </P>
    *
    */
   public static final int AUTHENTICATION_REQUEST = 3;

   /**
    * Event STOPPED, PPP has entered the stopped state.
    *
    * <P>
    * The <CODE>STOPPED</CODE> event is typically generated in response to an error condition. Some of the
    * possible sources of errors include problems negotiating connection options, rejections of authentication
    * information, rejection of PAP authentication protocol or the remote peer explicitly closing the connection.
    * The source of the error can determined by invoking the <CODE>PPP.getLastError</CODE> method on the <CODE>PPPEvent</CODE>
    * object passed to the event listener.
    * </P>
    *
    */
   public static final int STOPPED = 4;

    /**
     * Event STARTING, PPP has enetered the starting state.
     *
     * <P>
     * A <CODE>STARTING</CODE> event is generated by the application invoking the <CODE>PPP.open</CODE> method.
     * The <CODE>STARTING</CODE> event provides the application with a chance to bring up the physical communications
     * link. At a minimum this involves initializing the serial port that will be used by all PPP traffic. If a modem
     * is attached to the serial port the application also intializes the modem and either instructs it to dial a
     * remote modem or waits for the modem to answer an incomming call. After the physical communication link has been
     * established the application invokes <CODE>PPP.up</CODE> method. At this point PPP assumes
     * exclusive use of the serial port. Any other attempt to read from or write to the serial port could disrupt
     * the PPP connection and will likely result in a STOPPPED event being generated.
     * </P>
     *
     */
   public static final int STARTING = 5;




    /**
     * Error code NONE, no error.
     */
   public static final int NONE   = 0;

   /**
    * Error code ADDR, one or both of the IP addresses could not be negotiated.
    *
    * Some of the possible causes of address errors while using TINI PPP as a server:
    *   Not configuring local and remote addresses before starting a PPP server session.
    *   Remote peer attempting to use a static IP which does not match the configured address.
    *
    * Some of the possible causes of address errors while using TINI PPP as a client:
    *   Attempting to use a static IP while the server is configured for dynamic address configuration.
    *   Using the wrong static IP.
    */
   public static final int ADDR   = 1;

   /**
    * Error code AUTH, remote peer rejected this peers authentication credentials.
    *
    * Some of the possible causes of authentication errors are:
    *   Remote peer refuses to accept PAP authentication protocol, TINI PPP only implements PAP authentication protocol.
    *   Bad username and/or password.
    */
   public static final int AUTH   = 2;

   /**
    * Error code TIME, link negotiation timed out.
    *
    * This error occurs when the remote peer does not respond to configuration messages within the timeout period
    * specified by <CODE>PPP.setrestartPeriod</CODE>.
    *
    */
   public static final int TIME   = 3;

   /**
    * Error code REJECT, link options have been rejected by remote peer.
    *
    * This error occurs when the link configuration options have not converged with the number of negotiation attempts
    * specified by <CODE>PPP.setMaxFailure</CODE>.
    */
   public static final int REJECT = 4;



   private int             eventType;
   private int             error;


   /**
    * Construct a new PPPEvent object.
    *
    * @param srcPPP PPP object that generated this event
    * @param eventType type of event
    * @param error error asscociated with event
    *
    */
   PPPEvent (PPP srcPPP, int eventType, int error)
   {
      super(srcPPP);

      this.eventType = eventType;
      this.error = error;
   }

   /**
    *  Gets the event type for this PPPEvent
    *
    * @return event type
    *
    */
   public int getEventType ()
   {
      return eventType;
   }

   /**
    *  Gets the error type for this PPPEvent
    *
    * @return error type
    *
    */
   public int getLastError ()
   {
      return error;
   }

}
