/* ----------------------------------------------------------------------
 * FILE: emitter.c
 * PACKAGE: as31 - 8031/8051 Assembler.
 *
 * DESCRIPTION:
 *  This file contains the code to generate various
 *  object code formats. Provisions exist to
 *  support many types of output formats within the
 *  same executable.
 *
 * REVISION HISTORY:
 *  Jan. 19, 1990 - Created. (Ken Stauffer)
 *  Jan. 29, 1990 - Added S-records (Theo Deraadt)
 *  August   1999 - Converted to 24 bit addressing (Tom Chenot)
 *                - Added extended hex file support
 *  May      2000 - Added tbin (Tini BINary) support
 *                - Added tbin (Tini LIBrary) support
 *
 *
 * AUTHOR:
 *  Original code in this file written by Ken Stauffer (University of Calgary).
 *  January, 1990.
 *  24 bit addressing extensions written by Tom Chenot (Dallas Semiconductor).
 *  1999.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "a390.h"

#ifndef FALSE
#define FALSE 0
#define TRUE  1
#endif

#ifndef min
#define min(A,B) ((A) < (B))? (A): (B)
#endif

#define DebugDump(S) fprintf(stderr, "DEBUG: %s %s\n", "", S)
/**/
#define MEMORY_ALLOCATION_ERROR "Cannot allocate memory for buffer!"
#define BUFFER_SIZE 65536
#define STRING_SIZE 128
/**/
void  AddStringEntry(char *p_String);
void  AddAFixup(void);
void  AddLFixup(void);
void  AddDFixup(void);
void  Cleanup(void);
void  DumpFixups(void);
void  DumpStringTable(void);
/**/
typedef struct _FIXUP_TABLE_ENTRY
{
  long Address;
  struct _FIXUP_TABLE_ENTRY *Next;
} FIXUP_TABLE_ENTRY, *LPFIXUP_TABLE_ENTRY;
/**/
typedef struct _FIXUP_TABLE
{
  unsigned long Count;
  LPFIXUP_TABLE_ENTRY lpHead;
  LPFIXUP_TABLE_ENTRY lpCurrent;
} FIXUP_TABLE, *LPFIXUP_TABLE;
/**/
typedef struct _STRING_TABLE_ENTRY
{
  char String[STRING_SIZE+1];
  unsigned long Address;
  struct _STRING_TABLE_ENTRY *Next;
} STRING_TABLE_ENTRY, *LPSTRING_TABLE_ENTRY;
/**/
typedef struct _STRING_TABLE
{
  unsigned long Count;
  unsigned long ByteCount; 
  LPSTRING_TABLE_ENTRY lpHead;
  LPSTRING_TABLE_ENTRY lpCurrent;
} STRING_TABLE, *LPSTRING_TABLE;
/**/
typedef struct _DYNAMIC_BUFFER_ENTRY
{
  char *Data;
  struct _DYNAMIC_BUFFER_ENTRY *Next;
} DYNAMIC_BUFFER_ENTRY,*LPDYNAMIC_BUFFER_ENTRY;
typedef struct _DYNAMIC_BUFFER
{
  unsigned long TotalByteCount;
  LPDYNAMIC_BUFFER_ENTRY lpHead;
  LPDYNAMIC_BUFFER_ENTRY lpCurrent;
} DYNAMIC_BUFFER, LPDYNAMIC_BUFFER;
/**/
DYNAMIC_BUFFER g_DynamicBuffer = {0, NULL, NULL};
/**/
FIXUP_TABLE g_LFixups = {0, NULL},
            g_AFixups = {0, NULL},
            g_DFixups = {0, NULL};
/**/
STRING_TABLE g_StringTable = {0, 0, NULL};
/**/
extern char g_Native;
extern MODULE_STATS g_ModuleStats;

/* ----------------------------------------------------------------------
 * DECLARE your own open(), close(), addr(), and byte() routines here.
 *
 */
/*
static void open1(char *file, char *ftype, char *arg),    
            close1(),    
            addr1(unsigned long int a),    
            byte1(unsigned char b);
static void open2(char *file, char *ftype, char *arg),    
            close2(),    
            addr2(unsigned long int a),    
            byte2(unsigned char b);
static void open3(char *file, char *ftype, char *arg),
            close3(),    
            addr3(unsigned long int a),    
            byte3(unsigned char b);
static void open4(char *file, char *ftype, char *arg),    
            close4(),    
            addr4(unsigned long int a),    
            byte4(unsigned char b);
*/
static void OpenHex(char *file, char *ftype, char *arg),  
            CloseHex(),  
            AddrHex(unsigned long int a),  
            ByteHex(unsigned char b);
static void OpenTBin(char *file, char *ftype, char *arg), 
            CloseTBin(), 
            AddrTBin(unsigned long int a), 
            ByteTBin(unsigned char b);
static void OpenTLib(char *file, char *ftype, char *arg), 
            CloseTLib(), 
            AddrTLib(unsigned long int a), 
            ByteTLib(unsigned char b);
/**/
void finishline();
void dumpline(unsigned long a, unsigned char *b, int len);
/**/
extern char g_CallOperation;
/**/
#define INTEL_LINE_LENGTH 0x10
/**/
static  unsigned long g_CurrentAddress;
static  unsigned long g_StartAddress = 0; 
static  FILE          *g_lpOutputFile;
static  long int      g_Offset;
static  int           g_NewAddress;
static  int           g_Position     = -666;

/* Intel hex file variables */
static  unsigned  char g_IntelLineLength = INTEL_LINE_LENGTH;
static  unsigned  char g_IntelLineChecksum;
static  unsigned  long g_IntelStartAddress;
/**/
static  unsigned  char g_ByteBuffer[8+BUFFER_SIZE]; // sector size plus overhead
static  unsigned  int  g_ByteCount,
                       g_TCount;
/* intel stat vars */

char                    g_StatsOn                 = 0;
/**/
//static  long            g_MethodCount             = 0;
static  char            g_OutputFileName[512]     = {0};
static  char            g_IntelStatFileName[512]  = {0};
static  long            g_StatBanksWritten        = 0;
static  unsigned  char  g_CheckSum                = 0;
static  unsigned  char  g_BankSwitch              = 0;
static  unsigned  char  g_CurrentBank             = 0;
#define NUM_BANKS 64
static  unsigned  int   g_BytesInBank[NUM_BANKS]  = {0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0,
                                                     0,0,0,0,0,0,0,0};

/****************************************************************************
 *
 *   CRC16 generation. Used here only for table generation.
 *
 */
unsigned short CRC16TGen(unsigned char x, unsigned short crc)
{
  unsigned char bit,
                n;

  for(n = 0; n < 8; n++)
  {
    bit = (crc ^ x) & 1;
    crc >>= 1;

    if(bit)
      crc ^= 0xA001;

    x >>= 1;
  }

  return crc;
}
/**/
unsigned short CRCTable[256];
/**/
/****************************************************************************
 *
 *    Generate 256 word table for CRC16 lookup.
 *
 */
void GenCRC16Table()
{
  unsigned short i;

  for(i = 0; i < 256; ++i)
    CRCTable[i] = CRC16TGen((unsigned char) i, 0);
 }
/****************************************************************************
 *
 *    Fast CRC16 generation.
 *
 */
unsigned short CRC16(unsigned char x, unsigned short crc)
{
  return CRCTable[(crc ^ x) & 0xFF] ^ crc >> 8;
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 * ADD an entry to this table to register your
 * output format routines. Give your object format
 * a name to be specified with the -F option.
 *
 */
static int g_Format;
/**/
static struct 
{
  const char *name;
  void (*e_open)(char *, char *, char *);
  void (*e_close)();
  void (*e_addr)(unsigned long);
  void (*e_byte)(unsigned char);
} g_FormatTable[] = 
              { 
                { "tlib",   OpenTLib, CloseTLib, AddrTLib, ByteTLib },
                { "tbin",   OpenTBin, CloseTBin, AddrTBin, ByteTBin },
                { "hex",    OpenHex,  CloseHex,  AddrHex,  ByteHex  }, 
/*
                { "tdr",    open1,    close1,    addr1,    byte1    },
                { "byte",   open2,    close2,    addr2,    byte2    },
                { "od",     open3,    close3,    addr3,    byte3    },
                { "srec2",  open4,    close4,    addr4,    byte4    },
                { "srec3",  open4,    close4,    addr4,    byte4    },
                { "srec4",  open4,    close4,    addr4,    byte4    },
*/
              };

#define FORMTABSIZE  (sizeof(g_FormatTable)/sizeof(g_FormatTable[0]))
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitusage()
{
  int i;

  fprintf(stderr, "\n\t\tfmt is one of: ");
  
  for(i = 0; i < FORMTABSIZE; ) 
  {
    fprintf(stderr, "%s", g_FormatTable[i].name);

    if(++i < FORMTABSIZE)
      fprintf(stderr, ", ");
  }

  fprintf(stderr, ".\n");
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitopen(char *file, char *ftype, char *arg)
{
  int i;

  strcpy(g_OutputFileName, file);

  if(ftype) 
  {
    for(i = 0; i < FORMTABSIZE; i++) 
    {
      if(!strcmp(g_FormatTable[i].name, ftype)) 
      {
        g_Format = i;
        (*g_FormatTable[g_Format].e_open)(file, ftype, arg);
        return;
      }
    }

    fprintf(stderr, "warning: no format \"%s\", using \"%s\"\n",
      ftype, g_FormatTable[0].name);
  }
  /*
   * 0th entry is the default format type
   */
  g_Format = 0;
  (*g_FormatTable[g_Format].e_open)(file, ftype, arg);
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitclose()
{
  (*g_FormatTable[g_Format].e_close)();
  Cleanup();
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitdelete()
{
  remove(g_OutputFileName);
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitaddr(unsigned long int a)
{
  (*g_FormatTable[g_Format].e_addr)(a);
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void emitbyte(unsigned char b)
{
  (*g_FormatTable[g_Format].e_byte)(b);
}
/* ----------------------------------------------------------------------
 * Individual file format routines appear here:
 *  Each file format must define the following routines:
 *    open()  - Called ONCE before any of the others.
 *      It is passed with a filename and a format
 *      specific argument.
 *
 *    close() - Called ONCE when no more emit_byte()
 *      function calls will be made.
 *
 *    addr() - Called when ever a new address has been set
 *      in the assembler (ie. .org, .skip).
 *      This routine is also called once when the
 *      location counter is set to 0 at the very start of
 *      assembling.
 * 
 *    byte() - Called with each byte to be outputed.
 *
 */

/* ----------------------------------------------------------------------
 * "tdr" format. For tdr's 68008 system. Generates a
 * script file readable by a debugger.
 *  [addr] : [byte] [byte] ..
 *
 * arg: This is a number in decimal which specifies
 *  the offset, -Ftdr -A0000
 *
 *  These options specifies the tdr format, with an argument
 *  of 0. This becomes the offset used in generating the
 *  script file. The default if no A is present is 0x10000.
 * 
 */
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void OpenTLib(char *file, char *ftype, char *arg)
{
//DebugDump("OpenTLib"); 

  /* reset all tables */
  memset((char *)&g_DynamicBuffer, 0, sizeof(g_DynamicBuffer));
  memset((char *)&g_AFixups,       0, sizeof(g_AFixups));
  memset((char *)&g_LFixups,       0, sizeof(g_LFixups));
  memset((char *)&g_DFixups,       0, sizeof(g_DFixups));
  memset((char *)&g_StringTable,   0, sizeof(g_StringTable));
  
  if(!(g_DynamicBuffer.lpHead = malloc(sizeof(DYNAMIC_BUFFER_ENTRY))))
  {
    fprintf(stderr, MEMORY_ALLOCATION_ERROR);
    Cleanup();
    exit(1);
  }
  
  g_TCount = g_CurrentAddress = g_StartAddress = 0;
  memset(g_DynamicBuffer.lpHead, 0, sizeof(DYNAMIC_BUFFER_ENTRY));

  g_DynamicBuffer.lpCurrent      = g_DynamicBuffer.lpHead;
  g_DynamicBuffer.lpHead->Data   = g_ByteBuffer;
}
/* ----------------------------------------------------------------------
 *  CloseTLib - TODO: Add CRC support, Version support
 */
static void CloseTLib()
{
  /* this is where we format the data and write it to a file */
  char	   l_Temp[256] = "";
  int      l_CurrentByte;
  unsigned long i;
  unsigned long l_Length;

//DebugDump("CloseTLib");

  GenCRC16Table();
  remove(g_OutputFileName);
  g_lpOutputFile = fopen(g_OutputFileName, "wb");
 
//DebugDump("CloseTLiba");
  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writing.\n", g_OutputFileName);
    exit(1);
  }

  l_CurrentByte = 0;

//DebugDump("CloseTLibb");
  /* write magic */
  sprintf(l_Temp+l_CurrentByte, "TLIB");
  l_CurrentByte += 4;

  /* write legacy version bytes */
  /* NOTE: - this should probably be configurable */
  l_Temp[l_CurrentByte++] = l_Temp[l_CurrentByte++] = 0;
  
  /* write tag */
  l_Temp[l_CurrentByte++] = 0x01;

  l_Length = 3+  /* version */
             4 + /* Method count length */
             4 + g_StringTable.ByteCount + g_StringTable.Count*2 + /* String Table Length */
             4 + g_StringTable.Count*3 + /* Method address table length */
             4 + g_AFixups.Count*3 + /* A Fixup Table Length */ 
             4 + g_LFixups.Count*3 + /* L Fixup Table Length */ 
             4 + g_DFixups.Count*3 + /* D Fixup Table Length */ 
             4 + g_DynamicBuffer.TotalByteCount; /* Object code length */

//printf("Block Length: %d, 0x%04x\n", l_Length, l_Length);

  /* write block length */
  l_Temp[l_CurrentByte++]  = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++]  = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++]  = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++]  = (unsigned char)((l_Length>>24) & 0x0ff);

  /* write the version */
  /* HACK - Hard code for the first release - FIXME FIXME FIXME FIXME */
  l_Temp[l_CurrentByte++]  = 0x01; /* beta release  */
  l_Temp[l_CurrentByte++]  = 0x03; /* major version */
  l_Temp[l_CurrentByte++]  = 0x00; /* minor release */

//DebugDump("CloseTLibc");
//printf("Block Length: %d, 0x%04x\n", l_Length, l_Length);
  
  /* write method count */
  l_Temp[l_CurrentByte++] = (unsigned char)(g_StringTable.Count & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((g_StringTable.Count>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((g_StringTable.Count>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((g_StringTable.Count>>24) & 0x0ff);

  /* dump it to the file */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
  
//DebugDump("CloseTLibd");
  /* write the string table */
  l_CurrentByte = 0;
  l_Length = g_StringTable.Count*2 + g_StringTable.ByteCount;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

  /* dump it to the file */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
 
  g_StringTable.lpCurrent = g_StringTable.lpHead;

//printf("String Table Count: %d, 0x%08x\n", g_StringTable.Count, g_StringTable.Count);

  /* dump the string table */ 
  for(i = 0; i < g_StringTable.Count; i++)
  {
    l_CurrentByte = 0;
    l_Length = strlen(g_StringTable.lpCurrent->String);
    l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
    strcpy(l_Temp+l_CurrentByte, g_StringTable.lpCurrent->String);
//printf("%s\n", g_StringTable.lpCurrent->String);
    l_CurrentByte += l_Length;
    
    /* dump it to the file */
    fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
    g_StringTable.lpCurrent = g_StringTable.lpCurrent->Next; 
  }

//DebugDump("CloseTLibe");
  l_CurrentByte = 0;
  l_Length = g_StringTable.Count*3;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

  /* dump it to the file */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);

  g_StringTable.lpCurrent = g_StringTable.lpHead;

  /* dump the address table */ 
  for(i = 0; i < g_StringTable.Count; i++)
  {
    l_CurrentByte = 0;
    l_Temp[l_CurrentByte++] = (unsigned char)(g_StringTable.lpCurrent->Address & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_StringTable.lpCurrent->Address>>8) & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_StringTable.lpCurrent->Address>>16) & 0x0ff);
    
    /* dump it to the file */
    fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
    g_StringTable.lpCurrent = g_StringTable.lpCurrent->Next; 
  }
  
//DebugDump("CloseTLibf");
  /* now the A Fixups */
  /* first the length */
  l_CurrentByte       = 0;
  l_Length            = g_AFixups.Count*3;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

  /* dump it */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);

  g_AFixups.lpCurrent = g_AFixups.lpHead;

  /* now dump the A fixups */
  for(i = 0; i < g_AFixups.Count; i++)
  {
    l_CurrentByte = 0;
    l_Temp[l_CurrentByte++] = (unsigned char)(g_AFixups.lpCurrent->Address & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_AFixups.lpCurrent->Address>>8) & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_AFixups.lpCurrent->Address>>16) & 0x0ff);
    
    /* dump it */
    fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
    g_AFixups.lpCurrent = g_AFixups.lpCurrent->Next;
  }

//DebugDump("CloseTLibg");
  /* now the L Fixups */
  /* first the length */
  l_CurrentByte       = 0;
  l_Length            = g_LFixups.Count*3;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

  /* dump it */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);

  g_LFixups.lpCurrent = g_LFixups.lpHead;

  /* now dump the  fixups */
  for(i = 0; i < g_LFixups.Count; i++)
  {
    l_CurrentByte = 0;
    l_Temp[l_CurrentByte++] = (unsigned char)(g_LFixups.lpCurrent->Address & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_LFixups.lpCurrent->Address>>8) & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_LFixups.lpCurrent->Address>>16) & 0x0ff);
    
    /* dump it */
    fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
    g_LFixups.lpCurrent = g_LFixups.lpCurrent->Next;
  }

  /* now the D Fixups */
  /* first the length */
  l_CurrentByte       = 0;
  l_Length            = g_DFixups.Count*3;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

  /* dump it */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);

  g_DFixups.lpCurrent = g_DFixups.lpHead;

  /* now dump the  fixups */
  for(i = 0; i < g_DFixups.Count; i++)
  {
    l_CurrentByte = 0;
    l_Temp[l_CurrentByte++] = (unsigned char)(g_DFixups.lpCurrent->Address & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_DFixups.lpCurrent->Address>>8) & 0x0ff);
    l_Temp[l_CurrentByte++] = (unsigned char)((g_DFixups.lpCurrent->Address>>16) & 0x0ff);
    
    /* dump it */
    fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
    g_DFixups.lpCurrent = g_DFixups.lpCurrent->Next;
  }

//DebugDump("CloseTLibh");
  /* first the length */
  l_CurrentByte             = 0;
  l_Length                  = g_DynamicBuffer.TotalByteCount;
  l_Temp[l_CurrentByte++] = (unsigned char)(l_Length & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>8) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>16) & 0x0ff);
  l_Temp[l_CurrentByte++] = (unsigned char)((l_Length>>24) & 0x0ff);

//printf("Block size: %d,0x%04x\n", l_Length, l_Length);

  /* dump it */
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);

  g_DynamicBuffer.lpCurrent = g_DynamicBuffer.lpHead;

  for(i = 0; i < g_DynamicBuffer.TotalByteCount; i += min(BUFFER_SIZE, l_Length),
                                                 l_Length -= min(BUFFER_SIZE, l_Length))
  {

//printf("Block %d: size: %d,0x%04x", i, min(BUFFER_SIZE, l_Length), min(BUFFER_SIZE, l_Length));
    fwrite(g_DynamicBuffer.lpCurrent->Data, 1, min(BUFFER_SIZE, l_Length), g_lpOutputFile); 
    g_DynamicBuffer.lpCurrent = g_DynamicBuffer.lpCurrent->Next;
  }

  l_CurrentByte = 0;

//DebugDump("CloseTLibi");
  /* write CRC16 tag*/
  l_Temp[l_CurrentByte++] = 0x02;

  /* write CRC16 length */
  l_Temp[l_CurrentByte++] = 0x02;  
  l_Temp[l_CurrentByte++] = 0x00;  
  l_Temp[l_CurrentByte++] = 0x00;  
  l_Temp[l_CurrentByte++] = 0x00;  

  /* write the CRC16 */ /* HACK HACK HACK */
  l_Temp[l_CurrentByte++] = (unsigned char)0x055;
  l_Temp[l_CurrentByte++] = (unsigned char)0x055;

  /* write the end tag */
  l_Temp[l_CurrentByte++] = (unsigned char)0x0ff;
  fwrite(l_Temp, 1, l_CurrentByte, g_lpOutputFile);
  fclose(g_lpOutputFile);

//DebugDump("CloseTLibj");
//  Cleanup();
}
/* ----------------------------------------------------------------------
 *
 */
static void AddrTLib(unsigned long int a)
{
//DebugDump("AddrTLib");
  if(a == 0)
    return;

  /* since we are writing raw data to a buffer we have no concept       */
  /* of non-contiguous addressing... also, all addresses are considered */
  /* to be relative and will be fixed up                                */
  fprintf(stderr, "Absolute addressing is not available in this mode\n");
  Cleanup();
  exit(1);
}
/* ----------------------------------------------------------------------
 *
 */
static void ByteTLib(unsigned char b)
{
//DebugDump("ByteTLib");
  if(g_TCount == BUFFER_SIZE)
  {
    if(!(g_DynamicBuffer.lpCurrent->Next = malloc(sizeof(DYNAMIC_BUFFER_ENTRY))))
    {
      fprintf(stderr, MEMORY_ALLOCATION_ERROR);
      Cleanup();
      exit(1);
    }
    
    g_DynamicBuffer.lpCurrent = g_DynamicBuffer.lpCurrent->Next;
    memset(g_DynamicBuffer.lpCurrent, 0, sizeof(DYNAMIC_BUFFER_ENTRY));

    if(!(g_DynamicBuffer.lpCurrent->Data = malloc(BUFFER_SIZE)))
    {
      fprintf(stderr, MEMORY_ALLOCATION_ERROR);
      Cleanup();
      exit(1);
    }

    g_TCount = 0;
  }

  g_DynamicBuffer.lpCurrent->Data[g_TCount++] = b & 0x0ff;
  g_DynamicBuffer.TotalByteCount++;
  g_CurrentAddress += 1;
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void OpenTBin(char *file, char *ftype, char *arg)
{
  char *l_lpExtPointer;

  GenCRC16Table();
  remove(file);
  g_lpOutputFile = fopen(file, "wb");
 
  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writing.\n",file);
    exit(1);
  }

  g_TCount = g_CurrentAddress = g_StartAddress = 0;

  /* init stats */
  if(!g_StatsOn)
    return;

  strcpy(g_IntelStatFileName, file);

  if((l_lpExtPointer = strrchr(g_IntelStatFileName,'.')) == NULL)
     l_lpExtPointer = g_IntelStatFileName + strlen(g_IntelStatFileName);

  strcpy(l_lpExtPointer,".sta");

  g_BankSwitch  = 1;
  g_CheckSum = g_CurrentBank = (unsigned char)(g_StatBanksWritten = 0);
}
/* ----------------------------------------------------------------------
 *
 */
static void WriteTBinBlock()
{
  unsigned char  l_AddrLenBuff[5];
  unsigned char  l_CRC16Buff[2];
  unsigned int   l_BytesProcessed   = 0;
  unsigned short l_CRC = 0;
  int            l_RetVal;

  if(g_CurrentAddress == 0)
    return;

  if(g_TCount == 0)
    return;
  
  // copy start address
  l_AddrLenBuff[0] = (unsigned char)g_StartAddress;
  l_AddrLenBuff[1] = (unsigned char)(g_StartAddress>>8);
  l_AddrLenBuff[2] = (unsigned char)(g_StartAddress>>16);
  
  // copy byte count
  // the bytecount can be in the range 1 to 65536
  // and we will be representing the length with
  // two bytes so we will writing length - 1
  g_TCount--;
  l_AddrLenBuff[3] = g_TCount;
  l_AddrLenBuff[4] = g_TCount>>8;
  g_TCount++;

  for(l_BytesProcessed = 0; l_BytesProcessed < g_TCount; l_BytesProcessed++)
    l_CRC = CRC16(g_ByteBuffer[l_BytesProcessed], l_CRC); 

  // copy CRC 
  l_CRC16Buff[0] = (unsigned char)l_CRC;
  l_CRC16Buff[1] = l_CRC>>8;

  l_BytesProcessed = 0;

  fwrite(l_AddrLenBuff, 1, sizeof(l_AddrLenBuff), g_lpOutputFile);
  
  do
  {
    l_RetVal = fwrite(g_ByteBuffer+l_BytesProcessed, 1, g_TCount-l_BytesProcessed, g_lpOutputFile);

    if(l_RetVal < 1)
    {
      fprintf(stderr, "ERROR, File Write Failed!\n");
      fclose(g_lpOutputFile);
      exit(1);
    }

    l_BytesProcessed += l_RetVal;
  }while(l_BytesProcessed < g_TCount);

  fwrite(l_CRC16Buff, 1, sizeof(l_CRC16Buff), g_lpOutputFile);
  g_TCount = 0;
}
/* ----------------------------------------------------------------------
 *
 */
static void CloseTBin()
{
  WriteTBinBlock();
  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void AddrTBin(unsigned long int a)
{
  if(a == 0)
    return;

  WriteTBinBlock();
  g_StartAddress = g_CurrentAddress = a;
  g_CurrentBank = (unsigned char)((g_CurrentAddress >> 16) & 0xff);
}
/* ----------------------------------------------------------------------
 *
 */
static void BoundaryCheck()
{
  int i;
  
  if(!g_CallOperation)
    return;

  g_CallOperation = 0;

  for(i = 0; i < 4; i++)
  {
    if((g_TCount-1-i) >= 0)
    {
      if(g_ByteBuffer[g_TCount-1-i] == 0x012)// lcall
        printf("*** LCall Boundary Warning *** at address 0x%06x\n", g_CurrentAddress);
    
      if(i < 3)
        if((g_ByteBuffer[g_TCount-1-i] & 0x01f) == 0x011)// acall
          printf("*** ACall Boundary Warning *** at address 0x%06x\n", g_CurrentAddress);
    }
  }
}
/* ----------------------------------------------------------------------
 *
 */
static void ByteTBin(unsigned char b)
{
  if((g_CurrentAddress>>16) != (g_StartAddress>>16))
  {
    BoundaryCheck();
    WriteTBinBlock();
    g_StartAddress = g_CurrentAddress;
    g_BankSwitch = TRUE;
    g_CurrentBank = (unsigned char)((g_CurrentAddress >> 16) & 0xff);
  }

  g_ByteBuffer[g_TCount++] = b & 0x0ff;
  g_CurrentAddress += 1;

  if(g_ModuleStats.count >= 0)
    g_ModuleStats.modules[g_ModuleStats.count].size++;

  /* stats */
  if(!g_StatsOn)
    return;

  g_BytesInBank[g_CurrentBank]++;
  g_CheckSum += (b & 0x0ff);
 
  if(g_BankSwitch)
  {
    g_BankSwitch = FALSE;
    g_StatBanksWritten |= (1 << g_CurrentBank);
  }
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void OpenHex(char *file, char *ftype, char *arg)
{
  char *l_lpExtPointer;
  g_lpOutputFile = fopen(file, "w");
  
  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writting.\n",file);
    exit(1);
  }

  g_Position = -666;
  g_IntelStartAddress  = ~0L;/* bogus 32 bit address */
  g_ByteCount = 0;

  /* init stats */
  if(!g_StatsOn)
    return;

  strcpy(g_IntelStatFileName, file);

  if((l_lpExtPointer = strrchr(g_IntelStatFileName,'.')) == NULL)
     l_lpExtPointer = g_IntelStatFileName + strlen(g_IntelStatFileName);

  strcpy(l_lpExtPointer,".sta");

  g_BankSwitch  = 1;
  g_CheckSum = g_CurrentBank = (unsigned char)(g_StatBanksWritten = 0);
}
/* ----------------------------------------------------------------------
 *
 */
static void WriteCheckSum()
{
  unsigned int l_i;

  g_IntelLineChecksum = 0;
  g_ByteBuffer[0]   = g_ByteCount - 4;
  fprintf(g_lpOutputFile,":"); 

  for(l_i = 0; l_i < g_ByteCount; l_i ++)
  {
    fprintf(g_lpOutputFile,"%02X", g_ByteBuffer[l_i]); 
    g_IntelLineChecksum += g_ByteBuffer[l_i] & 0x0ff;
  }

  fprintf(g_lpOutputFile,"%02X\n", ((~g_IntelLineChecksum) + 1) & 0x0ff);
  g_ByteCount = 0;
  g_Position = g_IntelLineLength;
}
/* ----------------------------------------------------------------------
 *
 */
static void WriteExtendedLinearHeader()
{
  g_ByteCount = 1;

  g_ByteBuffer[g_ByteCount++] = 0;
  g_ByteBuffer[g_ByteCount++] = 0;
  g_ByteBuffer[g_ByteCount++] = 4;
  g_ByteBuffer[g_ByteCount++] = (unsigned char)((g_CurrentAddress >> 24) & 0xff);
  g_ByteBuffer[g_ByteCount++] = (unsigned char)((g_CurrentAddress >> 16) & 0xff);

  WriteCheckSum();

  /* stats */
  if(!g_StatsOn)
    return;

  g_BankSwitch  = 1;
  g_CurrentBank = (unsigned char)((g_CurrentAddress >> 16) & 0xff);
}
/* ----------------------------------------------------------------------
 *
 */
static void CloseHex()
{
  if(g_Position != g_IntelLineLength) 
    if(g_Position > 0)
      WriteCheckSum();

  fprintf(g_lpOutputFile,":00000001FF\n");
  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void AddrHex(unsigned long int a)
{
  g_CurrentAddress = a;

  if(g_IntelStartAddress >> 24)
  {
    g_IntelStartAddress = g_CurrentAddress;
    WriteExtendedLinearHeader();
  }
  else if((g_CurrentAddress>>16) != (g_IntelStartAddress>>16))
  {
    if(g_Position != -666)
      if(g_Position != g_IntelLineLength) 
        WriteCheckSum();

    g_IntelStartAddress = g_CurrentAddress;
    WriteExtendedLinearHeader();
  }

  g_IntelStartAddress = g_CurrentAddress;
  g_NewAddress = 1;
}
/* ----------------------------------------------------------------------
 *
 */
static void WriteDataHeader()
{
  g_ByteCount = 1;
  g_ByteBuffer[g_ByteCount++] = (unsigned char)(((g_CurrentAddress + g_Offset) >> 8) & 0x0ff);
  g_ByteBuffer[g_ByteCount++] = (unsigned char)((g_CurrentAddress + g_Offset) & 0x0ff);
  g_ByteBuffer[g_ByteCount++] = 0;
}
/* ----------------------------------------------------------------------
 *
 */
static void ByteHex(unsigned char b)
{
  if(g_IntelStartAddress >> 24)
  {
    g_IntelStartAddress = g_CurrentAddress;
    WriteExtendedLinearHeader();
  }

  if((g_CurrentAddress>>16) != (g_IntelStartAddress>>16))
  {
    if(g_Position != -666)
    {
      BoundaryCheck();

      if(g_Position != g_IntelLineLength) 
        WriteCheckSum();
    }

    g_IntelStartAddress = g_CurrentAddress;
    WriteExtendedLinearHeader();
  }

  if(g_NewAddress) 
  {
    if(g_Position != -666)
      if(g_Position != g_IntelLineLength) 
        WriteCheckSum();

    g_NewAddress = 0;
  }
   
  if(g_Position == g_IntelLineLength) 
    WriteDataHeader();

  g_ByteBuffer[g_ByteCount++] = b & 0xff;

  if(--g_Position == 0) 
    WriteCheckSum();

  g_CurrentAddress += 1;

  if(g_ModuleStats.count >= 0)
    g_ModuleStats.modules[g_ModuleStats.count].size++;

  /* stats */
  if(!g_StatsOn)
    return;

  g_CheckSum += (b & 0x0ff);
  g_BytesInBank[g_CurrentBank]++;
  
  if(g_BankSwitch)
  {
    g_BankSwitch = FALSE;
    g_StatBanksWritten |= (1 << g_CurrentBank);
  }
}
/* ----------------------------------------------------------------------
 *
 */
void WriteStats()
{
  int   i;
  long  l_TotalBytes  = 0,
        l_BanksUsed   = 0;

  if(!g_StatsOn)
    return;

  printf("\n\nFile Statistics\n");
  printf(    "===============\n");

  // write total bytes
  for(i = 0; i < NUM_BANKS; i++)
    l_TotalBytes += g_BytesInBank[i];

  printf("   Total Code Bytes Written: %d\n", l_TotalBytes);
  
  // bytes per bank
  printf("     Bytes Written Per Bank:\n");
  
  for(i = 0; i < NUM_BANKS; i++)
    if(g_BytesInBank[i])
    {
      printf("  Bank[%02d]: %d\n", i, g_BytesInBank[i]);
      l_BanksUsed++;
    }

// write banks used
  printf("\n           Total Banks Used: %d\n", l_BanksUsed);
  printf("              Code Checksum: 0x%02X\n", ((~g_CheckSum) + 1) & 0x0ff);
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void open1(char *file, char *ftype, char *arg)
{
  g_lpOutputFile = fopen(file,"w");

  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writting.\n",file);
    exit(1);
  }
  
  if(arg) 
  {
    g_Offset = atoi(arg);
  } 
  else
    g_Offset = 0x10000;
}
/* ----------------------------------------------------------------------
 *
 */
static void close1()
{
  if(g_Position != 15)
    fprintf(g_lpOutputFile,"\n");

  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void addr1(unsigned long int a)
{
  g_CurrentAddress = a;
  g_NewAddress = 1;
}
/* ----------------------------------------------------------------------
 *
 */
static void byte1(unsigned char b)
{
  if(g_NewAddress) 
  {
    if(g_Position != -666) 
      fprintf(g_lpOutputFile,"\n");

    g_NewAddress = 0;
    g_Position = 15;
    fprintf(g_lpOutputFile,"%06x: ", g_CurrentAddress + g_Offset);
  } 
  else if(g_Position == 15) 
  {
    fprintf(g_lpOutputFile,"%06x: ", g_CurrentAddress + g_Offset);
  }

  fprintf(g_lpOutputFile,"%02x ", b & 0xff);

  if(g_Position-- == 0) 
  {
    fprintf(g_lpOutputFile,"\n");
    g_Position = 15;
  }

  g_CurrentAddress += 1;
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 * "byte" format.
 *  Like "tdr" but each byte is on a line by itself.
 *  This is nice for debugging. No -A is used.
 */
static void open2(char *file, char *ftype, char *arg)
{
  g_lpOutputFile = fopen(file,"w");
  
  if( g_lpOutputFile == NULL ) 
  {
    fprintf(stderr,"Cannot open %s for writting.\n",file);
    exit(1);
  }
}
/* ----------------------------------------------------------------------
 *
 */
static void close2()
{
  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void addr2(unsigned long int a)
{
  g_CurrentAddress = a;
}
/* ----------------------------------------------------------------------
 *
 */
static void byte2(unsigned char b)
{
  fprintf(g_lpOutputFile,"%04x: %02x\n", g_CurrentAddress, b & 0xff);
  g_CurrentAddress += 1;
}
/* ----------------------------------------------------------------------
 * "od", this format shows 16 bytes per line, with address.
 *  It also includes ascii on one side.
 *
 * The format is similar to the od(1) program under Unix.
 *
 */
static int g_Position3;
static unsigned char buf[16];
static unsigned long saveaddr;
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void open3(char *file, char *ftype, char *arg)
{
  g_lpOutputFile = fopen(file,"w");
  
  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writting.\n",file);
    exit(1);
  }
}
/* ----------------------------------------------------------------------
 *
 */
static void close3()
{
  dumpline(saveaddr, buf, g_Position3 - 1);
  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void addr3(unsigned long int a)
{
  g_NewAddress = 1;
  g_CurrentAddress = a;
}
/* ----------------------------------------------------------------------
 *
 */
static void byte3(unsigned char b)
{
  if(g_NewAddress) 
  {
    dumpline(saveaddr, buf, g_Position3 - 1);
    g_Position3 = 0;
    g_NewAddress = 0;
    saveaddr = g_CurrentAddress;
  } 
  else if(g_Position3 == 16) 
  {
    dumpline(saveaddr, buf, g_Position3 - 1);
    g_Position3 = 0;
    saveaddr = g_CurrentAddress;
  }
  
  buf[g_Position3++] = b & 0x00ff;

  g_CurrentAddress += 1;
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void dumpline(unsigned long a, unsigned char *b, int len)
{
  int i;

  if(len <= 0) return;

  fprintf(g_lpOutputFile,"%06x: ",a);

  for(i = 0; i < 8; i++) 
  {
    if(i <= len)
      fprintf(g_lpOutputFile,"%02x ",b[i]);
    else
      fprintf(g_lpOutputFile,"   ");
  }

  fprintf(g_lpOutputFile,"- ");

  for(i = 8; i < 16; i++) 
  {
    if(i <= len)
      fprintf(g_lpOutputFile,"%02x ",b[i]);
    else
      fprintf(g_lpOutputFile,"   ");
  }

  fprintf(g_lpOutputFile,"   ");

  for(i = 0; i < 16; i++) 
  {
    if(i <= len)
      fprintf(g_lpOutputFile,"%c", ((b[i] >= ' ') && (b[i] <= '~')) ? b[i] : '.' );
    else
      break;
  }
  
  fprintf(g_lpOutputFile,"\n");
}
/* ----------------------------------------------------------------------
 * srecord format. This is called with "-Fsrec2", "-Fsrec3", or
 * "-Fsrec4"...
 *
 * arg: This is a number in decimal which specifies
 *  the offset, -Fsrec3 -A0000
 *
 *  These options specifies the tdr format, with an argument
 *  of 0. This becomes the offset used in generating the
 *  script file. The default if no A is present is 0x0000.
 * 
 */
#define SREC_BYTESPERLINE 32

static char g_Format4;
static int  check4, 
            index4;
static char buf4[SREC_BYTESPERLINE];
static long address4;
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
static void open4(char *file, char *ftype, char *arg)
{
  g_Format4 = ftype[4];    /* will be '2' -- '4' */

  g_lpOutputFile = fopen(file,"w");

  if(g_lpOutputFile == NULL) 
  {
    fprintf(stderr,"Cannot open %s for writing.\n",file);
    exit(1);
  }

  if(arg)  
    g_Offset = atoi(arg);
  else  
    g_Offset = 0;

  fprintf(g_lpOutputFile, "S0030000%02X\n", (~3 & 0xff));
}
/* ----------------------------------------------------------------------
 *
 */
static void close4()
{
  if(index4)
    finishline();
  
  switch(g_Format4) 
  {
    case '2':
      fprintf(g_lpOutputFile, "S9030000%02X\n", ~3 & 0xff);
    break;

    case '3':
      fprintf(g_lpOutputFile, "S804000000%02X\n", ~4 & 0xff);
    break;
    
    case '4':
      fprintf(g_lpOutputFile, "S70500000000%02X\n", ~5 & 0xff);
    break;
  }

  fclose(g_lpOutputFile);
}
/* ----------------------------------------------------------------------
 *
 */
static void addr4(unsigned long int a)
{
  if(index4 > 0)
    finishline();

  address4 = a + g_Offset;
}
/* ----------------------------------------------------------------------
 *
 */
static void byte4(unsigned char b)
{
  buf4[index4++] = b;

  if(index4==SREC_BYTESPERLINE) 
  {
    finishline();
    address4 += SREC_BYTESPERLINE;
  }
}
/* ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 *
 */
void finishline()
{
  int i;

  check4 = index4 + (address4 & 0xff) + ((address4 >> 8) & 0xff) + 4;

  switch(g_Format4) 
  {
    case '2':
      fprintf(g_lpOutputFile, "S1%02X%04X", index4 + 4,  address4 & 0xffff);
    break;

    case '3':
      fprintf(g_lpOutputFile, "S2%02X%06X", index4 + 6, address4 & 0xffffff);
      check4 += ((address4>>16) & 0xff) + 2;
    break;

    case '4':
      fprintf(g_lpOutputFile, "S3%02X%08X", index4 + 8, address4);
      check4 += ((address4>>16) & 0xff) + ((address4 >> 24) & 0xff) + 4;
    break;
  }

  for(i = 0; i < index4; i++) 
  {
    fprintf(g_lpOutputFile, "%02X", buf4[i] & 0xff);
    check4 += buf4[i];
  }

  fprintf(g_lpOutputFile, "%02X\n", (~check4 & 0xff) );
  index4 = 0;
}
/**********************************************************************
 * 
 */
char *StrNToUpper(char *p_lpString, int p_Count)
{
  int i;

  if(p_Count > (i = strlen(p_lpString)))
    p_Count = i;

  for(i = 0; i < p_Count; i++)
    p_lpString[i] = toupper(p_lpString[i]);

  return p_lpString;
}
/**********************************************************************
 * 
 */
void AddStringEntry(char *p_lpString)
{
  LPSTRING_TABLE_ENTRY l_lpStringTable;

  if(!g_Native)
    return;

  switch(*p_lpString)
  {
    case 'n':
    case 'N':
    break;

    default:
    return;
  }

  if(strncmp(StrNToUpper(p_lpString, strlen(p_lpString)), "NATIVE_", 7))
    return;

  if(!(l_lpStringTable = malloc(sizeof(STRING_TABLE_ENTRY))))
  {
    return;
  }

  memset(l_lpStringTable, 0, sizeof(STRING_TABLE_ENTRY));
  strcpy(l_lpStringTable->String, p_lpString);
  l_lpStringTable->Address = g_CurrentAddress;
  l_lpStringTable->Next    = g_StringTable.lpHead;
  g_StringTable.lpHead     = l_lpStringTable;
  g_StringTable.Count++;
  g_StringTable.ByteCount += strlen(p_lpString);
}
/**********************************************************************
 * 
 */
void DumpStringTable(void)
{
  LPSTRING_TABLE_ENTRY l_lpStringTable = g_StringTable.lpHead;

  printf("String table:\n");
  printf("String count = %d\n", g_StringTable.Count);
  printf("String byte count = %d\n", g_StringTable.ByteCount);

  while(l_lpStringTable)
  {
    printf("%s: 0x%06x\n", l_lpStringTable->String, l_lpStringTable->Address);
    l_lpStringTable = l_lpStringTable->Next;
  }
  
  printf("\n");
}
/**********************************************************************
 * 
 */
void AddAFixup(void)
{
  LPFIXUP_TABLE_ENTRY lpCurrentEntry;

  if(!g_Native)
    return;

  if(!(lpCurrentEntry = malloc(sizeof(FIXUP_TABLE_ENTRY))))
  {
    return;
  }
  
  lpCurrentEntry->Address = g_CurrentAddress;
  lpCurrentEntry->Next = g_AFixups.lpHead;
  g_AFixups.lpHead = lpCurrentEntry;
  g_AFixups.Count++;
}
/**********************************************************************
 * 
 */
void AddLFixup(void)
{
  LPFIXUP_TABLE_ENTRY lpCurrentEntry;

  if(!g_Native)
    return;

  if(!(lpCurrentEntry = malloc(sizeof(FIXUP_TABLE_ENTRY))))
  {
    return;
  }
  
  lpCurrentEntry->Address = g_CurrentAddress;
  lpCurrentEntry->Next = g_LFixups.lpHead;
  g_LFixups.lpHead = lpCurrentEntry;
  g_LFixups.Count++;
}
/**********************************************************************
 * 
 */
void AddDFixup(void)
{
  LPFIXUP_TABLE_ENTRY lpCurrentEntry;

  if(!g_Native)
    return;

  if(!(lpCurrentEntry = malloc(sizeof(FIXUP_TABLE_ENTRY))))
  {
    return;
  }
  
  lpCurrentEntry->Address = g_CurrentAddress;
  lpCurrentEntry->Next = g_DFixups.lpHead;
  g_DFixups.lpHead = lpCurrentEntry;
  g_DFixups.Count++;
}
/**********************************************************************
 * 
 */
void DumpFixups(void)
{
  unsigned long i;
  LPFIXUP_TABLE_ENTRY lpCurrentEntry;

  printf("Fixups:\n");
  printf("AFixup count = %d\n", g_AFixups.Count);

  lpCurrentEntry = g_AFixups.lpHead;

  for(i = 0; i < g_AFixups.Count; i++, lpCurrentEntry = lpCurrentEntry->Next)
    printf("0x%06x,%s", lpCurrentEntry->Address, ((i%8) == 7)? "\n": "");

  printf("\n\nLFixup count = %d\n", g_LFixups.Count);
  lpCurrentEntry = g_LFixups.lpHead;

  for(i = 0; i < g_LFixups.Count; i++, lpCurrentEntry = lpCurrentEntry->Next)
    printf("0x%06x,%s", lpCurrentEntry->Address, ((i%8) == 7)? "\n": "");

  printf("\n\nDFixup count = %d\n", g_DFixups.Count);
  lpCurrentEntry = g_DFixups.lpHead;

  for(i = 0; i < g_DFixups.Count; i++, lpCurrentEntry = lpCurrentEntry->Next)
    printf("0x%06x,%s", lpCurrentEntry->Address, ((i%8) == 7)? "\n": "");

  printf("\n");
}
/**********************************************************************
 * 
 */
void Cleanup(void)
{
  LPFIXUP_TABLE_ENTRY  l_lpCurrentEntry = NULL;
  LPSTRING_TABLE_ENTRY l_lpStringEntry  = NULL;

  if(!g_Native)
    return;

  while(g_AFixups.lpHead)
  {
    l_lpCurrentEntry = g_AFixups.lpHead;
    g_AFixups.lpHead = g_AFixups.lpHead->Next;
    free(l_lpCurrentEntry);
  }

  memset((char *)&g_AFixups, 0, sizeof(g_AFixups));

  while(g_LFixups.lpHead)
  {
    l_lpCurrentEntry = g_LFixups.lpHead;
    g_LFixups.lpHead = g_LFixups.lpHead->Next;
    free(l_lpCurrentEntry);
  }

  memset((char *)&g_LFixups, 0, sizeof(g_LFixups));
  
  while(g_DFixups.lpHead)
  {
    l_lpCurrentEntry = g_DFixups.lpHead;
    g_DFixups.lpHead = g_DFixups.lpHead->Next;
    free(l_lpCurrentEntry);
  }

  memset((char *)&g_DFixups, 0, sizeof(g_DFixups));
  g_AFixups.Count = g_LFixups.Count = g_DFixups.Count = 0;

  if(g_DynamicBuffer.lpHead)
  {
    // first block uses a static buffer so don't try to free it
    g_DynamicBuffer.lpCurrent = g_DynamicBuffer.lpHead->Next;
    free(g_DynamicBuffer.lpHead);

    while(g_DynamicBuffer.lpCurrent)
    {
      g_DynamicBuffer.lpHead = g_DynamicBuffer.lpCurrent;
      g_DynamicBuffer.lpCurrent = g_DynamicBuffer.lpCurrent->Next;
      free(g_DynamicBuffer.lpHead->Data);
      free(g_DynamicBuffer.lpHead);
    }
  }

  memset((char *)&g_DynamicBuffer, 0, sizeof(g_DynamicBuffer));
  l_lpStringEntry = g_StringTable.lpHead;

  while(l_lpStringEntry)
  {
    g_StringTable.lpHead = g_StringTable.lpHead->Next;
    free(l_lpStringEntry);
    l_lpStringEntry = g_StringTable.lpHead;
  }

  memset((char *)&g_StringTable, 0, sizeof(g_StringTable));
}
/**********************************************************************/
