/* ----------------------------------------------------------------------
 * FILE: main.c
 * PACKAGE: as31 - 8031/8051 Assembler.
 *
 * DESCRIPTION:
 *  The file contains main(). It handles the arguments and makes
 *  sure that pass 1 is done before pass 2 etc...
 *
 * REVISION HISTORY:
 *  Jan. 19, 1990 - Created. (Ken Stauffer)
 *    August 1999 - Converted to 24 bit addressing. (Tom Chenot, Dallas Semiconductor) 
 *                - Fixed 1-off bug in certain opcodes.
 *                - Added Intel Hex386 output generation.
 *                - Replaced lexical analyzer.
 *                - Added TBin output generation.
 *                - Added TLib output generation.
 *
 *
 * AUTHOR:
 *  Original code in this file written by Ken Stauffer (University of Calgary).
 *  January, 1990. 
 *  24 bit addressing extensions written by Tom Chenot (Dallas Semiconductor).
 *  August 1999.
 *
 */

/*
 $Workfile: main.c $
 $Revision: 20 $
 $Date: 2/26/02 8:48a $
 $Author: Chenot $
 $Modtime: 2/26/02 7:13a $
 $Log: /TINI/390 Assembler/source/a390/main.c $
 * 
 * 20    2/26/02 8:48a Chenot
 * added xrl a,#0 and xrl a, #0ffh to statistics
 * 
 * 19    2/12/02 12:13p Chenot
 * added support for native library version
 * 
 * 18    2/11/02 11:12a Chenot
 * 
 * 15    4/30/01 12:34p Chenot
 * 
 * 14    3/19/01 12:32p Chenot
 * 
 * 13    12/20/00 5:54a Chenot
 * 
 * 12    11/16/00 10:01a Chenot
 * 
 * 11    11/16/00 8:05a Chenot
 * 
 * 10    9/09/00 6:28p Chenot
 * 
 * 9     9/08/00 12:52p Chenot
 * 
 * 8     9/08/00 10:04a Chenot
 * 
 * 6     5/20/00 6:16p Chenot
 * 
 * 5     5/20/00 3:20p Chenot
 * 
 * 4     5/18/00 9:25p Chenot
 * 
 * 3     5/16/00 8:46p Chenot
 * 
 * 2     5/15/00 6:10a Chenot
 * 
 * 30    4/07/00 7:23a Chenot
 * 
 * 29    4/07/00 6:56a Chenot
 * 
 * 28    10/01/99 6:54a Tomc
 * 
 * 27    9/23/99 8:02a Tomc
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "a390.h"


#ifdef WIN32
#define DIR_SEP '\\'
#else
#define DIR_SEP '/'
#endif

/* TLib linkage info */
#define TINI_NATIVE_LEGACY -1
#define TINI_NATIVE_390 0
#define TINI_NATIVE_400 1

extern int           lineno;
extern int           pass,
                     fatal;
extern unsigned long lc,
                     cur_lc;
extern int           yydebug;
extern FILE          *yyin;
extern int           g_DebugOutput;
extern char          g_StatsOn;

extern int  yyparse(void);
extern void yyrestart(FILE *input_file);
extern void DumpSymbols(void);
extern char WriteStats(void);
extern void Cleanup(void);
extern void DumpFixups(void);
extern void DumpStringTable(void);
extern void emitusage(void);
extern void emitopen(char *file, char *ftype, char *arg);
extern void emitclose(void);
extern void emitdelete(void);
extern void emitaddr(unsigned long int a);
extern void emitbyte(unsigned char b);
extern void inclc(int i);

extern void syminit(void);
extern void InitPass(void);
extern void GenerateSymbolEquateFile(char *lpFileName, char *lpBanner);

extern void DumpExtraStatistics(void);

/**/
char    g_Native            = 1;
char    g_TINIModule        = TINI_NATIVE_LEGACY;
char    g_TINIStrings[2][8] = {"TINI390", "TINI400"};
char    *g_lpFirmwareVersion= NULL;
char    g_AllowDirectOverlap= 0;
char    g_ErrorsEncountered = 0;
jmp_buf main_env;
char    *asmfile;
int     dashl               = 0;
char    *dashS              = NULL;
FILE    *listing;
int     g_EndOfFile         = 0;
char    g_BannerTopString[] = 
#ifdef WIN32
          "ͻ\n";
#else
          "+------------------------------------------------------------------+\n";
#endif
char    g_BannerString[] =
#ifdef WIN32
          "";
#else
          "|";
#endif
char    g_BannerBottomString[] = 
#ifdef WIN32
          "ͼ\n";
#else
          "+------------------------------------------------------------------+\n";
#endif

#define DEFAULT_EXTENSION   "tlib"
#define ALTERNATE_EXTENSION "tbin"

#define VERSION_MAJOR 0
#define VERSION_MINOR 115
#define BUILD "$Revision: 20 $"
void PrintBanner()
{
  char  l_lpBuild[]      = BUILD;
  char  *l_lpTemp        = NULL;

  fprintf(stderr,
          "\n%s%s Dallas Semiconductor DS80C390 Assembler, Beta Version %d.%03d.%04u %s\n%s"
          "Last build %s at %s\n\n",
          g_BannerTopString,
          g_BannerString,
          VERSION_MAJOR, 
          VERSION_MINOR, 
          strtol(strtok(l_lpBuild, "evision: $R"), &l_lpTemp, 10),
          g_BannerString, 
          g_BannerBottomString,
          __DATE__, 
          __TIME__);

  if(dashS)
    printf("Generating linkage file tini.inc with banner \"%s\"\n", dashS);
}
/*
 */
void main(int argc, char **argv)
{
  FILE  *fin             = NULL;
  char  *p               = NULL,
        *dashF           = NULL,
        dashFAlternate[] = ALTERNATE_EXTENSION,
        *dashA           = NULL;
  char  objfile[260]     = "";
  char  lstfile[260]     = "";
  int   i                = 0;
  int   l_Count          = 0;
  int   l_CodeJumpOffset = 0;
  char  l_DisplayLineNo  = 1;
  char  l_VerboseMode    = 0;
  char  l_QuietMode      = 0;

  if(argc < 2)
  {
    PrintBanner();
    fprintf(stderr,
            "Usage: %s [-l] [-s] [-Ffmt] [-p <proc>] [-f <ver>] <input file>\n"\
            "\t-l Generate list file \t\t(default: no list file)\n"\
            "\t-s Print statistical information(default: off)\n"\
            "\t-d Allow direct declarations\t(default: do not allow)\n"\
            "\t-F Output format \t\t(default: tlib)\n"\
            "\t-p <tini processor> [390,400]\t(default: legacy)\n"\
            "\t-f <tini firmware version>\t(example: 1.02e)\n",
            strrchr(argv[0], DIR_SEP)? strrchr(argv[0], DIR_SEP)+1: argv[0]);
    emitusage();
    fprintf(stderr, "\n");
    exit(1);
  }
  
  InitPass();

  for(i = 1; i < argc; i++ ) 
  {
    if(argv[i][0] != '-') 
      break;

    switch(argv[i][1])
    {
      case 'v': case 'V':// verbose
        l_VerboseMode = 1;
      break;

      case 'l': // list file
        dashl = 1;
      break;

      case 's': // stats
        g_StatsOn = 1;
      break;

      case 'p': // processor
        if(argv[++i])
        {
          if(!strcmp(argv[i], "390"))
          {
            g_TINIModule = TINI_NATIVE_390;
            break;
          }
          else if(!strcmp(argv[i], "400"))
          {
            g_TINIModule = TINI_NATIVE_400;
            break;
          }
        }

        PrintBanner();
        fprintf(stderr,"*** Unknown processor.\n");
        fprintf(stderr, "\n");
        exit(1);        
      break;

      case 'f': // firmware version
        g_lpFirmwareVersion = argv[++i];
      break;

      case 'q': // quiet
        l_QuietMode = 1;
      break;

      case 'd': // allow directs and ORG statements
        g_Native = 0;
      break;

      case 'o': // allow directs to overlap
        g_AllowDirectOverlap = 1;
      break;

      case 'S': // String in tini.inc
        dashS = &argv[i][2];
      break;

      case 'F': // file type
        dashF = argv[i]+2;
      break;

      case 'A': // argument for emitopen
        dashA = argv[i]+2;
      break;

      default:
        PrintBanner();
        fprintf(stderr,"*** Unknown flag \'%c\'.\n",argv[i][1]);
        fprintf(stderr, "\n");
        exit(1);
      break;
    }// switch
  }

  if(i == argc) 
  {
    PrintBanner();
    fprintf(stderr, "*** Missing input file.\n");
    fprintf(stderr, "\n");
    exit(1);
  }

  if(!(p = strchr(argv[i], '.')))
  {
    PrintBanner();
    fprintf(stderr, "*** Error parsing command line\n");
    fprintf(stderr, "\n");
    exit(1);
  }

  if((g_TINIModule == TINI_NATIVE_LEGACY) && g_lpFirmwareVersion)
  {
    PrintBanner();
    fprintf(stderr, "*** Error, -f switch requires -p switch\n");
    fprintf(stderr, "\n");
    exit(1);
  }

  if(!l_QuietMode)
    PrintBanner();

  asmfile = argv[i];

  if((yyin = (fin = fopen(argv[i], "r"))) == NULL) 
  {
    fprintf(stderr, "*** Cannot open input file: %s\n", argv[i]);
    fprintf(stderr, "\n");
    exit(1);
  }

  if(dashl) 
  {
    strcpy(lstfile, argv[i]);
    strcpy(lstfile + (p - argv[i]),".lst");
    listing = fopen(lstfile, "w");
  
    if(listing == NULL) 
    {
      fprintf(stderr, "Cannot open file: %s for writting.\n", lstfile);
      exit(1);
    }
  }

  strcpy(objfile, argv[i]);
  strcpy(objfile + (p - argv[i]),".");
  dashF? strcat(objfile, dashF): g_Native? strcat(objfile, DEFAULT_EXTENSION): 
                                           (strcat(objfile, ALTERNATE_EXTENSION),
                                             dashF = dashFAlternate);

  emitopen(objfile, dashF, dashA);

  syminit();
  fatal   = 0;
  lineno  = 1;
  pass    = 0;
  cur_lc  = lc = 0x0000;
   
  if(setjmp(main_env)) 
  {
    emitclose();
    fclose(fin);
    unlink(objfile);
    exit(1);
  }

  /*
   * P A S S    1
   */
  if(g_Native && (g_TINIModule != TINI_NATIVE_LEGACY))
  {
    l_CodeJumpOffset = strlen(g_TINIStrings[g_TINIModule]) + // processor magic bytes
                       1 +                                   // asciiz
                       (g_lpFirmwareVersion?                 // version bytes
                          strlen(g_lpFirmwareVersion): 0)+
                       1;                                    // asciiz
    inclc(2+l_CodeJumpOffset);
  }

  g_EndOfFile = 0;

  if(!l_QuietMode)
    fprintf(stderr,"Starting pass 1...\n");

  do
  { /* none of that recursive stuff, do parse iteratively */
    /* our source blows the top off of the stack */
    if(yyparse())    /* parse a statement   */
    {
      break;
    }
    
    if(l_VerboseMode) /* this slows execution */
    {
      if(l_DisplayLineNo)
        fprintf(stderr,"Line: %d\r", lineno);
      else
      {
        switch((l_Count++)%4)
        {
          case 0:
            fprintf(stderr,"|\r");
          break;

          case 1:
            fprintf(stderr,"/\r");
          break;

          case 2:
            fprintf(stderr,"-\r");
          break;

          case 3:
            fprintf(stderr,"\\\r");
          break;
        }

        l_Count &= 3;
      }
    }
  } while (!g_EndOfFile);

  if(fatal)
    longjmp(main_env,1);

  rewind(fin);
  yyrestart(yyin);
  lineno  = 1;
  pass++;
  cur_lc  = lc  = 0x0000;
  emitaddr(lc);

  /*
   * P A S S    2
   */
  if(g_Native && (g_TINIModule != TINI_NATIVE_LEGACY))
  {
    unsigned char i;

    if(!l_QuietMode)
      fprintf(stderr,
              "Writing %d bytes of version information to tlib file\n",
               2+l_CodeJumpOffset);

    emitbyte(0x80); // sjmp
    emitbyte((char)l_CodeJumpOffset); // jump offset

    for(i = 0; i < strlen(g_TINIStrings[g_TINIModule]); i++)
      emitbyte(g_TINIStrings[g_TINIModule][i]);

    emitbyte(0); // asciiz

    if(g_lpFirmwareVersion)
    {
      for(i = 0; i < strlen(g_lpFirmwareVersion); i++)
        emitbyte(g_lpFirmwareVersion[i]);
    }

    emitbyte(0); // asciiz
    inclc(2+l_CodeJumpOffset);
  }

  g_EndOfFile = 0;

  if(!l_QuietMode)
    fprintf(stderr,"Starting pass 2...\n");
//yydebug=1;

  do
  { /* none of that recursive stuff, do parse iteratively */
    /* our source blows the top off of the stack */
    if(yyparse())    /* parse a statement   */
    {
      break;
    }

    if(l_VerboseMode) /* this slows execution */
    {
      if(l_DisplayLineNo)
        fprintf(stderr,"Line: %d\r", lineno);
      else
      {
        switch((l_Count++)%4)
        {
          case 0:
            fprintf(stderr,"|\r");
          break;

          case 1:
            fprintf(stderr,"/\r");
          break;

          case 2:
            fprintf(stderr,"-\r");
          break;

          case 3:
            fprintf(stderr,"\\\r");
          break;
        }

        l_Count &= 3;
      }
    }
  } while (!g_EndOfFile);

  if(l_DisplayLineNo)
    if(!l_QuietMode)
      fprintf(stderr,"Lines parsed: %d\n", lineno);
  
  if(dashS)
    GenerateSymbolEquateFile("tini.inc", dashS);
  
  emitclose();
  fclose(fin);

  if(dashl)
  {
    DumpSymbols();
    fclose(listing);
  }

  if(g_ErrorsEncountered)
  {
    fprintf(stderr,"*** Errors detected ***\n");
    emitdelete();
    exit(1);
  }
  else
  {
    WriteStats();
  }

  exit(0);
}
/***************************************************************
 * yywrap
 */
int   yywrap  (void)
{
  g_EndOfFile = 1;
  return 1;
}
